(**************************************************************************)
(*                                                                        *)
(*  Setup for SFtpServer                                                  *)
(*  Copyright (C) 2022   Peter Moylan                                     *)
(*                                                                        *)
(*  This program is free software: you can redistribute it and/or modify  *)
(*  it under the terms of the GNU General Public License as published by  *)
(*  the Free Software Foundation, either version 3 of the License, or     *)
(*  (at your option) any later version.                                   *)
(*                                                                        *)
(*  This program is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU General Public License for more details.                          *)
(*                                                                        *)
(*  You should have received a copy of the GNU General Public License     *)
(*  along with this program.  If not, see <http://www.gnu.org/licenses/>. *)
(*                                                                        *)
(*  To contact author:   http://www.pmoylan.org   peter@pmoylan.org       *)
(*                                                                        *)
(**************************************************************************)

IMPLEMENTATION MODULE BasicPage;

        (****************************************************************)
        (*                                                              *)
        (*                   PM Setup for SFtpServer                    *)
        (*                    Page 1 of the notebook                    *)
        (*                                                              *)
        (*        Last edited:    9 September 2022                      *)
        (*        Status:         OK                                    *)
        (*                                                              *)
        (****************************************************************)


FROM SYSTEM IMPORT ADDRESS, INT16, CARD8, CARD32, CAST, ADR;

IMPORT OS2, OS2RTL, DID, Strings, CommonSettings, INIData, RINIData;

FROM UserPage IMPORT
    (* proc *)  NumberOfUsers;

FROM FSUINI IMPORT
    (* type *)  FHINI,
    (* proc *)  GetFullINIname, OpenINIFiles, CloseINIFiles,
                SetHashMax, GetHashMax, OpenAlternative, DeleteAlternative, OpenRINIForUser,
                INIFetch, INIGetCard, INIGetString, ItemSize,
                RemoteConflict, INIPut, INIPutString;

FROM Languages IMPORT
    (* type *)  LangHandle,
    (* proc *)  StrToBuffer;

FROM FileOps IMPORT
    (* type *)  FilenameString, DirectoryEntry,
    (* proc *)  FirstDirEntry, NextDirEntry, DirSearchDone,
                DeleteFile, MoveFile, GetEXEDirectory;

FROM MiscFuncs IMPORT
    (* type *)  CharArrayPointer,
    (* proc *)  EVAL, ToLower, StringMatch;

FROM Conversions IMPORT
    (* proc *)  CardinalToStringLJ;

FROM Misc IMPORT
    (* proc *)  WinSetDlgItemCard, WinQueryDlgItemCard;

FROM Names IMPORT
    (* type *)  UserName, UserNameIndex;

FROM Timer IMPORT
    (* proc *)  Sleep;

FROM LowLevel IMPORT
    (* proc *)  IAND, IANDB;

FROM Storage IMPORT
    (* proc *)  ALLOCATE, DEALLOCATE;

(**************************************************************************)

CONST
    Nul = CHR(0);
    NewHashMax = 97;

TYPE
    TwoCard = ARRAY [0..1] OF CARDINAL;
    DirectoryString = ARRAY [0..511] OF CHAR;
    TextArray = ARRAY [0..31] OF CHAR;
    TextPointer = POINTER TO TextArray;
    LongTextPointer = POINTER TO ARRAY [0..MAX(CARDINAL) DIV 4] OF CHAR;

    AppPointer = POINTER TO AppRecord;
    KeyPointer = POINTER TO KeyRecord;

    AppRecord = RECORD
                    next: AppPointer;
                    firstkey: KeyPointer;
                    name: TextArray;
                END (*RECORD*);
    KeyRecord = RECORD
                    next: KeyPointer;
                    name: TextArray;
                    length: CARDINAL;
                    pval: LongTextPointer;
                END (*RECORD*);

VAR
    ChangeInProgress: BOOLEAN;
    notebookhandle, pagehandle: OS2.HWND;
    PageID: CARDINAL;
    FTP, SFTP: BOOLEAN;
    HashMax: CARDINAL;
    OldServerPort, OldSFTPPort: CARDINAL;
    OldMaxUsers, OldGuestLimit, OldSpaceThreshold: CARDINAL;
    OldTimeout: CARDINAL;
    OldTelnetDisable: BOOLEAN;
    OurDir, FTPdir, SFTPdir: FilenameString;

(************************************************************************)
(*                   WHICH SERVERS ARE WE CONFIGURING?                  *)
(************************************************************************)

PROCEDURE SetServer (configFTP, configSFTP: BOOLEAN);

    (* Specifies which server(s) we are configuring. *)

    BEGIN
        FTP := configFTP;
        SFTP := configSFTP;
    END SetServer;

(************************************************************************)
(*                    OPERATIONS ON DIALOGUE LABELS                     *)
(************************************************************************)

PROCEDURE EnableFields (hwnd: OS2.HWND);

    (* Enables or disables some of the dialogue fields. *)

    (********************************************************************)

    PROCEDURE DoOne (field: CARDINAL;  cond: BOOLEAN);

        VAR w: OS2.HWND;

        BEGIN
            w := OS2.WinWindowFromID (hwnd, field);

            (* I have never discovered why WinEnableWindow won't work   *)
            (* if the second parameter is a variable.                   *)

            IF cond THEN
                OS2.WinEnableWindow (w, TRUE);
            ELSE
                OS2.WinEnableWindow (w, FALSE);
            END (*IF*);
        END DoOne;

    (********************************************************************)

    BEGIN
        DoOne (DID.FtpPortLabel, FTP);
        DoOne (DID.FTPPort, FTP);
        DoOne (DID.SFtpPortLabel, SFTP);
        DoOne (DID.SFTPPort, SFTP);
        DoOne (DID.MaxGuestUsersLabel, FTP);
        DoOne (DID.MaxGuestUsers, FTP);
        DoOne (DID.FreeSpaceLabel, FTP);
        DoOne (DID.FreeSpace, FTP);
        DoOne (DID.TelnetDisable, FTP);
        DoOne (DID.FTPdirlabel, FTP);
        DoOne (DID.FTPdir, FTP);
        DoOne (DID.bremote, FTP);
        DoOne (DID.SFTPdirlabel, SFTP);
        DoOne (DID.SFTPdir, SFTP);
        DoOne (DID.sbremote, FTP);
    END EnableFields;

(************************************************************************)

PROCEDURE SetLanguage (lang: LangHandle);

    (* Relabels this page in the new language. *)

    VAR stringval: ARRAY [0..511] OF CHAR;

    (********************************************************************)

    PROCEDURE SetOne (item: ARRAY OF CHAR;  place: CARDINAL);

        BEGIN
            StrToBuffer (lang, item, stringval);
            OS2.WinSetDlgItemText (pagehandle, place, stringval);
        END SetOne;

    (********************************************************************)

    BEGIN
        StrToBuffer (lang, "Page1.tab", stringval);
        OS2.WinSendMsg (notebookhandle, OS2.BKM_SETTABTEXT,
                        CAST(ADDRESS,PageID), ADR(stringval));

        SetOne ("Page1.Global", DID.GlobalBox);
        SetOne ("Page1.FtpPort", DID.FtpPortLabel);
        SetOne ("Page1.SFtpPort", DID.SFtpPortLabel);
        SetOne ("Page1.MaxUsers", DID.MaxUsersLabel);
        SetOne ("Page1.MaxGuestUsers", DID.MaxGuestUsersLabel);
        SetOne ("Page1.FreeSpace", DID.FreeSpaceLabel);
        SetOne ("Page1.Timeout", DID.TimeOutLabel);
        SetOne ("Page1.MultiINI", DID.MultiINI);
        SetOne ("Page1.NoTelnet", DID.TelnetDisable);
        SetOne ("Page1.FTPdir", DID.FTPdirlabel);
        SetOne ("Initial.Remote", DID.bremote);
        SetOne ("Page1.SFTPdir", DID.SFTPdirlabel);
        SetOne ("Initial.Remote", DID.sbremote);

    END SetLanguage;

(************************************************************************)
(*                    MOVING USERS BETWEEN INI FILES                    *)
(************************************************************************)

PROCEDURE LoadApp (rhini: RINIData.RHINI;  p: AppPointer);

    (* Loads all data for one application from the currently opened     *)
    (* INI file.                                                        *)

    VAR state: RINIData.StringReadState;
        appname, keyname: TextArray;
        head, current: KeyPointer;

    BEGIN
        appname := p^.name;

        (* Find all keys and the corresponding values. *)

        RINIData.GetStringList (rhini, appname, '', state);
        head := NIL;  current := NIL;
        LOOP
            RINIData.NextString (state, keyname);
            IF keyname[0] = Nul THEN
                EXIT (*LOOP*);
            END (*IF*);
            IF head = NIL THEN
                NEW (head);  current := head;
            ELSE
                NEW (current^.next);  current := current^.next;
            END (*IF*);
            current^.next := NIL;
            current^.name := keyname;
            IF NOT RINIData.ItemSize (rhini, appname, keyname, current^.length) THEN
                current^.length := 0;
            END (*IF*);
            IF current^.length = 0 THEN
                current^.pval := NIL;
            ELSE
                ALLOCATE (current^.pval, current^.length);
                IF NOT RINIData.INIFetchBinary (rhini, appname, keyname,
                               current^.pval^, current^.length) THEN
                    DEALLOCATE (current^.pval, current^.length);
                    current^.length := 0;
                END (*IF*);
            END (*IF*);
        END (*LOOP*);
        RINIData.CloseStringList (state);
        p^.firstkey := head;

    END LoadApp;

(************************************************************************)

PROCEDURE StoreApp (rhini: RINIData.RHINI;  p: AppPointer);

    (* Stores all key/value pairs for one application to the currently  *)
    (* opened INI file.  As a side-effect, we deallocate the key list.  *)
    (* If this is a duplicate entry, we discard the data.               *)

    VAR appname: TextArray;
        current, temp: KeyPointer;
        storeit: BOOLEAN;

    BEGIN
        storeit := NOT RINIData.AppExists (rhini, p^.name);
        appname := p^.name;  current := p^.firstkey;
        p^.firstkey := NIL;
        WHILE current <> NIL DO
            IF current^.length = 0 THEN
                IF storeit THEN
                    RINIData.INIPutBinary (rhini, appname, current^.name, current, 0);
                END (*IF*);
            ELSE
                IF storeit THEN
                    RINIData.INIPutBinary (rhini, appname, current^.name, current^.pval^,
                                                      current^.length);
                END (*IF*);
                DEALLOCATE (current^.pval, current^.length);
            END (*IF*);
            temp := current;
            current := current^.next;
            DISPOSE (temp);
        END (*WHILE*);
    END StoreApp;

(************************************************************************)

PROCEDURE LoadAllUserData (rhini: RINIData.RHINI): AppPointer;

    (* Loads all user data from the currently opened INI file.  This is *)
    (* a destructive operation: we also delete the data from the file.  *)

    VAR state: RINIData.StringReadState;
        appname: TextArray;
        head, current: AppPointer;

    BEGIN
        (* Start by getting a linear list of all applications. *)

        RINIData.GetStringList (rhini, '', '', state);
        head := NIL;  current := NIL;
        LOOP
            RINIData.NextString (state, appname);
            IF appname[0] = Nul THEN
                EXIT (*LOOP*);
            END (*IF*);
            IF NOT Strings.Equal (appname, "$SYS") THEN
                IF head = NIL THEN
                    NEW (head);  current := head;
                ELSE
                    NEW (current^.next);  current := current^.next;
                END (*IF*);
                current^.next := NIL;
                current^.firstkey := NIL;
                current^.name := appname;
            END (*IF*);
        END (*LOOP*);
        RINIData.CloseStringList (state);

        (* Now pick up all values for each application. *)

        current := head;
        WHILE current <> NIL DO
            LoadApp (rhini, current);
            RINIData.INIDeleteApp (rhini, current^.name);
            current := current^.next;
        END (*WHILE*);

        RETURN head;

    END LoadAllUserData;

(************************************************************************)

PROCEDURE LoadUserList (rhini: RINIData.RHINI): AppPointer;

    (* Loads the names of all applications, excluding those whose name  *)
    (* starts with '$', from the currently opened INI file.             *)

    VAR state: RINIData.StringReadState;
        appname: TextArray;
        head, current: AppPointer;

    BEGIN

        RINIData.GetStringList (rhini, '', '', state);
        head := NIL;  current := NIL;
        LOOP
            RINIData.NextString (state, appname);
            IF appname[0] = Nul THEN
                EXIT (*LOOP*);
            ELSIF appname[0] <> '$' THEN
                IF head = NIL THEN
                    NEW (head);  current := head;
                ELSE
                    NEW (current^.next);  current := current^.next;
                END (*IF*);
                current^.next := NIL;
                current^.firstkey := NIL;
                current^.name := appname;
            END (*IF*);
        END (*LOOP*);
        RINIData.CloseStringList (state);

        RETURN head;

    END LoadUserList;

(************************************************************************)

PROCEDURE SwitchOneToSingleINI (j: CARDINAL;  conflict: BOOLEAN;  status: OS2.HWND);

    (* Switches the j'th INI set (j = 0 or 1) from multiple INI files   *)
    (* to a single INI file.                                            *)

    CONST CountSize = 16;

    VAR S: RINIData.SearchState;
        list, temp: AppPointer;
        count: CARDINAL;
        rhini, rhini2: RINIData.RHINI;
        remote, found: BOOLEAN;
        TotalBuffer: ARRAY [0..CountSize-1] OF CHAR;
        CountBuffer: ARRAY [0..2*CountSize] OF CHAR;
        mask: ARRAY [0..127] OF CHAR;
        fullname, name: FilenameString;
        N: CARDINAL;

    BEGIN
        (* Get the main INI name for set j. *)

        GetFullINIname (j, fullname, remote);
        IF (fullname[0] = Nul) OR (GetHashMax(j) = 0) THEN
            RETURN;
        END (*IF*);

        (* Turn the name into a wildcard search mask, to produce the    *)
        (* set of INI files we must check.                              *)

        count := 0;
        CardinalToStringLJ (NumberOfUsers(), TotalBuffer);
        HashMax := 0;
        SetHashMax (0);
        Strings.Assign (fullname, mask);
        N := Strings.Length(mask) - 4;
        Strings.Insert ('*', N, mask);
        rhini := RINIData.OpenINIFile (fullname, remote, conflict);
        found := RINIData.FirstDirEntry (mask, remote, FALSE, FALSE, TRUE, S, name);
        WHILE found DO

            (* The length test below is to exclude the file FTPD.?NI *)

            IF LENGTH (name) > 8 THEN
                rhini2 := OpenAlternative (j, name);
                list := LoadAllUserData (rhini2);
                RINIData.CloseINIFile (rhini2);
                DeleteAlternative (j, name, status);

                WHILE list <> NIL DO
                    ToLower (list^.name);
                    StoreApp (rhini, list);
                    INC (count);
                    temp := list;
                    list := list^.next;
                    DISPOSE (temp);
                END (*WHILE*);
                CardinalToStringLJ (count, CountBuffer);
                Strings.Append ('/', CountBuffer);
                Strings.Append (TotalBuffer, CountBuffer);
                OS2.WinSetWindowText (status, CountBuffer);

            END (*IF*);
            found := RINIData.NextDirEntry (S, name);
        END (*WHILE*);
        RINIData.DirSearchDone (S);

        RINIData.INIPut (rhini, '$SYS', 'HashMax', HashMax);
        RINIData.CloseINIFile (rhini);

    END SwitchOneToSingleINI;

(************************************************************************)

PROCEDURE SwitchToSingleINI (status: OS2.HWND);

    (* Switches from multiple INI files to a single INI file.     *)

    VAR conflict: BOOLEAN;

    BEGIN
        conflict := RemoteConflict();
        SwitchOneToSingleINI (0, conflict, status);
        SwitchOneToSingleINI (1, conflict, status);
    END SwitchToSingleINI;

(************************************************************************)

PROCEDURE SwitchOneToMultipleINI (j: CARDINAL;  conflict: BOOLEAN;  status: OS2.HWND);

    (* Switches the j'th INI set (j = 0 or 1) from a single INI file    *)
    (* to multiple INI files.                                           *)

    CONST GroupSize = 20;
          CountSize = 16;

    VAR list, current, temp: AppPointer;
        count, subcount: CARDINAL;
        rhini, rhini2: RINIData.RHINI;
        TotalBuffer: ARRAY [0..CountSize-1] OF CHAR;
        CountBuffer: ARRAY [0..2*CountSize] OF CHAR;
        fullname: FilenameString;
        remote: BOOLEAN;

    BEGIN
        count := 0;
        CardinalToStringLJ (NumberOfUsers(), TotalBuffer);
        GetFullINIname (j, fullname, remote);
        IF (fullname[0] = Nul) OR (GetHashMax(j) > 0) THEN
            (* Nothing to do *)
            RETURN;
        END (*IF*);
        HashMax := NewHashMax;
        SetHashMax (HashMax);
        rhini := RINIData.OpenINIFile (fullname, remote, conflict);
        RINIData.INIPut (rhini, '$SYS', 'HashMax', HashMax);
        list := LoadUserList (rhini);

        LOOP
            subcount := 0;
            IF list = NIL THEN
                EXIT (*LOOP*);
            END (*IF*);
            current := list;
            WHILE (subcount <= GroupSize) AND (current <> NIL) DO
                ToLower (current^.name);
                LoadApp (rhini, current);
                rhini2 := OpenRINIForUser (j, current^.name);
                StoreApp (rhini2, current);
                RINIData.CloseINIFile (rhini2);
                INC (subcount);
                current := current^.next;
            END (*WHILE*);

            (* We have now copied one batch of users to new INI files. *)
            (* Make a second pass through the list, deleting those     *)
            (* users from the main INI file.                           *)

            WHILE list <> current DO
                RINIData.INIDeleteApp (rhini, list^.name);
                temp := list;
                list := list^.next;
                DISPOSE (temp);
            END (*WHILE*);
            INC (count, subcount);
            CardinalToStringLJ (count, CountBuffer);
            Strings.Append ('/', CountBuffer);
            Strings.Append (TotalBuffer, CountBuffer);
            OS2.WinSetWindowText (status, CountBuffer);

        END (*LOOP*);

        RINIData.CloseINIFile (rhini);

    END SwitchOneToMultipleINI;

(************************************************************************)

PROCEDURE SwitchToMultipleINI (status: OS2.HWND);

    (* Switches from multiple INI files to a single INI file.     *)

    VAR conflict: BOOLEAN;

    BEGIN
        conflict := RemoteConflict();
        SwitchOneToMultipleINI (0, conflict, status);
        SwitchOneToMultipleINI (1, conflict, status);
    END SwitchToMultipleINI;

(************************************************************************)

PROCEDURE MoveUsers (hwnd: OS2.HWND;  NewMulti: BOOLEAN);

    (* Switches between a single INI file and multiple INI files,       *)
    (* depending on the state of the MultiINI checkbox.                 *)

    VAR status: OS2.HWND;

    BEGIN
        IF NewMulti <> (HashMax > 0) THEN
            status := OS2.WinWindowFromID (hwnd, DID.Page1Status);
            OS2.WinSetWindowText (status,
                                     "Converting INI data, please wait");
            IF NewMulti THEN
                SwitchToMultipleINI (status);
            ELSE
                SwitchToSingleINI (status);
            END (*IF*);
            OS2.WinSetWindowText (status, "")
        END (*IF*);
    END MoveUsers;

(************************************************************************)
(*                     LOADING DATA FROM THE INI FILE                   *)
(************************************************************************)

PROCEDURE LoadDirNames (hwnd: OS2.HWND);

    (* The two server directories are a special case, because their     *)
    (* names come from the Setup INI file.                              *)

    VAR SetupINI: FilenameString;
        hini: INIData.HINI;
        app: ARRAY [0..5] OF CHAR;
        server: CARD8;

    BEGIN
        CommonSettings.GetINIname (SetupINI);
        hini := INIData.OpenINIFile (SetupINI);

        app := "Setup";
        IF NOT INIData.INIGetString (hini, app, 'FTPDir', FTPdir) THEN
            FTPdir := OurDir;
        END (*IF*);
        OS2.WinSetDlgItemText (hwnd, DID.FTPdir, FTPdir);

        IF NOT INIData.INIGetString (hini, app, 'SFTPDir', SFTPdir) THEN
            SFTPdir := OurDir;
        END (*IF*);
        OS2.WinSetDlgItemText (hwnd, DID.SFTPdir, SFTPdir);

        (* Is either directory remote? *)

        IF NOT INIData.INIGet (hini, app, 'server', server) THEN
            server := 1;
        END (*IF*);
        IF IANDB (server, 2) = 0 THEN
            OS2.WinShowWindow (OS2.WinWindowFromID (hwnd, DID.bremote), FALSE);
        ELSE
            OS2.WinShowWindow (OS2.WinWindowFromID (hwnd, DID.bremote), TRUE);
        END (*IF*);
        IF IANDB (server, 8) = 0 THEN
            OS2.WinShowWindow (OS2.WinWindowFromID (hwnd, DID.sbremote), FALSE);
        ELSE
            OS2.WinShowWindow (OS2.WinWindowFromID (hwnd, DID.sbremote), TRUE);
        END (*IF*);

        INIData.CloseINIFile (hini);

    END LoadDirNames;

(************************************************************************)

PROCEDURE LoadValues (hwnd: OS2.HWND);

    (* Fills the dialogue elements on page 1 with data from the INI     *)
    (* file, or loads default values if they're not in the INI file.    *)

    VAR val0: CARDINAL;  status: OS2.HWND;
        fhini: FHINI;

    BEGIN
        LoadDirNames (hwnd);
        fhini := OpenINIFiles();

        (* Server ports. *)

        IF INIGetCard (fhini, '$SYS', 'ServerPort', val0) THEN
            OldServerPort := val0;
        ELSE
            val0 := 21;
        END (*IF*);
        WinSetDlgItemCard (hwnd, DID.FTPPort, val0);

        IF INIGetCard (fhini, '$SYS', 'SFTPPort', val0) THEN
            OldSFTPPort := val0;
        ELSE
            val0 := 22;
        END (*IF*);
        WinSetDlgItemCard (hwnd, DID.SFTPPort, val0);

        (* Maximum number of users. *)

        IF INIGetCard (fhini, '$SYS', 'MaxUsers', val0) THEN
            OldMaxUsers := val0;
        ELSE
            val0 := 10;
        END (*IF*);
        WinSetDlgItemCard (hwnd, DID.MaxUsers, val0);

        (* Maximum number of guest users. *)

        IF INIGetCard (fhini, '$SYS', 'GuestLimit', val0) THEN
            OldGuestLimit := val0;
        ELSE
            DEC(val0);
        END (*IF*);
        WinSetDlgItemCard (hwnd, DID.MaxGuestUsers, val0);

        (* Free space threshold. *)

        IF INIGetCard (fhini, '$SYS', 'SpaceThreshold', val0) THEN
            OldSpaceThreshold := val0;
        ELSE
            val0 := 10;
        END (*IF*);
        WinSetDlgItemCard (hwnd, DID.FreeSpace, val0);

        (* Timeout values. *)

        IF INIGetCard (fhini, '$SYS', 'TimeOut', val0) THEN
            OldTimeout := val0;
        ELSE
            val0 := 900;
        END (*IF*);
        WinSetDlgItemCard (hwnd, DID.TimeOut, val0);

        (* HashMax controls the MultiINI checkbox. *)

        IF NOT INIGetCard (fhini, '$SYS', 'HashMax', HashMax) THEN
            HashMax := 0;
        END (*IF*);

        (* Disable Telnet compatibility. *)

        IF NOT INIFetch (fhini, '$SYS', 'TelnetDisable', OldTelnetDisable) THEN
            OldTelnetDisable := FALSE;
        END (*IF*);
        OS2.WinSendDlgItemMsg (hwnd, DID.TelnetDisable, OS2.BM_SETCHECK,
                               OS2.MPFROMSHORT(ORD(OldTelnetDisable)), NIL);

        CloseINIFiles (fhini);

        (* Allowance for the possibility that HashMax is changing from  *)
        (* one nonzero value to another nonzero value, for example      *)
        (* because of a version change.                                 *)

        status := OS2.WinWindowFromID (hwnd, DID.Page1Status);
        OS2.WinSetWindowText (status, "");
        IF HashMax <> NewHashMax THEN
            OS2.WinSetWindowText (status,
                                     "Converting INI data, please wait");
            val0 := HashMax;
            SwitchToSingleINI (status);
            HashMax := val0;
            IF HashMax <> 0 THEN
                SwitchToMultipleINI (status);
            END (*IF*);
            OS2.WinSetDlgItemText (hwnd, DID.Page1Status, "");
        END (*IF*);
        OS2.WinSendDlgItemMsg (hwnd, DID.MultiINI, OS2.BM_SETCHECK,
                               OS2.MPFROMSHORT(ORD(HashMax<>0)), NIL);

        EnableFields (hwnd);

    END LoadValues;

(************************************************************************)
(*                      STORING DATA TO THE INI FILE                    *)
(************************************************************************)

PROCEDURE StoreData (hwnd: OS2.HWND);

    (* Stores the values on page 1 back into the INI file. *)

    CONST MaxTimeout = (MAX(CARD32) DIV 2) DIV 1000;

    VAR value: CARDINAL;  TelnetDisable: BOOLEAN;
        fhini: FHINI;

    BEGIN
        fhini := OpenINIFiles();

        (* Server ports. *)

        WinQueryDlgItemCard (hwnd, DID.FTPPort, value);
        IF value <> OldServerPort THEN
            INIPut (fhini, '$SYS', 'ServerPort', value);
        END (*IF*);

        WinQueryDlgItemCard (hwnd, DID.SFTPPort, value);
        IF value <> OldSFTPPort THEN
            INIPut (fhini, '$SYS', 'SFTPPort', value);
        END (*IF*);

        (* Maximum number of users. *)

        WinQueryDlgItemCard (hwnd, DID.MaxUsers, value);
        IF value <> OldMaxUsers THEN
            INIPut (fhini, '$SYS', 'MaxUsers', value);
        END (*IF*);

        (* Maximum number of guest users. *)

        WinQueryDlgItemCard (hwnd, DID.MaxGuestUsers, value);
        IF value <> OldGuestLimit THEN
            INIPut (fhini, '$SYS', 'GuestLimit', value);
        END (*IF*);

        (* Free space threshold. *)

        WinQueryDlgItemCard (hwnd, DID.FreeSpace, value);
        IF value <> OldSpaceThreshold THEN
            INIPut (fhini, '$SYS', 'SpaceThreshold', value);
        END (*IF*);

        (* Timeout value. *)

        WinQueryDlgItemCard (hwnd, DID.TimeOut, value);
        IF value > MaxTimeout THEN
            value := MaxTimeout;
            WinSetDlgItemCard (hwnd, DID.TimeOut, value);
            Sleep (1000);
        END (*IF*);
        IF value <> OldTimeout THEN
            INIPut (fhini, '$SYS', 'TimeOut', value);
        END (*IF*);

        (* Check the MultiINI checkbox and update HashMax.  This check  *)
        (* is probably redundant, because the dialogue procedure has    *)
        (* been updating HashMax as we've gone along, but it doesn't    *)
        (* hurt to have a final check.                                  *)

        IF OS2.LONGFROMMR (OS2.WinSendDlgItemMsg (hwnd, DID.MultiINI,
                                              OS2.BM_QUERYCHECK, NIL, NIL)) = 0 THEN
            value := 0;
        ELSE
            value := NewHashMax;
        END (*IF*);
        IF value <> HashMax THEN
            INIPut (fhini, 'fhini, $SYS', 'HashMax', value);
        END (*IF*);

        (* Disabling Telnet compatibility. *)

        TelnetDisable := OS2.LONGFROMMR (OS2.WinSendDlgItemMsg (hwnd, DID.TelnetDisable,
                                              OS2.BM_QUERYCHECK, NIL, NIL)) <> 0;
        IF TelnetDisable <> OldTelnetDisable THEN
            INIPut (fhini, '$SYS', 'TelnetDisable', TelnetDisable);
        END (*IF*);

        CloseINIFiles (fhini);

    END StoreData;

(**************************************************************************)

PROCEDURE ["SysCall"] DialogueProc (hwnd: OS2.HWND;  msg: OS2.ULONG;
                                      mp1, mp2: OS2.MPARAM): OS2.MRESULT;

    VAR id, code: OS2.USHORT;  NewMulti: BOOLEAN;
        ccode: CARDINAL;

    BEGIN
        IF msg = OS2.WM_INITDLG THEN

            OS2.WinSetWindowPos (hwnd, 0, 0, 0, 0, 0, OS2.SWP_MOVE);
            LoadValues (hwnd);
            RETURN NIL;

        ELSIF msg = OS2.WM_CONTROL THEN

            id := OS2.USHORT1FROMMP (mp1);
            code := OS2.USHORT2FROMMP (mp1);
            IF OS2.ULONGFROMMP(mp1) DIV 65536  = OS2.BN_CLICKED THEN
                CASE id OF
                  | DID.MultiINI:
                        IF NOT ChangeInProgress THEN
                            ChangeInProgress := TRUE;
                            NewMulti := OS2.LONGFROMMR (OS2.WinSendDlgItemMsg (hwnd, DID.MultiINI,
                                                              OS2.BM_QUERYCHECK, NIL, NIL)) <> 0;
                            IF NewMulti <> (HashMax <> 0)  THEN
                                IF OS2.WinMessageBox (OS2.HWND_DESKTOP, hwnd,
                                              "This is a slow operation. Are you sure?",
                                              "Warning", 0, OS2.MB_OKCANCEL) = OS2.MBID_OK THEN
                                    MoveUsers (hwnd, NewMulti);
                                    ChangeInProgress := FALSE;
                                    RETURN NIL;
                                ELSE
                                    NewMulti := NOT NewMulti;
                                    OS2.WinSendDlgItemMsg (hwnd, DID.MultiINI, OS2.BM_SETCHECK,
                                                           OS2.MPFROMSHORT(ORD(NewMulti)), NIL);
                                END (*IF*);
                            END (*IF*);
                            ChangeInProgress := FALSE;
                        END (*IF*);
                ELSE
                    RETURN OS2.WinDefDlgProc(hwnd, msg, mp1, mp2);
                END (*CASE*);

            ELSE
                RETURN OS2.WinDefDlgProc(hwnd, msg, mp1, mp2);
            END (*IF*);

        ELSIF msg = OS2.WM_CHAR THEN

            (* Explicitly ignore Esc key, because the default processing  *)
            (* was causing problems.                                      *)

            ccode := OS2.ULONGFROMMP(mp1);
            IF (IAND (ccode, OS2.KC_VIRTUALKEY) <> 0) THEN
                ccode := (OS2.ULONGFROMMP(mp2) DIV 65536) MOD 256;
                IF ccode = OS2.VK_ESC THEN
                    (* Say we've already handled it. *)
                    RETURN NIL;
                ELSE
                    RETURN OS2.WinDefDlgProc(hwnd, msg, mp1, mp2);
                END (*IF*);
            ELSE
                RETURN OS2.WinDefDlgProc(hwnd, msg, mp1, mp2);
            END (*IF*);

        ELSIF msg = OS2.WM_PRESPARAMCHANGED THEN

            IF NOT ChangeInProgress THEN
                ChangeInProgress := TRUE;
                CommonSettings.UpdateFontFrom (hwnd);
                ChangeInProgress := FALSE;
                RETURN NIL;
            END (*IF*);

        END (*IF*);

        (* Fall-through default for all cases not handled above. *)

        RETURN OS2.WinDefDlgProc (hwnd, msg, mp1, mp2);

    END DialogueProc;

(**************************************************************************)

PROCEDURE CreatePage (notebook: OS2.HWND;  VAR (*OUT*) pgID: CARDINAL): OS2.HWND;

    (* Creates page 1 and adds it to the notebook. *)

    VAR Label: ARRAY [0..31] OF CHAR;

    BEGIN
        notebookhandle := notebook;
        pagehandle := OS2.WinLoadDlg(notebook, notebook,
                       DialogueProc,    (* dialogue procedure *)
                       0,                   (* use resources in EXE *)
                       DID.BasicPage,                (* dialogue ID *)
                       NIL);                 (* creation parameters *)
        PageID := OS2.ULONGFROMMR (OS2.WinSendMsg (notebook, OS2.BKM_INSERTPAGE,
                         NIL, OS2.MPFROM2SHORT (OS2.BKA_MAJOR+OS2.BKA_AUTOPAGESIZE, OS2.BKA_LAST)));
        pgID := PageID;
        Label := "~Basic";
        OS2.WinSendMsg (notebook, OS2.BKM_SETTABTEXT,
                        CAST(ADDRESS,PageID), ADR(Label));
        OS2.WinSendMsg (notebook, OS2.BKM_SETPAGEWINDOWHWND,
                        CAST(ADDRESS,PageID), CAST(ADDRESS,pagehandle));
        RETURN pagehandle;
    END CreatePage;

(**************************************************************************)

PROCEDURE SetFont (VAR (*IN*) name: CommonSettings.FontName);

    (* Sets the font of the text on this page. *)

    CONST bufsize = CommonSettings.FontNameSize;

    BEGIN
        OS2.WinSetPresParam (pagehandle, OS2.PP_FONTNAMESIZE, bufsize, name);
    END SetFont;

(**************************************************************************)

BEGIN
    OldServerPort := 0;
    OldSFTPPort := 0;
    OldMaxUsers := 0;
    OldGuestLimit := 0;
    OldSpaceThreshold := 0;
    OldTimeout := 0;
    OldTelnetDisable := FALSE;
    ChangeInProgress := FALSE;
    GetEXEDirectory (OurDir);
END BasicPage.

