(**************************************************************************)
(*                                                                        *)
(*  Setup for SFtpServer                                                  *)
(*  Copyright (C) 2022   Peter Moylan                                     *)
(*                                                                        *)
(*  This program is free software: you can redistribute it and/or modify  *)
(*  it under the terms of the GNU General Public License as published by  *)
(*  the Free Software Foundation, either version 3 of the License, or     *)
(*  (at your option) any later version.                                   *)
(*                                                                        *)
(*  This program is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU General Public License for more details.                          *)
(*                                                                        *)
(*  You should have received a copy of the GNU General Public License     *)
(*  along with this program.  If not, see <http://www.gnu.org/licenses/>. *)
(*                                                                        *)
(*  To contact author:   http://www.pmoylan.org   peter@pmoylan.org       *)
(*                                                                        *)
(**************************************************************************)

IMPLEMENTATION MODULE InitialDialogue;

        (****************************************************************)
        (*                   PM Setup for SFtpServer                    *)
        (*                    Initial dialogue box                      *)
        (*                                                              *)
        (*    Last edited:    17 August 2022                            *)
        (*    Status:         Working                                   *)
        (*                                                              *)
        (****************************************************************)

IMPORT SYSTEM, OS2, OS2RTL, DID, FSUINI, RINIData, Remote, BigFrame, PMInit, Strings, CommonSettings;

FROM CommonSettings IMPORT
    (* proc *)  SetInitialLanguage, CurrentLanguage;

FROM Languages IMPORT
    (* type *)  LangHandle,
    (* proc *)  StrToBuffer;

FROM INIData IMPORT
    (* type *)  HINI,
    (* proc *)  OpenINIFile, CloseINIFile, INIGet, INIGetString, INIPut, INIPutString;

FROM FileOps IMPORT
    (* type *)  FilenameString,
    (* proc *)  GetCurrentDirectory;

FROM Timer IMPORT
    (* proc *)  Sleep;

FROM MiscFuncs IMPORT
    (* proc *)  EVAL;

FROM LowLevel IMPORT
    (* proc *) IANDB;

(************************************************************************)

CONST
    LangStringSize = 32;

VAR changed: BOOLEAN;
    UseTNI, TNIexplicit: BOOLEAN;
    SwitchData : OS2.SWCNTRL;     (* switch entry data *)
    SetupINIname, OurDir: FilenameString;

    server: SYSTEM.CARD8;
    FTPDir, SFTPDir: FilenameString;

(************************************************************************)
(*                    OPERATIONS ON DIALOGUE LABELS                     *)
(************************************************************************)

PROCEDURE SetLanguage (hwnd: OS2.HWND;  lang: LangHandle);

    (* Relabels this page in the new language. *)

    VAR stringval: ARRAY [0..511] OF CHAR;

    (********************************************************************)

    PROCEDURE SetOne (item: ARRAY OF CHAR;  place: CARDINAL);

        BEGIN
            StrToBuffer (lang, item, stringval);
            OS2.WinSetDlgItemText (hwnd, place, stringval);
        END SetOne;

    (********************************************************************)

    BEGIN

        StrToBuffer (lang, "Initial.Title", stringval);
        OS2.WinSetWindowText (hwnd, stringval);

        SetOne ("Initial.FTPBox", DID.FTPGroup);
        SetOne ("Initial.Edit", DID.FTPEdit);
        SetOne ("Initial.Remote", DID.FTPRemote);
        SetOne ("Initial.Dir", DID.FTPDirLabel);

        SetOne ("Initial.SFTPBox", DID.SFTPGroup);
        SetOne ("Initial.Edit", DID.SFTPEdit);
        SetOne ("Initial.Remote", DID.SFTPRemote);
        SetOne ("Initial.Dir", DID.SFTPDirLabel);

        SetOne ("Initial.language", DID.langlabel);
        SetOne ("Buttons.OK", DID.InitialOK);
        SetOne ("Buttons.Cancel", DID.InitialCancel);

    END SetLanguage;

(************************************************************************)

PROCEDURE EnableWindow (w: OS2.HWND;  enable: BOOLEAN);

    (* Enables or disables w. I hate to do it this way, but there is    *)
    (* a type incompatibility between C and Modula-2 Booleans.          *)
    (* (A C Boolean is a 32-bit type.)                                  *)

    BEGIN
        IF enable THEN
            OS2.WinEnableWindow (w, TRUE);
        ELSE
            OS2.WinEnableWindow (w, FALSE);
        END (*IF*);
    END EnableWindow;

(************************************************************************)

PROCEDURE EnableFields (hwnd: OS2.HWND;  SFTP, enable: BOOLEAN);

    BEGIN
        IF SFTP THEN
            EnableWindow (OS2.WinWindowFromID(hwnd, DID.SFTPRemote), enable);
            EnableWindow (OS2.WinWindowFromID(hwnd, DID.SFTPDirLabel), enable);
            EnableWindow (OS2.WinWindowFromID(hwnd, DID.SFTPDir), enable);
        ELSE
            EnableWindow (OS2.WinWindowFromID(hwnd, DID.FTPRemote), enable);
            EnableWindow (OS2.WinWindowFromID(hwnd, DID.FTPDirLabel), enable);
            EnableWindow (OS2.WinWindowFromID(hwnd, DID.FTPDir), enable);
        END (*IF*);
    END EnableFields;

(************************************************************************)
(*                     LOADING DATA FROM THE INI FILE                   *)
(************************************************************************)

PROCEDURE LoadValues (hwnd: OS2.HWND);

    (* Fills the dialogue elements on page 1 with data from the INI     *)
    (* file, or loads default values if they're not in the INI file.    *)
    (* Note that these are from the SETUP INI file, not the server      *)
    (* INI file.                                                        *)

    VAR hini: HINI;
        bool: BOOLEAN;
        app: ARRAY [0..7] OF CHAR;
        stringval: FilenameString;

    BEGIN
        changed := FALSE;
        hini := OpenINIFile (SetupINIname);
        app := "Setup";

        (* Language. *)

        IF NOT INIGetString (hini, app, 'Language', stringval) THEN
            changed := TRUE;
            stringval := "en";
        END (*IF*);
        OS2.WinSetDlgItemText (hwnd, DID.lang, stringval);

        (* Configure FTP/SFTP. The code is: *)
        (*      bit 0   enable FTP          *)
        (*      bit 1   FTP remote          *)
        (*      bit 2   enable SFTP         *)
        (*      bit 3   SFTP remote         *)

        IF NOT INIGet (hini, app, 'server', server) THEN
            server := 4;
        END (*IF*);
        bool := ODD (server);  server := server DIV 2;
        OS2.WinSendDlgItemMsg (hwnd, DID.FTPEdit, OS2.BM_SETCHECK,
                               OS2.MPFROMSHORT(ORD(bool)), NIL);
        EnableFields (hwnd, FALSE, bool);
        bool := ODD (server);  server := server DIV 2;
        OS2.WinSendDlgItemMsg (hwnd, DID.FTPRemote, OS2.BM_SETCHECK,
                               OS2.MPFROMSHORT(ORD(bool)), NIL);

        bool := ODD (server);  server := server DIV 2;
        OS2.WinSendDlgItemMsg (hwnd, DID.SFTPEdit, OS2.BM_SETCHECK,
                               OS2.MPFROMSHORT(ORD(bool)), NIL);
        EnableFields (hwnd, TRUE, bool);
        bool := ODD (server);  server := server DIV 2;
        OS2.WinSendDlgItemMsg (hwnd, DID.SFTPRemote, OS2.BM_SETCHECK,
                               OS2.MPFROMSHORT(ORD(bool)), NIL);

        (* Directories. *)

        IF NOT INIGetString (hini, app, 'FTPDir', FTPDir) THEN
            Strings.Assign (OurDir, FTPDir);
            changed := TRUE;
        END (*IF*);
        OS2.WinSetDlgItemText (hwnd, DID.FTPDir, FTPDir);

        IF NOT INIGetString (hini, app, 'SFTPDir', SFTPDir) THEN
            Strings.Assign (OurDir, SFTPDir);
            changed := TRUE;
        END (*IF*);
        OS2.WinSetDlgItemText (hwnd, DID.SFTPDir, SFTPDir);

        CloseINIFile (hini);

    END LoadValues;

(************************************************************************)

PROCEDURE DirectLoadValues;

    (* Like LoadValues, except that we don't fill in the dialogue fields. *)

    VAR hini: HINI;
        app: ARRAY [0..7] OF CHAR;

    BEGIN
        hini := OpenINIFile (SetupINIname);
        app := "Setup";

        (* The four-bit server code. *)

        IF NOT INIGet (hini, app, 'server', server) THEN
            server := 4;
        END (*IF*);

        (* Directories. *)

        IF NOT INIGetString (hini, app, 'FTPDir', FTPDir) THEN
            Strings.Assign (OurDir, FTPDir);
        END (*IF*);

        IF NOT INIGetString (hini, app, 'SFTPDir', SFTPDir) THEN
            Strings.Assign (OurDir, SFTPDir);
        END (*IF*);

        CloseINIFile (hini);

    END DirectLoadValues;

(************************************************************************)
(*                  STORING DATA BACK TO THE INI FILE                   *)
(************************************************************************)

PROCEDURE GetBool (hwnd: OS2.HWND;  ID: CARDINAL): BOOLEAN;

    BEGIN
        RETURN OS2.LONGFROMMR (OS2.WinSendDlgItemMsg (hwnd, ID,
                                    OS2.BM_QUERYCHECK, NIL, NIL)) <> 0;
    END GetBool;

(************************************************************************)

PROCEDURE MakeAbsPath (VAR (*INOUT*) name: FilenameString);

    (* Ensures that name is an absolute path, starting with a drive     *)
    (* letter.  This involves a guess in the case of remote files and   *)
    (* directories, but the changes we make if any in that case at      *)
    (* least serve to remind the user that an absolute path should      *)
    (* have been supplied. The result never ends with a slash, except   *)
    (* in the special case of the root directory of a volume.           *)

    VAR ourdisk, N: CARDINAL;
        ourdir: FilenameString;

    BEGIN
        IF name[1] <> ':' THEN

            (* A drive specification must be inserted. *)

            Strings.Insert ("C:", 0, name);
            OS2.DosQueryCurrentDisk (ourdisk, N);
            name[0] := CHR(ORD('A') + ourdisk - 1);
        END (*IF*);

        IF (name[2] <> '/') AND (name[2] <> '\') THEN

            (* We have a relative path.  Insert the name of the *)
            (* current working directory on the current drive.  *)

            N := SIZE (FilenameString);
            OS2.DosQueryCurrentDir (0, ourdir, N);
            Strings.Insert ('\', 0, ourdir);
            Strings.Append ('\', ourdir);
            Strings.Insert (ourdir, 2, name);

        END (*IF*);

        (* Remove a final slash if necessary. *)

        N := LENGTH(name);
        IF N > 3 THEN
            DEC (N);
            IF (name[N] = '/') OR (name[N] = '\') THEN
                name[N] := CHR(0);
            END (*IF*);
        END (*IF*);

    END MakeAbsPath;

(************************************************************************)

PROCEDURE StoreValues (hwnd: OS2.HWND);

    VAR hini: HINI;
        app: ARRAY [0..7] OF CHAR;
        stringval: FilenameString;

    BEGIN
        hini := OpenINIFile (SetupINIname);
        app := "Setup";

        (* Language. *)

        OS2.WinQueryDlgItemText (hwnd, DID.lang, 33, stringval);
        INIPutString (hini, app, 'Language', stringval);

        server := 0;
        IF GetBool (hwnd, DID.SFTPRemote) THEN INC(server) END (*IF*);
        server := 2*server;
        IF GetBool (hwnd, DID.SFTPEdit) THEN INC(server) END (*IF*);
        server := 2*server;
        IF GetBool (hwnd, DID.FTPRemote) THEN INC(server) END (*IF*);
        server := 2*server;
        IF GetBool (hwnd, DID.FTPEdit) THEN INC(server) END (*IF*);
        INIPut (hini, app, 'server', server);

        OS2.WinQueryDlgItemText (hwnd, DID.FTPDir, SIZE(FilenameString), FTPDir);
        MakeAbsPath (FTPDir);
        INIPutString (hini, app, 'FTPDir', FTPDir);
        OS2.WinQueryDlgItemText (hwnd, DID.SFTPDir, SIZE(FilenameString), SFTPDir);
        MakeAbsPath (SFTPDir);
        INIPutString (hini, app, 'SFTPDir', SFTPDir);

        CloseINIFile (hini);
        changed := FALSE;
    END StoreValues;

(************************************************************************)
(*               INFORMING THE SERVER THAT WE'RE FINISHED               *)
(************************************************************************)

PROCEDURE PostUpdated (semName: ARRAY OF CHAR);

    (* Posts on a public event semaphore. *)

    VAR changehev: OS2.HEV;

    BEGIN
        changehev := 0;
        IF OS2.DosOpenEventSem (semName, changehev) = OS2.ERROR_SEM_NOT_FOUND THEN
            OS2.DosCreateEventSem (semName, changehev, OS2.DC_SEM_SHARED, FALSE);
        END (*IF*);
        OS2.DosPostEventSem (changehev);
        OS2.DosCloseEventSem(changehev);
    END PostUpdated;

CONST UpdateSemName = "\SEM32\FTPSERVER\UPDATED";

(************************************************************************)
(*                      THE INITIAL DIALOGUE                            *)
(************************************************************************)

PROCEDURE ["SysCall"] InitialDialogueProc(hwnd     : OS2.HWND
                     ;msg      : OS2.ULONG
                     ;mp1, mp2 : OS2.MPARAM): OS2.MRESULT;

    VAR ButtonID, NotificationCode: CARDINAL;
        enabled, needremote: BOOLEAN;
        ourlang, NewLang: LangHandle;
        langname, message: ARRAY [0..63] OF CHAR;

    BEGIN
        CASE msg OF
            | OS2.WM_INITDLG:
                   OS2.WinSetDlgItemText (hwnd, DID.Status, "");
                   CurrentLanguage (ourlang, langname);
                   SetLanguage (hwnd, ourlang);
                   LoadValues (hwnd);
                   RETURN NIL;

           |  OS2.WM_CLOSE:
                   OS2.WinPostMsg(hwnd, OS2.WM_QUIT, NIL, NIL);

           |  OS2.WM_COMMAND:

                   CASE OS2.SHORT1FROMMP(mp1) OF

                     | DID.InitialOK:
                          StoreValues (hwnd);
                          IF IANDB(server, 5) = 0 THEN
                              (* Special case: no server enabled. *)

                              CurrentLanguage (ourlang, langname);
                              StrToBuffer (ourlang, "Initial.noserver", message);
                              OS2.WinMessageBox (OS2.HWND_DESKTOP, hwnd, message, NIL, 0, OS2.MB_ERROR + OS2.MB_OK);
                              OS2.WinPostMsg(hwnd, OS2.WM_QUIT, NIL, NIL);
                          END (*IF*);
                          needremote := (IANDB(server, 03H) = 03H) OR (IANDB(server, 0CH) = 0CH);
                          IF (NOT needremote) OR Remote.ConnectToServer (hwnd, DID.Status) THEN
                              FSUINI.SetDirs (server, FTPDir, SFTPDir, UseTNI, TNIexplicit);
                              CurrentLanguage (ourlang, langname);
                              StrToBuffer (ourlang, "Initial.loading", message);
                              OS2.WinSetDlgItemText (hwnd, DID.Status, message);
                              BigFrame.OpenBigFrame (hwnd);

                              (* The following code is executed after the BigFrame  *)
                              (* module has terminated.                             *)

                              IF needremote THEN
                                  EVAL(Remote.PostSemaphore (UpdateSemName));
                                  EVAL(Remote.ExecCommand ('Q'));
                              ELSE
                                  PostUpdated (UpdateSemName);
                              END (*IF*);
                              OS2.WinPostMsg(hwnd, OS2.WM_QUIT, NIL, NIL);
                          END (*IF*);

                     | DID.InitialCancel:
                          OS2.WinPostMsg(hwnd, OS2.WM_QUIT, NIL, NIL);

                   ELSE
                          RETURN OS2.WinDefDlgProc(hwnd, msg, mp1, mp2);
                   END (*CASE*);

           |  OS2.WM_CONTROL:
                   NotificationCode := OS2.ULONGFROMMP(mp1);
                   ButtonID := NotificationCode MOD 65536;
                   NotificationCode := NotificationCode DIV 65536;
                   IF NotificationCode = OS2.BN_CLICKED THEN
                       CASE ButtonID OF
                         | DID.FTPEdit:
                              enabled := GetBool (hwnd, ButtonID);
                              EnableFields (hwnd, FALSE, enabled);
                              changed := TRUE;
                         | DID.SFTPEdit:
                              enabled := GetBool (hwnd, ButtonID);
                              EnableFields (hwnd, TRUE, enabled);
                              changed := TRUE;
                         | DID.FTPRemote, DID.SFTPRemote:
                              IF GetBool (hwnd, ButtonID) THEN
                                  Remote.OpenSetupDialogue (hwnd);
                                  changed := TRUE;
                              END (*IF*);
                         ELSE
                              RETURN OS2.WinDefDlgProc(hwnd, msg, mp1, mp2);
                       END (*CASE*);

                   ELSIF NotificationCode = OS2.EN_CHANGE THEN

                       changed := TRUE;
                       IF ButtonID = DID.lang THEN
                           OS2.WinQueryDlgItemText (hwnd, DID.lang,
                                                    LangStringSize+1, langname);
                           IF CommonSettings.ChangeLanguageTo(langname) THEN
                               CommonSettings.CurrentLanguage (NewLang, langname);
                               SetLanguage (hwnd, NewLang);
                           END (*IF*);
                           RETURN NIL;
                       ELSE
                           RETURN OS2.WinDefDlgProc(hwnd, msg, mp1, mp2);
                       END (*IF*);
                   END (*IF*);

        ELSE    (* default must call WinDefWindowProc() *)
           RETURN OS2.WinDefDlgProc(hwnd, msg, mp1, mp2);
        END (*CASE*);

        RETURN NIL;

    END InitialDialogueProc;

(************************************************************************)

PROCEDURE CreateInitialDialogue (TNImode, explicit: BOOLEAN);

    (* Creates the main dialogue box.  *)

    (* TNImode specifies the INI/TNI mode that has already been chosen  *)
    (* for Setup.  It is this module's starting point for choosing      *)
    (* whether to use FTPD.INI or FTPD.TNI.  The variable explicit is   *)
    (* TRUE iff the Setup program has been called with a -I or -T       *)
    (* parameter.                                                       *)

    VAR hwnd: OS2.HWND;
        pid: OS2.PID;  tid: OS2.TID;

    BEGIN
        UseTNI := TNImode;  TNIexplicit := explicit;
        CommonSettings.GetINIname (SetupINIname);

        hwnd := OS2.WinLoadDlg(OS2.HWND_DESKTOP,    (* parent *)
                       OS2.HWND_DESKTOP,    (* owner *)
                       InitialDialogueProc, (* dialogue procedure *)
                       0,                   (* use resources in EXE *)
                       DID.FirstWindow,     (* dialogue ID *)
                       NIL);                (* creation parameters *)

        (* Put us on the visible task list. *)

        OS2.WinQueryWindowProcess (hwnd, pid, tid);
        SwitchData.hwnd := hwnd;
        WITH SwitchData DO
            hwndIcon      := 0;
            hprog         := 0;
            idProcess     := pid;
            idSession     := 0;
            uchVisibility := OS2.SWL_VISIBLE;
            fbJump        := OS2.SWL_JUMPABLE;
            szSwtitle     := "SFtpServer Setup";
            bProgType     := 0;
        END (*WITH*);
        OS2.WinCreateSwitchEntry (PMInit.OurHab(), SwitchData);

        CommonSettings.SetInitialWindowPosition (hwnd, "Initial");

        OS2.WinProcessDlg(hwnd);
        CommonSettings.StoreWindowPosition (hwnd, "Initial", TRUE);
        OS2.WinDestroyWindow (hwnd);

    END CreateInitialDialogue;

(**************************************************************************)
(*            THE CASE WHERE WE BYPASS THE INITIAL DIALOGUE               *)
(**************************************************************************)

PROCEDURE DirectProcess (TNImode, explicit: BOOLEAN);

    (* Moves directly to the main dialogue without going through the    *)
    (* initial dialogue.                                                *)

    VAR hwnd: OS2.HWND;
        ourlang: LangHandle;
        message: ARRAY [0..63] OF CHAR;
        needremote: BOOLEAN;

    BEGIN
        hwnd := OS2.NULLHANDLE;
        CommonSettings.GetINIname (SetupINIname);
        DirectLoadValues;

        IF IANDB(server, 5) = 0 THEN
            (* Special case: no server enabled. *)

            CurrentLanguage (ourlang, message);
            StrToBuffer (ourlang, "Initial.noserver", message);
            OS2.WinMessageBox (OS2.HWND_DESKTOP, OS2.HWND_DESKTOP, message, NIL, 0, OS2.MB_ERROR + OS2.MB_OK);
            OS2.WinPostMsg(hwnd, OS2.WM_QUIT, NIL, NIL);
        END (*IF*);

        needremote := (IANDB(server, 03H) = 03H) OR (IANDB(server, 0CH) = 0CH);
        IF (NOT needremote) OR Remote.ConnectToServer (hwnd, DID.Status) THEN

            FSUINI.SetDirs (server, FTPDir, SFTPDir, TNImode, explicit);
            BigFrame.OpenBigFrame (hwnd);

            (* The following code is executed after the BigFrame  *)
            (* module has terminated.                             *)

            IF needremote THEN
                EVAL(Remote.PostSemaphore (UpdateSemName));
                EVAL(Remote.ExecCommand ('Q'));
            ELSE
                PostUpdated (UpdateSemName);
            END (*IF*);
            OS2.WinPostMsg(hwnd, OS2.WM_QUIT, NIL, NIL);
        END (*IF*);
    END DirectProcess;

(**************************************************************************)

BEGIN
    GetCurrentDirectory (OurDir);
    SetInitialLanguage;
END InitialDialogue.

