(**************************************************************************)
(*                                                                        *)
(*  Support modules for SFTP Setup                                        *)
(*  Copyright (C) 2022   Peter Moylan                                     *)
(*                                                                        *)
(*  This program is free software: you can redistribute it and/or modify  *)
(*  it under the terms of the GNU General Public License as published by  *)
(*  the Free Software Foundation, either version 3 of the License, or     *)
(*  (at your option) any later version.                                   *)
(*                                                                        *)
(*  This program is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU General Public License for more details.                          *)
(*                                                                        *)
(*  You should have received a copy of the GNU General Public License     *)
(*  along with this program.  If not, see <http://www.gnu.org/licenses/>. *)
(*                                                                        *)
(*  To contact author:   http://www.pmoylan.org   peter@pmoylan.org       *)
(*                                                                        *)
(**************************************************************************)

IMPLEMENTATION MODULE RINIData;

        (************************************************************)
        (*                                                          *)
        (*                    INI File Operations                   *)
        (*                                                          *)
        (*     This module reads/writes user INI data either        *)
        (*     remotely (using INIServe) or locally.  The present   *)
        (*     version is intended for applications where           *)
        (*     up to two INI files can be open at any one time.     *)
        (*                                                          *)
        (*      WARNING: It contains a number of procedures that    *)
        (*      are not currently used, and which therefore might   *)
        (*      not yet have been updated to handle parallel        *)
        (*      INI files.                                          *)
        (*                                                          *)
        (*      Started:        13 January 2002                     *)
        (*      Last edited:    22 September 2022                   *)
        (*      Status:         OK                                  *)
        (*                                                          *)
        (************************************************************)


IMPORT OS2, Strings, FileSys, FileOps, INIData, Remote;

FROM SYSTEM IMPORT
    (* type *)  LOC, INT16;

FROM Storage IMPORT ALLOCATE, DEALLOCATE;

FROM FileOps IMPORT
    (* type *)  FilenameString;

FROM Remote IMPORT
    (* proc *)  SelectRemoteFile,
                SendCommand, ExecCommand, ExecCommand2, ReceiveLine,
                StartDirectoryListing, NextDirectoryEntry, FinishDirectoryListing;

FROM WildCard IMPORT
    (* proc *)  WildMatch;

FROM MiscFuncs IMPORT
    (* type *)  CharArrayPointer,
    (* proc *)  EVAL, HexEncodeArray, DecodeHex, DecodeHexString;

(********************************************************************************)

CONST Nul = CHR(0);

TYPE
    StringReadState = POINTER TO StateRecord;
    StateRecord = RECORD
                      pos, BufferSize: CARDINAL;
                      bufptr: CharArrayPointer;
                  END (*RECORD*);

    RHINI = POINTER TO
                RECORD
                    name: FilenameString;
                    conflict: BOOLEAN;
                    CASE remote: BOOLEAN OF
                        FALSE:
                            hini: INIData.HINI;
                      | TRUE:
                    END (*CASE*);
                END (*RECORD*);

(********************************************************************************)

(*
PROCEDURE DebugDump (rhini: RHINI);

    (* Writes the state of rhini to the Debug file *)

    BEGIN
        IF rhini = NIL THEN
            Debug.WriteString ("NIL");
        ELSE
            Debug.WriteString (rhini^.name);
            IF rhini^.remote THEN Debug.WriteString (" R") END (*IF*);
            IF rhini^.conflict THEN Debug.WriteString (" C") END (*IF*);
        END (*IF*);
        Debug.WriteLn;
    END DebugDump;
*)

(************************************************************************)
(*          DECIDING WHETHER TO SET INI OR TNI AS DEFAULT CHOICE        *)
(************************************************************************)

PROCEDURE FetchRemote (app, key: ARRAY OF CHAR;
                        VAR (*IN*) variable: ARRAY OF LOC): BOOLEAN;

    (* A specialised version of INIFetch (see below) for the remote case. *)

    VAR N: CARDINAL;
        bufptr: CharArrayPointer;
        buffer: ARRAY [0..31] OF CHAR;

    BEGIN
        EVAL (ExecCommand2 ("A", app));
        EVAL (ExecCommand2 ("K", key));
        EVAL (SendCommand ("S"));
        IF ReceiveLine (buffer) THEN
            IF buffer[0] = '+' THEN
                Strings.Delete (buffer, 0, 1);
                N := DecodeHex (buffer);
            ELSE
                RETURN FALSE;
            END (*IF*);
        ELSE
            RETURN FALSE;
        END (*IF*);
        IF N <> HIGH(variable) + 1 THEN
            RETURN FALSE;
        END (*IF*);
        ALLOCATE (bufptr, 2*N+3);
        IF ExecCommand2 ("A", app) AND ExecCommand2 ("K", key)
           AND SendCommand ("V") AND ReceiveLine (bufptr^)
              AND (bufptr^[0] = '+') THEN
            Strings.Delete (bufptr^, 0, 1);
            DecodeHexString (bufptr^, variable);
            DEALLOCATE (bufptr, 2*N+3);
            RETURN TRUE;
        ELSE
            DEALLOCATE (bufptr, 2*N+3);
            RETURN FALSE;
        END (*IF*);
    END FetchRemote;

(************************************************************************)

PROCEDURE PutRemote (app, key: ARRAY OF CHAR;  VAR (*IN*) variable: ARRAY OF LOC);

    (* A specialised version of INIPut (see below) for the remote case. *)

    VAR N: CARDINAL;
        bufptr: CharArrayPointer;

    BEGIN
        N := HIGH(variable)+1;
        EVAL (ExecCommand2 ("A", app));
        EVAL (ExecCommand2 ("K", key));
        ALLOCATE (bufptr, 2*N+1);
        HexEncodeArray (variable, N, bufptr^);
        EVAL (ExecCommand2 ("W", bufptr^));
        DEALLOCATE (bufptr, 2*N+1);
    END PutRemote;

(************************************************************************)
(*                          SOME DEBUG STUFF                            *)
(************************************************************************)

(*
CONST tracename = "TRACE.TXT";

VAR trace: FileOps.ChanId;

PROCEDURE OpenTraceFile;

    BEGIN
        trace := FileOps.OpenGeneral (tracename, FALSE, TRUE, TRUE, TRUE, FALSE);
    END OpenTraceFile;

PROCEDURE CloseTraceFile;

    BEGIN
        FileOps.CloseFile (trace);
    END CloseTraceFile;

PROCEDURE TraceMes (txt1, txt2: ARRAY OF CHAR);

    BEGIN
        FileOps.FWriteString (trace, txt1);
        FileOps.FWriteString (trace, txt2);
        FileOps.FWriteLn (trace);
    END TraceMes;

PROCEDURE SayCurrent;

    VAR dir: FilenameString;

    BEGIN
        FileOps.GetCurrentDirectory (dir);
        TraceMes ("Current directory is ", dir);
    END SayCurrent;
*)

(************************************************************************)
(*                      THE MAIN INI/TNI DECISION                       *)
(************************************************************************)

PROCEDURE DecideTNImode (dir, appname: ARRAY OF CHAR;
                                TNIexplicit, remote: BOOLEAN;
                                  VAR (*INOUT*) useTNI: BOOLEAN);

    (* Decides whether to use appname.INI or appname.TNI. *)

    VAR useINI, useTNI0, foundI, foundT, decided: BOOLEAN;
        app: ARRAY [0..5] OF CHAR;
        N, ourdisk, targetdisk, dummy: CARDINAL;
        saveddir, basename, name: FilenameString;
        Iname, Tname: FilenameString;

    BEGIN
        decided := FALSE;
        IF remote THEN

            Strings.Assign (dir, basename);
            IF dir[0] <> Nul THEN
                Strings.Append ('\', basename);
            END (*IF*);
            Strings.Append (appname, basename);

            IF TNIexplicit THEN

                (* Store the useTNI value in whichever of the two files exist,  *)
                (* or both if both exist.  If neither exists, create only one.  *)

                app := "$SYS";
                Strings.Assign (basename, name);
                Strings.Append (".TNI", name);
                IF SelectRemoteFile(name) OR useTNI THEN
                    PutRemote (app, "UseTNI", useTNI);
                END (*IF*);

                Strings.Assign (basename, name);
                Strings.Append (".INI", name);
                IF SelectRemoteFile(name) OR NOT useTNI THEN
                    PutRemote (app, "UseTNI", useTNI);
                END (*IF*);
                decided := TRUE;
            ELSE
                (* NOT TNIexplicit, so more tests are needed. *)

                Strings.Assign (basename, Iname);
                Strings.Assign (basename, Tname);
                Strings.Append (".INI", Iname);
                Strings.Append (".TNI", Tname);
                useTNI := SelectRemoteFile(Tname);
                useINI := SelectRemoteFile(Iname);

                (* If only one of the files exists, the decision is obvious.    *)
                (* If neither exists, default to using INI.                     *)

                IF NOT (useINI AND useTNI) THEN
                    decided  := TRUE;
                END (*IF*);

                IF NOT decided THEN

                    (* Both files exist.  In this case                      *)
                    (* we look up the entry ($SYS, useTNI) in each file.    *)

                    useTNI := FALSE;  useTNI0 := FALSE;
                    app := "$SYS";
                    foundI := FetchRemote (app, "UseTNI", useTNI0);

                    useTNI := SelectRemoteFile(Tname);
                    foundT := FetchRemote (app, "UseTNI", useTNI);

                    (* If both entries missing, default to using INI.  *)

                    IF NOT (foundI OR foundT) THEN
                        decided  := TRUE;
                    ELSE

                        (* If only one entry exists, use it. *)

                        IF foundI <> foundT THEN
                            IF foundI THEN
                                useTNI := useTNI0;
                            END (*IF*);
                            decided  := TRUE;
                        ELSE
                            (* Both entries exist. *)
                            decided  := useTNI0 = useTNI;
                        END (*IF*);
                    END (*IF*);

                END (*IF NOT decided*);

            END (*IF TNIexplicit*);

        ELSE

            (* The local, non-remote case.  We have to change       *)
            (* directory to use the INIData functions, which makes  *)
            (* me think that INIData needs a new function.          *)

            (* TraceMes ("In RINIData.DecideTNImode, dir = ", dir); SayCurrent; *)

            N := SIZE (FilenameString);
            OS2.DosQueryCurrentDisk (ourdisk, dummy);
            IF dir[1] = ':' THEN
                targetdisk := ORD(CAP(dir[0])) - ORD('A') + 1;
            ELSE
                targetdisk := ourdisk;
            END (*IF*);
            OS2.DosQueryCurrentDir(0, saveddir, N);
            Strings.Insert ('\', 0, saveddir);
            IF targetdisk <> ourdisk THEN
                OS2.DosSetDefaultDisk (targetdisk);
            END (*IF*);
            OS2.DosSetCurrentDir (dir);
            (*TraceMes ("saveddir is ", saveddir); SayCurrent; *)

            IF TNIexplicit THEN
                INIData.CommitTNIDecision (appname, useTNI);
                decided := TRUE;
            ELSE
                decided := INIData.ChooseDefaultINI (appname, useTNI);
            END;

            IF targetdisk <> ourdisk THEN
                OS2.DosSetDefaultDisk (ourdisk);
            END (*IF*);
            OS2.DosSetCurrentDir (saveddir);

            (*
            TraceMes ("Have now restored old directory from ", saveddir);
            SayCurrent;
            TraceMes ("End of traced section", "");
            *)

        END (*IF remote*);

        IF NOT decided THEN
            useTNI := FALSE;
        END (*IF*);
    END DecideTNImode;

(************************************************************************)
(*                  OPENING AND CLOSING THE INI FILE                    *)
(************************************************************************)

PROCEDURE UpdateOpenCount (rhini: RHINI;  amount: INTEGER);

    (* For consistency checking.  Updates a count in the file to see    *)
    (* whether we have matched opens and closes.                        *)

    VAR count: CARDINAL;  hini: INIData.HINI;
        app, key: ARRAY [0..15] OF CHAR;

    BEGIN
        IF NOT rhini^.remote THEN
            app := '$SYS';
            key := 'OpenCount';
            hini := rhini^.hini;
            IF NOT INIData.INIGet (hini, app, key, count) THEN
                count := 0;
            END (*IF*);
            IF amount > 0 THEN
                IF count < MAX(CARDINAL) THEN
                    INC (count, amount);
                END (*IF*);
            ELSE
                amount := ABS(amount);
                IF count >= VAL(CARDINAL,amount) THEN
                    DEC (count, amount);
                ELSE
                    count := MAX (CARDINAL);
                END (*IF*);
            END (*IF*);
            INIData.INIPut (hini, app, key, count);
        END (*IF*);
    END UpdateOpenCount;

(************************************************************************)

PROCEDURE OpenINIFile (filename: ARRAY OF CHAR;
                                    remote, conflict: BOOLEAN): RHINI;

    (* Opens the INI or TNI file.  The conflict parameter should be     *)
    (* TRUE if two remote files will be in use at the same time.        *)

    VAR rhini: RHINI;
        success: BOOLEAN;

    BEGIN
        NEW (rhini);
        Strings.Assign (filename, rhini^.name);
        rhini^.conflict := conflict;
        rhini^.remote := remote;
        IF remote THEN
            success := SelectRemoteFile (filename);
        ELSE
            rhini^.hini := INIData.OpenINIFile (filename);
            IF NOT INIData.INIValid (rhini^.hini) THEN
                rhini^.hini := INIData.CreateINIFile (filename);
            END (*IF*);
            success := INIData.INIValid (rhini^.hini);
        END (*IF*);
        IF success THEN
            UpdateOpenCount (rhini, +1);
        END (*IF*);
        IF NOT success THEN
            DISPOSE (rhini);
        END (*IF*);
        RETURN rhini;
    END OpenINIFile;

(********************************************************************************)

PROCEDURE CloseINIFile (VAR (*INOUT*) rhini: RHINI);

    (* Closes the file. *)

    BEGIN
        IF rhini <> NIL THEN
            UpdateOpenCount (rhini, -1);
            IF NOT rhini^.remote THEN
                INIData.CloseINIFile (rhini^.hini);
            END (*IF*);
            DISPOSE (rhini);
        END (*IF*);
    END CloseINIFile;

(********************************************************************************)
(*                               WRITING DATA                                   *)
(********************************************************************************)

PROCEDURE PutBinaryString (rhini: RHINI;  app, key: ARRAY OF CHAR;
                            VAR (*IN*) variable: ARRAY OF LOC;  N: CARDINAL);

    (* Writes N bytes of data to the INI file. *)

    VAR bufptr: CharArrayPointer;

    BEGIN
        IF rhini^.remote THEN
            IF rhini^.conflict THEN
                EVAL (SelectRemoteFile (rhini^.name));
            END (*IF*);
            EVAL (ExecCommand2 ("A", app));
            EVAL (ExecCommand2 ("K", key));
            ALLOCATE (bufptr, 2*N+1);
            HexEncodeArray (variable, N, bufptr^);
            EVAL (ExecCommand2 ("W", bufptr^));
            DEALLOCATE (bufptr, 2*N+1);
        ELSIF INIData.INIValid(rhini^.hini) THEN
            INIData.INIPutBinary (rhini^.hini, app, key, variable, N);
        END (*IF*);
    END PutBinaryString;

(********************************************************************************)

PROCEDURE INIPut (rhini: RHINI;  app, key: ARRAY OF CHAR;
                                        VAR (*IN*) value: ARRAY OF LOC);

    (* Writes data to the INI file. *)

    BEGIN
        PutBinaryString (rhini, app, key, value, HIGH(value)+1);
    END INIPut;

(********************************************************************************)

PROCEDURE INIPutBinary (rhini: RHINI;  app, key: ARRAY OF CHAR;
                    VAR (*IN*) variable: ARRAY OF LOC;  NumberOfBytes: CARDINAL);

    (* Writes a binary string of the specified length to the INI file. *)

    BEGIN
        PutBinaryString (rhini, app, key, variable, NumberOfBytes);
    END INIPutBinary;

(********************************************************************************)

PROCEDURE INIPutString (rhini: RHINI;  app, key: ARRAY OF CHAR;
                           VAR (*IN*) variable: ARRAY OF CHAR);

    (* Writes a character string to the INI file. *)

    BEGIN
        PutBinaryString (rhini, app, key, variable, LENGTH(variable));
    END INIPutString;

(********************************************************************************)
(*                               READING DATA                                   *)
(********************************************************************************)

PROCEDURE ItemSize (rhini: RHINI;  app, key: ARRAY OF CHAR;
                                        VAR (*OUT*) size: CARDINAL): BOOLEAN;

    (* Returns the number of bytes in the specified INI item.  If the item does *)
    (* not exist, returns FALSE with size set to zero.                          *)

    VAR success: BOOLEAN;
        buffer: ARRAY [0..31] OF CHAR;

    BEGIN
        success := FALSE;
        IF rhini^.remote THEN
            EVAL (ExecCommand2 ("A", app));
            EVAL (ExecCommand2 ("K", key));
            EVAL (SendCommand ("S"));
            IF ReceiveLine (buffer) THEN
                IF buffer[0] = '+' THEN
                    Strings.Delete (buffer, 0, 1);
                    size := DecodeHex (buffer);
                    success := TRUE;
                END (*IF*);
            END (*IF*);
        ELSE
            success := INIData.INIValid(rhini^.hini)
                 AND INIData.ItemSize (rhini^.hini, app, key, size);
        END (*IF*);
        IF NOT success THEN
            size := 0;
        END (*IF*);
        RETURN success;
    END ItemSize;

(********************************************************************************)

PROCEDURE INIFetchBinary (rhini: RHINI;  app, key: ARRAY OF CHAR;
                          VAR (*OUT*) result: ARRAY OF LOC;
                                            size: CARDINAL): BOOLEAN;

    (* Retrieves the value of a variable from the INI file.  Returns FALSE if   *)
    (* the variable was not found.  This the version in which we trust the      *)
    (* caller to have ensured that the size is correct.                         *)

    VAR bufptr: CharArrayPointer;

    BEGIN
        IF rhini^.remote THEN
            ALLOCATE (bufptr, 2*size+3);
            IF ExecCommand2 ("A", app) AND ExecCommand2 ("K", key)
               AND SendCommand ("V") AND ReceiveLine (bufptr^)
                  AND (bufptr^[0] = '+') THEN
                Strings.Delete (bufptr^, 0, 1);
                DecodeHexString (bufptr^, result);
                DEALLOCATE (bufptr, 2*size+3);
                RETURN TRUE;
            ELSE
                DEALLOCATE (bufptr, 2*size+3);
                RETURN FALSE;
            END (*IF*);
        ELSIF NOT INIData.INIValid(rhini^.hini) THEN
            RETURN FALSE;
        ELSE
            RETURN INIData.INIGetTrusted (rhini^.hini, app, key, result, size);
        END (*IF*);
    END INIFetchBinary;

(********************************************************************************)

PROCEDURE INIFetch (rhini: RHINI;  app, key: ARRAY OF CHAR;
                    VAR (*OUT*) variable: ARRAY OF LOC): BOOLEAN;

    (* Reads data from the INI file.  A size mismatch will cause failure.  *)

    VAR N: CARDINAL;

    BEGIN
        IF NOT ItemSize (rhini, app, key, N) THEN
            RETURN FALSE;
        END (*IF*);
        IF N <> HIGH(variable) + 1 THEN
            RETURN FALSE;
        END (*IF*);
        RETURN INIFetchBinary (rhini, app, key, variable, N);
    END INIFetch;

(********************************************************************************)

PROCEDURE INIGetCard (rhini: RHINI;  app, key: ARRAY OF CHAR;
                          VAR (*OUT*) val: CARDINAL): BOOLEAN;

    (* Gets a single cardinal value from the INI file, returns true if OK. *)

    BEGIN
        RETURN INIFetch (rhini, app, key, val);
    END INIGetCard;

(************************************************************************)

PROCEDURE INIGetString (rhini: RHINI;  app, key: ARRAY OF CHAR;
                          VAR (*OUT*) val: ARRAY OF CHAR): BOOLEAN;

    (* Gets a string value from the INI file, returns true if OK. *)

    VAR size: CARDINAL;

    BEGIN
        IF NOT ItemSize (rhini, app, key, size) THEN
            RETURN FALSE;
        END (*IF*);
        IF size > HIGH(val)+1 THEN
            RETURN FALSE;
        END (*IF*);
        IF INIFetchBinary (rhini, app, key, val, size) THEN
            IF size <= HIGH(val) THEN
                val[size] := Nul;
            END (*IF*);
            RETURN TRUE;
        ELSE
            RETURN FALSE;
        END (*IF*);
    END INIGetString;

(************************************************************************)
(*              READING A STRING OF STRINGS FROM AN INI FILE            *)
(************************************************************************)

PROCEDURE GetStringList (rhini: RHINI;  app, key: ARRAY OF CHAR;
                                   VAR (*OUT*) state: StringReadState);

    (* Initialisation in preparation for a "NextString" operation. *)

    BEGIN
        NEW (state);

        (* Pick up the list of all list names. *)

        state^.bufptr := NIL;
        IF NOT ItemSize (rhini, app, key, state^.BufferSize) THEN
            state^.BufferSize := 0;
        END (*IF*);
        IF state^.BufferSize > 0 THEN
            ALLOCATE (state^.bufptr, state^.BufferSize);
            IF NOT INIFetchBinary (rhini, app, key, state^.bufptr^,
                                       state^.BufferSize) THEN
                state^.bufptr^[0] := Nul;
            END (*IF*);
        END (*IF*);
        state^.pos := 0;

    END GetStringList;

(************************************************************************)

PROCEDURE NextString (state: StringReadState;  VAR (*OUT*) result: ARRAY OF CHAR);

    (* Reads the next character string from a string-of-strings field.  *)
    (* An empty string is returned when we have run out of strings.     *)

    VAR k: CARDINAL;

    BEGIN
        WITH state^ DO
            IF pos >= BufferSize THEN
                result[0] := Nul;
            ELSE
                k := 0;
                REPEAT
                    result[k] := bufptr^[pos];
                    INC (k);  INC (pos);
                UNTIL (pos >= BufferSize) OR (bufptr^[pos-1] = Nul)
                                        OR (k > HIGH(result));
                IF k <= HIGH(result) THEN
                    result[k] := Nul;
                END (*IF*);
            END (*IF*);
        END (*WITH*);
    END NextString;

(************************************************************************)

PROCEDURE CloseStringList (VAR (*INOUT*) state: StringReadState);

    (* Must be called to release the memory used in fetching a  *)
    (* string of strings.                                       *)

    BEGIN
        IF state <> NIL THEN
            IF state^.BufferSize > 0 THEN
                DEALLOCATE (state^.bufptr, state^.BufferSize);
            END (*IF*);
            DEALLOCATE (state, SIZE(StateRecord));
        END (*IF*);
    END CloseStringList;

(************************************************************************)
(*                      COPYING, DELETING, RENAMING                     *)
(************************************************************************)

PROCEDURE AppExists (rhini: RHINI;  app: ARRAY OF CHAR): BOOLEAN;

    (* Returns TRUE iff this application exists. *)

    VAR size: CARDINAL;

    BEGIN
        RETURN ItemSize (rhini, app, '', size) AND (size > 0);
    END AppExists;

(************************************************************************)

PROCEDURE INIDeleteApp (rhini: RHINI;  app: ARRAY OF CHAR);

    (* Deletes an application from the INI file. *)

    BEGIN
        IF rhini^.remote THEN
            IF rhini^.conflict THEN
                EVAL (SelectRemoteFile (rhini^.name));
            END (*IF*);
            EVAL (ExecCommand2 ("A", app) AND ExecCommand ("K")
                    AND ExecCommand ("D"));
        ELSIF INIData.INIValid(rhini^.hini) THEN
            INIData.INIDeleteApp (rhini^.hini, app);
        END (*IF*);
    END INIDeleteApp;

(************************************************************************)

PROCEDURE INIDeleteKey (rhini: RHINI;  app, key: ARRAY OF CHAR);

    (* Deletes a key from the INI file. *)

    VAR hini: INIData.HINI;

    BEGIN
        IF rhini^.remote THEN
            IF rhini^.conflict THEN
                EVAL (SelectRemoteFile (rhini^.name));
            END (*IF*);
            EVAL (ExecCommand2 ("A", app) AND ExecCommand2 ("K", key)
                    AND ExecCommand ("D"));
        ELSE
            hini := rhini^.hini;
            IF INIData.INIValid(hini) THEN
                INIData.INIDeleteKey (hini, app, key);
            END (*IF*);
        END (*IF*);
    END INIDeleteKey;

(************************************************************************)
(*                          FILE OPERATIONS                             *)
(************************************************************************)

(*
PROCEDURE AppendCard (N: CARDINAL;  VAR (*INOUT*) str: ARRAY OF CHAR);

    (* Needed temporarily while debugging. *)

    VAR pos: CARDINAL;

    BEGIN
        IF N > 9 THEN
            AppendCard (N DIV 10, str);
            N := N MOD 10;
        END (*IF*);
        pos := LENGTH(str);
        str[pos] := CHR(ORD('0') + N);
        str[pos+1] := CHR(0);
    END AppendCard;
*)

(************************************************************************)

PROCEDURE DeleteFile (filename: ARRAY OF CHAR;  remote: BOOLEAN;  status: OS2.HWND);

    (* Deletes a file on either the local or remote machine.  *)

    VAR rc: CARDINAL;
        (*
        debug: BOOLEAN;
        message: ARRAY [0..127] OF CHAR;
        *)

    BEGIN
        (*
        debug := TRUE;
        IF debug THEN
            message := "Deleting file ";
            Strings.Append (filename, message);
            OS2.WinSetWindowText (status, message);
        END (*IF*);
        *)
        IF remote THEN
            EVAL (ExecCommand2 ("X", filename));
        ELSE
            rc := OS2.DosDelete (filename);
            (*
            IF debug THEN
                message := "Delete status =";
                AppendCard (rc, message);
                OS2.WinSetWindowText (status, message);
            END;
            *)
        END (*IF*);
    END DeleteFile;

(************************************************************************)
(*                          DIRECTORY SEARCH                            *)
(************************************************************************)

TYPE
    NameList = POINTER TO NameRecord;

    NameRecord  =   RECORD
                        next: NameList;
                        this: FilenameString;
                    END (*RECORD*);

    SearchState = POINTER TO SearchRecord;

    SearchRecord =  RECORD
                        remote: BOOLEAN;
                        D: FileOps.DirectoryEntry;
                        list: NameList;
                    END (*RECORD*);

(************************************************************************)

PROCEDURE GetRemoteListing (mask: ARRAY OF CHAR;  Subdirectory: BOOLEAN): NameList;

    (* Gets a list of all files or subdirectories matching the criteria *)
    (* given in FirstDirEntry (see below).  If the mask contains        *)
    (* wildcard characters '*' and/or '?', these may only be in the     *)
    (* "file" part of the specification, not in the "directory" part.   *)
    (* That is, the search may not cross multiple directories.          *)

    VAR slashpos, pos2, L: CARDINAL;
        found, found2: BOOLEAN;
        dir, filename: FilenameString;
        result, tail: NameList;

    BEGIN
        (* Special case: empty mask means empty result. *)

        IF mask[0] = Nul THEN
            RETURN NIL;
        END (*IF*);

        (* Let slashpos be the position of the LAST slash. *)

        L := Strings.Length (mask) - 1;
        Strings.FindPrev ('/', mask, L, found, slashpos);
        Strings.FindPrev ('\', mask, L, found2, pos2);
        IF found THEN
            IF found2 AND (pos2 > slashpos) THEN
                slashpos := pos2;
            END (*IF*);
        ELSE
            found := found2;  slashpos := pos2;
        END (*IF*);

        (* Let dir be the part before the slash, and mask be the part after. *)

        IF found THEN
            Strings.Assign (mask, dir);
            dir[slashpos+1] := Nul;
            Strings.Delete (mask, 0, slashpos+1);
        ELSE
            dir[0] := Nul;
        END (*IF*);

        (* Ask IniServe for the directory listing. *)

        result := NIL;  tail := NIL;
        found := StartDirectoryListing (dir);
        WHILE found DO
            found := NextDirectoryEntry (filename);
            IF found THEN

                (* We still have to distinguish between directory   *)
                (* and non-directory files.                         *)

                L := Strings.Length (filename) - 1;
                found2 := filename[L] = '/';
                IF found2 THEN
                    filename[L] := Nul;
                END (*IF*);
                IF (found2 = Subdirectory) AND WildMatch (filename, mask) THEN

                    (* Finally, a name that meets all the criteria. *)

                    IF tail = NIL THEN
                        NEW (result);
                        tail := result;
                    ELSE
                        NEW (tail^.next);
                        tail := tail^.next;
                    END (*IF*);
                    tail^.next := NIL;
                    tail^.this := filename;
                END (*IF*);
            END (*IF*);
        END (*WHILE*);
        FinishDirectoryListing;
        RETURN result;

    END GetRemoteListing;

(************************************************************************)

PROCEDURE TakeFromList (VAR (*INOUT*) list: NameList;
                                VAR (*OUT*) name: ARRAY OF CHAR): BOOLEAN;

    (* Removes the first element of a NameList. *)

    VAR p: NameList;

    BEGIN
        p := list;
        IF p = NIL THEN
            name[0] := Nul;
            RETURN FALSE;
        ELSE
            Strings.Assign (p^.this, name);
            list := p^.next;
            DISPOSE (p);
            RETURN TRUE;
        END (*IF*);
    END TakeFromList;

(************************************************************************)

PROCEDURE FirstDirEntry (mask: ARRAY OF CHAR;  remote: BOOLEAN;
                            Either, Subdirectory, AllowHidden: BOOLEAN;
                                VAR (*OUT*) S: SearchState;
                                VAR (*OUT*) name: ARRAY OF CHAR): BOOLEAN;

    (* Gets the first directory entry satisfying the conditions:        *)
    (*  (a) if Subdirectory is FALSE, we want the first non-directory   *)
    (*      entry that matches "mask".                                  *)
    (*  (b) if Subdirectory is TRUE, we want the first directory that   *)
    (*      matches "mask".                                             *)
    (* If Either is TRUE then both directories and files are allowed,   *)
    (* and Subdirectory is ignored.  In all cases "mask" is a filename  *)
    (* specification that may include wildcards.  Hidden files are      *)
    (* included in the search iff AllowHidden is TRUE.  (In the remote  *)
    (* case, the AllowHidden parameter is ignored.)                     *)

    (* Restriction: if the mask contains wildcard characters '*' and/or *)
    (* '?', these may only be in the "file" part of the specification,  *)
    (* not in the "directory" part.  That is, the search may not cross  *)
    (* multiple directories.  Typically you would use a mask of the     *)
    (* form dirname/*, where dirname is the name of a directory in      *)
    (* either fully qualified form, or relative to the current directory. *)

    VAR found: BOOLEAN;

    BEGIN
        NEW (S);
        S^.remote := remote;
        IF remote THEN
            S^.list := GetRemoteListing (mask, Subdirectory);
            found := TakeFromList (S^.list, name);
        ELSE
            S^.list := NIL;
            found := FileOps.FirstDirEntry (mask, Either, Subdirectory,
                                                AllowHidden, S^.D);
            IF found THEN
                Strings.Assign (S^.D.name, name);
            END (*IF*);
        END (*IF*);
        RETURN found;
    END FirstDirEntry;

(************************************************************************)

PROCEDURE NextDirEntry (S: SearchState;
                                VAR (*OUT*) name: ARRAY OF CHAR): BOOLEAN;

    (* Read the next directory entry satisfying the search criteria     *)
    (* specified by the FirstDirEntry call.                             *)

    VAR found: BOOLEAN;

    BEGIN
        IF S = NIL THEN
            found := FALSE;
        ELSIF S^.remote THEN
            found := TakeFromList (S^.list, name);
        ELSE
            found := FileOps.NextDirEntry (S^.D);
            IF found THEN
                Strings.Assign (S^.D.name, name);
            END (*IF*);
        END (*IF*);
        RETURN found;
    END NextDirEntry;

(************************************************************************)

PROCEDURE DirSearchDone (VAR (*INOUT*) S: SearchState);

    (* Terminates the search. *)

    VAR p: NameList;

    BEGIN
        IF S <> NIL THEN
            IF S^.remote THEN
                p := S^.list;
                WHILE p <> NIL DO
                    S^.list := p^.next;
                    DISPOSE (p);
                END (*WHILE*);
            ELSE
                FileOps.DirSearchDone(S^.D);
            END (*IF*);
            DISPOSE (S);
        END (*IF*);
    END DirSearchDone;

(************************************************************************)

END RINIData.

