(**************************************************************************)
(*                                                                        *)
(*  Setup for SFtpServer                                                  *)
(*  Copyright (C) 2022   Peter Moylan                                     *)
(*                                                                        *)
(*  This program is free software: you can redistribute it and/or modify  *)
(*  it under the terms of the GNU General Public License as published by  *)
(*  the Free Software Foundation, either version 3 of the License, or     *)
(*  (at your option) any later version.                                   *)
(*                                                                        *)
(*  This program is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU General Public License for more details.                          *)
(*                                                                        *)
(*  You should have received a copy of the GNU General Public License     *)
(*  along with this program.  If not, see <http://www.gnu.org/licenses/>. *)
(*                                                                        *)
(*  To contact author:   http://www.pmoylan.org   peter@pmoylan.org       *)
(*                                                                        *)
(**************************************************************************)

IMPLEMENTATION MODULE SecurityS;

        (****************************************************************)
        (*                                                              *)
        (*                  PM Setup for SFtpServer                     *)
        (*            Second security page of the notebook              *)
        (*                                                              *)
        (*        Last edited:    2 October 2022                        *)
        (*        Status:         OK                                    *)
        (*                                                              *)
        (****************************************************************)


IMPORT OS2, OS2RTL, DID, CommonSettings, EditFilter, Strings, MiscFuncs;

FROM SYSTEM IMPORT ADR, ADDRESS, CAST;

FROM FSUINI IMPORT
    (* type *)  FHINI,
    (* proc *)  OpenINIFiles, CloseINIFiles, INIGetCard, INIPut;

FROM Languages IMPORT
    (* type *)  LangHandle,
    (* proc *)  StrToBuffer, UseLanguage;

FROM IPFilters IMPORT
    (* type *)  ListPtr,
    (* proc *)  AddrRecordToText, LoadIPFilterList, StoreIPFilterList;

FROM Clipboard IMPORT
    (* proc *)  SaveCopy, RetrieveCopy;

FROM Misc IMPORT
    (* proc *)  WinQueryDlgItemCard, WinSetDlgItemCard;

FROM LowLevel IMPORT
    (* proc *)  IAND;

FROM Storage IMPORT
    (* proc *)  ALLOCATE, DEALLOCATE;

(**************************************************************************)

VAR
    pagehandle, notebookhandle: OS2.HWND;
    PageID: CARDINAL;
    ourlang: LangHandle;

    OldSameIPLimit: CARDINAL;

    (* List head for the allow/exclude list. *)

    MasterListHead: ListPtr;

    (* Current position in the list. *)

    CurrentItem: ListPtr;
    CurrentIndex: INTEGER;

    (* Flag: TRUE if list has been altered. *)

    ListChanged: BOOLEAN;

    ChangeInProgress: BOOLEAN;

(************************************************************************)
(*                    OPERATIONS ON DIALOGUE LABELS                     *)
(************************************************************************)

PROCEDURE DisplayList (hwnd: OS2.HWND;  lang: LangHandle);

    (* Displays the main list.  We deal with it here because it has to  *)
    (* change each time the language changes.                           *)

    VAR current: ListPtr;
        text: ARRAY [0..63] OF CHAR;

    BEGIN
        OS2.WinSendDlgItemMsg (hwnd, DID.FilterListS, OS2.LM_DELETEALL, NIL, NIL);
        IF MasterListHead <> NIL THEN
            current := MasterListHead;
            WHILE current <> NIL DO
                AddrRecordToText (current, lang, text, FALSE);
                OS2.WinSendDlgItemMsg (hwnd, DID.FilterListS, OS2.LM_INSERTITEM,
                             OS2.MPFROMSHORT(OS2.LIT_END), ADR(text));
                current := current^.next;
            END (*WHILE*);

            OS2.WinSendDlgItemMsg (hwnd, DID.FilterListS, OS2.LM_SELECTITEM,
                             OS2.MPFROMSHORT(0), OS2.MPFROMSHORT(1));
            IF MasterListHead^.next = NIL THEN
                OS2.WinEnableWindow (OS2.WinWindowFromID(hwnd, DID.DeleteFilterButtonS), FALSE);
            END (*IF*);
        END (*IF*);
    END DisplayList;

(************************************************************************)

PROCEDURE SetLanguage (lang: LangHandle);

    (* Relabels this page in the new language. *)

    VAR stringval: ARRAY [0..511] OF CHAR;

    (********************************************************************)

    PROCEDURE SetOne (item: ARRAY OF CHAR;  place: CARDINAL);

        BEGIN
            StrToBuffer (lang, item, stringval);
            OS2.WinSetDlgItemText (pagehandle, place, stringval);
        END SetOne;

    (********************************************************************)

    BEGIN
        ourlang := lang;
        StrToBuffer (lang, "Security.tab", stringval);
        Strings.Insert ("SFTP: ", 0, stringval);
        OS2.WinSendMsg (notebookhandle, OS2.BKM_SETTABTEXT,
                        CAST(ADDRESS,PageID), ADR(stringval));
        SetOne ("Security.SameIPLimit", DID.SameIPLimitLabelS);
        SetOne ("Security.FilterListTitle", DID.FilterListTitleS);
        SetOne ("Buttons.Edit", DID.EditFilterButtonS);
        SetOne ("Buttons.Copy", DID.CopyFilterButtonS);
        SetOne ("Buttons.Paste", DID.PasteFilterButtonS);
        SetOne ("Buttons.Insert", DID.InsertFilterButtonS);
        SetOne ("Buttons.Promote", DID.PromoteFilterButtonS);
        SetOne ("Buttons.Delete", DID.DeleteFilterButtonS);
        DisplayList (pagehandle, ourlang);
    END SetLanguage;

(************************************************************************)

PROCEDURE SetServer (setFTP, setSFTP: BOOLEAN);

    (* Specifies which server(s) we are configuring. *)

    BEGIN
    END SetServer;

(************************************************************************)
(*               LOADING THE INITIAL DIALOGUE ITEM CONTENTS             *)
(************************************************************************)

PROCEDURE InitialDisplay (hwnd: OS2.HWND);

    (* Fills the dialogue elements on the security page with data from  *)
    (* the INI file, or loads default values if they're not in the INI file. *)

    VAR val: CARDINAL;  fhini: FHINI;

    BEGIN
        fhini := OpenINIFiles();

        (* Same IP limit. *)

        IF INIGetCard (fhini, '$SYS', 'SameIPLimitS', val) THEN
            OldSameIPLimit := val;
        ELSE
            val := MAX(CARDINAL);
        END (*IF*);
        CloseINIFiles (fhini);

        WinSetDlgItemCard (hwnd, DID.SameIPLimitS, val);

        (* The master list has already been loaded from the INI file.   *)
        (* We still have to copy it into the listbox.                   *)

        DisplayList (hwnd, ourlang);

    END InitialDisplay;

(************************************************************************)
(*                          STORING THE DATA                            *)
(************************************************************************)

PROCEDURE StoreData (hwnd: OS2.HWND);

    (* Stores the security data back into the INI file. *)

    VAR value: CARDINAL;  fhini: FHINI;

    BEGIN
        fhini := OpenINIFiles();

        (* Same IP Limit. *)

        WinQueryDlgItemCard (hwnd, DID.SameIPLimitS, value);
        IF value <> OldSameIPLimit THEN
            INIPut (fhini, '$SYS', 'SameIPLimitS', value);
        END (*IF*);

        (* The allow/deny list. *)

        IF ListChanged THEN
            StoreIPFilterList (fhini, "$SYS", "IPfilterS", MasterListHead);
        END (*IF*);

        CloseINIFiles (fhini);

    END StoreData;

(************************************************************************)
(*
PROCEDURE ConsistencyCheck (hwnd: OS2.HWND;  label: ARRAY OF CHAR);

    (* Debugging aid: checks links in the master list. *)

    VAR w: OS2.HWND;
        prev, p, next: ListPtr;
        pos, curpos: CARDINAL;
        corrupt: BOOLEAN;
        message: ARRAY [0..255] OF CHAR;

    BEGIN
        Strings.Assign (label, message);
        Strings.Append (' ', message);

        (* If master list has inconsistent links, add a '*' to message. *)

        prev := NIL;  p := MasterListHead;  pos := 0;  curpos := MAX(CARDINAL);  corrupt := FALSE;
        WHILE p <> NIL DO
            IF p^.previous <> prev THEN corrupt := TRUE END (*F*);
            next := p^.next;
            IF CurrentItem = p THEN curpos := pos END(*IF*);
            prev := p;
            p := next; INC(pos);
        END (*WHILE*);
        IF corrupt THEN
            Strings.Append ('*', message);
        END (*IF*);

        IF CurrentItem = NIL THEN
            Strings.Append ("CurrentItem is NIL, CurrentIndex=", message);
            MiscFuncs.AppendCard (CurrentIndex, message);
        ELSE
            Strings.Append ("CurrentItem is at position ", message);
            MiscFuncs.AppendCard (curpos, message);
        END (*IF*);
        w := OS2.WinWindowFromID (hwnd, DID.secSmes);
        OS2.WinSetWindowText (w, message);
    END ConsistencyCheck;
*)
(************************************************************************)
(*                         THE DIALOGUE PROCEDURE                       *)
(************************************************************************)

PROCEDURE SetCurrentItem (index: INTEGER);

    (* Uses index to set the values of global variables CurrentItem     *)
    (* and CurrentIndex.                                                *)

    BEGIN
        IF index = OS2.LIT_NONE THEN
            CurrentItem := NIL;
            CurrentIndex := 0;
        ELSE
            IF CurrentItem = NIL THEN
                CurrentItem := MasterListHead;
                CurrentIndex := 0;
            END (*IF*);
            WHILE CurrentIndex < index DO
                CurrentItem := CurrentItem^.next;  INC(CurrentIndex);
            END (*WHILE*);
            WHILE CurrentIndex > index DO
                CurrentItem := CurrentItem^.previous;  DEC(CurrentIndex);
            END (*WHILE*);
        END (*IF*);
    END SetCurrentItem;

(************************************************************************)

PROCEDURE CopyItems (hwnd: OS2.HWND;  VAR (*INOUT*) index: INTEGER);

    (* Makes a list containing the selected items in the window.  On    *)
    (* entry index identifies the first selected item.                  *)
    (* hwnd is the handle for this notebook page.                       *)

    VAR head, this, tail: ListPtr;

    BEGIN
        head := NIL;  tail := NIL;
        WHILE index <> OS2.LIT_NONE DO
            NEW (this);
            this^ := CurrentItem^;
            this^.previous := tail;
            this^.next := NIL;
            IF head = NIL THEN
                head := this;
            ELSE
                tail^.next := this;
            END (*IF*);
            tail := this;

            (* Deselect the current item, then move on to next. *)

            OS2.WinSendDlgItemMsg (hwnd, DID.FilterListS, OS2.LM_SELECTITEM,
                      OS2.MPFROMSHORT(index), OS2.MPFROMSHORT(ORD(FALSE)));
            index := OS2.LONGFROMMR(
                      OS2.WinSendDlgItemMsg (hwnd, DID.FilterListS,
                       OS2.LM_QUERYSELECTION, OS2.MPFROMSHORT(index), NIL));
            SetCurrentItem (index);
        END (*WHILE*);
        SaveCopy (head);
    END CopyItems;

(************************************************************************)

PROCEDURE InsertItems (hwnd: OS2.HWND;  items: ListPtr);

    (* Inserts a new sublist just before CurrentItem.  CurrentItem      *)
    (* remains unchanged, but CurrentIndex must be updated.             *)
    (* Special case: if CurrentItem = NIL (i.e. no item is currently    *)
    (* selected) we do the insertion at the head of the list.           *)

    VAR p, q, OldHead: ListPtr;
        j, N: CARDINAL;
        text: ARRAY [0..127] OF CHAR;

    BEGIN
        IF items <> NIL THEN

            (* First count the items. *)

            p := items;  N := 1;
            WHILE p^.next <> NIL DO
                INC (N);  p := p^.next;
            END (*WHILE*);

            (* Now insert the items into the MasterList.  Note that p   *)
            (* now points to the tail of the list to be inserted.       *)

            OldHead := MasterListHead;
            IF CurrentItem = NIL THEN
                q := NIL;
                CurrentIndex := 0;
            ELSE
                q := CurrentItem^.previous;
            END (*IF*);
            items^.previous := q;
            IF q = NIL THEN
                MasterListHead := items;
            ELSE
                q^.next := items;
            END (*IF*);

            IF CurrentItem = NIL THEN
                (* Note: OldHead can never be NIL. *)
                p^.next := OldHead;
                OldHead^.previous := p;
            ELSE
                p^.next := CurrentItem;
                CurrentItem^.previous := p;
            END (*IF*);

            (* Finally, update the display. *)

            q := items;
            FOR j := 1 TO N DO
                AddrRecordToText (q, ourlang, text, FALSE);
                OS2.WinSendDlgItemMsg (hwnd, DID.FilterListS, OS2.LM_INSERTITEM,
                       OS2.MPFROMSHORT(CurrentIndex), ADR(text));
                INC (CurrentIndex);
                q := q^.next;
            END (*FOR*);

            ListChanged := TRUE;

        END (*IF*);
    END InsertItems;

(************************************************************************)

PROCEDURE DeleteItems (hwnd: OS2.HWND);

    (* Deletes CurrentItem, and also any other items that happen to be  *)
    (* selected.                                                        *)

    VAR index: INTEGER;  p, next: ListPtr;
        Ndel, j: CARDINAL;
        ToDelete: ARRAY [0..32767] OF INTEGER;

    BEGIN
        index := CurrentIndex;  Ndel := 0;

        (* Initial pass.  Enumerate the items to be deleted.  *)

        WHILE index <> OS2.LIT_NONE DO
            ToDelete[Ndel] := index;  INC(Ndel);
            index := OS2.LONGFROMMR(
                      OS2.WinSendDlgItemMsg (hwnd, DID.FilterListS,
                       OS2.LM_QUERYSELECTION, OS2.MPFROMSHORT(index), NIL));
            SetCurrentItem (index);

            (* Special case: if the user has selected the last list     *)
            (* entry, ignore this selection.                            *)

            IF (index <> OS2.LIT_NONE) AND (CurrentItem^.next = NIL) THEN
                OS2.WinSendDlgItemMsg (hwnd, DID.FilterListS, OS2.LM_SELECTITEM,
                      OS2.MPFROMSHORT(index), OS2.MPFROMSHORT(ORD(FALSE)));
                index := OS2.LIT_NONE;
            END (*IF*);

        END (*WHILE*);

        (* Now do the deletions from both the master list and the listbox.  *)
        (* Doing it backwards means the index values remain valid.          *)

        IF Ndel > 0 THEN
            FOR j := Ndel-1 TO 0 BY -1 DO
                index := ToDelete[j];
                OS2.WinSendDlgItemMsg (hwnd, DID.FilterListS, OS2.LM_DELETEITEM,
                                          OS2.MPFROMSHORT(index), NIL);
                SetCurrentItem (index);

                (* Remove CurrentItem from the master list. *)

                p := CurrentItem;
                next := p^.next;
                IF p^.previous = NIL THEN
                    MasterListHead := next;
                ELSE
                    p^.previous^.next := next;
                END (*IF*);
                IF next <> NIL THEN
                    (* Defensive programming: in fact next can never be NIL *)
                    (* because we forbid deleting the last list entry.      *)
                    next^.previous := p^.previous;
                END (*IF*);
                DISPOSE (p);
            END (*FOR*);
        END (*IF*);

        ListChanged := TRUE;

    END DeleteItems;

(************************************************************************)

PROCEDURE ["SysCall"] DialogueProc(hwnd     : OS2.HWND
                     ;msg      : OS2.ULONG
                     ;mp1, mp2 : OS2.MPARAM): OS2.MRESULT;

    VAR ButtonID, NotificationCode, code: CARDINAL;
        index: INTEGER;  p: ListPtr;
        bool: BOOLEAN;
        listwindow: OS2.HWND;
        text: ARRAY [0..127] OF CHAR;

    BEGIN
        IF msg = OS2.WM_INITDLG THEN
            OS2.WinSetWindowPos (hwnd, 0, 0, 0, 0, 0, OS2.SWP_MOVE);
            InitialDisplay (hwnd);
            CurrentItem := MasterListHead;
            CurrentIndex := 0;
            RETURN NIL;
        END (*IF*);

        listwindow := OS2.WinWindowFromID(hwnd,DID.FilterListS);
        index := OS2.LONGFROMMR(
                  OS2.WinSendDlgItemMsg (hwnd, DID.FilterListS,
                   OS2.LM_QUERYSELECTION, OS2.MPFROMSHORT(OS2.LIT_NONE), NIL));
        SetCurrentItem (index);

        IF msg = OS2.WM_COMMAND THEN

            CASE OS2.SHORT1FROMMP(mp1) OF

              | DID.EditFilterButtonS:
                   IF EditFilter.Edit(listwindow, ourlang, CurrentItem) THEN
                       ListChanged := TRUE;
                       AddrRecordToText (CurrentItem, ourlang, text, FALSE);
                       OS2.WinSendDlgItemMsg (hwnd, DID.FilterListS, OS2.LM_SETITEMTEXT,
                              OS2.MPFROMSHORT(index), ADR(text));
                   END (*IF*);

              | DID.CopyFilterButtonS:
                   CopyItems (hwnd, index);

              | DID.PasteFilterButtonS:
                   InsertItems (hwnd, RetrieveCopy());

              | DID.InsertFilterButtonS:
                    (*ConsistencyCheck (hwnd, "Inserting,");*)
                    ListChanged := TRUE;
                    NEW (p);
                    WITH p^ DO
                        previous := CurrentItem^.previous;
                        next := CurrentItem;
                        allow := FALSE;  type := 2;
                        address := 0;  bitcount := 24;
                    END (*WITH*);
                    CurrentItem^.previous := p;
                    IF p^.previous = NIL THEN
                        MasterListHead := p;
                    ELSE
                        p^.previous^.next := p;
                    END (*IF*);
                    CurrentItem := p;
                    text := "";
                    OS2.WinSendDlgItemMsg (hwnd, DID.FilterListS, OS2.LM_INSERTITEM,
                           OS2.MPFROMSHORT(index), ADR(text));
                    OS2.WinSendDlgItemMsg (hwnd, DID.FilterListS, OS2.LM_SELECTITEM,
                          OS2.MPFROMSHORT(index), OS2.MPFROMSHORT(ORD(TRUE)));
                    OS2.WinSendDlgItemMsg (hwnd, DID.FilterListS, OS2.LM_SELECTITEM,
                           OS2.MPFROMSHORT(index+1), OS2.MPFROMSHORT(ORD(FALSE)));
                    IF EditFilter.Edit(listwindow, ourlang, CurrentItem) THEN
                        ListChanged := TRUE;
                        AddrRecordToText (CurrentItem, ourlang, text, FALSE);
                        OS2.WinSendDlgItemMsg (hwnd, DID.FilterListS, OS2.LM_SETITEMTEXT,
                               OS2.MPFROMSHORT(index), ADR(text));

                        (* Have we just done an insertion before the end of the list? *)
                        (* If so, might need to change "all" to "all others".         *)

                        p := CurrentItem^.next;
                        IF p^.type = 0 THEN
                            AddrRecordToText (p, ourlang, text, FALSE);
                            OS2.WinSendDlgItemMsg (hwnd, DID.FilterListS, OS2.LM_SETITEMTEXT,
                                   OS2.MPFROMSHORT(index+1), ADR(text));
                        END (*IF*);

                    ELSE
                        (* Editing was aborted, so delete the item we've just inserted. *)

                        OS2.WinSendMsg (hwnd, OS2.WM_COMMAND,
                                OS2.MPFROMSHORT(DID.DeleteFilterButtonS), NIL);
                    END (*IF*);
                    OS2.WinSetFocus (OS2.HWND_DESKTOP, listwindow);

              | DID.PromoteFilterButtonS:

                    (*ConsistencyCheck (hwnd, "Before promote");*)
                    p := CurrentItem^.previous;
                    IF p <> NIL THEN
                         (*
                         OS2.WinSendDlgItemMsg (hwnd, DID.FilterListS, OS2.LM_SELECTITEM,
                               OS2.MPFROMSHORT(CurrentIndex), OS2.MPFROMSHORT(ORD(FALSE)));
                         *)
                        ListChanged := TRUE;
                        p^.next := CurrentItem^.next;
                        p^.next^.previous := p;
                        CurrentItem^.next := p;
                        CurrentItem^.previous := p^.previous;
                        IF p^.previous = NIL THEN
                            MasterListHead := CurrentItem;
                        ELSE
                            p^.previous^.next := CurrentItem;
                        END (*IF*);
                        p^.previous := CurrentItem;
                        AddrRecordToText (p, ourlang, text, FALSE);
                        OS2.WinSendDlgItemMsg (hwnd, DID.FilterListS, OS2.LM_SETITEMTEXT,
                               OS2.MPFROMSHORT(CurrentIndex), ADR(text));
                        DEC (CurrentIndex);
                        AddrRecordToText (CurrentItem, ourlang, text, FALSE);
                        OS2.WinSendDlgItemMsg (hwnd, DID.FilterListS, OS2.LM_SETITEMTEXT,
                               OS2.MPFROMSHORT(CurrentIndex), ADR(text));
                        OS2.WinSendDlgItemMsg (hwnd, DID.FilterListS, OS2.LM_SELECTITEM,
                               OS2.MPFROMSHORT(CurrentIndex), OS2.MPFROMSHORT(ORD(TRUE)));
                        (*ConsistencyCheck (hwnd, "After promote");*)
                    END (*IF*);

              | DID.DeleteFilterButtonS:

                   ListChanged := TRUE;
                   DeleteItems (hwnd);

                   (* Are we now down to just one list item?                *)
                   (* If so, might need to change "all others" to "all".    *)

                   IF (CurrentItem^.type = 0) AND (CurrentItem^.previous = NIL) THEN
                       AddrRecordToText (CurrentItem, ourlang, text, FALSE);
                       OS2.WinSendDlgItemMsg (hwnd, DID.FilterListS, OS2.LM_SETITEMTEXT,
                              OS2.MPFROMSHORT(index), ADR(text));
                       OS2.WinSetFocus (OS2.HWND_DESKTOP,
                              OS2.WinWindowFromID(hwnd, DID.InsertFilterButtonS));
                   END (*IF*);

              | ELSE
                   RETURN OS2.WinDefDlgProc(hwnd, msg, mp1, mp2);

            END (*CASE*);

        ELSIF msg = OS2.WM_CHAR THEN
            (* Check for Esc.  I probably should be using an accelerator  *)
            (* table to do this, but I'll sort that out later.            *)

            code := OS2.ULONGFROMMP(mp1);
            IF (IAND (code, OS2.KC_VIRTUALKEY) <> 0) THEN
                code := (OS2.ULONGFROMMP(mp2) DIV 65536) MOD 256;
                IF code = OS2.VK_ESC THEN
                    (* Deselect everything in the list box. *)
                    OS2.WinSendDlgItemMsg (hwnd, DID.FilterListS, OS2.LM_SELECTITEM,
                                       OS2.MPFROMSHORT(OS2.LIT_NONE), NIL);
                    (*CurrentItem := NIL;*)    (* redundant statement, I think *)
                ELSE
                    RETURN OS2.WinDefDlgProc(hwnd, msg, mp1, mp2);
                END (*IF*);
            ELSE
                RETURN OS2.WinDefDlgProc(hwnd, msg, mp1, mp2);
            END (*IF*);

        ELSIF msg = OS2.WM_PRESPARAMCHANGED THEN

            IF ChangeInProgress THEN
                RETURN OS2.WinDefDlgProc(hwnd, msg, mp1, mp2);
            ELSE
                ChangeInProgress := TRUE;
                CommonSettings.UpdateFontFrom (hwnd);
                ChangeInProgress := FALSE;
                RETURN NIL;
            END (*IF*);

        ELSIF msg = OS2.WM_CONTROL THEN

            (*ConsistencyCheck (hwnd, "WM_CONTROL received");*)
            NotificationCode := OS2.ULONGFROMMP(mp1);
            ButtonID := NotificationCode MOD 65536;
            NotificationCode := NotificationCode DIV 65536;
            IF ButtonID = DID.FilterListS THEN
                IF NotificationCode = OS2.LN_SELECT THEN
                ELSIF NotificationCode = OS2.LN_ENTER THEN
                    (* Treat this one as if the edit button had been clicked. *)
                    OS2.WinSendMsg (hwnd, OS2.WM_COMMAND,
                          OS2.MPFROMSHORT(DID.EditFilterButtonS), NIL);
                ELSE
                    RETURN OS2.WinDefDlgProc(hwnd, msg, mp1, mp2);
                END (*IF*);
            ELSE
                (*RETURN OS2.WinDefDlgProc(hwnd, msg, mp1, mp2);*)
            END (*IF*);
        ELSE
            RETURN OS2.WinDefDlgProc(hwnd, msg, mp1, mp2);
        END (*CASE*);

        (* Recalculate CurrentItem, because some of the operations above    *)
        (* might have changed it.                                           *)

        index := OS2.LONGFROMMR(
                  OS2.WinSendDlgItemMsg (hwnd, DID.FilterListS,
                   OS2.LM_QUERYSELECTION,  OS2.MPFROMSHORT(OS2.LIT_NONE), NIL));
        SetCurrentItem (index);

        (* Enable or disable the buttons, as appropriate. *)

        (*ConsistencyCheck (hwnd, "When about to enable/disable buttons");*)
        IF CurrentItem = NIL THEN
            OS2.WinEnableWindow (OS2.WinWindowFromID(hwnd, DID.EditFilterButtonS), FALSE);
            OS2.WinEnableWindow (OS2.WinWindowFromID(hwnd, DID.CopyFilterButtonS), FALSE);
            OS2.WinEnableWindow (OS2.WinWindowFromID(hwnd, DID.InsertFilterButtonS), FALSE);
            OS2.WinEnableWindow (OS2.WinWindowFromID(hwnd, DID.PromoteFilterButtonS), FALSE);
            OS2.WinEnableWindow (OS2.WinWindowFromID(hwnd, DID.DeleteFilterButtonS), FALSE);
        ELSE

            (* After having so much trouble with this feature in the past, I    *)
            (* now find that I can pass a Boolean variable to WinEnableWindow.  *)
            (* The trick, I suspect, is to pass a variable rather than a        *)
            (* Boolean expression.                                              *)

            OS2.WinEnableWindow (OS2.WinWindowFromID(hwnd, DID.EditFilterButtonS), TRUE);
            OS2.WinEnableWindow (OS2.WinWindowFromID(hwnd, DID.InsertFilterButtonS), TRUE);

            bool := (CurrentItem^.previous <> NIL) AND (CurrentItem^.next <> NIL);
            OS2.WinEnableWindow (OS2.WinWindowFromID(hwnd, DID.PromoteFilterButtonS), bool);

            bool := CurrentItem^.next <> NIL;
            OS2.WinEnableWindow (OS2.WinWindowFromID(hwnd, DID.CopyFilterButtonS), bool);
            OS2.WinEnableWindow (OS2.WinWindowFromID(hwnd, DID.DeleteFilterButtonS), bool);
        END (*IF*);
        RETURN NIL;

    END DialogueProc;

(**************************************************************************)

PROCEDURE CreatePage (notebook: OS2.HWND;  VAR (*OUT*) pgid: CARDINAL): OS2.HWND;

    (* Creates the user page and adds it to the notebook. *)

    BEGIN
        ListChanged := FALSE;
        notebookhandle := notebook;
        MasterListHead := LoadIPFilterList ("$SYS", "IPfilterS");
        pagehandle := OS2.WinLoadDlg(notebook, notebook,
                       DialogueProc,    (* dialogue procedure *)
                       0,                   (* use resources in EXE *)
                       DID.SecurityS,                (* dialogue ID *)
                       NIL);                 (* creation parameters *)
        PageID := OS2.ULONGFROMMR (OS2.WinSendMsg (notebook, OS2.BKM_INSERTPAGE,
                         NIL, OS2.MPFROM2SHORT (OS2.BKA_MAJOR+OS2.BKA_AUTOPAGESIZE, OS2.BKA_LAST)));
        pgid := PageID;
        OS2.WinSendMsg (notebook, OS2.BKM_SETPAGEWINDOWHWND,
                        CAST(ADDRESS,PageID), CAST(ADDRESS,pagehandle));
        RETURN pagehandle;
    END CreatePage;

(**************************************************************************)

PROCEDURE SetFont (VAR (*IN*) name: CommonSettings.FontName);

    (* Sets the font of the text on this page. *)

    CONST bufsize = CommonSettings.FontNameSize;

    BEGIN
        OS2.WinSetPresParam (pagehandle, OS2.PP_FONTNAMESIZE, bufsize, name);
    END SetFont;

(**************************************************************************)

BEGIN
    OldSameIPLimit := 0;
    ChangeInProgress := FALSE;
    ourlang := UseLanguage ("Setup", "en");
END SecurityS.

