(**************************************************************************)
(*                                                                        *)
(*  Setup for SFtpServer                                                  *)
(*  Copyright (C) 2022   Peter Moylan                                     *)
(*                                                                        *)
(*  This program is free software: you can redistribute it and/or modify  *)
(*  it under the terms of the GNU General Public License as published by  *)
(*  the Free Software Foundation, either version 3 of the License, or     *)
(*  (at your option) any later version.                                   *)
(*                                                                        *)
(*  This program is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU General Public License for more details.                          *)
(*                                                                        *)
(*  You should have received a copy of the GNU General Public License     *)
(*  along with this program.  If not, see <http://www.gnu.org/licenses/>. *)
(*                                                                        *)
(*  To contact author:   http://www.pmoylan.org   peter@pmoylan.org       *)
(*                                                                        *)
(**************************************************************************)

IMPLEMENTATION MODULE UserPage;

        (****************************************************************)
        (*                                                              *)
        (*                   PM Setup for SFtpServer                    *)
        (*                 User page of the notebook                    *)
        (*                                                              *)
        (*        Last edited:    22 September 2022                     *)
        (*        Status:         OK                                    *)
        (*                                                              *)
        (****************************************************************)


FROM SYSTEM IMPORT CARD16, ADDRESS, CAST, ADR;

IMPORT OS2, OS2RTL, Strings, DID, RINIData, BasicPage, EditUser, CommonSettings,
                    WideUserDialogue, WU2Dialogue;

FROM FSUINI IMPORT
    (* type *)  FHINI, StringReadState,
    (* proc *)  SetHashMax, ChangeINIFilename, RestoreINIFilename,
                GetFullINIname,
                OpenINIFiles, OpenAlternative, OpenINIForUser, CloseINIFiles,
                INInameMask, ItemSize, INIFetch, INIGetCard,
                GetStringList, NextString, CloseStringList, INIPut,
                INIDeleteApp;

FROM Languages IMPORT
    (* type *)  LangHandle,
    (* proc *)  StrToBuffer, StrToBufferN;

FROM FileOps IMPORT
    (* type *)  FilenameString, DirectoryEntry,
    (* proc *)  FirstDirEntry, NextDirEntry, DirSearchDone;

FROM MiscFuncs IMPORT
    (* type *)  CharArrayPointer,
    (* proc *)  EVAL, ConvertCard, StringMatch;

FROM LowLevel IMPORT
    (* proc *)  IAND;

FROM Storage IMPORT
    (* proc *)  ALLOCATE, DEALLOCATE;

(************************************************************************)

CONST
    Nul = CHR(0);
    NameLength = 256;

TYPE
    NameString = ARRAY [0..NameLength-1] OF CHAR;

VAR
    pagehandle, notebookhandle: OS2.HWND;
    PageID: CARDINAL;
    ourlang: LangHandle;
    UserCount: CARDINAL;
    ChangeInProgress: BOOLEAN;

(************************************************************************)
(*                    OPERATIONS ON DIALOGUE LABELS                     *)
(************************************************************************)

PROCEDURE SetLanguage (lang: LangHandle);

    (* Relabels this page in the new language. *)

    VAR stringval: ARRAY [0..511] OF CHAR;

    (********************************************************************)

    PROCEDURE SetOne (item: ARRAY OF CHAR;  place: CARDINAL);

        BEGIN
            StrToBuffer (lang, item, stringval);
            OS2.WinSetDlgItemText (pagehandle, place, stringval);
        END SetOne;

    (********************************************************************)

    BEGIN
        ourlang := lang;
        StrToBuffer (lang, "Users.tab", stringval);
        OS2.WinSendMsg (notebookhandle, OS2.BKM_SETTABTEXT,
                        CAST(ADDRESS,PageID), ADR(stringval));
        SetOne ("Users.Label", DID.Users);
        SetOne ("Buttons.Add", DID.AddUserButton);
        SetOne ("Buttons.Clone", DID.CloneUserButton);
        SetOne ("Buttons.Edit", DID.EditUserButton);
        SetOne ("Buttons.Delete", DID.DeleteUserButton);
        SetOne ("Buttons.Sort", DID.SortButton);
        StrToBufferN (lang, "Users.total", UserCount, stringval);
        OS2.WinSetDlgItemText (pagehandle, DID.UserCount, stringval);
    END SetLanguage;

(************************************************************************)

(*
PROCEDURE CardToHex (N: CARDINAL;  VAR (*OUT*) text: ARRAY OF CHAR);

    (* For debugging.  32-bit cardinal to 8-char hexadecimal. *)

    VAR j, digit: CARDINAL;

    BEGIN
        text[8] := Nul;
        FOR j := 7 TO 0 BY -1 DO
            digit := N MOD 16;
            N := N DIV 16;
            IF digit < 10 THEN
                text[j] := CHR(ORD('0') + digit);
            ELSE
                text[j] := CHR(ORD('A') -10 + digit);
            END (*IF*);
        END (*FOR*);
    END CardToHex;

(************************************************************************)

PROCEDURE WriteHex (hwnd: OS2.HWND;  to: CARDINAL;  N: CARDINAL);

    (* For debugging.  Write N in hexadecimal to debug1 or debug2 field. *)

    VAR text: ARRAY [0..8] OF CHAR;
        w: OS2.HWND;

    BEGIN
        CardToHex (N, text);
        IF to = 1 THEN
            w := OS2.WinWindowFromID (hwnd, DID.debug1);
        ELSE
            w := OS2.WinWindowFromID (hwnd, DID.debug2);
        END (*IF*);
        OS2.WinSetWindowText (w, text);
    END WriteHex;
*)

(************************************************************************)

PROCEDURE NumberOfUsers (): CARDINAL;

    (* Returns the number of users. *)

    BEGIN
        RETURN UserCount;
    END NumberOfUsers;

(************************************************************************)

PROCEDURE DisplayUserCount (hwnd: OS2.HWND);

    (* Displays the value of global variable UserCount. *)

    VAR pos: CARDINAL;  buffer: ARRAY [0..63] OF CHAR;

    BEGIN
        buffer := "Total ";
        pos := LENGTH(buffer);
        ConvertCard (UserCount, buffer, pos);
        buffer[pos] := Nul;
        OS2.WinSetDlgItemText (hwnd, DID.UserCount, buffer);
    END DisplayUserCount;

(************************************************************************)
(*                                                                      *)
(*               LOADING AND STORING THE LIST OF USERS                  *)
(*                                                                      *)
(* The usernames may be scattered across multiple INI files, for two    *)
(* reasons:                                                             *)
(*  1.  The FSUINI module looks after two sets of INI data, one for     *)
(*      FTP and the other for SFTP.  If we are using this program to    *)
(*      configure both, we should have the same set of users in both    *)
(*      groups, but if not we should merge the two sets.                *)
(*  2.  Either or both sets might be using the system of spreading the  *)
(*      users over multiple INI files - the preferred choice if there   *)
(*      are many users.  If the two sets are inconsistent in this       *)
(*      respect then part of our job is to make them compatible.        *)
(*                                                                      *)
(* Some of the code here is similar to the BasicPage code that switches *)
(* between single-INI and multi-INI mode, but I have not so far         *)
(* attempted to pull out the common code.  We can live with the         *)
(* code duplication.                                                    *)
(*                                                                      *)
(************************************************************************)

TYPE NameList = POINTER TO
                    RECORD
                        next:   NameList;
                        seqnum: CARD16;
                        name:   NameString;
                    END (*RECORD*);

(********************************************************************)

PROCEDURE IsADuplicate (list: NameList;  name: NameString): BOOLEAN;

    (* Returns TRUE iff name is already in list.    *)

    VAR p: NameList;

    BEGIN
        p := list;
        LOOP
            IF p = NIL THEN RETURN FALSE END(*IF*);
            IF StringMatch (name, p^.name) THEN RETURN TRUE END(*IF*);
            p := p^.next;
        END (*LOOP*);
    END IsADuplicate;

(************************************************************************)

PROCEDURE MergeLists (VAR (*INOUT*) list, second: NameList): CARDINAL;

    (* Merges second into list, leaving second = NIL.  We assume that   *)
    (* the two lists are already sorted by sequence number.             *)

    VAR prev, current, next: NameList;
        count: CARDINAL;

    BEGIN
        prev := NIL;
        current := list;
        count := 0;
        WHILE second <> NIL DO
            IF IsADuplicate (list, second^.name) THEN
                next := second^.next;
                DISPOSE (second);
                second := next;
            ELSE
                WHILE (current <> NIL) AND (current^.seqnum < second^.seqnum) DO
                    INC (count);
                    prev := current;
                    current := current^.next;
                END (*WHILE*);

                (* Detach current^ and its successors from the main     *)
                (* list and make it the new second list.  The old       *)
                (* second list becomes the successor of prev^.          *)

                INC (count);        (* to account for second^ *)
                IF prev = NIL THEN
                    list := second;
                ELSE
                    prev^.next := second;
                END (*IF*);
                prev := second;
                second := current;
                current := prev^.next;

            END (*IF*);
        END (*WHILE*);

        (* Count the list elements not yet encountered. *)

        WHILE current <> NIL DO
            INC (count);
            current := current^.next;
        END (*WHILE*);

        RETURN count;

    END MergeLists;

(************************************************************************)

PROCEDURE SortList (VAR (*INOUT*) list: NameList;  N: CARDINAL): CARDINAL;

    (* Sorts list of N elements in increasing order of seqnum.  Returns *)
    (* the list size, after duplicates have been eliminated.            *)

    VAR j: CARDINAL;
        list2, prev, current: NameList;

    BEGIN
        IF N < 2 THEN
            RETURN N;
        END (*IF*);

        (* Split the list into two sublists. *)

        current := list;  prev := NIL;
        FOR j := 0 TO N DIV 2 - 1 DO
            prev := current;
            IF current <> NIL THEN
                current := current^.next;
            END (*IF*);
        END (*FOR*);
        list2 := current;
        IF prev <> NIL THEN
            prev^.next := NIL;
        END (*IF*);

        (* Sort the two sublists. *)

        EVAL (SortList (list, N DIV 2));
        EVAL (SortList (list2, N - N DIV 2));

        (* Merge the two sorted lists. *);

        RETURN MergeLists (list, list2);

    END SortList;

(************************************************************************)

PROCEDURE LoadOneUserList (rhini: RINIData.RHINI;  VAR (*OUT*) head: NameList);

    (* Loads the names of all applications, excluding those whose name  *)
    (* starts with '$', from the specified INI file.                    *)

    VAR state: RINIData.StringReadState;
        appname: NameString;
        count: CARDINAL;
        current: NameList;

    BEGIN
        RINIData.GetStringList (rhini, '', '', state);
        head := NIL;  current := NIL;  count := 0;
        LOOP
            RINIData.NextString (state, appname);
            IF appname[0] = Nul THEN
                EXIT (*LOOP*);
            ELSIF (appname[0] <> '$') AND NOT IsADuplicate (head, appname) THEN
                IF head = NIL THEN
                    NEW (head);  current := head;
                ELSE
                    NEW (current^.next);  current := current^.next;
                END (*IF*);
                INC (count);
                current^.next := NIL;
                current^.name := appname;

                (* Get sequence number. *)

                IF NOT RINIData.INIFetch (rhini, current^.name, "seqnum", current^.seqnum) THEN
                    current^.seqnum := MAX(CARD16);
                END (*IF*);

            END (*IF*);
        END (*LOOP*);
        RINIData.CloseStringList (state);

        EVAL (SortList (head, count));

    END LoadOneUserList;

(************************************************************************)

PROCEDURE LoadUserSet (j: CARDINAL;  VAR (*OUT*) list: NameList);

    (* Collects all the usernames from INI set j (j=0..1), sorted by    *)
    (* sequence number.                                                 *)

    VAR fullname, mask, name: FilenameString;
        state: RINIData.SearchState;
        rhini: RINIData.RHINI;
        list2: NameList;
        N: CARDINAL;
        remote, found: BOOLEAN;

    BEGIN
        list := NIL;

        (* Get the main INI name for set j. *)

        GetFullINIname (j, fullname, remote);
        IF fullname[0] = Nul THEN
            RETURN;
        END (*IF*);

        Strings.Assign (fullname, mask);

        (* Turn the name into a wildcard search mask, to produce the    *)
        (* set of INI files we must check.                              *)

        N := Strings.Length(mask) - 4;
        Strings.Insert ('*', N, mask);
        found := RINIData.FirstDirEntry (mask, remote, FALSE, FALSE, TRUE, state, name);
        WHILE found DO
            rhini := OpenAlternative (j, name);
            LoadOneUserList (rhini, list2);
            RINIData.CloseINIFile (rhini);
            EVAL (MergeLists (list, list2));
            found := RINIData.NextDirEntry (state, name);
        END (*WHILE*);
        RINIData.DirSearchDone (state);
    END LoadUserSet;

(************************************************************************)

PROCEDURE ListAllUsers (VAR (*OUT*) result0: NameList): CARDINAL;

    (* Produces a list of all users. *)

    (* ERROR at the point where two lists are merged *)

    VAR result1: NameList;
        count: CARDINAL;

    BEGIN
        LoadUserSet (0, result0);
        LoadUserSet (1, result1);
        count := MergeLists (result0, result1);
        RETURN count;
    END ListAllUsers;

(************************************************************************)

PROCEDURE LoadValues (hwnd: OS2.HWND);

    (* Fills the user list with data from the INI files.  Note that     *)
    (* there can be multiple INI files depending on the 'HashMax' INI   *)
    (* value.  The order in which we will put the entries into the      *)
    (* listbox will depend on recorded sequence numbers, which          *)
    (* unfortunately adds to the complexity of this procedure.          *)

    VAR list: NameList;
        username:  NameString;
        current, next: NameList;

    BEGIN
        list := NIL;
        UserCount := ListAllUsers (list);

        (* At this point we have collected all usernames in all of our  *)
        (* INI files, but we still haven't put them into the listbox.   *)
        (* That has been delayed until now because we want to look up   *)
        (* sequence numbers to work out the order.                      *)

        (* It looks as if the listbox doesn't like out-of-order         *)
        (* insertions, so we need to sort the list before doing the     *)
        (* listbox insertions.  Note that the sort will put all entries *)
        (* without sequence numbers at the end of the list.             *)

        UserCount := SortList (list, UserCount);

        (* Now that we have entries in the correct order, put them in   *)
        (* the listbox.  The sequence numbers are no longer needed,     *)
        (* because they will be recalculated at a later stage.          *)

        current := list;
        WHILE current <> NIL DO
            next := current^.next;
            Strings.Assign (current^.name, username);
            DISPOSE (current);
            current := next;

            (* Add name to the listbox. *)

            OS2.WinSendDlgItemMsg (hwnd, DID.userlist, OS2.LM_INSERTITEM,
                     OS2.MPFROMSHORT(OS2.LIT_END), ADR(username));
        END (*WHILE*);

        (* By now the list should be empty. *)

    END LoadValues;

(************************************************************************)

PROCEDURE StoreData (hwnd: OS2.HWND);

    (* Most of the INI file updating is done as part of the button      *)
    (* processing, but we take the opportunity to assign user sequence  *)
    (* numbers as Setup is about to exit.  That will ensure that the    *)
    (* next time we run setup the order will be unchanged.              *)

    CONST Skip = MAX(CARD16);

    VAR lbox: OS2.HWND;
        j, k, N: CARDINAL;
        fhini: FHINI;
        name: NameString;
        list, p, tail: NameList;

    BEGIN
        lbox := OS2.WinWindowFromID(hwnd,DID.userlist);

        (* Work out how many users there are in the list. *)

        N := CAST(CARDINAL, OS2.WinSendMsg (lbox, OS2.LM_QUERYITEMCOUNT, NIL, NIL));
        IF N = 0 THEN RETURN END(*IF*);

        (* Read the listbox into the NameList, assign sequence numbers. *)
        (* Skip any entries for which name = "?".                       *)

        k := 0;
        list := NIL;  tail := NIL;
        FOR j := 0 TO N-1 DO
            EVAL (OS2.WinSendMsg (lbox, OS2.LM_QUERYITEMTEXT,
                            OS2.MPFROM2SHORT(j, NameLength), ADR(name)));
            IF name[0] <> '?' THEN
                NEW (p);
                p^.next := NIL;
                p^.name := name;
                p^.seqnum := k;
                INC (k);
                IF list = NIL THEN
                    list := p;
                ELSE
                    tail^.next := p;
                END (*IF*);
                tail := p;
            END (*IF*);
        END (*FOR*);

        (* Now store sequence numbers from the linked list to the INI   *)
        (* file(s).  A sequence number of Skip (=MAX(CARD16)) means     *)
        (* that we have already dealt with this entry.                  *)

        WHILE list <> NIL DO
            IF list^.seqnum <> Skip THEN
                name := list^.name;
                IF name[0] <> '?' THEN
                    fhini := OpenINIForUser (name);
                    INIPut (fhini, name, "seqnum", list^.seqnum);
                    list^.seqnum := Skip;

                    (* Deal with any other users in this INI file. *)

                    p := list^.next;
                    WHILE p <> NIL DO
                        IF (p^.seqnum <> Skip)
                                    AND ItemSize (fhini, p^.name, '', k)
                                                    AND (k <> 0) THEN
                            INIPut (fhini, p^.name, "seqnum", p^.seqnum);
                            p^.seqnum := Skip;
                        END (*IF*);
                        p := p^.next;
                    END (*WHILE*);
                    CloseINIFiles (fhini);
                END (*IF*);
            END (*IF*);
            p := list^.next;
            DISPOSE (list);
            list := p;
        END (*WHILE*);

    END StoreData;

(************************************************************************)
(*                        ENABLE/DISABLE BUTTONS                        *)
(************************************************************************)

PROCEDURE ItemSelected (hwnd: OS2.HWND;  selected: BOOLEAN);

    (* Enables/disables some buttons, depending on "selected". *)

    BEGIN
        IF selected THEN
            OS2.WinEnableWindow (OS2.WinWindowFromID(hwnd, DID.EditUserButton), TRUE);
            OS2.WinEnableWindow (OS2.WinWindowFromID(hwnd, DID.CloneUserButton), TRUE);
            OS2.WinEnableWindow (OS2.WinWindowFromID(hwnd, DID.DeleteUserButton), TRUE);
        ELSE
            OS2.WinEnableWindow (OS2.WinWindowFromID(hwnd, DID.EditUserButton), FALSE);
            OS2.WinEnableWindow (OS2.WinWindowFromID(hwnd, DID.CloneUserButton), FALSE);
            OS2.WinEnableWindow (OS2.WinWindowFromID(hwnd, DID.DeleteUserButton), FALSE);
            OS2.WinSetFocus (OS2.HWND_DESKTOP,
                             OS2.WinWindowFromID(hwnd, DID.AddUserButton));
        END (*IF*);
    END ItemSelected;

(************************************************************************)
(*                          THE DIALOGUE PROCEDURE                      *)
(************************************************************************)

PROCEDURE ["SysCall"] DialogueProc(hwnd     : OS2.HWND
                     ;msg      : OS2.ULONG
                     ;mp1, mp2 : OS2.MPARAM): OS2.MRESULT;

    VAR ButtonID, NotificationCode: CARDINAL;
        index: INTEGER;
        code: CARDINAL;
        listwindow: OS2.HWND;
        fhini: FHINI;
        name: NameString;

    BEGIN
        IF msg = OS2.WM_INITDLG THEN
            OS2.WinSetWindowPos (hwnd, 0, 0, 0, 0, 0, OS2.SWP_MOVE);
            OS2.WinEnableWindow (OS2.WinWindowFromID(hwnd, DID.EditUserButton), FALSE);
            OS2.WinEnableWindow (OS2.WinWindowFromID(hwnd, DID.CloneUserButton), FALSE);
            OS2.WinEnableWindow (OS2.WinWindowFromID(hwnd, DID.DeleteUserButton), FALSE);
            LoadValues (hwnd);
            DisplayUserCount (hwnd);
            RETURN NIL;

        ELSIF msg = OS2.WM_COMMAND THEN

            listwindow := OS2.WinWindowFromID(hwnd,DID.userlist);
            index := OS2.LONGFROMMR(
                      OS2.WinSendDlgItemMsg (hwnd, DID.userlist, OS2.LM_QUERYSELECTION, NIL, NIL));
            CASE OS2.SHORT1FROMMP(mp1) OF

              | DID.AddUserButton:
                   name := "";
                   IF index = OS2.LIT_NONE THEN
                       index := 0;
                   ELSE
                       INC(index);
                   END (*IF*);
                   OS2.WinSendDlgItemMsg (hwnd, DID.userlist, OS2.LM_INSERTITEM,
                          OS2.MPFROMSHORT(index), ADR(name));
                   OS2.WinSendDlgItemMsg (hwnd, DID.userlist, OS2.LM_SELECTITEM,
                          OS2.MPFROMSHORT(index), OS2.MPFROMSHORT(ORD(TRUE)));
                   IF EditUser.Edit(listwindow, name, ourlang, FALSE) THEN
                       (* We have a new entry. *)
                       fhini :=OpenINIForUser (name);
                       CloseINIFiles (fhini);
                       OS2.WinSendMsg (listwindow, OS2.LM_SETITEMTEXT,
                                   OS2.MPFROMSHORT(index), ADR(name));
                       INC (UserCount);
                       DisplayUserCount (hwnd);
                   ELSE
                       (* Empty name, delete it. *)
                       OS2.WinSendMsg (hwnd, OS2.WM_COMMAND,
                             OS2.MPFROMSHORT(DID.DeleteUserButton), NIL);
                   END (*IF*);

              | DID.CloneUserButton:
                   OS2.WinSendMsg (listwindow, OS2.LM_QUERYITEMTEXT,
                                   OS2.MPFROM2USHORT(index, 32), ADR(name));
                   IF EditUser.Edit(listwindow, name, ourlang, TRUE) THEN
                       fhini := OpenINIForUser (name);
                       CloseINIFiles (fhini);
                       INC (index);
                       OS2.WinSendDlgItemMsg (hwnd, DID.userlist, OS2.LM_INSERTITEM,
                              OS2.MPFROMSHORT(index), ADR(name));
                       OS2.WinSendDlgItemMsg (hwnd, DID.userlist, OS2.LM_SELECTITEM,
                              OS2.MPFROMSHORT(index), OS2.MPFROMSHORT(ORD(TRUE)));
                       INC (UserCount);
                       DisplayUserCount (hwnd);
                   END (*IF*);

              | DID.EditUserButton:
                   OS2.WinSendMsg (listwindow, OS2.LM_QUERYITEMTEXT,
                                   OS2.MPFROM2USHORT(index, 32), ADR(name));
                   EVAL (EditUser.Edit(listwindow, name, ourlang, FALSE));
                   (* Name or category might have changed. *)
                   fhini := OpenINIForUser (name);
                   CloseINIFiles (fhini);
                   OS2.WinSendMsg (listwindow, OS2.LM_SETITEMTEXT,
                               OS2.MPFROMSHORT(index), ADR(name));

              | DID.DeleteUserButton:
                   OS2.WinSendMsg (listwindow, OS2.LM_QUERYITEMTEXT,
                                   OS2.MPFROM2USHORT(index, 32), ADR(name));
                   OS2.WinSendDlgItemMsg (hwnd, DID.userlist, OS2.LM_DELETEITEM,
                                          OS2.MPFROMSHORT(index), NIL);
                   ItemSelected (hwnd, FALSE);
                   IF name[0] <> Nul THEN
                       fhini := OpenINIForUser (name);
                       INIDeleteApp (fhini, name);
                       CloseINIFiles (fhini);
                       DEC (UserCount);
                       DisplayUserCount (hwnd);
                   END (*IF*);

              | DID.SortButton:
                   OS2.WinSendDlgItemMsg (hwnd, DID.userlist, OS2.LM_SELECTITEM,
                                          OS2.MPFROMSHORT(OS2.LIT_NONE), NIL);
                   WideUserDialogue.SortDlg (ourlang, hwnd, listwindow);

              | DID.SortButton2:

                   (* SortButton2 is an invisible button that cannot be clicked,    *)
                   (* so presumably WU2Dialogue is an obsolete module.  However I   *)
                   (* can't remember why I had that version.                        *)

                   OS2.WinSendDlgItemMsg (hwnd, DID.userlist, OS2.LM_SELECTITEM,
                                          OS2.MPFROMSHORT(OS2.LIT_NONE), NIL);
                   WU2Dialogue.SortDlg (hwnd, listwindow);

            ELSE
                RETURN OS2.WinDefDlgProc(hwnd, msg, mp1, mp2);
            END (*CASE*);

            RETURN NIL;

        ELSIF msg = OS2.WM_CONTROL THEN

            NotificationCode := OS2.ULONGFROMMP(mp1);
            ButtonID := NotificationCode MOD 65536;
            NotificationCode := NotificationCode DIV 65536;
            IF ButtonID = DID.userlist THEN
                IF NotificationCode = OS2.LN_SELECT THEN
                    ItemSelected (hwnd, TRUE);
                    RETURN NIL;
                ELSIF NotificationCode = OS2.LN_ENTER THEN
                    (* Treat this one as if the edit button had been clicked. *)
                    OS2.WinSendMsg (hwnd, OS2.WM_COMMAND,
                          OS2.MPFROMSHORT(DID.EditUserButton), NIL);
                    RETURN NIL;
                ELSE
                    RETURN OS2.WinDefDlgProc(hwnd, msg, mp1, mp2);
                END (*IF*);
            ELSE
                RETURN OS2.WinDefDlgProc(hwnd, msg, mp1, mp2);
            END (*IF*);

        ELSIF msg = OS2.WM_CHAR THEN
            (* Check for Esc.  I probably should be using an accelerator  *)
            (* table to do this, but I'll sort that out later.            *)

            (*
            WriteHex (hwnd, 1, OS2.ULONGFROMMP(mp1));
            WriteHex (hwnd, 2, OS2.ULONGFROMMP(mp2));
            *)

            code := OS2.ULONGFROMMP(mp1);
            IF (IAND (code, OS2.KC_VIRTUALKEY) <> 0) THEN
                code := (OS2.ULONGFROMMP(mp2) DIV 65536) MOD 256;
                IF code = OS2.VK_ESC THEN
                    (* Deselect everything in the list box. *)
                    OS2.WinSendDlgItemMsg (hwnd, DID.userlist, OS2.LM_SELECTITEM,
                                       OS2.MPFROMSHORT(OS2.LIT_NONE), NIL);
                    ItemSelected (hwnd, FALSE);
                    RETURN NIL;
                ELSE
                    RETURN OS2.WinDefDlgProc(hwnd, msg, mp1, mp2);
                END (*IF*);
            ELSE
                RETURN OS2.WinDefDlgProc(hwnd, msg, mp1, mp2);
            END (*IF*);

        ELSIF msg = OS2.WM_PRESPARAMCHANGED THEN

            IF ChangeInProgress THEN
                RETURN OS2.WinDefDlgProc(hwnd, msg, mp1, mp2);
            ELSE
                ChangeInProgress := TRUE;
                CommonSettings.UpdateFontFrom (hwnd);
                ChangeInProgress := FALSE;
                RETURN NIL;
            END (*IF*);

        ELSE
            RETURN OS2.WinDefDlgProc(hwnd, msg, mp1, mp2);
        END (*CASE*);
    END DialogueProc;

(************************************************************************)

PROCEDURE CreatePage (notebook: OS2.HWND;  VAR (*OUT*) pgid: CARDINAL): OS2.HWND;

    (* Creates the user page and adds it to the notebook. *)

    VAR Label: ARRAY [0..31] OF CHAR;

    BEGIN
        notebookhandle := notebook;
        pagehandle := OS2.WinLoadDlg(notebook, notebook,
                       DialogueProc,    (* dialogue procedure *)
                       0,                   (* use resources in EXE *)
                       DID.UserPage,                (* dialogue ID *)
                       NIL);                 (* creation parameters *)
        PageID := OS2.ULONGFROMMR (OS2.WinSendMsg (notebook, OS2.BKM_INSERTPAGE,
                         NIL, OS2.MPFROM2SHORT (OS2.BKA_MAJOR+OS2.BKA_AUTOPAGESIZE, OS2.BKA_LAST)));
        pgid := PageID;
        Label := "~Users";
        OS2.WinSendMsg (notebook, OS2.BKM_SETTABTEXT,
                        CAST(ADDRESS,PageID), ADR(Label));
        OS2.WinSendMsg (notebook, OS2.BKM_SETPAGEWINDOWHWND,
                        CAST(ADDRESS,PageID), CAST(ADDRESS,pagehandle));
        RETURN pagehandle;
    END CreatePage;

(************************************************************************)

PROCEDURE SetFont (VAR (*IN*) name: CommonSettings.FontName);

    (* Sets the font of the text on this page. *)

    CONST bufsize = CommonSettings.FontNameSize;

    BEGIN
        OS2.WinSetPresParam (pagehandle, OS2.PP_FONTNAMESIZE, bufsize, name);
    END SetFont;

(************************************************************************)

BEGIN
    ChangeInProgress := FALSE;
END UserPage.

