(**************************************************************************)
(*                                                                        *)
(*  SSH Authentication layer                                              *)
(*  Copyright (C) 2022   Peter Moylan                                     *)
(*                                                                        *)
(*  This program is free software: you can redistribute it and/or modify  *)
(*  it under the terms of the GNU General Public License as published by  *)
(*  the Free Software Foundation, either version 3 of the License, or     *)
(*  (at your option) any later version.                                   *)
(*                                                                        *)
(*  This program is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU General Public License for more details.                          *)
(*                                                                        *)
(*  You should have received a copy of the GNU General Public License     *)
(*  along with this program.  If not, see <http://www.gnu.org/licenses/>. *)
(*                                                                        *)
(*  To contact author:   http://www.pmoylan.org   peter@pmoylan.org       *)
(*                                                                        *)
(**************************************************************************)

IMPLEMENTATION MODULE AuthLayer;

        (********************************************************)
        (*                                                      *)
        (*              Authentication layer of SSH             *)
        (*                                                      *)
        (*  Programmer:         P. Moylan                       *)
        (*  Started:            29 May 2021                     *)
        (*  Last edited:        6 October 2022                  *)
        (*  Status:             OK                              *)
        (*                                                      *)
        (********************************************************)


FROM SYSTEM IMPORT ADDRESS, CARD8, ADR, CAST;
FROM Arith64 IMPORT CARD64;

IMPORT Services, Strings, Channels, PacketQueues;

FROM STextIO IMPORT
    (* proc *)  WriteString, WriteLn;

FROM TransportLayer IMPORT
    (* proc *)  ClosePort;

FROM PacketIO IMPORT
    (* type *)  IOPort,
    (* proc *)  PeerAddr, SendPacket;

FROM PacketQueues IMPORT
    (* type *)  PacketQ,
    (* proc *)  CreateQueue, GetPacket, PutPacket;

FROM FDUsers IMPORT
    (* type *)  User, UserCategory,
    (* proc *)  ReadUserData, AddressForbidden, PasswordOK, DestroyUserData, NoteLoginTime;

FROM Consts IMPORT
    (* const*)  SSH_MSG_USERAUTH_REQUEST, SSH_MSG_USERAUTH_FAILURE, SSH_MSG_USERAUTH_SUCCESS,
                SSH_MSG_TERMINATE;

FROM VarStrings IMPORT
    (* type *)  ByteStr,
    (* proc *)  MakeBS, DiscardBS, GetCharString;

FROM MiscFuncs IMPORT
    (* proc *)  AppendCard, Swap4;

FROM TransLog IMPORT
    (* type *)  TransactionLogID,
    (* proc *)  LogTransaction, LogTransactionL;

FROM TaskControl IMPORT
    (* proc *)  CreateTask1;

FROM LowLevel IMPORT
    (* proc *)  Copy, EVAL;

FROM Storage IMPORT
    (* proc *)  ALLOCATE, DEALLOCATE;

(************************************************************************)

CONST NilUser = CAST (User, NIL);

(************************************************************************)
(*                      AUTHENTICATION CHECKING                         *)
(************************************************************************)

PROCEDURE FailureReplyPacket(): ByteStr;

    (* Constructs a USERAUTH_FAILURE packet.  *)

    VAR k, size, N: CARDINAL;
        reply: ByteStr;
        methodname: ARRAY [0..7] OF CHAR;

    BEGIN
        methodname := "password";
        k := Strings.Length (methodname);
        size := 1 + 4 + k + 1;
        MakeBS (reply, size);
        reply.data^[0] := SSH_MSG_USERAUTH_FAILURE;
        N := Swap4(k);
        Copy (ADR(N), ADR(reply.data^[1]), 4);
        Copy (ADR(methodname), ADR(reply.data^[5]), k);
        reply.data^[5+k] := 0;
        RETURN reply;
    END FailureReplyPacket;

(************************************************************************)

PROCEDURE GetUserInfo (username: ARRAY OF CHAR;  peeraddr: CARDINAL;
                            VAR (*OUT*) category: UserCategory): User;

    (* Identifies a user and returns its category. *)

    BEGIN
        IF AddressForbidden (peeraddr, username) THEN
            category := NoSuchUser;
            RETURN NIL;
        ELSE
            RETURN ReadUserData (username, category);
        END (*IF*);
    END GetUserInfo;

(************************************************************************)

PROCEDURE TryToAuthenticate (port: IOPort;  pkt: ByteStr;
                                            logID: TransactionLogID): User;

    (* If successful, returns a User record.  The result is NilUser *)
    (* for authentication failure.                                  *)

    VAR pos, addr: CARDINAL;
        success: BOOLEAN;
        user: User;
        category: UserCategory;
        reply: ByteStr;
        username, servicename, methodname, password: ARRAY [0..63] OF CHAR;
        message: ARRAY [0..127] OF CHAR;

    BEGIN
        success := FALSE;       (* default value *)
        pos := 1;

        (* Extract the fields from the packet.  *)

        GetCharString (pkt, pos, username);
        GetCharString (pkt, pos, servicename);
        GetCharString (pkt, pos, methodname);

        (* While testing, log the parameters. *)

        (*
        message := "user, service, method = ";
        Strings.Append (username, message);  Strings.Append (", ", message);
        Strings.Append (servicename, message);  Strings.Append (", ", message);
        Strings.Append (methodname, message);
        LogTransaction (logID, message);
        *)

        addr := PeerAddr (port);
        user := GetUserInfo (username, addr, category);

        IF Services.IdentifyService (servicename) = Services.none THEN
            message := "Unsupported service ";
            Strings.Append (servicename, message);
            LogTransaction (logID, message);
            success := FALSE;

        ELSIF (category = NoPasswordNeeded) OR (category = GuestUser) THEN
            success := TRUE;

        (* In the current version, the only method we handle is "password". *)

        ELSIF Strings.Equal (methodname, "password") THEN
            INC (pos);
            GetCharString (pkt, pos, password);
            success := (user <> NilUser) AND PasswordOK (user, password);
            IF NOT success THEN
                Strings.Assign ("Auth failure, username=", message);
                Strings.Append (username, message);
                Strings.Append (", password=", message);
                Strings.Append (password, message);
                LogTransaction (logID, message);
                DestroyUserData (user);
            END (*IF*);
        ELSIF NOT Strings.Equal (methodname, "none") THEN
            message := "Unsupported authentication method ";
            Strings.Append (methodname, message);
            LogTransaction (logID, message);
            success := FALSE;
        END (*IF*);
        DiscardBS (pkt);

        (* Send the reply.  *)

        IF success THEN
            MakeBS (reply, 1);
            reply.data^[0] := SSH_MSG_USERAUTH_SUCCESS;
        ELSE
            reply := FailureReplyPacket();
            DestroyUserData (user);
            user := NilUser;
        END (*IF*);
        EVAL(SendPacket (port, reply));

        RETURN user;

    END TryToAuthenticate;

(************************************************************************)
(*                THE AUTHENTICATION SESSION TASK CODE                  *)
(************************************************************************)

TYPE
    (* Data used in creating a new instance of the session handler task. *)

    NewSessionPointer = POINTER TO
                           RECORD
                               port: IOPort;
                               queue: PacketQ;
                               logID: TransactionLogID;
                           END (*RECORD*);

(************************************************************************)

PROCEDURE SessionHandler (arg: ADDRESS);

    (* The server task that handles authentication.   *)

    CONST
        NilQ = CAST(PacketQ, NIL);

    VAR NSP: NewSessionPointer;
        ID: TransactionLogID;
        Q, upQ: PacketQ;
        port: IOPort;
        pkt: ByteStr;
        user: User;
        failures: CARDINAL;
        opcode: CARD8;
        authenticated, forceshutdown: BOOLEAN;
        message: ARRAY [0..63] OF CHAR;

    BEGIN                   (* Body of SessionHandler *)
        NSP := arg;
        Q := NSP^.queue;
        port := NSP^.port;
        ID := NSP^.logID;
        DISPOSE (NSP);
        authenticated := FALSE;
        user := NIL;
        upQ := NilQ;
        failures := 0;
        forceshutdown := FALSE;

        LOOP
            pkt := GetPacket(Q);
            opcode := pkt.data^[0];
                        (*
                        Strings.Assign ("AuthLayer received packet type ", message);
                        AppendCard (opcode, message);
                        LogTransaction (ID, message);
                        *)
            IF opcode = SSH_MSG_USERAUTH_REQUEST THEN
                IF NOT authenticated THEN
                    user := TryToAuthenticate (port, pkt, ID);
                    authenticated := user <> NilUser;
                    IF authenticated THEN
                        NoteLoginTime (user);
                    ELSE
                        INC (failures);
                        IF failures >= 3 THEN
                            forceshutdown := TRUE;
                            EXIT (*LOOP*);
                        END (*IF*);
                    END (*IF*);
                END (*IF*);
            ELSIF opcode = SSH_MSG_TERMINATE THEN
                EXIT (*LOOP*);
            ELSIF opcode >= 80 THEN
                IF authenticated THEN

                    (* Kick this packet upstairs. *)

                    IF upQ = NilQ THEN
                        (*LogTransactionL (ID, "Have created new queue");*)
                        upQ := CreateQueue();
                        IF NOT Channels.NewSession (port, upQ, user, ID) THEN
                            LogTransactionL (ID, "Cannot start connection protocol");
                        END (*IF*);
                    END (*IF*);
                    PutPacket (upQ, pkt);
                ELSE
                    (* User not authenticated, refuse request. *)

                    DiscardBS (pkt);
                    pkt := FailureReplyPacket();
                    EVAL(SendPacket (port, pkt));
                END (*IF*);
            ELSE
                Strings.Assign ("Cannot currently handle packet type ", message);
                AppendCard (opcode, message);
                LogTransaction (ID, message);
                DiscardBS (pkt);
            END (*IF*);
        END (*LOOP*);

        (* We can reach this point in only two ways. If we've exited the    *)
        (* above loop because of an SSH_MSG_TERMINATE message, then the     *)
        (* transport layer is already shutting down. If instead we have had *)
        (* too many login attempts, then we are triggering the shutdown     *)
        (* ourselves.                                                       *)

        IF forceshutdown THEN
            ClosePort (port);
        ELSE
            (* Pass the termination message upstairs, and tidy up our       *)
            (* own layer.                                                   *)

            IF upQ = NilQ THEN
                DiscardBS (pkt);
            ELSE
                PacketQueues.PutPacket (upQ, pkt);
            END (*IF*);
            PacketQueues.CloseQueue (Q);

        END (*IF*);

    END SessionHandler;

(************************************************************************)

PROCEDURE NewSession (sshport: IOPort;  Q: PacketQ;  ID: TransactionLogID): BOOLEAN;

    (* Starts and runs a session to serve one client.  The session runs *)
    (* in a separate thread; this procedure returns after starting the  *)
    (* session, it does not wait until the session is over.             *)

    VAR NSP: NewSessionPointer;  success: BOOLEAN;

    BEGIN
        NEW (NSP);
        WITH NSP^ DO
            port := sshport;  queue := Q;  logID := ID;
        END (*WITH*);
        success := CreateTask1 (SessionHandler, 3, "SSH Auth", NSP);
        IF NOT success THEN
            DISPOSE (NSP);
        END (*IF*);
        RETURN success;
    END NewSession;

(************************************************************************)

END AuthLayer.

