(**************************************************************************)
(*                                                                        *)
(*  SSH Connection Protocol                                               *)
(*  Copyright (C) 2022   Peter Moylan                                     *)
(*                                                                        *)
(*  This program is free software: you can redistribute it and/or modify  *)
(*  it under the terms of the GNU General Public License as published by  *)
(*  the Free Software Foundation, either version 3 of the License, or     *)
(*  (at your option) any later version.                                   *)
(*                                                                        *)
(*  This program is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU General Public License for more details.                          *)
(*                                                                        *)
(*  You should have received a copy of the GNU General Public License     *)
(*  along with this program.  If not, see <http://www.gnu.org/licenses/>. *)
(*                                                                        *)
(*  To contact author:   http://www.pmoylan.org   peter@pmoylan.org       *)
(*                                                                        *)
(**************************************************************************)

IMPLEMENTATION MODULE Channels;

        (********************************************************)
        (*                                                      *)
        (*               Connection layer of SSH                *)
        (*       This is the server end of the connection       *)
        (*                                                      *)
        (*  Programmer:         P. Moylan                       *)
        (*  Started:            18 July 2021                    *)
        (*  Last edited:        18 April 2022                   *)
        (*  Status:             Working                         *)
        (*                                                      *)
        (********************************************************)


FROM SYSTEM IMPORT ADDRESS, CARD8, CAST;

IMPORT Strings, PacketQueues;

FROM FTPSession IMPORT
    (* proc *)  StartFTPSession;

FROM TransportLayer IMPORT
    (* proc *)  ClosePort;

FROM FDUsers IMPORT
    (* type *)  User;

FROM sshApps IMPORT
    (* type *)  SessionData,
    (* proc *)  ParseChannelRequest, DiscardSessionData;

FROM PacketIO IMPORT
    (* type *)  IOPort,
    (* proc *)  SendPacket;

FROM PacketQueues IMPORT
    (* type *)  PacketQ,
    (* proc *)  CreateQueue, GetPacket;

FROM Consts IMPORT
    (* const*)  SSH_MSG_CHANNEL_OPEN, SSH_MSG_CHANNEL_OPEN_CONFIRMATION, SSH_MSG_CHANNEL_OPEN_FAILURE,
                SSH_MSG_CHANNEL_WINDOW_ADJUST, SSH_MSG_CHANNEL_DATA, SSH_MSG_CHANNEL_EXTENDED_DATA,
                SSH_MSG_CHANNEL_EOF, SSH_MSG_CHANNEL_CLOSE,
                SSH_MSG_CHANNEL_REQUEST, SSH_MSG_CHANNEL_SUCCESS, SSH_MSG_CHANNEL_FAILURE,
                SSH_MSG_TERMINATE;

FROM VarStrings IMPORT
    (* type *)  ByteStr,
    (* proc *)  MakeBS, DiscardBS, CopyOfBS, BSDelete, GetCharString, GetCard, AddCard,
                AddByteString, AddCharString;

FROM MiscFuncs IMPORT
    (* proc *)  AppendCard, EVAL;

FROM TransLog IMPORT
    (* type *)  TransactionLogID,
    (* proc *)  LogTransaction, LogTransactionL;

FROM TaskControl IMPORT
    (* proc *)  CreateTask1;

FROM LowLevel IMPORT
    (* proc *)  Copy;

FROM Storage IMPORT
    (* proc *)  ALLOCATE, DEALLOCATE;

(************************************************************************)

CONST
    SSH_OPEN_ADMINISTRATIVELY_PROHIBITED    = 1;
    SSH_OPEN_CONNECT_FAILED                 = 2;
    SSH_OPEN_UNKNOWN_CHANNEL_TYPE           = 3;
    SSH_OPEN_RESOURCE_SHORTAGE              = 4;

(************************************************************************)
(*                  THE POOL OF AVAILABLE CHANNELS                      *)
(************************************************************************)

CONST
    MaxChannelNo = 256;
    NotAChannel = MAX(CARDINAL);
    NilQ = CAST (PacketQ, NIL);

TYPE
    (* prevfree and nextfree form a linked list of available channels.  *)

    ChannelInfoType =   RECORD
                            prevfree, nextfree: CARDINAL;
                            peer: CARDINAL;
                            closingphase: CARDINAL;
                            port: IOPort;
                            SD: SessionData;
                            upQ: PacketQ;
                        END (*RECORD*);

    (* Remark: the SD field is never used, so we can probably simplify  *)
    (* or eliminate module sshApps.                                     *)

VAR
    (* Head and tail of list of available channels. *)

    FreeChans:  RECORD
                    head, tail: CARDINAL;
                END (*RECORD*);

    (* The status of all channels. *)

    ChannelStatus: ARRAY [1..MaxChannelNo] OF ChannelInfoType;

(************************************************************************)

PROCEDURE ClearChannelData;

    (* Initialises the ChannelStatus array.  *)

    VAR chan: CARDINAL;

    BEGIN
        FOR chan := 1 TO MaxChannelNo DO
            WITH ChannelStatus[chan] DO
                prevfree := chan - 1;
                nextfree := chan + 1;
                peer := NotAChannel;
                port := NIL;
                closingphase := 0;
                SD := NIL;
                upQ := NIL;
            END (*WITH*);
        END (*FOR*);
        ChannelStatus[MaxChannelNo].nextfree := 0;
        FreeChans.head := 1;
        FreeChans.tail := MaxChannelNo;
    END ClearChannelData;

(************************************************************************)

PROCEDURE ObtainChannel(): CARDINAL;

    (* Takes a channel from the free list, or returns 0 if no channel available. *)

    VAR chan: CARDINAL;

    BEGIN
        chan := FreeChans.head;
        IF chan <> 0 THEN
            FreeChans.head := ChannelStatus[chan].nextfree;
            ChannelStatus[chan].nextfree := 0;
            ChannelStatus[chan].prevfree := 0;
        END (*IF*);
        IF FreeChans.head = 0 THEN
            FreeChans.tail := 0;
        END (*IF*);
        ChannelStatus[chan].closingphase := 0;
        RETURN chan;
    END ObtainChannel;

(************************************************************************)

PROCEDURE ReleaseChannel (chan: CARDINAL);

    (* Returns a channel to the free list. *)

    BEGIN
        (* Beware of duplicate release. *)

        IF ChannelStatus[chan].peer <> NotAChannel THEN
            ChannelStatus[chan].peer := NotAChannel;
            ChannelStatus[chan].port := NIL;
            ChannelStatus[chan].closingphase := 0;
            ChannelStatus[chan].SD := NIL;
            ChannelStatus[chan].upQ := NIL;

            (* Put this channel on the head of the free list. *)

            IF FreeChans.head = 0 THEN
                FreeChans.tail := chan;
            END (*IF*);
            ChannelStatus[chan].prevfree := 0;
            ChannelStatus[chan].nextfree := FreeChans.head;
            FreeChans.head := chan;
        END (*IF*);

    END ReleaseChannel;

(************************************************************************)

PROCEDURE PortAlive (port: IOPort): BOOLEAN;

    (* Returns TRUE iff there is still a channel using this port.   *)

    VAR j: CARDINAL;

    BEGIN
        j := 1;
        LOOP
            IF ChannelStatus[j].port = port THEN
                RETURN TRUE;
            ELSIF j = MaxChannelNo THEN
                RETURN FALSE;
            END (*IF*);
            INC (j);
        END (*LOOP*);
    END PortAlive;

(************************************************************************)

PROCEDURE FirstChan (port: IOPort): CARDINAL;

    (* Returns the channel number of a channel using this port,     *)
    (* or 0 if there is no such channel.                            *)

    VAR j: CARDINAL;

    BEGIN
        j := 1;
        LOOP
            IF ChannelStatus[j].port = port THEN
                RETURN j;
            ELSIF j = MaxChannelNo THEN
                RETURN 0;
            END (*IF*);
            INC (j);
        END (*LOOP*);
    END FirstChan;

(************************************************************************)
(*                   OPENING AND CLOSING A CHANNEL                      *)
(************************************************************************)

PROCEDURE OpenAChannel (port: IOPort;  pkt: ByteStr;  logID: TransactionLogID);

    VAR pos: CARDINAL;
        ourchannelno, peerchannelno, windowsize, packetsize: CARDINAL;
        success: BOOLEAN;
        reply: ByteStr;
        channeltype, message, reason: ARRAY [0..63] OF CHAR;

    BEGIN
        pos := 1;

        (* Extract the request parameters from the packet.  *)

        GetCharString (pkt, pos, channeltype);
        peerchannelno := GetCard (pkt, pos);
        windowsize := GetCard (pkt, pos);
        packetsize := GetCard (pkt, pos);
        DiscardBS (pkt);
        ourchannelno := ObtainChannel();
        success := ourchannelno <> 0;

        (* In the current version, we are not yet storing the window    *)
        (* size or max packet size. *)

        IF success THEN
            Strings.Assign ("Opening channel ", message);
            AppendCard (ourchannelno, message);
            Strings.Append (", type ", message);
            Strings.Append (channeltype, message);
            LogTransaction (logID, message);
            Strings.Assign ("  peer channel # = ", message);
            AppendCard (peerchannelno, message);
            LogTransaction (logID, message);
            Strings.Assign ("  window size = ", message);
            AppendCard (windowsize, message);
            LogTransaction (logID, message);
            Strings.Assign ("  max packet size = ", message);
            AppendCard (packetsize, message);
            LogTransaction (logID, message);

            IF ChannelStatus[ourchannelno].upQ = NilQ THEN
                ChannelStatus[ourchannelno].upQ := CreateQueue();
            END (*IF*);

        ELSE
            LogTransactionL (logID, "Run out of free channels");
        END (*IF*);

        (* Send the reply.  *)

        IF success THEN
            ChannelStatus[ourchannelno].peer := peerchannelno;
            ChannelStatus[ourchannelno].port := port;
            MakeBS (reply, 17);
            reply.data^[0] := SSH_MSG_CHANNEL_OPEN_CONFIRMATION;
            pos := 1;
            AddCard (reply, pos, peerchannelno);
            AddCard (reply, pos, ourchannelno);
            AddCard (reply, pos, windowsize);
            AddCard (reply, pos, packetsize);
        ELSE
            reason := "no available free channels";
            MakeBS (reply, 9 + 4 + LENGTH(reason) + 6);
            reply.data^[0] := SSH_MSG_CHANNEL_OPEN_FAILURE;
            pos := 1;
            AddCard (reply, pos, peerchannelno);
            AddCard (reply, pos, SSH_OPEN_RESOURCE_SHORTAGE);
            AddCharString (reply, pos, reason);
            AddCharString (reply, pos, "en");
            DiscardBS (pkt);
        END (*IF*);
        EVAL(SendPacket (port, reply));

    END OpenAChannel;

(************************************************************************)

PROCEDURE CloseChannel (channel: CARDINAL);

    (* Closes the specified channel. This procedure is either called by *)
    (* another module, to initiate the closing of a channel, or after   *)
    (* receiving a close request from the peer.                         *)

    (* In the SFTP case, the close request always comes from the client *)
    (* and never from the server. To keep the code clean, though, we    *)
    (* want to cover both possibilities. In both cases, a channel close *)
    (* message is sent in both directions.                              *)

    (* The required response to a "close" message is another "close"    *)
    (* message. That is, there should be exactly one close message      *)
    (* from each end. Our closingphase variable counts how many have    *)
    (* been sent or received. When the count gets to 2, we know we can  *)
    (* proceed to shutting down the channel.                            *)

    VAR pos: CARDINAL;
        port: IOPort;
        reply: ByteStr;

    BEGIN
        port := ChannelStatus[channel].port;

        (* We have to send a close message to the peer, unless we have  *)
        (* already done so.                                             *)

        IF ChannelStatus[channel].closingphase < 2 THEN
            MakeBS (reply, 5);
            reply.data^[0] := SSH_MSG_CHANNEL_CLOSE;
            pos := 1;
            AddCard (reply, pos, ChannelStatus[channel].peer);
            EVAL(SendPacket (port, reply));
            INC (ChannelStatus[channel].closingphase);
        END (*IF*);

        IF ChannelStatus[channel].closingphase = 2 THEN

            (* We and the peer have both sent close messages. If this   *)
            (* is the last channel using the same port, shut down the   *)
            (* port. The connection layer will respond with an          *)
            (* SSH_MSG_TERMINATE message, at which point we can shut    *)
            (* down all channels using that port.                       *)

            ChannelStatus[channel].port := NIL;
            IF NOT PortAlive(port) THEN
                ClosePort (port);
            END (*IF*);
            ChannelStatus[channel].port := port;

            (* We cannot yet release this channel, because we will  *)
            (* still need the upQ field of the channel record.      *)

        END (*IF*);

    END CloseChannel;

(************************************************************************)

PROCEDURE AcceptEOF (pkt: ByteStr);

    VAR pos, ourchan: CARDINAL;

    BEGIN
        pos := 1;
        ourchan := GetCard (pkt, pos);
        DiscardBS (pkt);

        (* Nothing more to do here, but in a later version we will  *)
        (* probably have to hand the EOF on to an application.      *)
    END AcceptEOF;

(************************************************************************)
(*                      CLOSE REQUEST FROM PEER                         *)
(************************************************************************)

PROCEDURE HandleCloseRequest (pkt: ByteStr);

    (* Deals with a CHANNEL_CLOSE message from the peer. *)

    VAR pos, ourchan: CARDINAL;

    BEGIN
        pos := 1;
        ourchan := GetCard (pkt, pos);
        DiscardBS (pkt);
        INC (ChannelStatus[ourchan].closingphase);
        CloseChannel (ourchan);
    END HandleCloseRequest;

(************************************************************************)
(*                          CHANNEL REQUEST                             *)
(************************************************************************)

PROCEDURE HandleChannelRequest (port: IOPort;  pkt: ByteStr;
                                    U: User;  logID: TransactionLogID);

    (* Handles SSH_MSG_CHANNEL_REQUEST messages. It is not yet clear to *)
    (* me which of these need to be handled here and which need to be   *)
    (* handled elsewhere - see module sshApps - so this code is still   *)
    (* in a state of flux.                                              *)

    VAR pos: CARDINAL;
        ourchan, peerchan: CARDINAL;
        wantreply, success: BOOLEAN;
        reply: ByteStr;
        reqtype, subsystem: ARRAY [0..63] OF CHAR;
        message: ARRAY [0..63] OF CHAR;

    BEGIN
        pos := 1;

        (* Extract the request parameters from the packet.  *)

        ourchan := GetCard (pkt, pos);
        peerchan := ChannelStatus[ourchan].peer;
        message := "CHANNEL_REQUEST, channel ";
        AppendCard (ourchan, message);
        Strings.Append (", peer channel ", message);
        AppendCard (peerchan, message);
        LogTransaction (logID, message);

        (* Pick up the next parameters, which we might or might not need. *)

        GetCharString (pkt, pos, reqtype);
        wantreply := pkt.data^[pos] <> 0;  INC(pos);

        (*
        message := "Channel request type ";
        Strings.Append (reqtype, message);
        LogTransaction (logID, message);
        *)

        IF Strings.Equal (reqtype, "subsystem") THEN
            GetCharString (pkt, pos, subsystem);
            (*
            message := "Request for subsystem ";  Strings.Append (subsystem, message);
            LogTransactionL (logID, message);
            *)
            IF Strings.Equal (subsystem, "sftp") THEN
                IF ChannelStatus[ourchan].upQ = NilQ THEN
                    ChannelStatus[ourchan].upQ := CreateQueue();
                END (*IF*);
                success := StartFTPSession (ourchan, ChannelStatus[ourchan].upQ, U, logID);
            ELSE
                success := FALSE;
            END (*IF*);
        ELSE
            success := ParseChannelRequest (ChannelStatus[ourchan].SD, pkt, port, logID);
        END (*IF*);

        (* Send the reply if one is wanted.  *)

        IF wantreply THEN
            MakeBS (reply, 5);
            IF success THEN
                reply.data^[0] := SSH_MSG_CHANNEL_SUCCESS;
            ELSE
                reply.data^[0] := SSH_MSG_CHANNEL_FAILURE;
            END (*IF*);
            pos := 1;
            AddCard (reply, pos, peerchan);
            EVAL(SendPacket (port, reply));
        END (*IF*);

    END HandleChannelRequest;

(************************************************************************)
(*                           DATA TRANSFER                              *)
(************************************************************************)

PROCEDURE ReceiveData (pkt: ByteStr;  ID: TransactionLogID);

    (* Accepts data from peer. *)

    VAR ourchan, pos: CARDINAL;
        message: ARRAY [0..63] OF CHAR;

    BEGIN
        pos := 1;
        ourchan := GetCard (pkt, pos);
        message := "CHANNEL_DATA, channel ";
        AppendCard (ourchan, message);
        LogTransaction (ID, message);

        (* The first 9 bytes are SSH data rather than FTP packet data. *)

        BSDelete (pkt, 0, 9);

        (* Send the rest upstairs. *);

        PacketQueues.PutPacket (ChannelStatus[ourchan].upQ, pkt);

    END ReceiveData;

(************************************************************************)

PROCEDURE SendData (channel: CARDINAL;  pkt: ByteStr);

    (* Sends a data packet to peer. *)

    VAR pos: CARDINAL;
        tosend: ByteStr;

    BEGIN
        MakeBS (tosend, 9 + pkt.size);
        tosend.data^[0] := SSH_MSG_CHANNEL_DATA;
        pos := 1;
        AddCard (tosend, pos, ChannelStatus[channel].peer);
        AddByteString (tosend, pos, pkt.data^, pkt.size);
        DiscardBS (pkt);
        EVAL(SendPacket (ChannelStatus[channel].port, tosend));
    END SendData;

(************************************************************************)
(*                   THE SESSION COMMAND DISPATCHER                     *)
(************************************************************************)

TYPE
    (* Data used in creating a new instance of the session handler task. *)

    NewSessionPointer = POINTER TO
                           RECORD
                               ourport: IOPort;
                               queue: PacketQ;
                               userID: User;
                               logID: TransactionLogID;
                           END (*RECORD*);

(************************************************************************)

PROCEDURE SessionHandler (arg: ADDRESS);

    (* The server task that handles connection protocol requests.   *)

    CONST
        NilQ = CAST(PacketQ, NIL);

    VAR NSP: NewSessionPointer;
        ID: TransactionLogID;
        Q, upQ: PacketQ;
        port: IOPort;
        user: User;
        pkt, pktterm: ByteStr;
        channel: CARDINAL;
        opcode: CARD8;
        message: ARRAY [0..63] OF CHAR;

    BEGIN                   (* Body of SessionHandler *)
        NSP := arg;
        Q := NSP^.queue;
        port := NSP^.ourport;
        user := NSP^.userID;
        ID := NSP^.logID;
        DISPOSE (NSP);

        LOOP
            pkt := GetPacket(Q);
            opcode := pkt.data^[0];
                        (*
                        Strings.Assign ("Connection layer received packet type ", message);
                        AppendCard (opcode, message);
                        LogTransaction (ID, message);
                        *)
            IF opcode = SSH_MSG_CHANNEL_OPEN THEN
                OpenAChannel (port, pkt, ID);
            ELSIF opcode = SSH_MSG_CHANNEL_REQUEST THEN
                HandleChannelRequest (port, pkt, user, ID);
            ELSIF opcode = SSH_MSG_CHANNEL_DATA THEN
                ReceiveData (pkt, ID);
            ELSIF opcode = SSH_MSG_CHANNEL_EOF THEN
                AcceptEOF (pkt);
            ELSIF opcode = SSH_MSG_CHANNEL_CLOSE THEN
                HandleCloseRequest (pkt);
            ELSIF opcode = SSH_MSG_TERMINATE THEN
                EXIT (*LOOP*);
            ELSE
                Strings.Assign ("Channels cannot currently handle packet type ", message);
                AppendCard (opcode, message);
                LogTransaction (ID, message);
                DiscardBS (pkt);
            END (*IF*);
        END (*LOOP*);

        (* An SSH_MSG_TERMINATE means that the transport layer is       *)
        (* closing down, so we should clean up our layer.               *)

        DiscardBS (pkt);
        MakeBS(pktterm, 5);
        pktterm.data^[0] := 1;  pktterm.data^[1] := 0;
        pktterm.data^[2] := 0;  pktterm.data^[3] := 0;
        pktterm.data^[4] := SSH_MSG_TERMINATE;

        (* In principle there can be several channels using the same    *)
        (* port.  Shut them all down.                                   *)

        LOOP
            channel := FirstChan (port);
            IF channel = 0 THEN EXIT(*LOOP*) END(*IF*);
            upQ := ChannelStatus[channel].upQ;

            (* Send the termination message upstairs. *)

            IF upQ <> NilQ THEN
                pkt := CopyOfBS (pktterm);
                (*
                Strings.Assign ("Putting a termination packet, code ", message);
                AppendCard (pkt.data^[4], message);
                LogTransaction (ID, message);
                *)
                PacketQueues.PutPacket (upQ, pkt);
            END (*IF*);

            (* Complete the closedown of the channel. *)

            DiscardSessionData (ChannelStatus[channel].SD);
            ReleaseChannel (channel);

        END (*LOOP*);
        DiscardBS (pktterm);

        PacketQueues.CloseQueue (Q);
        (*LogTransactionL (ID, "Exiting channel layer");*)

    END SessionHandler;

(************************************************************************)

PROCEDURE NewSession (port: IOPort;  Q: PacketQ;  user: User;
                                        ID: TransactionLogID): BOOLEAN;

    (* Starts and runs a session to serve one client.  The session runs *)
    (* in a separate thread; this procedure returns after starting the  *)
    (* session, it does not wait until the session is over.             *)

    VAR NSP: NewSessionPointer;  success: BOOLEAN;

    BEGIN
        NEW (NSP);
        WITH NSP^ DO
            ourport := port;  queue := Q;  userID := user;  logID := ID;
        END (*WITH*);
        success := CreateTask1 (SessionHandler, 3, "SSH Channels", NSP);
        IF NOT success THEN
            DISPOSE (NSP);
        END (*IF*);
        RETURN success;
    END NewSession;

(************************************************************************)

BEGIN
    ClearChannelData;
END Channels.

