(**************************************************************************)
(*                                                                        *)
(*  My own version of SSH                                                 *)
(*  Copyright (C) 2021   Peter Moylan                                     *)
(*                                                                        *)
(*  This program is free software: you can redistribute it and/or modify  *)
(*  it under the terms of the GNU General Public License as published by  *)
(*  the Free Software Foundation, either version 3 of the License, or     *)
(*  (at your option) any later version.                                   *)
(*                                                                        *)
(*  This program is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU General Public License for more details.                          *)
(*                                                                        *)
(*  You should have received a copy of the GNU General Public License     *)
(*  along with this program.  If not, see <http://www.gnu.org/licenses/>. *)
(*                                                                        *)
(*  To contact author:   http://www.pmoylan.org   peter@pmoylan.org       *)
(*                                                                        *)
(**************************************************************************)

IMPLEMENTATION MODULE Compression;

        (********************************************************)
        (*                                                      *)
        (*                   Data compression                   *)
        (*                                                      *)
        (*  Programmer:         P. Moylan                       *)
        (*  Started:            6 January 2021                  *)
        (*  Last edited:        24 May 2021                     *)
        (*  Status:             OK, but no compression          *)
        (*                        algorithms implemented yet    *)
        (*                                                      *)
        (********************************************************)


IMPORT Strings;

FROM NameLists IMPORT
    (* proc *)  ListAll, MatchNames;

FROM VarStrings IMPORT
    (* type *)  ByteStr;

(************************************************************************)

TYPE
    NameArray = ARRAY CompressionType OF ARRAY [0..15] OF CHAR;

CONST
    Names = NameArray {"none"};

(************************************************************************)
(*                IDENTIFIERS FOR COMPRESSION ALGORITHMS                *)
(************************************************************************)

PROCEDURE NameOf (ct: CompressionType;  VAR (*OUT*) name: ARRAY OF CHAR);

    (* Returns the RFC 4253 name of mt. *)

    BEGIN
        Strings.Assign (Names[ct], name);
    END NameOf;

(************************************************************************)

PROCEDURE AllNames (VAR (*OUT*) list: ARRAY OF CHAR);

    (* Returns a comma-separated list of our algorithm names.  *)

    CONST max = ORD(MAX(CompressionType));

    BEGIN
        ListAll (Names, max, list);
    END AllNames;

(************************************************************************)

PROCEDURE FindMatch (VAR (*IN*) list: ARRAY OF CHAR): CompressionType;

    (* Input is a comma-separated list of algorithm names.  *)
    (* Returns the first one we support.                    *)

    CONST max = ORD(MAX(CompressionType));

    VAR result: CARDINAL;

    BEGIN
        result := MatchNames (Names, max, list);
        IF result > max THEN RETURN none;
        ELSE RETURN VAL(CompressionType, result);
        END (*IF*);
    END FindMatch;

(********************************************************************************)
(*                                  COMPRESSION                                 *)
(********************************************************************************)

PROCEDURE Compress (ct: CompressionType;
                       VAR (*IN*) indata: ByteStr;  VAR (*OUT*) outdata: ByteStr);

    (* Compressed result is outdata, indata is discarded. *)

    BEGIN
        IF ct = none THEN
            outdata := indata;
            indata.allocated := 0;
            indata.size := 0;
            indata.data := NIL;
        ELSE
            (* Unavailable compression type. Can this ever happen? *)
        END (*IF*);
    END Compress;

(********************************************************************************)
(*                                 DECOMPRESSION                                *)
(********************************************************************************)

PROCEDURE Decompress (ct: CompressionType;
                       VAR (*IN*) indata: ByteStr;  VAR (*OUT*) outdata: ByteStr);

    (* Compressed result is outdata, indata is discarded. *)

    BEGIN
        IF ct = none THEN
            outdata := indata;
            indata.allocated := 0;
            indata.size := 0;
            indata.data := NIL;
        ELSE
            (* Unavailable compression type. Can this ever happen? *)
        END (*IF*);
    END Decompress;

(********************************************************************************)

END Compression.

