(**************************************************************************)
(*                                                                        *)
(*  My own version of SSH                                                 *)
(*  Copyright (C) 2021   Peter Moylan                                     *)
(*                                                                        *)
(*  This program is free software: you can redistribute it and/or modify  *)
(*  it under the terms of the GNU General Public License as published by  *)
(*  the Free Software Foundation, either version 3 of the License, or     *)
(*  (at your option) any later version.                                   *)
(*                                                                        *)
(*  This program is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU General Public License for more details.                          *)
(*                                                                        *)
(*  You should have received a copy of the GNU General Public License     *)
(*  along with this program.  If not, see <http://www.gnu.org/licenses/>. *)
(*                                                                        *)
(*  To contact author:   http://www.pmoylan.org   peter@pmoylan.org       *)
(*                                                                        *)
(**************************************************************************)

IMPLEMENTATION MODULE Encryption;

        (********************************************************)
        (*                                                      *)
        (*                   Data encryption                    *)
        (*                                                      *)
        (*  Programmer:         P. Moylan                       *)
        (*  Started:            6 January 2021                  *)
        (*  Last edited:        11 October 2021                 *)
        (*  Status:             Done, working                   *)
        (*                                                      *)
        (********************************************************)


IMPORT Strings, AES, DES;

FROM SYSTEM IMPORT
    (* type *)  LOC, CARD8, CARD32,
    (* proc *)  ADR;

FROM NameLists IMPORT
    (* proc *)  MatchNames, ListAll;

FROM VarStrings IMPORT
    (* type *)  ByteStringPtr;

FROM LowLevel IMPORT
    (* proc *)  Copy;

FROM Storage IMPORT
    (* proc *)  ALLOCATE;

(********************************************************************************)

TYPE
    CardArray = ARRAY CipherType OF CARDINAL;
    NameArray = ARRAY CipherType OF ARRAY [0..15] OF CHAR;

CONST
    BlkSiz = CardArray {16, 16, 16, 16, 16, 16, 8, 1};
    Keylength = CardArray {16, 24, 32, 16, 24, 32, 24, 1};
    IVlength = CardArray {16, 16, 16, 16, 16, 16, 8, 1};
    Names = NameArray {"aes128-ctr", "aes192-ctr", "aes256-ctr",
                        "aes128-cbc", "aes192-cbc", "aes256-cbc", "3des-cbc", "none"};

TYPE
    (* Note: a CipherInfo record is good for only one transmission direction.   *)
    (* A full connection required two such records.                             *)

    CipherInfo = POINTER TO
                    RECORD
                        cipher: CipherType;
                        incoming: BOOLEAN;
                        AESctx: AES.AEScontext;
                        DESctx: DES.DEScontext;
                    END (*RECORD*);

(********************************************************************************)
(*                      SOME PROPERTIES OF THE CIPHERS                          *)
(********************************************************************************)

PROCEDURE BlockSize (et: CipherType): CARDINAL;

    (* Returns the block size for the specified cipher. *)

    BEGIN
        RETURN BlkSiz[et];
    END BlockSize;

(************************************************************************)

PROCEDURE KeySize (et: CipherType): CARDINAL;

    (* Returns the number of bytes needed for the encryption key.    *)

    BEGIN
        RETURN Keylength[et];
    END KeySize;

(************************************************************************)

PROCEDURE IVSize (et: CipherType): CARDINAL;

    (* Returns the number of bytes needed for the initialisation        *)
    (* vector, including nonce if required.                             *)

    BEGIN
        RETURN IVlength[et];
    END IVSize;

(************************************************************************)

PROCEDURE NameOf (et: CipherType;  VAR (*OUT*) name: ARRAY OF CHAR);

    (* Returns the RFC 4253 name of mt. *)

    BEGIN
        Strings.Assign (Names[et], name);
    END NameOf;

(************************************************************************)

PROCEDURE AllNames (VAR (*OUT*) list: ARRAY OF CHAR);

    (* Returns a comma-separated list of our algorithm names.  *)

    CONST max = ORD(MAX(CipherType));

    BEGIN
        ListAll (Names, max, list);
    END AllNames;

(************************************************************************)

PROCEDURE FindMatch (VAR (*IN*) list: ARRAY OF CHAR): CipherType;

    (* Input is a comma-separated list of algorithm names.  *)
    (* Returns the first one we support.                    *)

    CONST max = ORD(MAX(CipherType));

    VAR result: CARDINAL;

    BEGIN
        result := MatchNames (Names, max, list);
        IF result > max THEN RETURN none;
        ELSE RETURN VAL(CipherType, result);
        END (*IF*);
    END FindMatch;

(********************************************************************************)
(*                   OPENING AND CLOSING A CipherInfo RECORD                    *)
(********************************************************************************)

PROCEDURE InitCipherInfo (incoming: BOOLEAN): CipherInfo;

    (* Creates a record to be used for the whole session. *)

    VAR result: CipherInfo;

    BEGIN
        NEW (result);
        result^.cipher := none;
        result^.incoming := incoming;
        RETURN result;
    END InitCipherInfo;

(********************************************************************************)

PROCEDURE CloseCipherInfo (VAR (*INOUT*) info: CipherInfo);

    BEGIN
        DISPOSE (info);
    END CloseCipherInfo;

(********************************************************************************)
(*                         CHANGING CIPHER ALGORITHMS                           *)
(********************************************************************************)

PROCEDURE StopEncryption (CI: CipherInfo);

    VAR output: ARRAY [0..31] OF CARD8;
        outpos: CARDINAL;

    (* Remark: in principle this operation could produce some final output, *)
    (* which we receive and discard. But there should be no partial block   *)
    (* left, because an SSH packet size is an integer multiple of the       *)
    (* encryption block size.                                               *)

    BEGIN
        outpos := 0;
        CASE CI^.cipher OF
            |   aes128ctr, aes192ctr, aes256ctr,
                 aes128cbc, aes192cbc, aes256cbc:
                    AES.AESfinal (CI^.AESctx, output, outpos);
            |   des3cbc:
                    DES.DESfinal (CI^.DESctx, output, outpos);
            |   none:
                    (* do nothing *)
        END (*CASE*);
    END StopEncryption;

(********************************************************************************)

PROCEDURE ChangeEncryption (CI: CipherInfo;  newcipher: CipherType;
                                    VAR (*IN*) key, IV: ARRAY OF CARD8);

    (* Switches to using a new cipher.  This needs to be called twice,      *)
    (* once for each direction.                                             *)

    (* Note that RFC 4344 defines the CTR modes differently from the        *)
    (* original AES-CTR mode in RFC 3686, and in particular there is no     *)
    (* nonce.  I have now changed the code to use the RFC 4344 version.     *)

    VAR L: CARDINAL;

    BEGIN
        StopEncryption (CI);
        CI^.cipher := newcipher;
        L := 8*Keylength[newcipher];
        CASE newcipher OF
            |   aes128ctr, aes192ctr, aes256ctr:
                    CI^.AESctx := AES.AES_SDCTRopen (L, key, IV);
            |   aes128cbc, aes192cbc, aes256cbc:
                    CI^.AESctx := AES.AES_CBCopen (CI^.incoming, FALSE, L, key, IV);
            |   des3cbc:
                    CI^.DESctx := DES.DES_CBCopen (CI^.incoming, FALSE, 24, key, IV);
            |   none:
                    (* do nothing *)
        END (*CASE*);
    END ChangeEncryption;

(********************************************************************************)
(*                                   ENCRYPTION                                 *)
(********************************************************************************)

PROCEDURE Encrypt (CI: CipherInfo;  amount: CARDINAL;
                            VAR indata, outdata: ARRAY OF CARD8);

    (* Encrypted result is outdata.  We assume that the caller has allocated    *)
    (* enough space for outdata.                                                *)

    VAR pos: CARDINAL;

    BEGIN
        pos := 0;
        CASE CI^.cipher OF
            |   aes128ctr, aes192ctr, aes256ctr, aes128cbc, aes192cbc, aes256cbc:
                    AES.AESprocess (CI^.AESctx, amount, indata, outdata, pos);
            |   des3cbc:
                    DES.DESEncrypt (CI^.DESctx, amount, indata, outdata, pos);
            |   none:
                    Copy (ADR(indata), ADR(outdata), amount);
        END (*CASE*);
    END Encrypt;

(********************************************************************************)
(*                                  DECRYPTION                                  *)
(********************************************************************************)

PROCEDURE Decrypt (CI: CipherInfo;  amount: CARDINAL;
                            VAR indata, outdata: ARRAY OF CARD8);

    (* Decrypted result is outdata.  We assume that the caller has allocated    *)
    (* enough space for outdata.                                                *)

    VAR pos: CARDINAL;

    BEGIN
        pos := 0;
        CASE CI^.cipher OF
            |   aes128ctr, aes192ctr, aes256ctr, aes128cbc, aes192cbc, aes256cbc:
                    AES.AESprocess (CI^.AESctx, amount, indata, outdata, pos);
            |   des3cbc:
                    DES.DESEncrypt (CI^.DESctx, amount, indata, outdata, pos);
            |   none:
                    Copy (ADR(indata), ADR(outdata), amount);
        END (*CASE*);
    END Decrypt;

(********************************************************************************)

END Encryption.

