(**************************************************************************)
(*                                                                        *)
(*  SFTPD server                                                          *)
(*  Copyright (C) 2021   Peter Moylan                                     *)
(*                                                                        *)
(*  This program is free software: you can redistribute it and/or modify  *)
(*  it under the terms of the GNU General Public License as published by  *)
(*  the Free Software Foundation, either version 3 of the License, or     *)
(*  (at your option) any later version.                                   *)
(*                                                                        *)
(*  This program is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU General Public License for more details.                          *)
(*                                                                        *)
(*  You should have received a copy of the GNU General Public License     *)
(*  along with this program.  If not, see <http://www.gnu.org/licenses/>. *)
(*                                                                        *)
(*  To contact author:   http://www.pmoylan.org   peter@pmoylan.org       *)
(*                                                                        *)
(**************************************************************************)

IMPLEMENTATION MODULE FDFiles;

        (********************************************************)
        (*                                                      *)
        (*                    File operations                   *)
        (*                                                      *)
        (*  Programmer:         P. Moylan                       *)
        (*  Started:            15 November 2021                *)
        (*  Last edited:        7 October 2022                  *)
        (*  Status:             OK                              *)
        (*                                                      *)
        (********************************************************)

IMPORT Strings, OS2;

FROM SYSTEM IMPORT CARD16, ADR, LOC;

FROM Arith64 IMPORT
    (* type *)  CARD64,
    (* proc *)  ToAPI64;

FROM FDUsers IMPORT
    (* type *)  User, Permission, FileAttributes, DirListPtr, FName,
    (* proc *)  FindAttributes, CreateDirectoryListing, MakeRealName,
                NameMatch, MakeFName, SameDrive, EmptyDirectory;

FROM FileOps IMPORT
    (* const*)  NoSuchChannel,
    (* type *)  ChanId,
    (* type *)  FilenameString, FileAttribute,
    (* proc *)  OpenGeneral, SetPosition;

FROM VarStrings IMPORT
    (* type *)  ByteStr,
    (* proc *)  MakeBS, AddCard, AddByteChunk, AddCharString,
                GetCard;

FROM MiscFuncs IMPORT
    (* proc *)  ConvertCard64RJ;

FROM SysClock IMPORT
    (* type *)  DateTime,
    (* proc *)  GetClock;

FROM MyClock IMPORT
    (* proc *)  UnpackDateTime, CorrectToGMT, FormatDateTime;

FROM LowLevel IMPORT
    (* proc *)  IAND;

FROM Storage IMPORT
    (* proc *)  ALLOCATE, DEALLOCATE;

(************************************************************************)

CONST
    Nul = CHR(0);
    FileToken = 5555;
    DirToken = 1234567;

TYPE
    FileHandle = POINTER TO
                    RECORD
                        token: CARDINAL;
                        name: FilenameString;
                        CASE directory: BOOLEAN OF
                            FALSE:
                                cid: ChanId;
                          | TRUE:
                                list: DirListPtr;
                        END (*CASE*);
                    END (*RECORD*);

(************************************************************************)
(*              CONVERT FILESYSTEM DATE/TIME TO UNIX FORMAT             *)
(*                Unix format is seconds since 1 Jan 1970               *)
(************************************************************************)

PROCEDURE ConvertDateTime (date, time: CARD16): CARDINAL;

    (* Converts a DOS date/time to seconds since 1 Jan 1970. *)

    TYPE MonthData = ARRAY [1..12] OF CARDINAL;

    CONST
        SecsPerDay = 86400;
        SecsPerYear = 365 * SecsPerDay;
        DaysInMonth = MonthData { 31, 28, 31, 30, 31, 30,
                                  31, 31, 30, 31, 30, 31};

    VAR secs, k: CARDINAL;
        dt: DateTime;

    BEGIN
        secs := 10*SecsPerYear      (* to get to a 1980 base *)
                    + 2*SecsPerDay; (* for two leap years *)

        dt := UnpackDateTime (date, time);
        CorrectToGMT (dt);

        (* Count the seconds from 1 Jan 1980. *)

        FOR k := 1980 TO dt.year-1 DO
            INC (secs, SecsPerYear);
            IF k MOD 4 = 0 THEN
                INC (secs, SecsPerDay);
            END (*IF*);
        END (*FOR*);
        FOR k := 1 TO dt.month-1 DO
            INC (secs, SecsPerDay*DaysInMonth[k]);
            IF (k = 2) AND (dt.year MOD 4 = 0) THEN
                INC (secs, SecsPerDay);
            END (*IF*);
        END (*FOR*);
        FOR k := 1 TO dt.day-1 DO
            INC (secs, SecsPerDay);
        END (*FOR*);

        (* Add the seconds in the last day. *)

        INC (secs, 3600*dt.hour + 60*dt.minute + dt.second);

        RETURN secs;

    END ConvertDateTime;

(************************************************************************)
(*                      SIMPLIFYING A PATH STRING                       *)
(************************************************************************)

PROCEDURE StripTail (VAR (*INOUT*) name: ARRAY OF CHAR): BOOLEAN;

    (* Deletes the last subdirectory specification from name, e.g.      *)
    (* we change A/B/C/D to A/B/C.  If name happens to end with a '/'   *)
    (* then we just remove the '/'.  If there is no '/' anywhere in the *)
    (* name we can't go up in the tree, so we report an error.  If the  *)
    (* found '/' is the first character of name, this too is a special  *)
    (* case.                                                            *)

    VAR pos: CARDINAL;  found: BOOLEAN;

    BEGIN
        IF name[0] <> Nul THEN
            Strings.FindPrev ('/', name, Strings.Length(name)-1, found, pos);
            IF found THEN
                IF pos = 0 THEN
                    IF name[1] = Nul THEN
                        RETURN FALSE;
                    END (*IF*);
                    name[1] := Nul;
                ELSE
                    name[pos] := Nul;
                END (*IF*);
            ELSE
                RETURN FALSE;
            END (*IF*);
        END (*IF*);
        RETURN TRUE;
    END StripTail;

(************************************************************************)

PROCEDURE ReducePath (U: User;  VAR (*INOUT*) path: ARRAY OF CHAR): BOOLEAN;

    (* Canonicalises a path name to an absolute path, by eliminating    *)
    (* '.' and '..' entries.  Returns FALSE if the result is invalid.   *)

    (* SFTP has no concept of "current directory".  Paths are either    *)
    (* absolute or relative to the user's home directory.  In the       *)
    (* FtpServer file system model, the user's home directory is        *)
    (* identical to the root directory, so '.' means the same as '/'.   *)

    VAR result, head: FilenameString;
        pos: CARDINAL;
        haveslash, found: BOOLEAN;

    BEGIN
        (* Initial pass: change every '\' to '/'. *)

        pos := 0;
        REPEAT
            Strings.FindNext ('\', path, pos, found, pos);
            IF found THEN
                path[pos] := '/';
            END (*IF*);
        UNTIL NOT found;

        (* For simplicity, make all results start with '/'.  *)

        result := '/';  haveslash := TRUE;
        IF (path[0] = '.') AND ((path[1] = '/') OR (path[1] = Nul)) THEN
            Strings.Delete (path, 0, 1);
        END (*IF*);
        IF path[0] = '/' THEN
            Strings.Delete (path, 0, 1);
        END (*IF*);

        (* Now process what is left of path. *)

        WHILE path[0] <> Nul DO
            Strings.FindNext ('/', path, 0, found, pos);
            Strings.Assign (path, head);
            IF found THEN
                head[pos] := Nul;
                Strings.Delete (path, 0, pos+1);
            ELSE
                path[0] := Nul;
            END (*IF*);

            IF Strings.Equal (head, '..') THEN
                IF NOT StripTail (result) THEN
                    RETURN FALSE;
                END (*IF*);
                haveslash := result[1] = Nul;
            ELSE
                IF NOT haveslash THEN
                    Strings.Append ('/', result);
                END (*IF*);
                Strings.Append (head, result);
                haveslash := FALSE;
            END (*IF*);
        END (*WHILE*);

        Strings.Assign (result, path);
        RETURN TRUE;

    END ReducePath;

(************************************************************************)
(*                           FILE ATTRIBUTES                            *)
(************************************************************************)

PROCEDURE ParseAttrs (pkt: ByteStr;  VAR (*INOUT*) pos: CARDINAL): AttrRecord;

    (* Collect the requested attributes from a client packet.  *)

    VAR att: AttrRecord;
        flags: CARDINAL;

    BEGIN
        (* Clear all attributes to give a default value. *)

        att.size := CARD64{0,0};
        att.uid := 0;
        att.gid := 0;
        att.permissions := 0;
        att.atime := 0;
        att.mtime := 0;

        flags := GetCard (pkt, pos);
        att.flags := flags;
        IF ODD(flags) THEN
            att.size.high := GetCard (pkt, pos);
            att.size.low := GetCard (pkt, pos);
        END (*IF*);
        flags := flags DIV 2;
        IF ODD(flags) THEN
            att.uid := GetCard (pkt, pos);
            att.gid := GetCard (pkt, pos);
        END (*IF*);
        flags := flags DIV 2;
        IF ODD(flags) THEN
            att.permissions := GetCard (pkt, pos);
        END (*IF*);
        flags := flags DIV 2;
        IF ODD(flags) THEN
            att.atime := GetCard (pkt, pos);
            att.mtime := GetCard (pkt, pos);
        END (*IF*);

        RETURN att;

    END ParseAttrs;

(************************************************************************)

PROCEDURE ConvertAttributes (attr: FileAttributes;
                                    VAR (*OUT*) result: AttrRecord);

    (* Converts a FileAttributes record to the data needed by ATTRS. *)

    CONST dirflag = 4000H;      (* 16384 seems to work *)
          regularflag = 8000H;

    VAR permissions: CARDINAL;

    BEGIN
        result.flags := 13;
        result.size := attr.D.size;
        result.uid := 0;
        result.gid := 0;

        (* permissions is a bit mask as defined by posix.               *)
        (* A 3-bit Posix permission is rwx, for read, write, execute.   *)
        (* But we use the x flag to mean "delete".                      *)
        (* That extends to 9 bits for owner, group, world.              *)

        permissions := 0;
        IF AllowRead IN attr.permissions THEN permissions := 4 END(*IF*);
        IF AllowWrite IN attr.permissions THEN INC (permissions, 2) END(*IF*);
        IF AllowDelete IN attr.permissions THEN INC (permissions) END(*IF*);
        permissions := (64+8+1) * permissions;    (* extend to 9 bits *)
        IF attr.IsADir THEN
            INC (permissions, dirflag);
        ELSE
            INC (permissions, regularflag);
        END (*IF*);
        result.permissions := permissions;

        result.size := attr.D.size;

        result.atime := ConvertDateTime (attr.D.dateAcc, attr.D.timeAcc);
        result.mtime := ConvertDateTime (attr.D.datePkd, attr.D.timePkd);

    END ConvertAttributes;

(************************************************************************)

PROCEDURE GetAttributes (U: User;  filename: ARRAY OF CHAR;
                            VAR (*OUT*) sshattr: AttrRecord): BOOLEAN;

    (* From the filename, extracts the data needed by ATTRS. *)

    CONST dirflag = 16384;      (* 16384 seems to work *)

    VAR attr: FileAttributes;

    BEGIN
        attr := FindAttributes (U, filename);
        IF NOT attr.exists THEN
            RETURN FALSE;
        END (*IF*);
        ConvertAttributes (attr, sshattr);
        RETURN TRUE;
    END GetAttributes;

(************************************************************************)
(*                           FILE HANDLES                               *)
(************************************************************************)

CONST
    (* pflags values to say what kind of open is requested. *)

    SSH_FXF_READ        =   01H;
    SSH_FXF_WRITE       =   02H;
    SSH_FXF_APPEND      =   04H;
    SSH_FXF_CREAT       =   08H;
    SSH_FXF_TRUNC       =   10H;
    SSH_FXF_EXCL        =   20H;

    (* Codes used in SSH_FX_STATUS replies.  There don't seem to    *)
    (* be codes for most of the conditions I'd like to report.      *)

    SSH_FX_OK                   =   0;
    SSH_FX_EOF                  =   1;
    SSH_FX_NO_SUCH_FILE         =   2;
    SSH_FX_PERMISSION_DENIED    =   3;
    SSH_FX_FAILURE              =   4;
    SSH_FX_BAD_MESSAGE          =   5;
    SSH_FX_NO_CONNECTION        =   6;
    SSH_FX_CONNECTION_LOST      =   7;
    SSH_FX_OP_UNSUPPORTED       =   8;

(************************************************************************)

PROCEDURE MakeFileHandle (U: User;  filename: ARRAY OF CHAR;
                            pflags: CARDINAL;  attrs: AttrRecord;
                              VAR (*OUT*) handle: FileHandle): CARDINAL;

    (* Creates a FileHandle for a non-directory file.   *)
    (* The return values are:                           *)
    (*      0   OK                                      *)
    (*      2   no such file                            *)
    (*      3   permission denied                       *)
    (*      5   bad message                             *)

    VAR ourattr: FileAttributes;
        code: CARDINAL;
        cid: ChanId;
        wantread, wantwrite, wantappend, wantcreate,
                                    wanttrunc, failifexists: BOOLEAN;
        realname: FilenameString;

    BEGIN
        handle := NIL;
        ourattr := FindAttributes (U, filename);
        wantread := IAND(pflags, SSH_FXF_READ) <> 0;
        wantwrite := IAND(pflags, SSH_FXF_WRITE) <> 0;
        wantappend := IAND(pflags, SSH_FXF_APPEND) <> 0;
        wantcreate := IAND(pflags, SSH_FXF_CREAT) <> 0;
        wanttrunc := IAND(pflags, SSH_FXF_TRUNC) <> 0;
        failifexists := IAND(pflags, SSH_FXF_EXCL) <> 0;
        code := 0;

        (* Rule out some contradictions. *)

        IF (wanttrunc OR failifexists) AND NOT wantcreate THEN
            code := 5;
        END (*IF*);

        (* Now we have to do the legality checks.  Note that ourattr    *)
        (* is as determined in our own file system, and attrs are what  *)
        (* is specified in the open request.  At present it looks to me *)
        (* as if we have to ignore the supplied attrs, because we don't *)
        (* want to go against the file permissions as set by the system *)
        (* manager.                                                     *)

        IF (code = 0) AND (ourattr.exists AND ourattr.IsADir) THEN
            code := 3;
        END (*IF*);

        (* Read request *)

        IF (code = 0) AND wantread THEN
            IF NOT ourattr.exists THEN
                code := 2;
            ELSIF NOT (AllowRead IN ourattr.permissions) THEN
                code := 3;
            END (*IF*);
        END (*IF*);

        (* Write request *)

        IF (code = 0) AND wantwrite THEN
            IF NOT (AllowWrite IN ourattr.permissions) THEN
                code := 3;
            ELSIF ourattr.exists AND NOT (AllowDelete IN ourattr.permissions) THEN
                code := 3;
            END (*IF*);
        END (*IF*);

        (* Append request *)

        IF (code = 0) AND wantappend THEN
            IF AllowWrite IN ourattr.permissions THEN
                wantwrite := TRUE;
            ELSE
                code := 3;
            END (*IF*);
        END (*IF*);

        (* Create request *)

        IF (code = 0) AND wantcreate THEN
            IF AllowWrite IN ourattr.permissions THEN
                wantwrite := TRUE;
            ELSE
                code := 3;
            END (*IF*);
        END (*IF*);

        (* Truncate request *)

        IF (code = 0) AND wanttrunc THEN
            IF AllowWrite IN ourattr.permissions THEN
                wantwrite := TRUE;
            ELSE
                code := 3;
            END (*IF*);
        END (*IF*);

        (* Fail if file exists *)

        IF (code = 0) AND failifexists THEN
            IF ourattr.exists THEN
                wantwrite := TRUE;
            ELSE
                code := 3;
            END (*IF*);
        END (*IF*);

        (* After that tedious set of checks (is there a better way to   *)
        (* do them?), we can finally try to open the file.              *)

        cid := NoSuchChannel;
        IF code = 0 THEN
            MakeRealName (U, filename, realname);
            cid := OpenGeneral (realname, wantread, wantwrite,
                                        wantappend, wantcreate, wanttrunc);
            IF cid = NoSuchChannel THEN
                code := 2;
            END (*IF*);
        END (*IF*);

        (* Finally, create the handle. *)

        IF code = 0 THEN
            NEW (handle);
            handle^.token := FileToken;
            Strings.Assign (filename, handle^.name);
            handle^.directory := FALSE;
            handle^.cid := cid;
        END (*IF*);

        RETURN code;

    END MakeFileHandle;

(************************************************************************)

PROCEDURE MakeDirHandle (U: User;  dirname: ARRAY OF CHAR;
                                VAR (*OUT*) handle: FileHandle): CARDINAL;

    (* Creates a FileHandle for a directory.    *)
    (* The return values are:                   *)
    (*      0   OK                              *)
    (*      1   file is not a directory         *)
    (*      2   directory does not exist        *)
    (*      3   you don't have read permission  *)

    VAR attr: FileAttributes;

    BEGIN
        handle := NIL;
        attr := FindAttributes (U, dirname);
        IF NOT attr.exists THEN
            RETURN 2;
        ELSIF NOT attr.IsADir THEN
            RETURN 1;
        ELSIF NOT (AllowRead IN attr.permissions) THEN
            RETURN 3;
        END (*IF*);

        NEW (handle);
        handle^.token := DirToken;
        Strings.Assign (dirname, handle^.name);
        handle^.directory := TRUE;
        handle^.list := CreateDirectoryListing (U, dirname);
        RETURN 0;
    END MakeDirHandle;

(************************************************************************)

PROCEDURE CloseHandle (VAR (*INOUT*) handle: FileHandle): CARDINAL;

    (* Closes a file or directory handle.       *)
    (* The return values are:                   *)
    (*      0   OK                              *)
    (*      2   invalid handle                  *)

    VAR p, next: DirListPtr;

    BEGIN
        IF (handle = NIL) OR ((handle^.token <> FileToken)
                                    AND (handle^.token <> DirToken)) THEN
            RETURN 2;
        END (*IF*);

        handle^.token := 0;
        IF handle^.directory THEN
            p := handle^.list;
            WHILE p <> NIL DO
                next := p^.next;
                DISPOSE (p);
                p := next;
            END (*WHILE*);
            handle^.list := NIL;
        ELSE
            OS2.DosClose (handle^.cid);
        END (*IF*);
        DISPOSE (handle);
        RETURN 0;

    END CloseHandle;

(************************************************************************)

PROCEDURE FilenameFromHandle (handle: FileHandle;
                                    VAR (*OUT*) filename: ARRAY OF CHAR);

    (* Returns the file or directory name associated with this handle.  *)
    (* Returns an empty string for an invalid handle.                   *)

    BEGIN
        IF (handle = NIL) OR ((handle^.token <> FileToken)
                                    AND (handle^.token <> DirToken)) THEN
            filename[0] := Nul;
        ELSE
            Strings.Assign (handle^.name, filename);
        END (*IF*);
    END FilenameFromHandle;

(************************************************************************)
(*                          DATA TRAMSFER                               *)
(************************************************************************)

PROCEDURE ReadData (handle: FileHandle;  offset: CARD64;  length: CARDINAL;
                                      VAR (*OUT*) data: ByteStr): CARDINAL;

    (* Reads a chunk of data from a file. The return values are:    *)
    (*      0   OK                                                  *)
    (*      1   end of file                                         *)
    (*      4   read failure                                        *)
    (*      5   bad handle                                          *)

    VAR code, rc, NumberRead: CARDINAL;

    BEGIN
        IF (handle = NIL) OR (handle^.token <> FileToken) THEN
            RETURN 5;
        END (*IF*);
        MakeBS (data, length);
        SetPosition (handle^.cid, ToAPI64(offset));
        rc := OS2.DosRead (handle^.cid, data.data, length, NumberRead);
        IF NumberRead < length THEN
            data.size := NumberRead;
        END (*IF*);
        IF rc = 232 THEN
            code := 1;
        ELSIF rc <> 0 THEN
            code := 4;
        ELSIF NumberRead = 0 THEN
            code := 1;
        ELSE
            code := 0;
        END (*IF*);
        RETURN code;
    END ReadData;

(************************************************************************)

PROCEDURE WriteData (handle: FileHandle;  offset: CARD64; length: CARDINAL;
                                  VAR (*IN*) data: ARRAY OF LOC): CARDINAL;

    (* Writes a chunk of data to a file. The return values are:     *)
    (*      0   OK                                                  *)
    (*      1   end of file                                         *)
    (*      4   write failure                                       *)
    (*      5   bad handle                                          *)

    VAR code, rc, actual: CARDINAL;

    BEGIN
        IF (handle = NIL) OR (handle^.token <> FileToken) THEN
            RETURN 5;
        END (*IF*);
        SetPosition (handle^.cid, ToAPI64(offset));
        rc := OS2.DosWrite (handle^.cid, data, length, actual);
        IF (rc <> 0) OR (actual < length) THEN
            code := 4;
        ELSE
            code := 0;
        END (*IF*);
        RETURN code;
    END WriteData;

(************************************************************************)

PROCEDURE RemoveFile (U: User;  VAR (*IN*) name: ARRAY OF CHAR): CARDINAL;

    (* Deletes a file, if possible. The return values are:  *)
    (*      0   OK                                          *)
    (*      2   no such file                                *)
    (*      3   permission denied                           *)
    (*      4   operation failed                            *)

    VAR attr: FileAttributes;
        realname: FilenameString;

    BEGIN
        attr := FindAttributes (U, name);
        IF NOT attr.exists THEN
            RETURN 2;
        ELSIF attr.IsADir THEN
            RETURN 3;
        ELSIF NOT (AllowDelete IN attr.permissions) THEN
            RETURN 3;
        END (*IF*);

        MakeRealName (U, name, realname);
        IF OS2.DosDelete (realname) = 0 THEN RETURN 0;
        ELSE RETURN 4;
        END (*IF*);

    END RemoveFile;

(************************************************************************)
(*                             RENAME                                   *)
(************************************************************************)

PROCEDURE ParentOf (VAR (*IN*) file:ARRAY OF CHAR;
                                      VAR (*OUT*) parent: ARRAY OF CHAR);

    (* Returns the name of the parent directory of file.  *)

    VAR pos: CARDINAL;  found: BOOLEAN;

    BEGIN
        Strings.Assign (file, parent);
        Strings.FindPrev ('/', parent, LENGTH(file), found, pos);
        IF found THEN
            IF pos = 0 THEN
                parent[1] := Nul;
            ELSE
                parent[pos] := Nul;
            END (*IF*);
        ELSE
            (* Should not happen, produce a rubbish answer. *)
            Strings.Assign ("#", parent);
        END (*IF*);
    END ParentOf;

(************************************************************************)

PROCEDURE DoRename (U: User;  VAR (*IN*) name1, name2: ARRAY OF CHAR): CARDINAL;

    (* Renames name1 to name2, if possible. The return values are:  *)
    (*      0   OK                                                  *)
    (*      2   no such file                                        *)
    (*      3   permission denied                                   *)
    (*      4   operation failed                                    *)
    (* name1 can be the name of a file or a directory.              *)

    VAR attr1, attr2: FileAttributes;
        rc: CARDINAL;
        parent1, parent2, realname1, realname2: FilenameString;
        fname1, fname2: FName;

    BEGIN
        (* It is an error if name2 already exists, except in the        *)
        (* special case where name1 and name2 are identical except for  *)
        (* character case.                                              *)

        IF NOT NameMatch (name1, name2) THEN
            attr2 := FindAttributes (U, name2);
            IF attr2.exists THEN
                RETURN 3;
            END (*IF*);
        END (*IF*);

        (* Permissions depend on the parent directory. *)

        MakeRealName (U, name1, realname1);
        MakeRealName (U, name2, realname2);
        ParentOf (name1, parent1);
        ParentOf (name2, parent2);
        attr1 := FindAttributes (U, parent1);

        IF NameMatch (parent1, parent2) THEN

            (* No change in directory, so renaming only requires    *)
            (* rename permission in parent1.                        *)

            IF AllowRename IN attr1.permissions THEN
                rc := OS2.DosMove (realname1, realname2);
                IF rc = 0 THEN RETURN 0;
                ELSE RETURN 4;
                END (*IF*);
            ELSE
                RETURN 3;
            END (*IF*);
        ELSE
            (* Moving across directories does not require rename        *)
            (* permission; instead, it needs delete permission in the   *)
            (* source directory and write permission in the target.     *)
            (* In addition, the way to do the move depends on whether   *)
            (* the two locations are on the same drive.                 *)

            attr2 := FindAttributes (U, parent2);
            IF (AllowDelete IN attr1.permissions)
                        AND (AllowWrite IN attr2.permissions) THEN

                fname1 := MakeFName (U, name1);
                fname2 := MakeFName (U, name2);
                IF SameDrive (fname1, fname2) THEN
                    rc := OS2.DosMove (realname1, realname2);
                    IF rc = 0 THEN RETURN 0;
                    ELSE RETURN 4;
                    END (*IF*);
                ELSE
                    rc := OS2.DosCopy (realname1, realname2, 0);
                    IF rc = 0 THEN
                        rc := OS2.DosDelete (realname1);
                        IF rc = 0 THEN
                            RETURN 0;
                        ELSE
                            RETURN 4;
                        END (*IF*);
                    ELSE RETURN 4;
                    END (*IF*);
                END (*IF*);
            ELSE
                RETURN 3;
            END (*IF*);
        END (*IF*);

    END DoRename;

(************************************************************************)
(*                   CREATING AND DELETING DIRECTORIES                  *)
(************************************************************************)

PROCEDURE MkDir (U: User;  VAR (*IN*) name: ARRAY OF CHAR): CARDINAL;

    (* Makes a directory, if possible. Return values are:   *)
    (*      0   OK                                          *)
    (*      3   permission denied                           *)
    (*      4   operation failed                            *)

    VAR attr: FileAttributes;
        rc: CARDINAL;
        realname, parent: FilenameString;

    BEGIN
        (* It is an error if name already exists.  *)

        attr := FindAttributes (U, name);
        IF attr.exists THEN
            RETURN 3;
        END (*IF*);

        (* Otherwise, permission depends on the parent directory. *)

        MakeRealName (U, name, realname);
        ParentOf (name, parent);
        attr := FindAttributes (U, parent);
        IF AllowWrite IN attr.permissions THEN
            rc := OS2.DosCreateDir (realname, NIL);
            IF rc = 0 THEN
                RETURN 0;
            ELSE
                RETURN 4;
            END (*IF*);
        ELSE
            RETURN 3;
        END (*IF*);

    END MkDir;

(************************************************************************)

PROCEDURE RmDir (U: User;  VAR (*IN*) name: ARRAY OF CHAR;
                            VAR (*OUT*) message: ARRAY OF CHAR): CARDINAL;

    (* Deletes a directory, if possible. A nonzero return value is an   *)
    (* error code, and message describes the error.                     *)

    VAR attr: FileAttributes;
        rc: CARDINAL;
        realname, parent: FilenameString;

    BEGIN
        (* It is an error if name does not exist, or is not a directory.  *)

        attr := FindAttributes (U, name);
        IF NOT attr.exists THEN
            Strings.Assign ("directory does not exist", message);
            RETURN 3;
        ELSIF NOT attr.IsADir THEN
            Strings.Assign ("not a directory", message);
            RETURN 3;
        END (*IF*);

        (* Operation will fail unless directory is empty. *)

        IF NOT EmptyDirectory (U, name) THEN
            Strings.Assign ("directory is not empty", message);
            RETURN 3;
        END (*IF*);

        (* Otherwise, permission depends on the parent directory. *)

        MakeRealName (U, name, realname);
        ParentOf (name, parent);
        attr := FindAttributes (U, parent);
        IF AllowDelete IN attr.permissions THEN
            rc := OS2.DosDeleteDir (realname);
            IF rc = 0 THEN
                Strings.Assign ("OK", message);
                RETURN 0;
            ELSE
                Strings.Assign ("operation failed", message);
                RETURN 4;
            END (*IF*);
        ELSE
            Strings.Assign ("permission denied", message);
            RETURN 3;
        END (*IF*);

    END RmDir;

(************************************************************************)
(*                   CONSTRUCTING A DIRECTORY LISTING                   *)
(*                                                                      *)
(* The SSH_FXP_NAME response has the following format (page 21):        *)
(*                                                                      *)
(*          uint32     id                                               *)
(*          uint32     count                                            *)
(*          repeats count times:                                        *)
(*              string     filename                                     *)
(*              string     longname                                     *)
(*              ATTRS      attrs                                        *)
(*                                                                      *)
(* The caller looks after the id field.  The procedures below build a   *)
(* ByteStr that contains everything else.                               *)
(*                                                                      *)
(************************************************************************)

CONST LongnameTemplate =
        "drwxrwxrwx   1 ftp      -A-HN           0 Dec 31  2019 ";
      (* 1234567890 123 12345678 12345678 12345678 123456789012 *)
       LongnameP22 =
        "-rwxr-xr-x   1 mjos     staff      348911 Mar 25 14:29 t-filexfer";
      (* 1234567890 123 12345678 12345678 12345678 123456789012 *)

(************************************************************************)

PROCEDURE ListSize (list: DirListPtr;  VAR (*OUT*) length: CARDINAL;
                                           CountMax: CARDINAL): CARDINAL;

    (* The function return value is the number of entries in the list,  *)
    (* and length is returned as the sum of the sizes that will be      *)
    (* needed for each directory entry when we construct it.  If the    *)
    (* list is very long then we report only the results for the first  *)
    (* CountMax (see below) entries.                                    *)

    VAR count: CARDINAL;
        p: DirListPtr;

    BEGIN
        count := 0;
        length := 4;        (* for the count field itself *)
        p := list;
        WHILE (p <> NIL) AND (count < CountMax) DO
            INC (count);
            INC (length, 4 + LENGTH(p^.name));
            INC (length, 4 + LENGTH(LongnameTemplate) + 4 + LENGTH(p^.name));
            INC (length, 24);       (* the ATTRS data length. *)
            p := p^.next;
        END (*WHILE*);

        RETURN count;

    END ListSize;

(************************************************************************)

PROCEDURE AddLongname (result: ByteStr;  VAR (*INOUT*) pos: CARDINAL;
                                                         p: DirListPtr);

    (* Adds the LONGNAME string for p^.  *)

    VAR longname: ARRAY [0..255] OF CHAR;
        format, field: ARRAY [0..15] OF CHAR;
        spos: CARDINAL;
        dt, Now: DateTime;

(* Examples:
        "drwxrwxrwx   1 ftp      -A-HN           0 Dec 31  2019 ";
      (* 1234567890 123 12345678 12345678 12345678 123456789012 *)
        "dr--r--r--   1 ftp      -----           0 23 Nov 05:39 dir001"
         -rwxr-xr-x   1 mjos     staff      348911 Mar 25 14:29 t-filexfer
*)

    BEGIN
        (* Start with the 'd' attribute. *)

        IF p^.attr.IsADir THEN
            longname := 'd';
        ELSE
            longname := '-'
        END (*IF*);

        (* Now the rwx field. *)

        field := "rwx";
        IF NOT (AllowRead IN p^.attr.permissions) THEN field[0] := '-' END(*IF*);
        IF NOT (AllowWrite IN p^.attr.permissions) THEN field[1] := '-' END(*IF*);
        IF NOT (AllowDelete IN p^.attr.permissions) THEN field[2] := '-' END(*IF*);
        Strings.Append (field, longname);
        Strings.Append (field, longname);
        Strings.Append (field, longname);

        (* Now the dummy link count and dummy owner. *)

        Strings.Append ("   1 ftp      ", longname);

        (* Next comes the RASHN field. *)

        field := "RASHN    ";
        IF NOT (readonly IN p^.attr.D.attr) THEN field[0] := '-' END(*IF*);
        IF NOT (archive IN p^.attr.D.attr)  THEN field[1] := '-' END(*IF*);
        IF NOT (system IN p^.attr.D.attr)   THEN field[2] := '-' END(*IF*);
        IF NOT (hidden IN p^.attr.D.attr)   THEN field[3] := '-' END(*IF*);
        IF NOT (AllowDelete IN p^.attr.permissions) THEN field[4] := '-' END(*IF*);
        Strings.Append (field, longname);

        (* The size takes 8 characters. *)

        spos := 0;
        ConvertCard64RJ (p^.attr.D.size, field, 8, spos);
        field[8] := ' ';  field[9] := Nul;
        Strings.Append (field, longname);

        (* For the date last modified we have 12 characters, followed by one space. *)

        GetClock (Now);
        dt := UnpackDateTime (p^.attr.D.datePkd, p^.attr.D.timePkd);
        IF dt.year = Now.year THEN
            format := "dd mmm HH:MM ";
        ELSE
            format := "dd mmm  yyyy ";
        END (*IF*);
        FormatDateTime (dt, format, TRUE, field);
        Strings.Append (field, longname);

        (* Finally, the UTF-8 form of the file name. *)

        Strings.Append (p^.name, longname);

        (* Store the result. *)

        AddCharString (result, pos, longname);

    END AddLongname;

(************************************************************************)

PROCEDURE MakeDirListing (U: User;  handle: FileHandle;  VAR (*OUT*) result: ByteStr);

    (* result is a directory listing, starting with a count of entries. *)

    CONST CountMax = 128;

    VAR list, p, next: DirListPtr;
        sshattr: AttrRecord;
        count, length, pos: CARDINAL;

    BEGIN
        IF (handle = NIL) OR (handle^.token <> DirToken) THEN
            result.allocated := 0;  result.size := 0;
            result.data := NIL;
            RETURN;
        END (*IF*);

        (*path := handle^.path;*)
        list := handle^.list;

        IF list = NIL THEN
            (* Return empty packet, which the caller will turn into *)
            (* an EOF status message.                               *)
            count := 0;
            MakeBS (result, 0);
        ELSE
            count := ListSize (list, length, CountMax);
            MakeBS (result, length);
            pos := 0;
            AddCard (result, pos, count);
        END (*IF*);

        p := list;
        WHILE (p <> NIL) AND (count > 0) DO
            AddCharString (result, pos, p^.name);
            AddLongname (result, pos, p);

            (* Now we have to add the ATTR fields. *)

            ConvertAttributes (p^.attr, sshattr);
            AddCard (result, pos, sshattr.flags);
            AddCard (result, pos, sshattr.size.high);
            AddCard (result, pos, sshattr.size.low);
            AddCard (result, pos, sshattr.permissions);
            AddCard (result, pos, sshattr.atime);
            AddCard (result, pos, sshattr.mtime);

            DEC (count);
            next := p^.next;
            DISPOSE (p);
            p := next;

        END (*WHILE*);

        (* Leave the trailing part of the list, if any, for next call. *)

        handle^.list := p;

    END MakeDirListing;

(************************************************************************)

END FDFiles.

