(**************************************************************************)
(*                                                                        *)
(*  SFTPD FTP daemon                                                      *)
(*  Copyright (C) 2022   Peter Moylan                                     *)
(*                                                                        *)
(*  This program is free software: you can redistribute it and/or modify  *)
(*  it under the terms of the GNU General Public License as published by  *)
(*  the Free Software Foundation, either version 3 of the License, or     *)
(*  (at your option) any later version.                                   *)
(*                                                                        *)
(*  This program is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU General Public License for more details.                          *)
(*                                                                        *)
(*  You should have received a copy of the GNU General Public License     *)
(*  along with this program.  If not, see <http://www.gnu.org/licenses/>. *)
(*                                                                        *)
(*  To contact author:   http://www.pmoylan.org   peter@pmoylan.org       *)
(*                                                                        *)
(**************************************************************************)

IMPLEMENTATION MODULE FDUsers;

        (********************************************************)
        (*                                                      *)
        (*  SFTP server: module to deal with user access rights *)
        (*                                                      *)
        (*  Programmer:         P. Moylan                       *)
        (*  Started:            14 November 2021                *)
        (*  Last edited:        4 October 2022                  *)
        (*  Status:             OK                              *)
        (*                                                      *)
        (********************************************************)


(************************************************************************)
(*                                                                      *)
(*     Syntax for the directory permissions in the INI file:            *)
(*                                                                      *)
(*     <result>     ->  <dirlist>                                       *)
(*     <dirlist>    ->  <diritem> { , <diritem> }*                      *)
(*     <diritem>    ->  <dirname> <dirrule>                             *)
(*     <dirname>    ->  <namestring>  |  <namestring> = <namestring>    *)
(*     <dirrule>    -> { <permission> }* { ( <dirlist> ) }              *)
(*     <permission> ->  V+ | V- | R+ | R- | W- | W+ | D- | D+ | N+ | N- *)
(*     <namestring> ->  <string1> | " <string2> " | ' <string3> '       *)
(*     <string1>    -> any string not including space char              *)
(*     <string2>    -> any string not including double quote            *)
(*     <string3>    -> any string not including single quote            *)
(*                                                                      *)
(*  Notation: {} means optional, {}* means zero or more, {}+ means      *)
(*            one or more.                                              *)
(*                                                                      *)
(************************************************************************)

FROM SYSTEM IMPORT ADR, CARD8, CARD16;

IMPORT FileSys, Strings, SysClock, OS2;

FROM Arith64 IMPORT
    (* const*)  Max64,
    (* type *)  CARD64;

FROM LowLevel IMPORT
    (* proc *)  EVAL;

FROM TaskControl IMPORT
    (* type *)  Lock,
    (* proc *)  CreateLock, Obtain, Release;

FROM Semaphores IMPORT
    (* type *)  Semaphore,
    (* proc *)  Signal;

FROM FileOps IMPORT
    (* const*)  NoSuchChannel,
    (* type *)  ChanId, FilenameString, FileAttr, FileAttribute, DirectoryEntry,
    (* proc *)  FileExists, FirstDirEntry, NextDirEntry, DirSearchDone,
                FreeSpace, OpenOldFile, OpenNewFile, CloseFile,
                SetPosition, EndPosition;

FROM Volume IMPORT
    (* proc *)  CreateDir, RmvDir;

FROM Storage IMPORT
    (* proc *)  ALLOCATE, DEALLOCATE;

FROM FtpdINI IMPORT
    (* proc *)  OpenINIForUser, CloseINIFile;

FROM INIData IMPORT
    (* type *)  HINI, StringReadState,
    (* proc *)  INIValid, ItemSize, INIPut, INIGet, INIGetTrusted, INIGetString,
                INIDeleteApp, INIDeleteKey, GetStringList, NextString,
                CloseStringList;

FROM IPFilter IMPORT
    (* proc *)  CheckUserIPFilter;

FROM MyClock IMPORT
    (* proc *)  CorrectToGMT, CurrentTimeToString;

FROM MiscFuncs IMPORT
    (* proc *)  AppendString, ConvertCardRJ,
                ConvertCard64RJ, ConvertCardZ, AddEOL;

FROM MD5 IMPORT
    (* type *)  MD5_CTX,
    (* proc *)  MD5Init, MD5Update, MD5Final, MD5DigestToString;

FROM CodePage IMPORT
    (* proc *)  TranslateToUTF8;

FROM Sockets IMPORT
    (* type *)  Socket,
    (* proc *)  send;

FROM WildCard IMPORT
    (* proc *)  WildMatch;

FROM Names IMPORT
    (* type *)  PassString, HostName;

(********************************************************************************)

CONST Nul = CHR(0);

TYPE
    CharSet = SET OF CHAR;
    UserName = FilenameString;

    (* The following declarations relate to the data structure that we keep,    *)
    (* for a logged-in user, to show which directories and files are accessible.*)
    (* For each directory, we have a linked list of subdirectories.             *)

    (* A HideList is a linear list of filename masks, to indicate filenames     *)
    (* that should always be invisible regardless of what directory they're in. *)

    HideListPtr = POINTER TO HideListRecord;
    HideListRecord =
               RECORD
                   next: HideListPtr;
                   mask: FilenameString;
               END (*RECORD*);

    LinkData = RECORD
                   parent, this: FilenameString;
               END (*RECORD*);

    DirEntryPtr = POINTER TO DirEntry;

    (* A DirEntry describes one directory.  If IsADir is FALSE then it          *)
    (* describes a file rather than a directory.                                *)
    (*      flags       The user's permissions for this directory               *)
    (*      parent      Parent directory                                        *)
    (*      FirstChild  First subdirectory of this directory                    *)
    (*      next        Next subdirectory of the parent                         *)
    (*      name        Name of this directory or file (not the full path)      *)
    (*      exists      TRUE if this is either a pseudo-directory or a file     *)
    (*                    or directory that exists in our file system           *)
    (*      IsADir      TRUE if this is a directory, not a non-directory file   *)
    (*      categorised TRUE if the values of the "exists" and "IsADir" field   *)
    (*                    are valid.                                            *)
    (*      virtual     TRUE iff this is a symbolic link                        *)
    (*      link        Pointer to full path name for the parent directory and  *)
    (*                    the current directory, in the case of a link.         *)
    (* NOTE: For a symbolic link, the parent link always ends with a '/', but   *)
    (* the "this" link never ends with a '/'.                                   *)

    (* Note that there are two "virtual" cases. If link^.this is an empty       *)
    (* string we have a pseudo-directory, i.e. one that doesn't exist on the    *)
    (* disk. Otherwise we have a symbolic link to a real file or directory.     *)

    DirEntry = RECORD
                   flags: PermissionSet;
                   parent, FirstChild, next: DirEntryPtr;
                   name: FilenameString;
                   exists, IsADir, categorised, virtual: BOOLEAN;
                   link: POINTER TO LinkData;
               END (*RECORD*);

    (* An FName structure is our internal representation of a file name.        *)
    (*     dir          Directory string, a complete path name.  Always ends    *)
    (*                    with a '/', unless it's an empty string.              *)
    (*     vdir         like dir, except that it's the path name as seen by     *)
    (*                    the client, which can be different from the path      *)
    (*                    name as seen by the file system.  The component       *)
    (*                    separator is always a '/', never a '\'.               *)
    (*     finalname    file name within the directory.  This is an empty       *)
    (*                    string if the file in question is a directory.        *)
    (*     EntryPtr     points to the DirEntry record for the "dir" field, or   *)
    (*                    NIL if we've detected an invalid file name.           *)

    FName = POINTER TO FNameRecord;
    FNameRecord = RECORD
                      dir, vdir, finalname: FilenameString;
                      EntryPtr: DirEntryPtr;
                  END (*RECORD*);

    (* The fields in a UserPermission record have the following meanings.       *)
    (*      Name         The user's login name, which can be a plain username   *)
    (*                       or possibly a name of the form user@domain         *)
    (*      Password     The user's password                                    *)
    (*      EncryptPassword   TRUE iff the password is stored in encrypted form *)
    (*      category     What kind of user                                      *)
    (*      UserNumber   A serial number to use in welcome messages.            *)
    (*      UserLimit    Maximum number of instances of this user.  Storing     *)
    (*                   a copy of this for each instance might seem to be      *)
    (*                   redundant; but this allows us to cover the possibility *)
    (*                   that the limit has changed between instances.          *)
    (*      CurrentDir   Current directory, an absolute path.  This always      *)
    (*                      ends with a '/', except in the special case where   *)
    (*                      it's a null string.                                 *)
    (*      CurrentVDir  Same as CurrentDir, except that this is the full       *)
    (*                      pathname as seen by the client, not the pathname    *)
    (*                      as seen by the file system.                         *)
    (*      TreeRoot     Root of the permissions tree for this user.            *)
    (*      PosInTree    Pointer to current position within tree                *)
    (*      HideList     A list of file masks, for files that should be         *)
    (*                      invisible no matter where they are in the tree.     *)
    (*                                                                          *)
    (* The user data in the INI file can also include a LoginLimit field which, *)
    (* if nonzero, specifies how many logins will be allowed before this user   *)
    (* account is destroyed or deactivated.  We don't have to include that in   *)
    (* the permission record because we deal with it immediately on reading     *)
    (* the INI data.                                                            *)

    User = POINTER TO UserPermission;

    UserPermission = RECORD
                         Name: UserName;
                         Password: PassString;
                         category: UserCategory;
                         TreeRoot, PosInTree: DirEntryPtr;
                         HideList: HideListPtr;
                         CurrentDir: FilenameString;
                         CurrentVDir: FilenameString;
                         UserNumber, UserLimit: CARDINAL;
                         EncryptPassword: BOOLEAN;
                     END (*RECORD*);

    (* A structure for maintaining a linked list of user counts. *)

    CountList = POINTER TO CountListRecord;
    CountListRecord = RECORD
                          next: CountList;
                          count: CARDINAL;
                          name: UserName;
                      END (*RECORD*);

(********************************************************************************)

VAR
    (* A linked list showing, for each username, how many users are currently   *)
    (* logged in under that name.                                               *)

    UserCount: CountList;

    (* Critical section protection for this list. *)

    UserCountLock: Lock;

(********************************************************************************)
(*                         CHECK FOR TOO MANY USERS                             *)
(********************************************************************************)

PROCEDURE AllocateUserNumber (username: UserName;  limit: CARDINAL): CARDINAL;

    (* Works out which instance we are of this username.  If the result goes    *)
    (* over the limit, the returned result is 0.                                *)

    VAR current, newelement: CountList;
        result: CARDINAL;

    BEGIN
        Obtain (UserCountLock);

        (* Find the list entry for this username. *)

        current := UserCount;
        WHILE (current <> NIL) AND NOT Strings.Equal(current^.name, username) DO
            current := current^.next;
        END (*WHILE*);

        (* If this username is not already in the list, add it. *)

        IF current = NIL THEN
            NEW (newelement);
            WITH newelement^ DO
                next := UserCount;
                count := 0;
                name := username;
            END (*WITH*);
            UserCount := newelement;
            current := newelement;
        END (*IF*);

        (* Increment the count, unless it's already at the limit. *)

        IF current^.count < limit THEN
            INC (current^.count);
            result := current^.count;
        ELSE
            result := 0;
        END (*IF*);

        Release (UserCountLock);
        RETURN result;

    END AllocateUserNumber;

(********************************************************************************)
(*                               THE HIDE LIST                                  *)
(********************************************************************************)

PROCEDURE DiscardHideList (U: User);

    (* Disposes of this user's HideList. *)

    VAR p, next: HideListPtr;

    BEGIN
        p := U^.HideList;
        WHILE p <> NIL DO
            next := p^.next;
            DISPOSE (p);
            p := next;
        END (*WHILE*);
    END DiscardHideList;

(********************************************************************************)
(*                          CATEGORISING A TREE NODE                            *)
(********************************************************************************)

PROCEDURE CategoriseNode (p: DirEntryPtr;  parentpath: ARRAY OF CHAR);

    (* Checks whether p^ describes a file, a directory, or something that       *)
    (* doesn't exist on the disk.  (A pseudo-directory is considered to be a    *)
    (* directory that exists.)  The parentpath argument, which should end with  *)
    (* a '/' unless it is empty, is the physical path of this node's parent     *)
    (* directory.  On exit values have been assigned to p^.exists and p^.IsADir.*)

    VAR fullname: FilenameString;
        InfoBuf: OS2.FSALLOCATE;

    BEGIN
        OS2.DosError (OS2.FERR_DISABLEHARDERR);
        IF p^.virtual THEN
            fullname := p^.link^.this;
        ELSE
            Strings.Assign (parentpath, fullname);
            Strings.Append (p^.name, fullname);
        END (*IF*);
        IF fullname[0] = Nul THEN

            (* This is a pseudo-directory. *)

            p^.exists := TRUE;  p^.IsADir := TRUE;

        ELSIF (fullname[2] = Nul) AND (fullname[1] = ':') THEN

            (* Root of a filesystem volume. *)

            p^.exists := OS2.DosQueryFSInfo (ORD(CAP(fullname[0])) - ORD('A') + 1,
                           1, ADR(InfoBuf), SIZE(InfoBuf)) = 0;
            p^.IsADir := p^.exists;

        ELSE
            (* We have a real name, but we still have to check three    *)
            (* possibilities: this is a file, this is a directory, this *)
            (* is something that doesn't exist on our disk.             *)

            p^.exists := FileExists (fullname, p^.IsADir);

        END (*IF*);
        p^.categorised := TRUE;

    END CategoriseNode;

(********************************************************************************)
(*                                PARSER                                        *)
(********************************************************************************)

PROCEDURE LoadUserData (name: ARRAY OF CHAR;
                          VAR (*OUT*) category: UserCategory): User;

    (* Fetches the password, etc., for the user whose username is specified     *)
    (* as the argument.  Returns with category = NoSuchUser if the user's data  *)
    (* could not be found or if this is a user category that does not permit    *)
    (* logging in.  The result is NIL in this case, and also in the             *)
    (* case of an overflow user.                                                *)

    CONST Space = " ";

    VAR NextChar: CHAR;
        hini: HINI;
        bufptr: POINTER TO ARRAY [0..MAX(CARDINAL) DIV 4] OF CHAR;
        BufferPosition, BufferSize: CARDINAL;

    (****************************************************************************)

    PROCEDURE Scan;

        (* Puts the next input character into variable NextChar. *)

        BEGIN
            IF BufferPosition >= BufferSize THEN
                NextChar := Nul;
            ELSE
                NextChar := bufptr^[BufferPosition];
                INC (BufferPosition);
            END (*IF*);
        END Scan;

    (****************************************************************************)

    PROCEDURE LoadString (VAR (*OUT*) result: ARRAY OF CHAR;  Stoppers: CharSet);

        (* Reads a string up to but not including a character in Stoppers.      *)

        VAR j: CARDINAL;

        BEGIN
            j := 0;
            WHILE (j <= HIGH(result)) AND NOT (NextChar IN Stoppers) DO
                result[j] := NextChar;  INC(j);
                Scan;
            END (*WHILE*);
            IF j <= HIGH(result) THEN result[j] := Nul END(*IF*);
        END LoadString;

    (****************************************************************************)

    PROCEDURE GetNameString (VAR (*OUT*) result: ARRAY OF CHAR;  Stoppers: CharSet);

        (*     <namestring> ->  <string1> | " <string2> " | ' <string3> '       *)
        (*     <string1>   -> any string not including space char               *)
        (*     <string2>   -> any string not including double quote             *)
        (*     <string3>   -> any string not including single quote             *)
        (* The strings in <string1> also may not contain characters in Stoppers.*)

        CONST SingleQuote = "'";  DoubleQuote = '"';

        VAR Delimiter: CHAR;

        BEGIN
            INCL (Stoppers, Nul);
            IF (NextChar = SingleQuote) OR (NextChar = DoubleQuote) THEN
                Delimiter := NextChar;
                Stoppers := CharSet {Nul, Delimiter};
                Scan;
                LoadString (result, Stoppers);
                IF NextChar = Delimiter THEN
                    Scan;
                END (*IF*);
            ELSE
                INCL (Stoppers, Space);
                LoadString (result, Stoppers);
            END (*IF*);
        END GetNameString;

    (****************************************************************************)

    PROCEDURE DirItem (mother: DirEntryPtr;
                       MotherPath: ARRAY OF CHAR): DirEntryPtr;  FORWARD;

    (****************************************************************************)

    PROCEDURE DirList (mother: DirEntryPtr;
                            MotherPath: ARRAY OF CHAR): DirEntryPtr;

        (*     <dirlist>   ->  <diritem> { , <diritem> }*       *)
        (* Result returned: a linked list of directory nodes.   *)

        CONST Comma = ",";

        VAR result, lastnode: DirEntryPtr;

        BEGIN
            result := DirItem (mother, MotherPath);  lastnode := result;
            WHILE (NextChar = Comma) OR (NextChar = ';') DO
                Scan;
                lastnode^.next := DirItem (mother, MotherPath);
                lastnode := lastnode^.next;
            END (*WHILE*);
            RETURN result;
        END DirList;

    (****************************************************************************)

    PROCEDURE DirRule (pnode: DirEntryPtr;  OurPath: ARRAY OF CHAR);

        (* Fills in the permissions and subdirectory info for pnode.            *)
        (*     <dirrule>   -> { <permission> }* { ( <dirlist> ) }               *)
        (*     <permission> ->  V+ | V- | R+ | R- | W- | W+ | D- | D+ | N- | N+ *)

        VAR option: Permission;

        BEGIN
            (* Default flags are inherited from parent. *)

            IF pnode^.parent = NIL THEN
                pnode^.flags := PermissionSet {Visible, AllowRead}
            ELSE
                pnode^.flags := pnode^.parent^.flags
            END (*IF*);

            (* Look for optional permission codes. *)

            WHILE CAP(NextChar) IN CharSet {'V', 'R', 'W', 'D', 'N'} DO
                CASE CAP(NextChar) OF
                  | 'V':  option := Visible;
                  | 'R':  option := AllowRead;
                  | 'W':  option := AllowWrite;
                  | 'D':  option := AllowDelete;
                  | 'N':  option := AllowRename;
                END (*CASE*);
                Scan;
                IF NextChar = '-' THEN
                    Scan;
                    EXCL (pnode^.flags, option);
                ELSIF NextChar = '+' THEN
                    Scan;
                    INCL (pnode^.flags, option);
                END (*IF*);

            END (*WHILE*);

            (* Look for optional list of subdirectories. *)

            IF NextChar = '(' THEN
                Scan;
                IF NextChar <> ')' THEN
                    pnode^.FirstChild := DirList(pnode, OurPath);
                END (*IF*);
                IF NextChar = ')' THEN
                    Scan;
                END (*IF*);
            END (*IF*);

        END DirRule;

    (****************************************************************************)

    PROCEDURE DirName (mother: DirEntryPtr): DirEntryPtr;

        (*     <dirname>    ->  <namestring>  |  <namestring> = <namestring>    *)
        (* The alternative with the '=' sign means that we have to create       *)
        (* a symbolic link.                                                     *)

        VAR result: DirEntryPtr;  k: CARDINAL;

        BEGIN
            NEW (result);
            WITH result^ DO
                flags := PermissionSet {Visible, AllowRead};
                parent := mother;  FirstChild := NIL;  next := NIL;
                GetNameString (name, CharSet{});
                virtual := FALSE;  categorised := FALSE;
                link := NIL;
                IF NextChar = '=' THEN
                    Scan;
                    virtual := TRUE;
                    NEW (link);
                    link^.parent := "";      (* fill this in later *)
                    GetNameString (link^.this, CharSet{});

                    (* Transition arrangement: get rid of the trailing '/'.  *)

                    k := LENGTH (link^.this);
                    IF k > 0 THEN
                        DEC (k);
                        IF (link^.this[k] = '/') OR (link^.this[k] = '\') THEN
                            link^.this[k] := Nul;
                        END (*IF*);
                    END (*IF*);

                ELSIF name[0] = Nul THEN

                    (* Assume bare "" means ""="". *)

                    virtual := TRUE;
                    NEW (link);
                    link^.parent := "";  link^.this := "";

                END (*IF*);
            END (*WITH*);
            RETURN result;
        END DirName;

    (****************************************************************************)

    PROCEDURE DirItem (mother: DirEntryPtr;
                       MotherPath: ARRAY OF CHAR): DirEntryPtr;

        (*     <diritem>    ->  <dirname> <dirrule>                             *)

        VAR result: DirEntryPtr;  OurPath: FilenameString;

        BEGIN
            result := DirName(mother);
            IF result^.virtual THEN
                Strings.Assign (MotherPath, result^.link^.parent);
                OurPath := result^.link^.this;
            ELSE
                Strings.Assign (MotherPath, OurPath);
                Strings.Append (result^.name, OurPath);
            END (*IF*);
            IF OurPath[0] <> Nul THEN
                Strings.Append ('/', OurPath);
            END (*IF*);
            DirRule (result, OurPath);
            RETURN result;
        END DirItem;

    (****************************************************************************)

    VAR Active: BOOLEAN;

    (****************************************************************************)

    PROCEDURE IsActiveUser (hini: HINI;  username: ARRAY OF CHAR): BOOLEAN;

        (* Returns TRUE iff the "useSFTP" flag is set for this user.            *)

        VAR size: CARDINAL;  result: BOOLEAN;

        BEGIN
            IF (username[0] = Nul) OR (username[0] = '$') OR NOT INIValid(hini) THEN
                result := FALSE;
            ELSE
                result := ItemSize (hini, username, '', size)
                                       AND (size <> 0);
                IF result AND INIGet (hini, username, "useSFTP", Active) THEN
                    result := Active;
                END (*IF*);
            END (*IF*);
            RETURN result;
        END IsActiveUser;

    (****************************************************************************)

    PROCEDURE LoadHideList (hini: HINI;  U: User);

        (* Loads the HideList for this user. *)

        VAR state: StringReadState;
            previous, p: HideListPtr;
            key: ARRAY [0..8] OF CHAR;

        BEGIN
            key := "HideList";
            GetStringList (hini, U^.Name, key, state);
            p := NIL;
            REPEAT
                previous := p;
                NEW (p);
                IF previous = NIL THEN U^.HideList := p;
                ELSE previous^.next := p;
                END (*IF*);
                NextString (state, p^.mask);
            UNTIL p^.mask[0] = Nul;
            CloseStringList (state);

            (* Correct for the overshoot. *)

            IF previous = NIL THEN U^.HideList := NIL;
            ELSE previous^.next := NIL;
            END (*IF*);
            DISPOSE (p);

        END LoadHideList;

    (************************************************************************)
    (*                     BODY OF LOADUSERDATA                             *)
    (*                                                                      *)
    (*     <result>  ->  <userclass> <password> { <volumeinfo> }*           *)
    (*     <userclass> -> G | U | N | M | T                                 *)
    (*     <password>  -> <namestring>                                      *)
    (*                                                                      *)
    (* Note: for simplicity, a password must always be specified, even if   *)
    (* it is not going to be used.                                          *)
    (*                                                                      *)
    (************************************************************************)

    VAR result: User;  root: DirEntryPtr;
        UseTemplate: BOOLEAN;
        App: UserName;   key: ARRAY [0..6] OF CHAR;

    BEGIN       (* Body of LoadUserData *)

        NEW (result);
        Strings.Assign (name, result^.Name);
        result^.TreeRoot := NIL;  result^.PosInTree := NIL;  result^.HideList := NIL;

        (* Read the user category and password from the INI file. *)

        hini := OpenINIForUser(name, FALSE);
        IF IsActiveUser (hini, result^.Name)
                AND INIGet (hini, result^.Name, "Category", result^.category)
                AND (result^.category <= Manager) THEN

            category := result^.category;

            (* Password. *)

            IF category = GuestUser THEN
                result^.Password := '@';
            ELSE
                EVAL (INIGetString (hini, result^.Name, "Password", result^.Password));
            END (*IF*);
            IF (category < NormalUser) OR NOT INIGet (hini, result^.Name, "EncryptPassword", result^.EncryptPassword) THEN
                result^.EncryptPassword := FALSE;
            END (*IF*);

            (* Hide list. *)

            LoadHideList (hini, result);

            (* User limit. *)

            IF NOT INIGet (hini, result^.Name, "UserLimit", result^.UserLimit) THEN
                result^.UserLimit := MAX(CARDINAL);
            END (*IF*);

            (* Should we read volume information from a template? *)

            IF NOT INIGet (hini, result^.Name, "UseTemplate", UseTemplate)
                     OR NOT UseTemplate
                     OR NOT INIGetString (hini, result^.Name, "TemplateName", App)
                     OR (App[0] = Nul) THEN
                Strings.Assign (result^.Name, App);
            END (*IF*);

            (* Check for too many users. *)

            result^.UserNumber := AllocateUserNumber (result^.Name, result^.UserLimit);
            IF result^.UserNumber = 0 THEN
                DiscardHideList (result);
                DISPOSE (result);
                category := OverflowUser;
            END (*IF*);

        ELSE

            DiscardHideList (result);
            DISPOSE (result);
            category := NoSuchUser;

        END (*IF*);

        (* If the volume information exists, allocate a buffer for it. *)

        IF (result <> NIL)
                AND ItemSize (hini, App, "Volume", BufferSize)
                AND (BufferSize <> 0) THEN
            ALLOCATE (bufptr, BufferSize);
            key := "Volume";
            EVAL (INIGetTrusted (hini, App, key, bufptr^, BufferSize));
        ELSE
            bufptr := NIL;
            BufferSize := 0;
        END (*IF*);
        BufferPosition := 0;
        Scan;

        IF result <> NIL THEN
            (* Give the user a root directory. *)

            NEW (root);
            result^.TreeRoot := root;
            WITH root^ DO
                flags := PermissionSet {Visible, AllowRead};
                name := "";
                parent := NIL;  FirstChild := NIL;  next := NIL;
                exists := TRUE;  IsADir := TRUE;
                virtual := TRUE;  categorised := TRUE;
                NEW (link);
                link^.parent := "";
                link^.this := "";
            END (*WITH*);

            (* Load the volume information. *)

            root^.FirstChild := DirList (root, "");
            result^.CurrentVDir := '/';

            (* Discard the INI file raw data. *)

            IF bufptr <> NIL THEN
                DEALLOCATE (bufptr, BufferSize);
            END (*IF*);

            (* Decide whether we really need a virtual directory as     *)
            (* the root directory.                                      *)

            IF root^.FirstChild <> NIL THEN
                IF root^.FirstChild^.next <> NIL THEN
                    result^.CurrentDir := "";
                ELSE
                    DISPOSE (root^.link);
                    root := root^.FirstChild;
                    DISPOSE (result^.TreeRoot);
                    result^.TreeRoot := root;
                    root^.parent := NIL;
                    IF root^.virtual THEN
                        result^.CurrentDir := root^.link^.this;
                        IF result^.CurrentDir[0] <> Nul THEN
                            Strings.Append ('/', result^.CurrentDir);
                        END (*IF*);
                    ELSE
                        result^.CurrentDir := "";
                    END (*IF*);
                    CategoriseNode (root, "");
                END (*IF*);
            END (*IF*);
            result^.PosInTree := root;

        END (*IF*);

        CloseINIFile (hini);

        RETURN result;

    END LoadUserData;

(********************************************************************************)

PROCEDURE ReadUserData (name: ARRAY OF CHAR;
                          VAR (*OUT*) category: UserCategory): User;

    (* Loads the password, etc., for the user whose username is specified       *)
    (* as the argument.  Returns with category = NoSuchUser if the user's data  *)
    (* could not be found or if this is a user category that does not permit    *)
    (* logging in.  The result is NIL in this case, and also in the             *)
    (* case of an overflow user.                                                *)

    BEGIN
        RETURN LoadUserData (name, category);
    END ReadUserData;

(********************************************************************************)

PROCEDURE CheckLimitedUse (hini: HINI;  VAR (*IN*) name: UserName);

    (* Checks whether this is a limited-use account, and takes appropriate      *)
    (* action if so. This procedure should be called after the user has been    *)
    (* authenticated, i.e. once we are certain that this is a genuine login.    *)

    VAR LoginLimit: CARDINAL;
        SingleUse, DeleteExpired: BOOLEAN;

    BEGIN
        LoginLimit := 0;

        (* Limited-use account? *)

        IF INIGet (hini, name, "SingleUse", SingleUse) THEN
            (* Obsolete option *)
            IF SingleUse THEN
                LoginLimit := 1;
            END (*IF*);
            INIDeleteKey (hini, name, "SingleUse");
        ELSE EVAL (INIGet(hini, name, "LoginLimit", LoginLimit));
        END (*IF*);
        IF NOT INIGet (hini, name, "DeleteExpired", DeleteExpired) THEN
            DeleteExpired := FALSE;
        END (*IF*);

        (* If this is a 'limited logins' account, decrement the     *)
        (* number of logins allowed.  If the count reaches zero,    *)
        (* keep the user data in memory but either delete it from   *)
        (* the INI data or make the account inactive, depending on  *)
        (* the DeletedExpired flag in the INI file.                 *)
        (*       Note: LoginLimit=0 means no limit.                 *)

        IF LoginLimit > 0 THEN
            DEC (LoginLimit);
            IF LoginLimit = 0 THEN
                IF DeleteExpired THEN
                    INIDeleteApp (hini, name);
                ELSE
                    INIPut (hini, name, "Active", FALSE);
                END (*IF*);
            ELSE
                INIPut (hini, name, "LoginLimit", LoginLimit);
            END (*IF*);
        END (*IF*);

    END CheckLimitedUse;

(********************************************************************************)

PROCEDURE NoteLoginTime (U: User);

    (* Puts the current date/time as the "last login time" for this user.  This *)
    (* is also the point at which we check whether this is a limited-use        *)
    (* account.                                                                 *)

    VAR hini: HINI;  timestamp: ARRAY [0..18] OF CHAR;

    BEGIN
        CurrentTimeToString (timestamp);
        hini := OpenINIForUser (U^.Name, FALSE);
        IF INIValid(hini) THEN
            INIPut (hini, U^.Name, "LastLoginS", timestamp);
            CheckLimitedUse (hini, U^.Name);
            CloseINIFile (hini);
        END (*IF*);
    END NoteLoginTime;

(********************************************************************************)
(*                        DISCARDING A USER RECORD                              *)
(********************************************************************************)

PROCEDURE KillList (VAR (*INOUT*) p: DirEntryPtr);

    (* Discards a directory tree.  *)

    VAR q: DirEntryPtr;

    BEGIN
        WHILE p <> NIL DO
            KillList (p^.FirstChild);
            q := p^.next;
            IF p^.virtual THEN
                DISPOSE (p^.link);
            END (*IF*);
            DISPOSE (p);
            p := q;
        END (*IF*);
    END KillList;

(********************************************************************************)

PROCEDURE DestroyUserData (VAR (*INOUT*) U: User);

    (* Logs off the user and discards the data structure.  *)

    VAR current: CountList;

    BEGIN
        IF U <> NIL THEN

            (* Decrement the number of instances of this username. *)

            Obtain (UserCountLock);
            current := UserCount;
            WHILE (current <> NIL) AND NOT Strings.Equal(current^.name, U^.Name) DO
                current := current^.next;
            END (*WHILE*);
            IF current <> NIL THEN
                DEC (current^.count);
            END (*IF*);
            Release (UserCountLock);
            KillList (U^.TreeRoot);
            DiscardHideList (U);
            DISPOSE (U);
        END (*IF*);
    END DestroyUserData;

(********************************************************************************)
(*                    PERMISSIBLE ADDRESSES FOR THIS USER                       *)
(********************************************************************************)

PROCEDURE AddressForbidden (ClientIPAddress: CARDINAL;  user: ARRAY OF CHAR): BOOLEAN;

    (* Returns TRUE if this address is not on our list of acceptable addresses for this username. *)

    VAR result: BOOLEAN;
        hini: HINI;

    BEGIN
        result := FALSE;
        hini := OpenINIForUser(user, FALSE);
        IF INIValid (hini) THEN
            result := NOT CheckUserIPFilter (hini, user, ClientIPAddress);
            CloseINIFile (hini);
        END (*IF*);
        RETURN result;
    END AddressForbidden;

(********************************************************************************)
(*                           PASSWORD CHECKING                                  *)
(********************************************************************************)

PROCEDURE PasswordOK (U: User;  pass: ARRAY OF CHAR): BOOLEAN;

    (* Returns TRUE iff password is correct. *)

    VAR dummy: CARDINAL;  found: BOOLEAN;
        ctx: MD5_CTX;  digest: ARRAY[0..15] OF CARD8;
        buffer: PassString;

    BEGIN
        IF U = NIL THEN
            found := FALSE;
        ELSIF (U^.Password[0] = '@') AND (U^.Password[1] = Nul) THEN
            Strings.FindNext ('@', pass, 0, found, dummy);
        ELSE
            IF U^.EncryptPassword THEN
                ctx := MD5Init();
                MD5Update (ctx, pass, LENGTH(pass));
                MD5Final (ctx, digest);
                MD5DigestToString (digest, buffer);
                found := Strings.Equal (buffer, U^.Password);
            ELSE
                found := Strings.Equal (pass, U^.Password);
            END (*IF*);
        END (*IF*);
        RETURN found;
    END PasswordOK;

(********************************************************************************)
(*                           STRING COMPARISONS                                 *)
(********************************************************************************)

PROCEDURE CompareStrings (first, second: ARRAY OF CHAR): INTEGER;

    (* Returns +1 if first>second, 0 if first=second, and -1 if first<second.   *)
    (* Character case is ignored when making the comparison.                    *)

    VAR k: CARDINAL;

    BEGIN
        k := 0;
        LOOP
            IF k > HIGH(second) THEN
                IF k > HIGH(first) THEN RETURN 0
                ELSE RETURN +1
                END (*IF*);
            ELSIF k > HIGH(first) THEN RETURN -1
            ELSIF first[k] = Nul THEN
                IF second[k] = Nul THEN RETURN 0
                ELSE RETURN -1
                END (*IF*);
            ELSIF CAP(first[k]) > CAP(second[k]) THEN RETURN +1
            ELSIF CAP(first[k]) < CAP(second[k]) THEN RETURN -1
            ELSE INC(k);
            END (*IF*);
        END (*LOOP*);
    END CompareStrings;

(********************************************************************************)

PROCEDURE NameMatch (VAR (*IN*) first, second: ARRAY OF CHAR): BOOLEAN;

    (* Equality test, modulo upper/lower case differences.  That is, the result *)
    (* is TRUE even if the letters in first and second don't have the same      *)
    (* case, as long as the strings are otherwise equal.                        *)

    BEGIN
        RETURN CompareStrings(first, second) = 0;
    END NameMatch;

(********************************************************************************)

PROCEDURE InHideList (U: User;  name: ARRAY OF CHAR): BOOLEAN;

    (* Returns TRUE if this name is in the user's hide list.   *)
    (* Special case: ".." is never hidden.                     *)

    VAR p: HideListPtr;
        dotdot: ARRAY [0..1] OF CHAR;

    BEGIN
        dotdot := "..";
        IF WildMatch (name, dotdot) THEN
            RETURN FALSE;
        ELSE
            p := U^.HideList;
            WHILE (p <> NIL) AND NOT WildMatch (name, p^.mask) DO
                p := p^.next;
            END (*WHILE*);
            RETURN p <> NIL;
        END (*IF*);
    END InHideList;

(********************************************************************************)
(*                          FILE NAME PROCESSING                                *)
(********************************************************************************)

PROCEDURE MakeFullName (Name: FName;  VAR (*OUT*) FullName: ARRAY OF CHAR);

    (* Converts a relative file name to a full physical file name.  *)

    VAR k: CARDINAL;

    BEGIN
        IF (Name = NIL) OR (Name^.EntryPtr = NIL) OR (Name^.dir[0] = Nul) THEN
            FullName[0] := Nul;
            RETURN;
        ELSE
            Strings.Assign (Name^.dir, FullName);
        END (*IF*);
        IF Name^.finalname[0] = Nul THEN

            (* Strip trailing '/'. *)

            k := LENGTH (FullName);
            IF k > 0 THEN
                DEC (k);
                IF FullName[k] = '/' THEN
                    FullName[k] := Nul;
                END (*IF*);
            END (*IF*);
        ELSE
            Strings.Append (Name^.finalname, FullName);
        END (*IF*);

    END MakeFullName;

(********************************************************************************)

PROCEDURE SplitHead (VAR (*INOUT*) src: ARRAY OF CHAR;  VAR (*OUT*) head: ARRAY OF CHAR);

    (* Finds the first '/' character in src.  Everything up to the separator is *)
    (* assigned to head, the separator itself is discarded, the remainder is    *)
    (* assigned to src.  If there is no '/', src becomes the null string.       *)

    (* (Any '\' found is treated as a '/'.)                                     *)

    VAR k: CARDINAL;  ch: CHAR;

    BEGIN
        k := 0;  ch := Nul;
        LOOP
            IF k > HIGH(src) THEN
                EXIT (*LOOP*);
            END (*IF*);
            ch := src[k];
            IF (ch = Nul) OR (ch = '/') OR (ch = '\')
                       OR (k > HIGH(head)) THEN
                EXIT (*LOOP*);
            ELSE
                head[k] := ch;  INC(k);
            END (*IF*);
        END (*LOOP*);
        IF k <= HIGH(head) THEN
            head[k] := Nul;
        END (*IF*);
        IF ch <> Nul THEN
            INC (k);
        END (*IF*);
        IF k > HIGH(src) THEN
            Strings.Assign ("", src);
        ELSE
            Strings.Delete (src, 0, k);
        END (*IF*);
    END SplitHead;

(********************************************************************************)

PROCEDURE SplitTail (VAR (*INOUT*) src: ARRAY OF CHAR;
                              VAR (*OUT*) head: ARRAY OF CHAR);

    (* Finds the last '/' character in src.  Everything up to the separator is  *)
    (* assigned to head, the separator is discarded, and src is set to whatever *)
    (* is left after the separator.  If there is no '/', head becomes the null  *)
    (* string.  Special case: if the last '/' is actually the first character   *)
    (* of src, then we return with head = "/".                                  *)

    (* (Any '\' found is treated as a '/'.)                                     *)

    VAR k: CARDINAL;  ch: CHAR;

    BEGIN
        k := Strings.Length (src);  ch := Nul;
        LOOP
            IF k = 0 THEN EXIT(*LOOP*) END(*IF*);
            DEC (k);
            ch := src[k];
            IF ch = '\' THEN ch := '/' END(*IF*);
            IF ch = '/' THEN EXIT (*LOOP*) END(*IF*);
        END (*LOOP*);
        IF ch = '/' THEN
            Strings.Assign (src, head);
            Strings.Delete (src, 0, k+1);
            IF k = 0 THEN INC(k) END(*IF*);
        END (*IF*);
        head[k] := Nul;
    END SplitTail;

(********************************************************************************)

PROCEDURE SplitTail1 (VAR (*INOUT*) src: ARRAY OF CHAR;
                              VAR (*OUT*) tail: ARRAY OF CHAR);

    (* Finds the last '/' character in src.  Everything up to the separator is  *)
    (* assigned to src, the separator is discarded, and tail is set to whatever *)
    (* is left after the separator.  If there is no '/', src becomes the null   *)
    (* string.  Special case: if the last '/' is actually the first character   *)
    (* of src, then we return with src = "/".                                   *)

    (* (Any '\' found is treated as a '/'.)                                     *)

    VAR k: CARDINAL;  ch: CHAR;

    BEGIN
        k := Strings.Length (src);  ch := Nul;
        LOOP
            IF k = 0 THEN EXIT(*LOOP*) END(*IF*);
            DEC (k);
            ch := src[k];
            IF ch = '\' THEN ch := '/' END(*IF*);
            IF ch = '/' THEN EXIT (*LOOP*) END(*IF*);
        END (*LOOP*);
        IF ch = '/' THEN
            Strings.Assign (src, tail);
            Strings.Delete (tail, 0, k+1);
            IF k = 0 THEN INC(k) END(*IF*);
        ELSE
            Strings.Assign (src, tail);
        END (*IF*);
        src[k] := Nul;
    END SplitTail1;

(********************************************************************************)

PROCEDURE StripTail (VAR (*INOUT*) name: ARRAY OF CHAR);

    (* Deletes the last subdirectory specification from name, for example       *)
    (* we change A/B/C/D/ to A/B/C/.                                            *)

    VAR k: CARDINAL;

    BEGIN
        k := Strings.Length (name);
        IF k > 0 THEN
            REPEAT
                DEC (k);
            UNTIL (k = 0) OR (name[k-1] = '/') OR (name[k-1] = '\');
        END (*IF*);
        name[k] := Nul;
    END StripTail;

(********************************************************************************)

PROCEDURE TailMatch (first, second: ARRAY OF CHAR): BOOLEAN;

    (* Tests whether the tail of first matches second. *)

    VAR tail: FilenameString;

    BEGIN
        SplitTail1 (first, tail);
        RETURN NameMatch (tail, second);
    END TailMatch;

(********************************************************************************)
(*                              DIRECTORIES                                     *)
(********************************************************************************)

PROCEDURE NameIsReserved (name: ARRAY OF CHAR): BOOLEAN;

    (* Returns TRUE iff name is a device name such as CLOCK$. *)

    CONST
        List3 = "CONNULPRN";
        List4 = "COM1COM2COM3COM4KBD$LPT1LPT2LPT3";
        List6 = "CLOCK$MOUSE$";
        List7 = "SCREEN$";
        List8 = "POINTER$";

    (********************************************************************)

    PROCEDURE Match (length: CARDINAL;  list: ARRAY OF CHAR): BOOLEAN;

        (* List is the concatenation of strings of 'length' characters. *)
        (* We return TRUE iff name matches one of them.                 *)

        VAR j, k, limit: CARDINAL;

        BEGIN
            limit := LENGTH(list);
            k := 0;
            LOOP
                IF k >= limit THEN RETURN FALSE END(*IF*);
                j := 0;
                LOOP
                    IF j = length THEN RETURN TRUE END(*IF*);
                    IF CAP(name[j]) <> list[k] THEN EXIT(*LOOP*) END(*IF*);
                    INC(j);  INC(k);
                END (*LOOP*);
                INC (k, length-j);
            END (*LOOP*);
        END Match;

    (********************************************************************)

    BEGIN
        CASE LENGTH(name) OF
          |  3:  RETURN Match(3, List3);
          |  4:  RETURN Match(4, List4);
          |  6:  RETURN Match(6, List6);
          |  7:  RETURN Match(7, List7);
          |  8:  RETURN Match(8, List8);
          |  ELSE
                 RETURN FALSE;
        END (*CASE*);
    END NameIsReserved;

(********************************************************************************)

PROCEDURE FileIsADirectory (parent: FName;  name: ARRAY OF CHAR): BOOLEAN;

    (* Returns TRUE iff the file exists and is a directory.   *)

    VAR FullName: FilenameString;  D: DirectoryEntry;  result: BOOLEAN;

    BEGIN
        IF parent^.EntryPtr = NIL THEN result := FALSE
        ELSIF parent^.dir[0] = Nul THEN result := FALSE
        ELSIF name[0] = Nul THEN result := TRUE;
        ELSE
            FullName := parent^.dir;
            Strings.Append (name, FullName);

            (* Watch out for wildcard matches! *)

            result := FirstDirEntry (FullName, FALSE, TRUE, TRUE, D)
                         AND (directory IN D.attr)
                         AND TailMatch (name, D.name);
            DirSearchDone (D);
        END (*IF*);

        RETURN result;

    END FileIsADirectory;

(********************************************************************************)

PROCEDURE FindDirEntry (parent: FName;  name: ARRAY OF CHAR;
                                                  NoCheck: BOOLEAN): DirEntryPtr;

    (* Looks for name in the directory described by the directory part of       *)
    (* parent.  (The finalname component of parent is not used.)  If            *)
    (* subdirectory name is not already in the directory tree, but name exists  *)
    (* and is a directory, an entry is added to the tree.  If the NoCheck flag  *)
    (* is set, we trust the caller's assertion that name is indeed a directory, *)
    (* so we don't do the check again.                                          *)
    (* Returns NIL if the file doesn't exist.                                   *)

    (* Note: new entries are added at the tail of the appropriate list, to      *)
    (* ensure that already-existing entries have priority over those added      *)
    (* during a directory scan.                                                 *)

    (* Special case: we don't want ".." records wasting memory, so we deal      *)
    (* with them in such a way that they don't get added to the tree.           *)

    VAR father, previous, current: DirEntryPtr;

    BEGIN
        father := parent^.EntryPtr;
        IF Strings.Equal (name, "..") THEN
            RETURN father^.parent;
        END (*IF*);

        previous := NIL;  current := father^.FirstChild;
        LOOP
            IF current = NIL THEN EXIT(*LOOP*) END(*IF*);
            IF NameMatch (name, current^.name) THEN EXIT(*LOOP*) END(*IF*);
            previous := current;  current := current^.next;
        END (*LOOP*);

        IF current = NIL THEN
            IF NoCheck OR FileIsADirectory (parent, name) THEN
                NEW (current);
                Strings.Assign (name, current^.name);
                current^.parent := father;
                current^.FirstChild := NIL;  current^.next := NIL;
                current^.flags := father^.flags;
                IF previous = NIL THEN
                    father^.FirstChild := current;
                ELSE
                    previous^.next := current;
                END (*IF*);
                current^.exists := TRUE;  current^.IsADir := TRUE;
                current^.categorised := TRUE;  current^.virtual := FALSE;
                current^.link := NIL;
            END (*IF*);
        END (*IF*);

        IF (current <> NIL) AND NOT current^.categorised THEN
            CategoriseNode (current, parent^.dir);
        END (*IF*);

        RETURN current;

    END FindDirEntry;

(********************************************************************************)
(*                         THE INTERNAL DATA TYPE FName                         *)
(********************************************************************************)

PROCEDURE MakeFName (U: User;  filename: ARRAY OF CHAR): FName;

    (* Creates an FName descriptor from the filename.  The result is NIL for    *)
    (* a name that implies a nonexistent directory.  Does not check whether the *)
    (* finalname part of the name is an existing file.                          *)

    VAR result: FName;
        head: FilenameString;
        NewEntryPtr: DirEntryPtr;  FinalSlash, finished: BOOLEAN;
        k: CARDINAL;

    BEGIN
        (* For later reference, we want to know whether the name ends with '/'. *)

        k := LENGTH (filename);
        FinalSlash := (k > 0) AND ((filename[k-1] = '/') OR (filename[k-1] = '\'));

        NEW (result);
        WITH result^ DO
            Strings.Assign (filename, finalname);

            (* Start at the user's root directory or the user's current     *)
            (* directory, depending on whether the name starts with a '/'.  *)

            IF (finalname[0] = '/') OR (finalname[0] = '\') THEN
                EntryPtr := U^.TreeRoot;
                IF EntryPtr^.virtual THEN
                    dir := EntryPtr^.link^.this;
                    IF dir[0] <> Nul THEN
                        Strings.Append ('/', dir);
                    END (*IF*);
                ELSE
                    dir := "";
                END (*IF*);
                vdir := '/';
                Strings.Delete (finalname, 0, 1);
            ELSE
                dir := U^.CurrentDir;
                vdir := U^.CurrentVDir;
                EntryPtr := U^.PosInTree;
            END (*IF*);

            (* Now strip all leading directory information from finalname,  *)
            (* and move it to dir, updating EntryPtr as we go.              *)

            finished := (EntryPtr = NIL) OR (finalname[0] = Nul);
            WHILE NOT finished DO

                SplitHead (finalname, head);
                IF Strings.Equal (head, "..") THEN

                    (* Change to parent directory. *)

                    IF EntryPtr^.virtual THEN
                        dir := EntryPtr^.link^.parent;
                        StripTail (vdir);
                        EntryPtr := EntryPtr^.parent;
                    ELSE
                        EntryPtr := EntryPtr^.parent;
                        IF EntryPtr <> NIL THEN
                            StripTail (dir);
                            StripTail (vdir);
                        END (*IF*);
                    END (*IF*);

                ELSIF (head[0] = '.') AND (head[1] = Nul) THEN

                    (* Change to current directory, i.e. do nothing. *)

                ELSIF NameIsReserved (head) THEN

                    EntryPtr := NIL;

                ELSE
                    (* Try to move to a subdirectory of the current directory. *)

                    NewEntryPtr := FindDirEntry (result, head, FALSE);
                    IF (NewEntryPtr = NIL) OR NOT NewEntryPtr^.exists THEN

                        (* 'head' is not the name of a subdirectory. *)

                        finished := TRUE;
                        IF (finalname[0] = Nul) AND NOT FinalSlash THEN

                            (* Still OK, we must have arrived at a file name or *)
                            (* mask instead of a subdirectory name.             *)

                            finalname := head;

                        ELSE

                            (* Illegal file name, clear EntryPtr to show error. *)

                            EntryPtr := NIL;

                        END (*IF*);

                    ELSIF NewEntryPtr^.virtual THEN

                        (* We've reached a symbolic link.  Is it a link to an   *)
                        (* ordinary file or to a directory?                     *)

                        EntryPtr := NewEntryPtr;
                        dir := EntryPtr^.link^.this;
                        IF EntryPtr^.IsADir THEN
                            Strings.Append (head, vdir);
                            Strings.Append ('/', vdir);
                        ELSIF (finalname[0] <> Nul) OR FinalSlash THEN
                            EntryPtr := NIL;
                        ELSE
                            finalname := dir;
                            SplitTail (finalname, dir);
                            finished := TRUE;
                        END (*IF*);
                        IF dir[0] <> Nul THEN
                            Strings.Append ('/', dir);
                        END (*IF*);

                    ELSIF NewEntryPtr^.IsADir THEN

                        (* Move to subdirectory. *)

                        EntryPtr := NewEntryPtr;
                        Strings.Append (head, dir);
                        Strings.Append ('/', dir);
                        Strings.Append (head, vdir);
                        Strings.Append ('/', vdir);

                    ELSIF (finalname[0] <> Nul) OR FinalSlash THEN

                        (* Attempt to treat a file as a directory. *)

                        EntryPtr := NIL;

                    ELSE

                        (* We've reached a file rather than a directory. *)

                        EntryPtr := NewEntryPtr;
                        finalname := head;
                        finished := TRUE;

                    END (*IF*);

                END (*IF*);

                finished := finished OR (EntryPtr = NIL) OR (finalname[0] = Nul);

            END (*WHILE*);

        END (*WITH*);

        IF result^.EntryPtr = NIL THEN
            DISPOSE (result);
        END (*IF*);

        (*
        WriteString ("MakeFName: ");
        IF result = NIL THEN
            WriteString ("NIL result");  WriteLn;
        ELSE
            WriteString ("dir = ");  WriteString (result^.dir);
            WriteString (", vdir = ");  WriteString (result^.vdir);
            WriteString (", finalname = ");  WriteString (result^.finalname);
            WriteLn;
            IF result^.EntryPtr^.virtual THEN
                WriteString ("  (virtual) parent = ");  WriteString (result^.EntryPtr^.link^.parent);
                WriteString (", this = ");  WriteString (result^.EntryPtr^.link^.this);
                WriteLn;
            END (*IF*);
        END (*IF*);
        *)

        RETURN result;

    END MakeFName;

(********************************************************************************)

PROCEDURE DiscardFName (VAR (*INOUT*) name: FName);

    (* Disposes of the storage used by name. *)

    BEGIN
        IF name <> NIL THEN
            DISPOSE (name);
        END (*IF*);
    END DiscardFName;

(********************************************************************************)

PROCEDURE MakeRealName (U: User;  name: ARRAY OF CHAR;
                                            VAR (*OUT*) RealName: ARRAY OF CHAR);

    (* Converts a  name in the user's file system to a full physical name. *)

    VAR fname: FName;

    BEGIN
        fname := MakeFName (U, name);
        MakeFullName (fname, RealName);
        DiscardFName (fname);
    END MakeRealName;

(********************************************************************************)

PROCEDURE SameDrive (name1, name2: FName): BOOLEAN;

    (* Checks whether two file names are in the same physical drive. *)

    BEGIN
        IF name1 = NIL THEN
            RETURN name2 = NIL;
        ELSIF name2 = NIL THEN
            RETURN FALSE;
        ELSIF (name1^.dir[1] <> ':') OR (name2^.dir[1] <> ':') THEN
            RETURN FALSE;
        ELSE
            RETURN CAP(name1^.dir[0]) = CAP(name2^.dir[0]);
        END (*IF*);
    END SameDrive;

(********************************************************************************)
(*                         FILE OR DIRECTORY ATTRIBUTES                         *)
(********************************************************************************)

PROCEDURE FindAttributes (U: User;  filename: ARRAY OF CHAR): FileAttributes;

    (* Returns the properties of filename. *)

    CONST
        NYD = 021H;           (* New Year's Day, packed *)
        DummyD = DirectoryEntry {0, FileAttr{}, 0, NYD, 0, NYD, 0, NYD, CARD64{0,0}, 0, ""};

    VAR fname: FName;  p: DirEntryPtr;
        D: DirectoryEntry;
        result: FileAttributes;
        FullName: FilenameString;

    BEGIN
        fname := MakeFName (U, filename);
        IF fname = NIL THEN
            result.permissions := PermissionSet{};
            result.exists := FALSE;
            result.IsADir := FALSE;
            RETURN result;
        END (*IF*);

        p := fname^.EntryPtr;
        IF NOT p^.categorised THEN
            CategoriseNode (p, p^.parent^.name);
        END (*IF*);
        result.permissions := p^.flags;
        result.exists := p^.exists;
        result.IsADir := p^.IsADir;

        IF p^.virtual AND (fname^.finalname[0] = Nul) THEN
            (* Create a dummy result.  *)

            result.D := DummyD;
            IF p^.IsADir THEN
                result.D.attr  := FileAttr{directory};
            END (*IF*);

        ELSE
            (* We need a filesystem lookup to get the D data.  *)

            MakeFullName (fname, FullName);
            result.exists := result.exists AND FirstDirEntry (FullName, TRUE, TRUE, TRUE, D);
            result.D := D;
            result.IsADir := directory IN D.attr;
            DirSearchDone (D);
        END (*IF*);

        IF InHideList (U, fname^.finalname) THEN
            EXCL (result.permissions, Visible);
        END (*IF*);

        (* If the file doesn't exist, we still want to return the directory *)
        (* permissions, because for example a caller might want to know     *)
        (* whether we have permission to create a new file there.           *)

        IF NOT result.exists THEN
            result.IsADir := FALSE;
            result.D := DummyD;
        END (*IF*);

        RETURN result;

    END FindAttributes;

(********************************************************************************)
(*               PUTTING OUT A DIRECTORY LISTING, SFTP VERSION                  *)
(********************************************************************************)

TYPE
    (* We produce a directory "listing" as a linked list of items.  Another     *)
    (* module will turn this into a format suitable for transmission.           *)

    DirQ = RECORD
               head, tail: DirListPtr;
           END (*RECORD*);

(********************************************************************************)

PROCEDURE PostprocessDirList (U: User;  VAR (*INOUT*) list: DirQ);

    (* Invisible items are originally allowed to be put in the list of files    *)
    (* in a directory listing, in case they mask out items with the same name.  *)
    (* We remove them here as a final step.                                     *)
    (* File names in the listing are also UTF-8 encoded at this point.          *)

    VAR previous, p, next: DirListPtr;
        rawname: FilenameString;

    BEGIN
        p := list.head;  previous := NIL;
        WHILE p <> NIL DO
            IF (Visible IN p^.attr.permissions) AND NOT InHideList(U, p^.name) THEN
                rawname := p^.name;
                TranslateToUTF8 (rawname, p^.name);
                previous := p;
                p := p^.next;
            ELSE
                (* Delete this item from the list. *)
                next := p^.next;
                IF previous = NIL THEN
                    list.head := next;
                ELSE
                    previous^.next := next;
                END (*IF*);
                IF next = NIL THEN
                    list.tail := previous;
                END (*IF*);
                DISPOSE (p);
                p := next;
            END (*IF*);
        END (*WHILE*);

    END PostprocessDirList;

(********************************************************************************)

PROCEDURE AddItemToList (VAR (*INOUT*) list: DirQ;  D: DirectoryEntry;
                                                    flags: PermissionSet);

    (* Adds one directory entry to the list of items waiting to be listed.      *)
    (* Returns FALSE if this duplicates an item already present in the list.    *)
    (* We keep the list sorted by name.                                         *)

    VAR previous, current, this: DirListPtr;  test: INTEGER;

    BEGIN
        NEW (this);
        this^.next := NIL;
        this^.name := D.name;
        this^.attr.D := D;
        this^.attr.permissions := flags;
        this^.attr.exists := TRUE;
        this^.attr.IsADir := directory IN D.attr;

        (* REMARK 1: p is, I think, a pointer to the parent directory, so we    *)
        (* can probably take only its flags.                                    *)
        (* REMARK 2: the exists field is probably redundant, because I don't    *)
        (* think we can get to this point with an item that doesn't exist.      *)

        previous := NIL;  current := list.head;  test := +1;
        LOOP
            IF current = NIL THEN EXIT(*LOOP*) END(*IF*);
            test := CompareStrings (D.name, current^.name);
            IF test <= 0 THEN
                EXIT (*LOOP*);
            ELSE
                previous := current;  current := current^.next;
            END (*IF*);
        END (*LOOP*);

        IF test <> 0 THEN

            (* Insert new item between previous and current. *);

            IF previous = NIL THEN list.head := this
            ELSE previous^.next := this;
            END (*IF*);
            this^.next := current;
            IF current = NIL THEN
                list.tail := this;
            END (*IF*);

        END (*IF*);

    END AddItemToList;

(********************************************************************************)

PROCEDURE ListRealDirectory (arg: FName;  VAR (*INOUT*) list: DirQ;
                                                    manager, MayExpand: BOOLEAN);

    (* Sends a directory listing, as specified by arg.                      *)
    (* The listed file names, as added to list, are in UTF-8 encoding, but  *)
    (* arg uses file names in the native character set; translation to      *)
    (* UTF-8 does not occur until the final step.                           *)

    VAR D: DirectoryEntry;
        success: BOOLEAN;

    (****************************************************************************)

    PROCEDURE HandleOneEntry (flags: PermissionSet);

        (* Sends one line of a directory listing.  If updatesublist is TRUE,    *)
        (* and the entry is a subdirectory, then the subdirectory is added to   *)
        (* sublist.  The node we are handling is described by 'arg', and the    *)
        (* caller has put its directory details into D.                         *)

        VAR SendIt: BOOLEAN;  p: DirEntryPtr;

        BEGIN
            SendIt := TRUE;

            (* Skip over the "." and ".." entries.   *)

            IF D.name[0] = '.' THEN
                SendIt := (D.name[1] <> Nul) AND
                                 ((D.name[1] <> '.') OR (D.name[2] <> Nul));
            END (*IF*);

            (* Work out whether we want to send this entry. *)

            IF SendIt THEN
                SendIt := manager
                          OR NOT ((hidden IN D.attr) OR (system IN D.attr));
            END (*IF*);

            (* Don't send invisible directories.  Non-directory files are       *)
            (* different: we might encounter this file more than once in our    *)
            (* processing, and so we want to keep the first instance in our     *)
            (* "to send" list -- even though we finally won't send it if it's   *)
            (* invisible -- so that it will take precedence over any future     *)
            (* attempts to put it in the list.                                  *)

            IF SendIt AND (directory IN D.attr) THEN

                (* I'm not sure whether this call to FindDirEntry is redundant  *)
                (* by now, but this works so I'll leave it alone.               *)

                p := FindDirEntry (arg, D.name, TRUE);
                flags := p^.flags;
                SendIt := Visible IN flags;
            END (*IF*);

            (* Queue this item on the list of things to be listed. *)

            IF SendIt THEN AddItemToList (list, D, flags) END (*IF*);

        END HandleOneEntry;

    (****************************************************************************)

    CONST DummyTime = 0;  NYD = 33;  (* 01 Jan 1980 *)

    VAR Name, temp, mask: FilenameString;  q: DirEntryPtr;
        OK: BOOLEAN;  j: CARDINAL;

    BEGIN                       (* body of ListRealDirectory *)

        (* An empty filename mask means "entire directory", unless the     *)
        (* caller has set MayExpand = FALSE.                               *)

        IF (arg^.finalname[0] = Nul) AND MayExpand THEN
            arg^.finalname[0] := '*';  arg^.finalname[1] := Nul;
        END (*IF*);

        (* Update our directory tree to ensure that it contains all     *)
        (* subdirectories of the current directory that match our mask. *)

        q := arg^.EntryPtr;
        IF (NOT q^.virtual) OR (q^.link^.this[0] <> Nul) THEN
            MakeFullName (arg, Name);
            success := FirstDirEntry (Name, FALSE, TRUE, TRUE, D);
            WHILE success DO
                OK := directory IN D.attr;
                IF OK AND (D.name[0] = '.') THEN
                    OK := (D.name[1] <> Nul) AND
                               ((D.name[1] <> '.') OR (D.name[2] <> Nul));
                END (*IF*);

                IF OK THEN
                    EVAL (FindDirEntry (arg, D.name, TRUE));
                END (*IF*);
                success := NextDirEntry (D);

            END (*WHILE*);
            DirSearchDone (D);
        END (*IF*);

        (* First handle the items that are explicitly listed in our     *)
        (* directory tree, including the directories we've just added.  *)

        q := arg^.EntryPtr;
        IF arg^.finalname[0] = Nul THEN
            mask := arg^.vdir;
            j := LENGTH(mask);
            IF (j > 0) AND (mask[j-1] = '/') THEN
                mask[j-1] := Nul;
            END (*IF*);
            SplitTail (mask, temp);
        ELSE
            q := q^.FirstChild;
            mask := arg^.finalname;
        END (*IF*);

        WHILE q <> NIL DO

            IF NOT q^.categorised THEN
                CategoriseNode (q, arg^.dir);
            END (*IF*);

            IF q^.exists AND WildMatch(q^.name, mask) THEN

                IF q^.virtual THEN

                    (* This entry describes a link. *)

                    IF (q^.link^.this[0] = Nul) OR
                      ((q^.link^.this[1] = ':') AND (q^.link^.this[2] = Nul)) THEN

                        (* We have a disk root directory or a pseudo-directory.     *)
                        (* Construct a dummy DirectoryEntry record for it.          *)

                        WITH D DO
                            dirHandle := 0;
                            attr      := FileAttr{directory};
                            timePkd   := DummyTime;
                            datePkd   := NYD;
                            timeCre   := DummyTime;
                            dateCre   := NYD;
                            timeAcc   := DummyTime;
                            dateAcc   := NYD;
                            size      := CARD64{0,0};
                        END (*WITH*);
                        Strings.Assign (q^.name, D.name);
                        HandleOneEntry (q^.flags);

                    ELSE

                        IF FirstDirEntry (q^.link^.this, TRUE, FALSE, TRUE, D) THEN
                            Strings.Assign (q^.name, D.name);
                            HandleOneEntry (q^.flags);
                        END (*IF*);
                        DirSearchDone (D);

                    END (*IF*);

                ELSE

                    (* Not a link, this is an ordinary directory or file. *)

                    temp := arg^.finalname;
                    IF temp[0] <> Nul THEN
                        arg^.finalname := q^.name;
                    END (*IF*);
                    MakeFullName (arg, Name);
                    IF temp[0] <> Nul THEN
                        arg^.finalname := temp;
                    END (*IF*);

                    IF FirstDirEntry (Name, TRUE, FALSE, TRUE, D) THEN
                        HandleOneEntry (q^.flags);
                    END (*IF*);
                    DirSearchDone (D);

                END (*IF*);

            END (*IF*);

            IF arg^.finalname[0] = Nul THEN
                q := NIL;
            ELSE
                q := q^.next;
            END (*IF*);

        END (*WHILE*);

        (* Now the main part of the listing.  (Omit this part if we're in a     *)
        (* pseudo-directory.)  We can ignore directories at this stage, because *)
        (* these have already been listed above.                                *)

        q := arg^.EntryPtr;
        IF (NOT q^.virtual) OR (q^.link^.this[0] <> Nul) THEN
            MakeFullName (arg, Name);
            success := FirstDirEntry (Name, FALSE, FALSE, TRUE, D);
            WHILE success DO

                IF NOT (directory IN D.attr) THEN
                    HandleOneEntry (q^.flags);
                END (*IF*);
                success := NextDirEntry (D);

            END (*WHILE*);
            DirSearchDone (D);
        END (*IF*);

    END ListRealDirectory;

(********************************************************************************)

PROCEDURE ListDirectory (U: User;  VAR (*INOUT*) arg: FName): DirListPtr;

    (* Sends a directory listing, as specified by arg.  On return               *)
    (* arg has been disposed of.  Parameter arg uses the native character set,  *)
    (* but the file names returned in the list are in UTF-8 encoding.           *)

    VAR result: DirQ;
        MayExpand: BOOLEAN;

    BEGIN
        MayExpand := TRUE;
        result.head := NIL;
        result.tail := NIL;

        (* Special case: if the root node is invisible, we allow an empty   *)
        (* listing to be returned.                                          *)

        IF (arg^.EntryPtr^.parent = NIL) AND NOT (Visible IN arg^.EntryPtr^.flags) THEN
            MayExpand := FALSE;
        END (*IF*);

        ListRealDirectory (arg, result, U^.category=Manager, MayExpand);
        PostprocessDirList (U, result);

        DISPOSE (arg);
        RETURN result.head;

    END ListDirectory;

(********************************************************************************)

PROCEDURE CreateDirectoryListing (U: User;  dirname: ARRAY OF CHAR): DirListPtr;

    (* Creates a directory listing in the form of a linked list.    *)
    (* File names returned in the list are in UTF-8 encoding.       *)

    VAR arg: FName;

    BEGIN
        arg := MakeFName (U, dirname);
        RETURN ListDirectory (U, arg);
    END CreateDirectoryListing;

(********************************************************************************)
(*                             IS A DIRECTORY EMPTY?                            *)
(********************************************************************************)

PROCEDURE EmptyDirectory (U: User;  dir: ARRAY OF CHAR): BOOLEAN;

    (* Returns TRUE if dir contains no files. *)

    VAR arg: FName;
        D: DirectoryEntry;
        q: DirEntryPtr;
        mask: FilenameString;

    BEGIN
        arg := MakeFName (U, dir);
        IF (arg = NIL) OR (arg^.finalname[0] <> Nul) THEN
            RETURN FALSE;
        END (*IF*);
        q := arg^.EntryPtr;
        IF (q = NIL) OR (q^.FirstChild <> NIL) THEN
            RETURN FALSE;
        END (*IF*);

        (* In the case of a pseudo-directory, that is all we need to know. *)

        IF q^.virtual AND (q^.link^.this[0] = Nul) THEN
            RETURN TRUE;
        END (*IF*);

        (* That takes care of the simple cases.  Now we have to look for a file *)
        (* in the directory that is not '.' or '..'.                            *)

        MakeFullName (arg, mask);
        Strings.Append ("\*", mask);
        IF NOT FirstDirEntry (mask, TRUE, FALSE, TRUE, D) THEN
            DirSearchDone (D);
            RETURN TRUE;
        END (*IF*);
        LOOP
            IF (D.name <> '.') AND (D.name <> '..') THEN
                DirSearchDone (D);
                RETURN FALSE;
            END (*IF*);
            IF NOT NextDirEntry (D) THEN
                DirSearchDone (D);
                RETURN TRUE;
            END (*IF*);
        END (*LOOP*);
    END EmptyDirectory;

(********************************************************************************)
(*                              INITIALISATION                                  *)
(********************************************************************************)

BEGIN
    UserCount := NIL;
    CreateLock (UserCountLock);
END FDUsers.

