(**************************************************************************)
(*                                                                        *)
(*  SFTPD server                                                          *)
(*  Copyright (C) 2022   Peter Moylan                                     *)
(*                                                                        *)
(*  This program is free software: you can redistribute it and/or modify  *)
(*  it under the terms of the GNU General Public License as published by  *)
(*  the Free Software Foundation, either version 3 of the License, or     *)
(*  (at your option) any later version.                                   *)
(*                                                                        *)
(*  This program is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU General Public License for more details.                          *)
(*                                                                        *)
(*  You should have received a copy of the GNU General Public License     *)
(*  along with this program.  If not, see <http://www.gnu.org/licenses/>. *)
(*                                                                        *)
(*  To contact author:   http://www.pmoylan.org   peter@pmoylan.org       *)
(*                                                                        *)
(**************************************************************************)

IMPLEMENTATION MODULE FTPSession;

        (********************************************************)
        (*                                                      *)
        (*                SFTPD session handler                 *)
        (*                                                      *)
        (*  Programmer:         P. Moylan                       *)
        (*  Started:            11 November 2021                *)
        (*  Last edited:        18 April 2022                   *)
        (*  Status:             Complete                        *)
        (*                                                      *)
        (* We support SFTP version 3, as described in           *)
        (*         draft-ietf-secsh-filexfer-02                 *)
        (* Draft standards for higher version numbers exist,    *)
        (* but are not widely supported.                        *)
        (*                                                      *)
        (********************************************************)


FROM SYSTEM IMPORT ADDRESS, CARD8, CAST, MOVE, ADR;

IMPORT Strings;

FROM PacketQueues IMPORT
    (* type *)  PacketQ,
    (* proc *)  GetPacket, CloseQueue;

FROM Channels IMPORT
    (* proc *)  SendData(*, HandleChannelCloseRequest*);

FROM FDFiles IMPORT
    (* type *)  FileHandle, AttrRecord,
    (* proc *)  ReducePath, GetAttributes, MakeDirHandle, CloseHandle, MakeDirListing,
                ParseAttrs, MakeFileHandle, FilenameFromHandle, ReadData, WriteData,
                RemoveFile, DoRename, MkDir, RmDir;

FROM FDUsers IMPORT
    (* type *)  User,
    (* proc *)  DestroyUserData;

FROM CodePage IMPORT
    (* type *)  TranslateFromUTF8, TranslateFromUTF8r;

FROM FileOps IMPORT
    (* type *)  FilenameString;

FROM Consts IMPORT
    (* const*)  SSH_FXP_INIT, SSH_FXP_VERSION, SSH_FXP_OPEN, SSH_FXP_CLOSE,
                SSH_FXP_READ, SSH_FXP_WRITE, SSH_FXP_LSTAT, SSH_FXP_FSTAT,
                SSH_FXP_SETSTAT, SSH_FXP_FSETSTAT, SSH_FXP_OPENDIR, SSH_FXP_READDIR,
                SSH_FXP_REMOVE, SSH_FXP_MKDIR, SSH_FXP_RMDIR, SSH_FXP_REALPATH,
                SSH_FXP_STAT, SSH_FXP_RENAME, SSH_FXP_READLINK, SSH_FXP_SYMLINK,
                SSH_FXP_STATUS, SSH_FXP_HANDLE, SSH_FXP_DATA,
                SSH_FXP_NAME, SSH_FXP_ATTRS, SSH_MSG_TERMINATE;

FROM VarStrings IMPORT
    (* type *)  ByteStr,
    (* proc *)  MakeBS, DiscardBS, GetCard, GetCharString, GetByteChunk, GetByteString,
                AddByte, AddCard, AddByteChunk, AddByteString, AddCharString;

FROM Arith64 IMPORT
    (* type *)  CARD64;

FROM MiscFuncs IMPORT
    (* proc *)  AppendCard, Swap4, EVAL;

FROM TransLog IMPORT
    (* type *)  TransactionLogID,
    (* proc *)  LogTransaction, LogTransactionL;

FROM TaskControl IMPORT
    (* proc *)  CreateTask1;

FROM Storage IMPORT
    (* proc *)  ALLOCATE, DEALLOCATE;

(************************************************************************)

CONST
    Nul = CHR(0);

TYPE
    ErrMess = ARRAY [0..79] OF CHAR;

    (* The state of a session.  *)

    Session = POINTER TO
                RECORD
                    chan: CARDINAL;
                    user: User;
                    logID: TransactionLogID;
                END (*RECORD*);

(************************************************************************)
(* NOTE: all packets, in either direction, have the form                *)
(*          uint32      length                                          *)
(*          byte        type                                            *)
(*          bytes       payload                                         *)
(* In the majority of cases the payload begins with a CARD32 id         *)
(************************************************************************)

(************************************************************************)
(*                     RETURNING FILE ATTRIBUTES                        *)
(************************************************************************)

CONST
    (* Flags for attribute response. *)

    SSH_FILEXFER_ATTR_SIZE          = 00000001H;
    SSH_FILEXFER_ATTR_UIDGID        = 00000002H;
    SSH_FILEXFER_ATTR_PERMISSIONS   = 00000004H;
    SSH_FILEXFER_ATTR_ACMODTIME     = 00000008H;
    SSH_FILEXFER_ATTR_EXTENDED      = 80000000H;

    (* Status codes. *)

    SSH_FX_OK                       =     0;
    SSH_FX_EOF                      =     1;
    SSH_FX_NO_SUCH_FILE             =     2;
    SSH_FX_PERMISSION_DENIED        =     3;
    SSH_FX_FAILURE                  =     4;
    SSH_FX_BAD_MESSAGE              =     5;
    SSH_FX_NO_CONNECTION            =     6;
    SSH_FX_CONNECTION_LOST          =     7;
    SSH_FX_OP_UNSUPPORTED           =     8;

(************************************************************************)

PROCEDURE AttrsResponse (U: User;  fname: ARRAY OF CHAR;  id: CARDINAL): ByteStr;

    (* Creates a response packet which is an SSH_FXP_ATTRS response for *)
    (* the file or directory fname, or an SSH_FXP_STATUS response in    *)
    (* case of error.  The fname argument is in our native charset.     *)

    VAR N, pos: CARDINAL;
        attrs: AttrRecord;
        message: ARRAY [0..63] OF CHAR;
        exists: BOOLEAN;
        reply: ByteStr;

    BEGIN
        (* First we must find the file attributes. *)

        exists := GetAttributes (U, fname, attrs);

        (* I would expect the flags to be ATTR_SIZE + ATTR_PERMISSIONS + ACDMODTIME = 13 *)
        (* This adds up to 6 32-bit words. *)

        IF exists THEN
            N := 1 + 4 + 24;
        ELSE
            message := "No such file";
            N := 1 + 4 + 4 + 4 + LENGTH(message) + 6;
        END (*IF*);

        IF exists THEN
            MakeBS (reply, N + 4);
            pos := 0;
            AddCard (reply, pos, N);
            AddByte (reply, pos, SSH_FXP_ATTRS);
            AddCard (reply, pos, id);
            AddCard (reply, pos, attrs.flags);
            AddCard (reply, pos, attrs.size.high);
            AddCard (reply, pos, attrs.size.low);
            AddCard (reply, pos, attrs.permissions);
            AddCard (reply, pos, attrs.atime);
            AddCard (reply, pos, attrs.mtime);
        ELSE
            MakeBS (reply, N + 4);
            pos := 0;
            AddCard (reply, pos, N);
            AddByte (reply, pos, SSH_FXP_STATUS);
            AddCard (reply, pos, id);
            AddCard (reply, pos, SSH_FX_NO_SUCH_FILE);
            AddCharString (reply, pos, message);
            AddCharString (reply, pos, "en");
        END (*IF*);

        RETURN reply;

    END AttrsResponse;

(************************************************************************)
(*                        GENERIC STATUS REPLY                          *)
(************************************************************************)

PROCEDURE StatusReply (chan: CARDINAL;  id, code: CARDINAL);

    (* Sends an SSH_FXP_STATUS message to the client.  A weakness of    *)
    (* the SFTP standard is that there is no provision for giving more  *)
    (* detail about why an operation failed.                            *)

    VAR pos, N: CARDINAL;
        reply: ByteStr;
        message: ARRAY [0..511] OF CHAR;

    BEGIN
        IF code = 0 THEN message := "OK";
        ELSIF code = 2 THEN message := "no such file";
        ELSIF code = 3 THEN message := "permission denied";
        ELSIF code = 4 THEN message := "operation failed";
        ELSIF code = 5 THEN message := "Invalid path string";
        END (*IF*);
        N := 1 + 4 + 4 + 4 + LENGTH(message) + 6;
        MakeBS (reply, N + 4);
        pos := 0;
        AddCard (reply, pos, N);
        AddByte (reply, pos, SSH_FXP_STATUS);
        AddCard (reply, pos, id);
        AddCard (reply, pos, code);
        AddCharString (reply, pos, message);
        AddCharString (reply, pos, "en");

        SendData (chan, reply);

    END StatusReply;

(************************************************************************)

PROCEDURE StatusReply2 (chan: CARDINAL;  id, code: CARDINAL;
                                            message: ErrMess);

    (* Sends an SSH_FXP_STATUS message to the client.  A weakness of    *)
    (* the SFTP standard is that there is no provision for giving more  *)
    (* detail about why an operation failed, so we allow the caller to  *)
    (* override the default message.                                    *)

    VAR pos, N: CARDINAL;
        reply: ByteStr;

    BEGIN
        IF message[0] = Nul THEN
            (* Use the default message for this code. *)

            IF code = 0 THEN message := "OK";
            ELSIF code = 1 THEN message := "end of file";
            ELSIF code = 2 THEN message := "no such file";
            ELSIF code = 3 THEN message := "permission denied";
            ELSIF code = 4 THEN message := "operation failed";
            ELSIF code = 5 THEN message := "Invalid path string";
            END (*IF*);
        END (*IF*);
        N := 1 + 4 + 4 + 4 + LENGTH(message) + 6;
        MakeBS (reply, N + 4);
        pos := 0;
        AddCard (reply, pos, N);
        AddByte (reply, pos, SSH_FXP_STATUS);
        AddCard (reply, pos, id);
        AddCard (reply, pos, code);
        AddCharString (reply, pos, message);
        AddCharString (reply, pos, "en");

        SendData (chan, reply);

    END StatusReply2;

(************************************************************************)
(*                     1. SESSION INITIALISATION                        *)
(************************************************************************)

PROCEDURE InitFTPSession (sess: Session;  pkt: ByteStr);

    (* Beginning of a new session.  *)

    CONST ourversion = 3;

    VAR pos, clientversion: CARDINAL;
        reply: ByteStr;
        (*message: ARRAY [0..63] OF CHAR;*)

    BEGIN
        pos := 5;
        clientversion := GetCard (pkt, pos);
        DiscardBS (pkt);

        (*
        message := "Client SFTP version is ";  AppendCard (clientversion, message);
        LogTransaction (sess^.logID, message);
        *)
        IF pkt.size > 9 THEN
            LogTransactionL (sess^.logID, "Extension data have been ignored");
        END (*IF*);

        MakeBS (reply, 9 (*+ 12 + 7*));
        pos := 0;
        AddCard (reply, pos, 5);
        reply.data^[pos] := SSH_FXP_VERSION;  INC (pos);
        AddCard (reply, pos, ourversion);

        SendData (sess^.chan, reply);

    END InitFTPSession;

(************************************************************************)
(*                           2. VERSION                                 *)
(************************************************************************)

PROCEDURE ReceiveVersion (sess: Session;  pkt: ByteStr);

    (* We should never receive this, so silently discard the packet. *)

    BEGIN
        DiscardBS (pkt);
    END ReceiveVersion;

(************************************************************************)
(*                             3. OPEN                                  *)
(************************************************************************)

PROCEDURE OpenFile (sess: Session;  pkt: ByteStr);

    (* Opens a non-directory file, returns a handle to client. *)

    VAR pos, id, code, N, pflags: CARDINAL;
        handle: FileHandle;
        attrs: AttrRecord;
        reply: ByteStr;
        filename, tempname, message: FilenameString;

    BEGIN
        pos := 5;
        id := GetCard (pkt, pos);
        GetCharString (pkt, pos, filename);
        pflags := GetCard (pkt, pos);
        attrs := ParseAttrs (pkt, pos);
        DiscardBS (pkt);
        Strings.Assign (filename, tempname);
        TranslateFromUTF8 (tempname, filename);
        message := "OPEN ";
        Strings.Append (filename, message);
        LogTransaction (sess^.logID, message);
        code := MakeFileHandle (sess^.user, filename, pflags, attrs, handle);
        IF code = 0 THEN

             (* SSH_FXP_HANDLE response *)

            N := 1 + 4 + 4 + SIZE(handle);
            MakeBS (reply, N + 4);
            pos := 0;
            AddCard (reply, pos, N);
            AddByte (reply, pos, SSH_FXP_HANDLE);
            AddCard (reply, pos, id);
            AddByteString (reply, pos, handle, SIZE(handle));

        ELSE
            (* SSH_FXP_STATUS response *)

            IF code = 2 THEN
                Strings.Assign (filename, message);
                Strings.Append (" does not exist", message);
            ELSIF code = 3 THEN
                Strings.Assign ("permission denied", message);
            ELSIF code = 5 THEN
                Strings.Assign ("badly specified request", message);
            END (*IF*);
            N := 1 + 4 + 4 + 4 + LENGTH(message) + 6;
            MakeBS (reply, N + 4);
            pos := 0;
            AddCard (reply, pos, N);
            AddByte (reply, pos, SSH_FXP_STATUS);
            AddCard (reply, pos, id);
            AddCard (reply, pos, code);
            AddCharString (reply, pos, message);
            AddCharString (reply, pos, "en");
        END (*IF*);

        SendData (sess^.chan, reply);

    END OpenFile;

(************************************************************************)
(*                             4. CLOSE                                 *)
(*     This means close a file or directory, not close a session.       *)
(************************************************************************)

PROCEDURE CloseOperation (sess: Session;  pkt: ByteStr);

    VAR pos, id, code, N: CARDINAL;
        H: ARRAY [0..3] OF CARD8;
        handle: FileHandle;
        reply: ByteStr;
        path, message: FilenameString;

    BEGIN
        pos := 5;
        id := GetCard (pkt, pos);
        EVAL (GetByteString (pkt, pos, H));
        DiscardBS (pkt);
        handle := CAST(FileHandle, H);
        LogTransactionL (sess^.logID, "CLOSE");
        code := CloseHandle (handle);

        (* SSH_FXP_STATUS response *)

        Strings.Assign (path, message);
        IF code = 0 THEN
            message := "OK";
        ELSIF code = 2 THEN
            message := "Invalid handle";
        ELSIF code = 5 THEN
            message := "Handle has incorrect length";
        END (*IF*);
        N := 1 + 4 + 4 + 4 + LENGTH(message) + 6;
        MakeBS (reply, N + 4);
        pos := 0;
        AddCard (reply, pos, N);
        AddByte (reply, pos, SSH_FXP_STATUS);
        AddCard (reply, pos, id);
        AddCard (reply, pos, code);
        AddCharString (reply, pos, message);
        AddCharString (reply, pos, "en");

        SendData (sess^.chan, reply);

    END CloseOperation;

(************************************************************************)
(*                             5. READ                                  *)
(************************************************************************)

PROCEDURE ReadFile (sess: Session;  pkt: ByteStr);

    (* Reads data from a file whose handle is specified.  *)

    VAR pos, id, length, code, N: CARDINAL;
        offset: CARD64;
        H: ARRAY [0..3] OF CARD8;
        message, filename: FilenameString;
        handle: FileHandle;
        data, reply: ByteStr;

    BEGIN
        pos := 5;
        id := GetCard (pkt, pos);
        EVAL (GetByteString (pkt, pos, H));
        handle := CAST(FileHandle, H);
        offset.high := GetCard (pkt, pos);
        offset.low := GetCard (pkt, pos);
        length := GetCard (pkt, pos);
        DiscardBS (pkt);

        (* Reply should be an SSH_FXP_DATA message, or SSH_FXP_STATUS   *)
        (* in case of failure or EOF.                                   *)

        code := ReadData (handle, offset, length, data);

        IF code <> 0 THEN
            IF code = 1 THEN
                Strings.Assign ("Downloaded ", message);
            ELSE
                Strings.Assign ("Failed to download ", message);
            END (*IF*);
            FilenameFromHandle (handle, filename);
            Strings.Append (filename, message);
            LogTransaction (sess^.logID, message);
        END (*IF*);

        IF code = 0 THEN
            (* SSH_FXP_DATA response *)

            N := 1 + 4 + 4 + data.size;
            MakeBS (reply, N+4);
            pos := 0;
            AddCard (reply, pos, N);
            AddByte (reply, pos, SSH_FXP_DATA);
            AddCard (reply, pos, id);
            AddCard (reply, pos, data.size);
            MOVE (data.data, ADR(reply.data^[pos]), data.size);
        ELSE
            (* SSH_FXP_STATUS response *)

            IF code = 1 THEN
                message := "End of file";
            ELSIF code = 4 THEN
                message := "read failure";
            ELSIF code = 5 THEN
                message := "bad handle";
            END (*IF*);
            N := 1 + 4 + 4 + 4 + LENGTH(message) + 6;
            MakeBS (reply, N + 4);
            pos := 0;
            AddCard (reply, pos, N);
            AddByte (reply, pos, SSH_FXP_STATUS);
            AddCard (reply, pos, id);
            AddCard (reply, pos, code);
            AddCharString (reply, pos, message);
            AddCharString (reply, pos, "en");
        END (*IF*);

        DiscardBS (data);
        SendData (sess^.chan, reply);

    END ReadFile;

(************************************************************************)
(*                              6. WRITE                                *)
(************************************************************************)

PROCEDURE WriteFile (sess: Session;  pkt: ByteStr);

    (* Writes data to a file whose handle is specified.  *)

    VAR pos, id, length, code, N: CARDINAL;
        offset: CARD64;
        H: ARRAY [0..3] OF CARD8;
        p: ADDRESS;
        handle: FileHandle;
        reply: ByteStr;
        message: FilenameString;

    BEGIN
        pos := 5;
        id := GetCard (pkt, pos);
        EVAL (GetByteString (pkt, pos, H));
        handle := CAST(FileHandle, H);
        offset.high := GetCard (pkt, pos);
        offset.low := GetCard (pkt, pos);
        length := GetCard (pkt, pos);
        LogTransactionL (sess^.logID, "WRITE");
        p := ADR(pkt.data^[pos]);
        code := WriteData (handle, offset, length, p^);
        DiscardBS (pkt);

        (* Reply should be SSH_FXP_STATUS   *)

       (* SSH_FXP_STATUS response *)

        IF code = 1 THEN
            message := "End of file";
        ELSIF code = 4 THEN
            message := "write failure";
        ELSIF code = 5 THEN
            message := "bad handle";
        END (*IF*);
        N := 1 + 4 + 4 + 4 + LENGTH(message) + 6;
        MakeBS (reply, N + 4);
        pos := 0;
        AddCard (reply, pos, N);
        AddByte (reply, pos, SSH_FXP_STATUS);
        AddCard (reply, pos, id);
        AddCard (reply, pos, code);
        AddCharString (reply, pos, message);
        AddCharString (reply, pos, "en");

        SendData (sess^.chan, reply);

    END WriteFile;

(************************************************************************)
(*                               7. LSTAT                               *)
(*                             see 17. STAT                             *)
(************************************************************************)

(************************************************************************)
(*                               8. FSTAT                               *)
(************************************************************************)

PROCEDURE FStatRequest (sess: Session;  pkt: ByteStr);

    (* Returns the attributes of a file specified by its handle. *)

    VAR pos, id: CARDINAL;
        H: ARRAY [0..3] OF CARD8;
        handle: FileHandle;
        reply: ByteStr;
        filename, message: FilenameString;

    BEGIN
        pos := 5;
        id := GetCard (pkt, pos);
        EVAL (GetByteString (pkt, pos, H));
        handle := CAST(FileHandle, H);
        DiscardBS (pkt);

        FilenameFromHandle (handle, filename);
        message := "FSTAT ";
        Strings.Append (filename, message);
        LogTransaction (sess^.logID, message);
        IF filename[0] = Nul THEN
            StatusReply2 (sess^.chan, id, 2, "invalid handle");
        ELSE
            reply := AttrsResponse (sess^.user, filename, id);
            SendData (sess^.chan, reply);
        END (*IF*);

    END FStatRequest;

(************************************************************************)
(*                              9. SETSTAT                              *)
(*                             10. FSETSTAT                             *)
(************************************************************************)

PROCEDURE Unimplemented (sess: Session;  pkt: ByteStr);

    (* We don't support these, so just return an error message. *)

    VAR pos, id: CARDINAL;

    BEGIN
        pos := 5;
        id := GetCard (pkt, pos);
        DiscardBS (pkt);
        StatusReply2 (sess^.chan, id, SSH_FX_OP_UNSUPPORTED,
                                "This operation not supported");
    END Unimplemented;

(************************************************************************)
(*                              11. OPENDIR                             *)
(************************************************************************)

PROCEDURE OpenDirectory (sess: Session;  pkt: ByteStr);

    VAR pos, id, code, N: CARDINAL;
        handle: FileHandle;
        reply: ByteStr;
        path, tempname, message: FilenameString;

    BEGIN
        pos := 5;
        id := GetCard (pkt, pos);
        GetCharString (pkt, pos, path);
        DiscardBS (pkt);
        Strings.Assign (path, tempname);
        TranslateFromUTF8 (tempname, path);
        message := "OPENDIR ";
        Strings.Append (path, message);
        LogTransaction (sess^.logID, message);
        code := MakeDirHandle (sess^.user, path, handle);
        IF code = 0 THEN

             (* SSH_FXP_HANDLE response *)

            N := 13;
            MakeBS (reply, N + 4);
            pos := 0;
            AddCard (reply, pos, N);
            AddByte (reply, pos, SSH_FXP_HANDLE);
            AddCard (reply, pos, id);
            AddByteString (reply, pos, handle, SIZE(handle));

        ELSE
            (* SSH_FXP_STATUS response *)

            Strings.Assign (path, message);
            IF code = 1 THEN
                Strings.Append (" is not a directory", message);
                code := 3;
            ELSIF code = 2 THEN
                Strings.Append (" does not exist", message);
            ELSIF code = 3 THEN
                Strings.Append (" does not have read permission", message);
            END (*IF*);
            N := 1 + 4 + 4 + 4 + LENGTH(message) + 6;
            MakeBS (reply, N + 4);
            pos := 0;
            AddCard (reply, pos, N);
            AddByte (reply, pos, SSH_FXP_STATUS);
            AddCard (reply, pos, id);
            AddCard (reply, pos, code);
            AddCharString (reply, pos, message);
            AddCharString (reply, pos, "en");
        END (*IF*);

        SendData (sess^.chan, reply);

    END OpenDirectory;

(************************************************************************)
(*                           12. READDIR                                *)
(************************************************************************)

PROCEDURE ListDirectory (sess: Session;  pkt: ByteStr);

    (* Returns a directory listing (or part of a listing) from a list   *)
    (* that has already been prepared by an OpenDirectory call.         *)

    VAR pos, id, N: CARDINAL;
        H: ARRAY [0..3] OF CARD8;
        handle: FileHandle;
        listing: ByteStr;
        reply: ByteStr;
        message: FilenameString;

    BEGIN
        pos := 5;
        id := GetCard (pkt, pos);
        EVAL (GetByteString (pkt, pos, H));
        DiscardBS (pkt);
        handle := CAST(FileHandle, H);

        MakeDirListing (sess^.user, handle, listing);
        IF listing.size = 0 THEN
            (* SSH_FXP_STATUS response *)

            message := "End of listing";
            N := 1 + 4 + 4 + 4 + LENGTH(message) + 6;
            MakeBS (reply, N + 4);
            pos := 0;
            AddCard (reply, pos, N);
            AddByte (reply, pos, SSH_FXP_STATUS);
            AddCard (reply, pos, id);
            AddCard (reply, pos, SSH_FX_EOF);
            AddCharString (reply, pos, message);
            AddCharString (reply, pos, "en");

        ELSE
            (* SSH_FXP_NAME response *)
            (* In the listing the file names are already in UTF-8,  *)
            (* so no further translation is necessary.              *)

            N := 1 + 4 + listing.size;
            MakeBS (reply, N+4);
            pos := 0;
            AddCard (reply, pos, N);
            AddByte (reply, pos, SSH_FXP_NAME);
            AddCard (reply, pos, id);
            AddByteChunk (reply, pos, listing.data^, listing.size);
            DiscardBS (listing);
        END (*IF*);

        SendData (sess^.chan, reply);

    END ListDirectory;

(************************************************************************)
(*                            13. REMOVE                                *)
(************************************************************************)

PROCEDURE DeleteFile (sess: Session;  pkt: ByteStr);

    (* Deletes a file, if we have permission to do so.  *)
    (* Possible responses are:                          *)
    (*      0       OK                                  *)
    (*      2       no such file                        *)
    (*      3       permission denied                   *)
    (*      4       operation failed                    *)
    (*      5       invalid path string                 *)

    VAR pos, id, code: CARDINAL;
        name, message: FilenameString;

    BEGIN
        pos := 5;
        id := GetCard (pkt, pos);
        GetCharString (pkt, pos, name);
        DiscardBS (pkt);
        TranslateFromUTF8r (name);
        message := "REMOVE ";
        Strings.Append (name, message);
        LogTransaction (sess^.logID, message);
        IF ReducePath (sess^.user, name) THEN
            code := RemoveFile (sess^.user, name);
        ELSE
            code := 5;
        END (*IF*);

        (* reply with SSH_FXP_STATUS message *)

        StatusReply (sess^.chan, id, code);

    END DeleteFile;

(************************************************************************)
(*                             14. MKDIR                                *)
(************************************************************************)

PROCEDURE MakeDirectory (sess: Session;  pkt: ByteStr);

    (* Creates a new directory. *)

    VAR pos, id, code: CARDINAL;
        path, message: ARRAY [0..511] OF CHAR;

    BEGIN
        pos := 5;
        id := GetCard (pkt, pos);
        GetCharString (pkt, pos, path);

        (* An ATTRs section follows, but in this version we are going   *)
        (* to ignore it.                                                *)

        DiscardBS (pkt);
        TranslateFromUTF8r (path);
        message := "MKDIR ";
        Strings.Append (path, message);
        LogTransaction (sess^.logID, message);
        IF ReducePath (sess^.user, path) THEN
            code := MkDir (sess^.user, path);
        ELSE
            code := 5;
        END (*IF*);

        (* reply with SSH_FXP_STATUS message *)

        StatusReply (sess^.chan, id, code);

    END MakeDirectory;

(************************************************************************)
(*                             15. RMDIR                                *)
(************************************************************************)

PROCEDURE RemoveDirectory (sess: Session;  pkt: ByteStr);

    (* Deletes a directory. *)

    VAR pos, id, code: CARDINAL;
        path: ARRAY [0..511] OF CHAR;
        message: ErrMess;

    BEGIN
        pos := 5;
        id := GetCard (pkt, pos);
        GetCharString (pkt, pos, path);

        (* An ATTRs section follows, but in this version we are going   *)
        (* to ignore it.                                                *)

        DiscardBS (pkt);
        TranslateFromUTF8r (path);
        message := "RMDIR ";
        Strings.Append (path, message);
        LogTransaction (sess^.logID, message);
        IF ReducePath (sess^.user, path) THEN
            code := RmDir (sess^.user, path, message);
        ELSE
            Strings.Assign ("malformed path string", message);
            code := 5;
        END (*IF*);

        (* reply with SSH_FXP_STATUS message *)

        StatusReply2 (sess^.chan, id, code, message);

    END RemoveDirectory;

(************************************************************************)
(*                            16. REALPATH                              *)
(************************************************************************)

PROCEDURE RealPath (sess: Session;  pkt: ByteStr);

    (* Returns the canonical version of a path name.  Both the path as  *)
    (* supplied by the client and the result we return are in UTF-8,    *)
    (* so no character set translation is needed.                       *)

    VAR pos, id, N: CARDINAL;
        reply: ByteStr;
        path, longname, message: ARRAY [0..511] OF CHAR;

    BEGIN
        pos := 5;
        id := GetCard (pkt, pos);
        GetCharString (pkt, pos, path);
        DiscardBS (pkt);
        message := "REALPATH ";
        Strings.Append (path, message);
        Strings.Append (" -> ", message);

        (* We are now supposed to simplify path, and reply with an  *)
        (* SSH_FXP_NAME message (with dummy attributes).            *)

        IF ReducePath (sess^.user, path) THEN
            Strings.Append (path, message);
            LogTransaction (sess^.logID, message);
            longname := path;
            N := 4 + 1 + 4 + 4 + 8 + LENGTH(path) + LENGTH(longname) + 4;

            MakeBS (reply, N+4);
            pos := 0;
            AddCard (reply, pos, N);
            reply.data^[pos] := SSH_FXP_NAME;  INC (pos);
            AddCard (reply, pos, id);
            AddCard (reply, pos, 1);
            AddCharString (reply, pos, path);
            AddCharString (reply, pos, longname);

            (* Dummy attributes consist of a single 0 flag word. *)

            AddCard (reply, pos, 0);
        ELSE
            (* SSH_FXP_STATUS response *)

            message := "Invalid path string";
            N := 1 + 4 + 4 + 4 + LENGTH(message) + 6;
            MakeBS (reply, N + 4);
            pos := 0;
            AddCard (reply, pos, N);
            AddByte (reply, pos, SSH_FXP_STATUS);
            AddCard (reply, pos, id);
            AddCard (reply, pos, 4);
            AddCharString (reply, pos, message);
            AddCharString (reply, pos, "en");
        END (*IF*);

        SendData (sess^.chan, reply);

    END RealPath;

(************************************************************************)
(*                               ( 7. LSTAT)                            *)
(*                                17. STAT                              *)
(************************************************************************)

PROCEDURE StatRequest (sess: Session;  pkt: ByteStr);

    (* Returns the attributes of a supplied filestring path. *)

    VAR pos, id, N: CARDINAL;
        reply: ByteStr;
        path, tempname, message: FilenameString;

    BEGIN
        pos := 5;
        id := GetCard (pkt, pos);
        GetCharString (pkt, pos, path);
        DiscardBS (pkt);
        message := "STAT ";
        Strings.Append (path, message);
        LogTransaction (sess^.logID, message);

        IF ReducePath (sess^.user, path) THEN
            Strings.Assign (path, tempname);
            TranslateFromUTF8 (tempname, path);
            reply := AttrsResponse (sess^.user, path, id);
        ELSE
            (* SSH_FXP_STATUS response *)

            message := "Invalid path string";
            N := 1 + 4 + 4 + 4 + LENGTH(message) + 6;
            MakeBS (reply, N + 4);
            pos := 0;
            AddCard (reply, pos, N);
            AddByte (reply, pos, SSH_FXP_STATUS);
            AddCard (reply, pos, id);
            AddCard (reply, pos, 4);
            AddCharString (reply, pos, message);
            AddCharString (reply, pos, "en");
        END (*IF*);

        SendData (sess^.chan, reply);

    END StatRequest;

(************************************************************************)
(*                               18. RENAME                             *)
(************************************************************************)

PROCEDURE Rename (sess: Session;  pkt: ByteStr);

    (* Rename a file or directory. *)

    VAR pos, id, code, N: CARDINAL;
        reply: ByteStr;
        name1, name2: FilenameString;
        message: FilenameString;

    BEGIN
        pos := 5;
        id := GetCard (pkt, pos);
        GetCharString (pkt, pos, name1);
        GetCharString (pkt, pos, name2);
        DiscardBS (pkt);
        TranslateFromUTF8r (name1);
        TranslateFromUTF8r (name2);
        message := "RENAME ";
        Strings.Append (name1, message);
        Strings.Append (" ", message);
        Strings.Append (name2, message);
        LogTransaction (sess^.logID, message);
        code := DoRename (sess^.user, name1, name2);

        (* reply with SSH_FXP_STATUS message *)

        IF code = 0 THEN message := "OK";
        ELSIF code = 2 THEN message := "no such file";
        ELSIF code = 3 THEN message := "permission denied";
        ELSIF code = 4 THEN message := "operation failed";
        ELSIF code = 5 THEN message := "Invalid path string";
        END (*IF*);
        N := 1 + 4 + 4 + 4 + LENGTH(message) + 6;
        MakeBS (reply, N + 4);
        pos := 0;
        AddCard (reply, pos, N);
        AddByte (reply, pos, SSH_FXP_STATUS);
        AddCard (reply, pos, id);
        AddCard (reply, pos, code);
        AddCharString (reply, pos, message);
        AddCharString (reply, pos, "en");

        SendData (sess^.chan, reply);

    END Rename;

(************************************************************************)
(*                             19. READLINK                             *)
(*                             20. SYMLINK                              *)
(*                  See procedure Unimplemented above                   *)
(************************************************************************)

TYPE
    (* Data used in creating a new instance of the session handler task. *)

    NewSessionPointer = POINTER TO
                           RECORD
                               chan: CARDINAL;
                               queue: PacketQ;
                               user: User;
                               logID: TransactionLogID;
                           END (*RECORD*);

(************************************************************************)
(*                   THE SESSION COMMAND DISPATCHER                     *)
(************************************************************************)

PROCEDURE SessionHandler (arg: ADDRESS);

    (* The server task that handles SFTP requests.   *)

    (********************************************************************)
    (*                                                                  *)
    (* NOTE on session termination.                                     *)
    (*                                                                  *)
    (* There is no provision in the draft SFTP standard for closing     *)
    (* a session. In practice a session will be terminated in one of    *)
    (* two ways:                                                        *)
    (*                                                                  *)
    (*  1.  The connection is lost (or deliberately aborted, e.g. by    *)
    (*      a timeout) in the SSH transport layer.  In that case the    *)
    (*      transport layer sends us an SSH_MSG_TERMINATE message       *)
    (*      (my personal extension to the SSH standard), which tells us *)
    (*      that lower protocol layers of this connection have already  *)
    (*      been closed down. All that remains for us to do is dispose  *)
    (*      of the user data and dispose of our message queue.          *)
    (*                                                                  *)
    (*      (What about sess^.logID? This needs to be checked.)         *)
    (*                                                                  *)
    (*  2.  At the channel layer (connection protocol layer), the       *)
    (*      client sends a "close channel" request, and this will       *)
    (*      trigger the shutdown of the lower protocol layers.  As in   *)
    (*      case 1 the end result is that an SSH_MSG_TERMINATE message  *)
    (*      is sent upwards. As far as we are concerned this is the     *)
    (*      same as case 1.                                             *)
    (*                                                                  *)
    (********************************************************************)

    VAR NSP: NewSessionPointer;
        sess: Session;
        Q: PacketQ;
        pkt: ByteStr;
        type: CARD8;
        message: ARRAY [0..63] OF CHAR;

    (********************************************************************)

    (*
    PROCEDURE LogType;

        (* For debugging, displays value of variable "type". *)

        VAR msg: ARRAY [0..31] OF CHAR;

        BEGIN
            CASE type OF
                SSH_FXP_INIT:       msg := "SSH_FXP_INIT";
                |
                SSH_FXP_REALPATH:   msg := "SSH_FXP_REALPATH";
                |
                SSH_MSG_TERMINATE:  msg := "SSH_MSG_TERMINATE";
            ELSE
                msg := "message ";
                AppendCard (type, msg);
            END (*CASE*);
            LogTransaction (sess^.logID, msg);
        END LogType;
    *)

    (********************************************************************)

    BEGIN                   (* Body of SessionHandler *)
        NSP := arg;
        Q := NSP^.queue;

        NEW (sess);
        sess^.chan := NSP^.chan;
        sess^.user := NSP^.user;
        sess^.logID := NSP^.logID;
        DISPOSE (NSP);

        LOOP
            (*LogTransactionL (sess^.logID, "FTS: about to call GetPacketCmp");*)

            pkt := GetPacket (Q);
            (*LogTransactionL (sess^.logID, "FTS: returned from GetPacketCmp");*)

            (* The length field is redundant, so skip it. *)

            type := pkt.data^[4];
                (*
                Strings.Assign ("FTP session received command type ", message);
                AppendCard (type, message);
                LogTransaction (sess^.logID, message);
                *)
            (*LogType;*)
            CASE type OF
              | SSH_FXP_INIT:                       (*  1 *)
                    InitFTPSession (sess, pkt);
              | SSH_FXP_VERSION:                    (*  2 *)
                    ReceiveVersion (sess, pkt);
              | SSH_FXP_OPEN:                       (*  3 *)
                    OpenFile (sess, pkt);
              | SSH_FXP_CLOSE:                      (*  4 *)
                    CloseOperation (sess, pkt);
              | SSH_FXP_READ:                       (*  5 *)
                    ReadFile (sess, pkt);
              | SSH_FXP_WRITE:                      (*  6 *)
                    WriteFile (sess, pkt);
              | SSH_FXP_LSTAT:                      (*  7 *)
                    StatRequest (sess, pkt);
              | SSH_FXP_FSTAT:                      (*  8 *)
                    FStatRequest (sess, pkt);
              | SSH_FXP_SETSTAT:                    (*  9 *)
                    Unimplemented (sess, pkt);
              | SSH_FXP_FSETSTAT:                   (* 10 *)
                    Unimplemented (sess, pkt);
              | SSH_FXP_OPENDIR:                    (* 11 *)
                    OpenDirectory (sess, pkt);
              | SSH_FXP_READDIR:                    (* 12 *)
                    ListDirectory (sess, pkt);
              | SSH_FXP_REMOVE:                     (* 13 *)
                    DeleteFile (sess, pkt);
              | SSH_FXP_MKDIR:                      (* 14 *)
                    MakeDirectory (sess, pkt);
              | SSH_FXP_RMDIR:                      (* 15 *)
                    RemoveDirectory (sess, pkt);
              | SSH_FXP_REALPATH:                   (* 16 *)
                    RealPath (sess, pkt);
              | SSH_FXP_STAT:                       (* 17 *)
                    StatRequest (sess, pkt);
              | SSH_FXP_RENAME:                     (* 18 *)
                    Rename (sess, pkt);
              | SSH_FXP_READLINK:                   (* 19 *)
                    Unimplemented (sess, pkt);
              | SSH_FXP_SYMLINK:                    (* 20 *)
                    Unimplemented (sess, pkt);
              | SSH_MSG_TERMINATE:
                    EXIT (*LOOP*);
            ELSE
                Strings.Assign ("Cannot currently handle FTP command type ", message);
                AppendCard (type, message);
                LogTransaction (sess^.logID, message);
                DiscardBS (pkt);
            END (*IF*);
        END (*LOOP*);

        CloseQueue (Q);
        (*LogTransactionL (sess^.logID, "Exiting FTP session handler");*)
        DestroyUserData (sess^.user);
        DISPOSE (sess);

    END SessionHandler;

(************************************************************************)

PROCEDURE StartFTPSession (channel: CARDINAL;  Q: PacketQ;  U: User;
                                        ID: TransactionLogID): BOOLEAN;

    (* Starts and runs an SFTP server session to serve one client.  The *)
    (* session runs in a separate thread; this procedure returns after  *)
    (* starting the session, it does not wait until the session is over.*)

    VAR NSP: NewSessionPointer;  success: BOOLEAN;

    BEGIN
        NEW (NSP);
        WITH NSP^ DO
            chan := channel;  queue := Q;  user := U;  logID := ID;
        END (*WITH*);
        success := CreateTask1 (SessionHandler, 3, "FTP Layer", NSP);
        IF NOT success THEN
            DISPOSE (NSP);
        END (*IF*);
        RETURN success;
    END StartFTPSession;

(************************************************************************)

END FTPSession.

