IMPLEMENTATION MODULE HashAlgs;

        (********************************************************)
        (*                                                      *)
        (*                    Hash algorithms                   *)
        (*                                                      *)
        (*  Programmer:         P. Moylan                       *)
        (*  Started:            14 February 2021                *)
        (*  Last edited:        3 November 2021                 *)
        (*  Status:             Complete, working               *)
        (*                                                      *)
        (********************************************************)


IMPORT SHA1, SHA2, SHA512, VarStrings;


FROM SYSTEM IMPORT LOC, CARD8, ADR;

FROM BigNum IMPORT
    (* type *)  BN,
    (* proc *)  Discard;

FROM VarStrings IMPORT
    (* type *)  ByteStr;

FROM Inet2Misc IMPORT
    (* proc *)  Swap4;

FROM LowLevel IMPORT
    (* proc *)  Copy;

FROM Storage IMPORT
    (* proc *)  ALLOCATE;

(************************************************************************)

TYPE
    HashCtx = POINTER TO
                RECORD
                    CASE alg: HashAlgorithm OF
                        sha1:
                            algctx1: SHA1.SHA1_CTX;
                      |
                        sha2_256:
                            algctx256: SHA2.SHA2_CTX;
                      |
                        sha2_512:
                            algctx512: SHA512.SHA512_CTX;
                    END (*CASE*);
                END (*RECORD*);

     LengthArray = ARRAY HashAlgorithm OF CARDINAL;

CONST
    DigestLength = LengthArray {20, 32, 64};

(************************************************************************)

PROCEDURE HashLength (alg: HashAlgorithm): CARDINAL;

    (* Returns the number of bytes in a hashed result. *)

    BEGIN
        RETURN DigestLength[alg];
    END HashLength;

(************************************************************************)
(*                       STARTING A NEW OPERATION                       *)
(************************************************************************)

PROCEDURE HashInit (alg: HashAlgorithm): HashCtx;

    (* Initialization.  Creates a context for a new hash operation. *)

    VAR result: HashCtx;

    BEGIN
        NEW (result);
        result^.alg := alg;
        CASE alg OF
            sha1:
                result^.algctx1 := SHA1.SHA1Init();
          | sha2_256:
                result^.algctx256 := SHA2.SHA256Init();
          | sha2_512:
                result^.algctx512 := SHA512.SHA512Init();
        END (*CASE*);
        RETURN result;
    END HashInit;

(************************************************************************)
(*                          INPUTTING MORE DATA                         *)
(************************************************************************)

PROCEDURE HashUpdate (context: HashCtx;
                    VAR (*IN*) data: ARRAY OF LOC;  length: CARDINAL);

    (* Block update operation. Continues hashing,  adding more data     *)
    (* to what has already been processed for this context.             *)

    BEGIN
        CASE context^.alg OF
            sha1:
                SHA1.SHA1Update (context^.algctx1, data, length);
          | sha2_256:
                SHA2.SHA2Update (context^.algctx256, data, length);
          | sha2_512:
                SHA512.SHA512Update (context^.algctx512, data, length);
        END (*CASE*);
    END HashUpdate;

(************************************************************************)
(*                      RETRIEVING THE HASHED RESULT                    *)
(************************************************************************)

PROCEDURE HashFinal (VAR (*INOUT*) context: HashCtx;
                        VAR (*OUT*) digest: ARRAY OF CARD8);

    (* Finalization. Ends a hashing operation, returning    *)
    (* the hashed result and discarding the context.        *)

    BEGIN
        CASE context^.alg OF
            sha1:
                SHA1.SHA1Final (context^.algctx1, digest);
          | sha2_256:
                SHA2.SHA2Final (context^.algctx256, digest);
          | sha2_512:
                SHA512.SHA512Final (context^.algctx512, digest);
        END (*CASE*);
        DISPOSE (context);
    END HashFinal;

(************************************************************************)
(*                           KEY GENERATION                             *)
(************************************************************************)

TYPE
    KeygenCtx = POINTER TO
                    RECORD
                        hash: HashAlgorithm;
                        K, H, session_id: ByteStr;
                    END (*RECORD*);

(************************************************************************)

PROCEDURE InitKeyGeneration (hashalg: HashAlgorithm;  VAR (*INOUT*) K: BN;
                                VAR (*INOUT*) H, sess_id: ByteStr): KeygenCtx;

    (* The caller passes in the parameters that will be needed for a    *)
    (* series of key generation operations.  We destroy the caller's    *)
    (* copy of K, H, and sess_id.                                       *)

    VAR result: KeygenCtx;

    BEGIN
        NEW (result);
        result^.hash := hashalg;
        result^.K := VarStrings.BNtoByteStr (K);
        Discard (K);
        result^.H := H;
        H.allocated := 0;  H.data := NIL;
        result^.session_id := sess_id;
        sess_id.allocated := 0;  sess_id.data := NIL;
        RETURN result;
    END InitKeyGeneration;

(************************************************************************)

PROCEDURE GenerateKeyData (ctx: KeygenCtx;  code: ARRAY OF CHAR;
                            VAR (*OUT*) data: ARRAY OF CARD8;  N: CARDINAL);

    (* Generates N bytes of key data.  This is initially generated as   *)
    (*      K1 = HASH (K || H || code || session_id )                   *)
    (* If this does not give enough bytes then we extend it using       *)
    (*      Kj = HASH (K || H || K1 || K2 || ... || K(j-1)              *)

    (* DISCOVERY: the lengths of H, code, and session_id should not be  *)
    (* included in the calculation.  The standard is ambiguous on       *)
    (* this point.                                                      *)

    VAR L, Lbig, total, amount, temp: CARDINAL;
        hctx: HashCtx;
        chunk: ARRAY [0..255] OF CARD8;
        bigchunk: ARRAY [0..40955] OF CARD8;

    BEGIN
        (* Copy K and H to bigchunk.  Include the size of K but not *)
        (* the size of H.                                           *)

        L := HashLength (ctx^.hash);
        amount := ctx^.K.size;
        temp := Swap4 (amount);
        Copy (ADR(temp), ADR(bigchunk), 4);
        Copy (ctx^.K.data, ADR(bigchunk[4]), amount);
        Lbig := amount + 4;

        amount := ctx^.H.size;    (* which is equal to L, in fact *)
        Copy (ctx^.H.data, ADR(bigchunk[Lbig]), amount);
        INC (Lbig, amount);

        (* Initial hash. *)

        hctx := HashInit (ctx^.hash);
        HashUpdate (hctx, bigchunk, Lbig);
        HashUpdate (hctx, code, 1);
        HashUpdate (hctx, ctx^.session_id.data^, ctx^.session_id.size);
        HashFinal (hctx, chunk);

        amount := L;
        IF amount >= N THEN amount := N; END(*IF*);
        Copy (ADR(chunk), ADR(data), amount);
        total := amount;

        (* Further hashes if needed. *)

        WHILE total < N DO

            Copy (ADR(chunk), ADR(bigchunk[Lbig]), L);
            INC (Lbig, L);

            hctx := HashInit (ctx^.hash);
            HashUpdate (hctx, bigchunk, Lbig);
            HashFinal (hctx, chunk);

            amount := L;
            IF total + amount > N THEN amount := N - total; END(*IF*);
            Copy (ADR(chunk), ADR(data[total]), amount);
            INC (total, amount);

        END (*WHILE*);

    END GenerateKeyData;

(************************************************************************)

PROCEDURE EndKeyGeneration (VAR (*INOUT*) ctx: KeygenCtx);

    (* Disposes of the ctx record. *)

    BEGIN
        VarStrings.DiscardBS (ctx^.K);
        VarStrings.DiscardBS (ctx^.H);
        VarStrings.DiscardBS (ctx^.session_id);
        DISPOSE (ctx);
    END EndKeyGeneration;

(************************************************************************)

END HashAlgs.

