IMPLEMENTATION MODULE KeyExchange;

        (********************************************************)
        (*                                                      *)
        (*                Key exchange operations               *)
        (*                                                      *)
        (*  Programmer:         P. Moylan                       *)
        (*  Started:            19 January 2021                 *)
        (*  Last edited:        10 January 2022                 *)
        (*  Status:             OK                              *)
        (*                                                      *)
        (********************************************************)

IMPORT Strings, BigNum;

FROM NameLists IMPORT
    (* proc *)  MatchNames, ListAll;

FROM DiffieHellman IMPORT
    (* proc *)  DHClientStartKeyExchange, DHClientKxphase2, DHSrvDoKeyExchange;

FROM PacketIO IMPORT
    (* type *)  IOPort;

FROM PublicKeys IMPORT
    (* type *)  PKAType;

FROM HashAlgs IMPORT
    (* type *)  KeygenCtx;

FROM BigNum IMPORT
    (* type *)  BN;

FROM VarStrings IMPORT
    (* type *)  ByteStr;

FROM TransLog IMPORT
    (* type *)  TransactionLogID;

(************************************************************************)

TYPE
    NameArray = ARRAY KexType OF ARRAY [0..31] OF CHAR;

CONST
    Names = NameArray {"diffie-hellman-group1-sha1", "diffie-hellman-group14-sha1",
                       "diffie-hellman-group14-sha256",  "diffie-hellman-group15-sha512",
                       "none"};

(************************************************************************)
(*                IDENTIFIERS FOR KEY EXCHANGE ALGORITHMS               *)
(************************************************************************)

PROCEDURE NameOf (kx: KexType;  VAR (*OUT*) name: ARRAY OF CHAR);

    (* Returns the RFC 4253 name of mt. *)

    BEGIN
        Strings.Assign (Names[kx], name);
    END NameOf;

(************************************************************************)

PROCEDURE AllNames (VAR (*OUT*) list: ARRAY OF CHAR);

    (* Returns a comma-separated list of our algorithm names.  *)

    CONST max = ORD(MAX(KexType));

    BEGIN
        ListAll (Names, max, list);
    END AllNames;

(************************************************************************)

PROCEDURE FindMatch (VAR (*IN*) list: ARRAY OF CHAR): KexType;

    (* Input is a comma-separated list of algorithm names.  *)
    (* Returns the first one we support.                    *)

    CONST max = ORD(MAX(KexType));

    VAR result: CARDINAL;

    BEGIN
        result := MatchNames (Names, max, list);
        IF result > max THEN RETURN none;
        ELSE RETURN VAL(KexType, result);
        END (*IF*);
    END FindMatch;

(************************************************************************)
(*                         KEY EXCHANGE OPERATIONS                      *)
(************************************************************************)

PROCEDURE ClientStartKeyExchange (sess: IOPort;  kex: KexType;
                                pka: PKAType;  ID: TransactionLogID): BN;

    (* The client triggers a key exchange. *)

    BEGIN
        CASE kex OF
            dh1, dh14, dh14_sha256, dh15_sha512:
                    RETURN DHClientStartKeyExchange (sess, kex, pka, ID);
            |   none:
                        (* Not sure what to do in this case *)
                        (* Probably should give an error response *)
        END (*CASE*);
    END ClientStartKeyExchange;

(************************************************************************)

PROCEDURE ClientKxphase2 (sess: IOPort;  ID: TransactionLogID;
                        packet: ByteStr;  kex: KexType;  pka: PKAType;
                        VAR (*INOUT*) kxprivate: BN): KeygenCtx;

    (* Client response to an SSH_MSG_KEXDH_REPY message.    *)
    (* We dispose of kxprivate after using it.              *)

    VAR result: KeygenCtx;

    BEGIN
        CASE kex OF
            dh1, dh14, dh14_sha256, dh15_sha512:
                    result := DHClientKxphase2 (sess, ID, packet, kex, pka, kxprivate);
                    BigNum.Discard (kxprivate);
                    RETURN result;
            |   none:
                    RETURN NIL;
                        (* Not sure what to do in this case *)
                        (* Probably should give an error response *)
        END (*CASE*);
    END ClientKxphase2;

(************************************************************************)

PROCEDURE SrvDoKeyExchange (sess: IOPort; kex: KexType;  pka: PKAType;
                                             packet: ByteStr): KeygenCtx;

    (* Responds to a SSH_MSG_KEX_ECDH_INIT message.    *)

    BEGIN
        CASE kex OF
            dh1, dh14, dh14_sha256, dh15_sha512:
                    RETURN DHSrvDoKeyExchange (sess, kex, pka, packet);
            |   none:
                    RETURN NIL;
                        (* Not sure what to do in this case *)
                        (* Probably should give an error response *)
        END (*CASE*);
    END SrvDoKeyExchange;

(************************************************************************)

END KeyExchange.

