IMPLEMENTATION MODULE MAC;

        (********************************************************)
        (*                                                      *)
        (*              Message authentication code             *)
        (*                                                      *)
        (*  Programmer:         P. Moylan                       *)
        (*  Started:            16 January 2021                 *)
        (*  Last edited:        5 January 2022                  *)
        (*  Status:             Working                         *)
        (*                                                      *)
        (********************************************************)


FROM SYSTEM IMPORT LOC, ADDRESS, CARD8, ADR;

IMPORT Strings;

FROM NameLists IMPORT
    (* proc *)  MatchNames, ListAll;

FROM HMAC IMPORT
    (* proc *)  HMAC_SHA1, HMAC_SHA256, HMAC_SHA384, HMAC_SHA512, HMAC_MD5;

FROM VarStrings IMPORT ByteStr, ByteStringPtr;

FROM MiscFuncs IMPORT
    (* proc *)  Swap4;

FROM LowLevel IMPORT
    (* proc *)  Copy, AddOffset;

FROM Storage IMPORT
    (* proc *)  ALLOCATE, DEALLOCATE;

(************************************************************************)

TYPE
    CardArray = ARRAY MACType OF CARDINAL;
    NameArray = ARRAY MACType OF ARRAY [0..31] OF CHAR;

VAR
    Size: CardArray;
    Names: NameArray;

(************************************************************************)
(*                    IDENTIFIERS FOR MAC ALGORITHMS                    *)
(************************************************************************)

PROCEDURE NameOf (mt: MACType;  VAR (*OUT*) name: ARRAY OF CHAR);

    (* Returns the RFC 4253 name of mt. *)

    BEGIN
        Strings.Assign (Names[mt], name);
    END NameOf;

(************************************************************************)

PROCEDURE AllNames (VAR (*OUT*) list: ARRAY OF CHAR);

    (* Returns a comma-separated list of our algorithm names.  *)

    CONST max = ORD(MAX(MACType));

    BEGIN
        ListAll (Names, max, list);
    END AllNames;

(************************************************************************)

PROCEDURE FindMatch (VAR (*IN*) list: ARRAY OF CHAR): MACType;

    (* Input is a comma-separated list of algorithm names.  *)
    (* Returns the first one we support.                    *)

    CONST max = ORD(MAX(MACType));

    VAR result: CARDINAL;

    BEGIN
        result := MatchNames (Names, max, list);
        IF result > max THEN RETURN none;
        ELSE RETURN VAL(MACType, result);
        END (*IF*);
    END FindMatch;

(************************************************************************)
(*                      THE LENGTH OF A MAC DIGEST                      *)
(************************************************************************)

PROCEDURE MAClength (mt: MACType): CARDINAL;

    (* Returns the number of bytes in the MAC. *)

    BEGIN
        RETURN Size[mt];
    END MAClength;

(************************************************************************)
(*                           GENERATING A MAC                           *)
(************************************************************************)

PROCEDURE MakeCopy (seqno: CARDINAL;  VAR (*IN*) packet: ByteStr;
                    VAR (*OUT*) presult: ADDRESS): CARDINAL;

    (* Produces presult^, which is seqno followed by the packet data.   *)
    (* The function result is the byte count for the combination.       *)
    (* Unfortunately HMAC is designed in such a way that you can't feed *)
    (* it data a section at a time, so I can't see any way of inserting *)
    (* the sequence number other than making a copy.  (Or writing a     *)
    (* radically redesigned HMAC implementation.)                       *)

    VAR size: CARDINAL;  q: ADDRESS;

    BEGIN
        size := packet.size + 4;
        ALLOCATE (presult, size);
        seqno := Swap4 (seqno);
        Copy (ADR(seqno), presult, 4);
        q := AddOffset (presult, 4);
        Copy (packet.data, q, packet.size);
        RETURN size;
    END MakeCopy;

(************************************************************************)

PROCEDURE ComputeMac (mt: MACType;  VAR (*IN*) key: ARRAY OF CARD8;
                       keylength, seqno: CARDINAL;
                        indata: ByteStr;
                         VAR (*OUT*) outdata: ARRAY OF CARD8);

    (* The MAC is outdata, and its size can be found by calling *)
    (* MAClength.  The input is an unencrypted packet.          *)

    VAR p: ByteStringPtr;  length: CARDINAL;
        digest: ARRAY [0..127] OF CARD8;

    BEGIN
        IF mt = none THEN
            RETURN;
        END (*IF*);

        length := MakeCopy (seqno, indata, p);
        CASE mt OF
            | hmacsha1, hmacsha1_96:
                HMAC_SHA1 (key, keylength, p^, length, digest);
                Copy (ADR(digest), ADR(outdata), Size[mt]);
            | hmacsha256, hmacsha256_96:
                HMAC_SHA256 (key, keylength, p^, length, digest);
                Copy (ADR(digest), ADR(outdata), Size[mt]);
            | hmacsha384, hmacsha384_96:
                HMAC_SHA384 (key, keylength, p^, length, digest);
                Copy (ADR(digest), ADR(outdata), Size[mt]);
            | hmacsha512, hmacsha512_96:
                HMAC_SHA512 (key, keylength, p^, length, digest);
                Copy (ADR(digest), ADR(outdata), Size[mt]);
            | hmacmd5, hmacmd5_96:
                HMAC_MD5 (key, keylength, p^, length, digest);
                Copy (ADR(digest), ADR(outdata), Size[mt]);
            | none:
                (* Do nothing; should never happen *)
        END (*CASE*);

        (* Discard the copy. *)

        IF p <> NIL THEN
            DEALLOCATE (p, length);
        END (*IF*);

    END ComputeMac;

(************************************************************************)
(*                       MODULE INITIALISATION                          *)
(************************************************************************)

PROCEDURE SetNamesAndSize;

    (* I used to have the "Size" and "Names" arrays declared as         *)
    (* constants, but doing it this way is safer given the risk that a  *)
    (* revision might alter the order of MAC algorithms.                *)

    VAR k: MACType;

    BEGIN
        FOR k := MIN (MACType) TO MAX (MACType) DO
            CASE k OF
                | hmacsha1:
                    Size[k] := 20;  Names[k] := "hmac-sha1";
                | hmacsha1_96:
                    Size[k] := 12;  Names[k] := "hmac-sha1-96";
                | hmacsha256:
                    Size[k] := 32;  Names[k] := "hmac-sha2-256";
                | hmacsha256_96:
                    Size[k] := 12;  Names[k] := "hmac-sha2-256-96";
                | hmacsha384:
                    Size[k] := 48;  Names[k] := "hmac-sha2-384";
                | hmacsha384_96:
                    Size[k] := 12;  Names[k] := "hmac-sha2-384-96";
                | hmacsha512:
                    Size[k] := 64;  Names[k] := "hmac-sha2-512";
                | hmacsha512_96:
                    Size[k] := 12;  Names[k] := "hmac-sha2-512-96";
                | hmacmd5:
                    Size[k] := 16;  Names[k] := "hmac-md5";
                | hmacmd5_96:
                    Size[k] := 12;  Names[k] := "hmac-md5-96";
                | none:
                    Size[k] := 0;  Names[k] := "none";
            END (*CASE*);
        END (*FOR*);
    END SetNamesAndSize;

(************************************************************************)

BEGIN
    SetNamesAndSize;
END MAC.

