(**************************************************************************)
(*                                                                        *)
(*  My own version of SSH                                                 *)
(*  Copyright (C) 2022   Peter Moylan                                     *)
(*                                                                        *)
(*  This program is free software: you can redistribute it and/or modify  *)
(*  it under the terms of the GNU General Public License as published by  *)
(*  the Free Software Foundation, either version 3 of the License, or     *)
(*  (at your option) any later version.                                   *)
(*                                                                        *)
(*  This program is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU General Public License for more details.                          *)
(*                                                                        *)
(*  You should have received a copy of the GNU General Public License     *)
(*  along with this program.  If not, see <http://www.gnu.org/licenses/>. *)
(*                                                                        *)
(*  To contact author:   http://www.pmoylan.org   peter@pmoylan.org       *)
(*                                                                        *)
(**************************************************************************)

IMPLEMENTATION MODULE PacketIO;

        (********************************************************)
        (*                                                      *)
        (*        Sending and receiving encrypted packets       *)
        (*                                                      *)
        (*  Programmer:         P. Moylan                       *)
        (*  Started:            6 January 2021                  *)
        (*  Last edited:        12 July 2022                    *)
        (*  Status:             Apparently working and complete *)
        (*                                                      *)
        (********************************************************)


FROM STextIO IMPORT WriteString, WriteLn;     (* while debugging *)

FROM SYSTEM IMPORT
    (* type *)  LOC, ADDRESS, CARD8, CARD32,
    (* proc *)  ADR, CAST;

IMPORT SFV, Strings, Encryption, MAC, Compression;

FROM RandomData IMPORT
    (* proc *)  RandomBytes;

FROM Sockets IMPORT
    (* const *) NotASocket,
    (* type *)  Socket, SockAddr,
    (* proc *)  send, getpeername, so_cancel, soclose;

FROM VarStrings IMPORT
    (* type *)  ByteStr, ByteStringPtr,
    (* proc *)  MakeBS, CopyOfBS, CharStrToByteStr, BSDelete, DiscardBS;

FROM Compression IMPORT
    (* type *)  CompressionType,
    (* proc *)  Compress, Decompress;

FROM Encryption IMPORT
    (* proc *)  Encrypt, Decrypt;

FROM TransLog IMPORT
    (* type *)  TransactionLogID,
    (* proc *)  LogTransaction, LogTransactionL;

FROM MiscFuncs IMPORT
    (* proc *)  AppendCard, AddEOL;

FROM Inet2Misc IMPORT
    (* proc *)  Swap4, WaitForSocket, Receive;

FROM Timer IMPORT
    (* proc *)  Sleep;

FROM Semaphores IMPORT
    (* type *)  Semaphore,
    (* proc *)  CreateSemaphore, DestroySemaphore, Wait, Signal;

FROM LowLevel IMPORT
    (* proc *)  Copy, EVAL, AddOffset;

FROM Storage IMPORT
    (* proc *)  ALLOCATE, DEALLOCATE;

(********************************************************************************)

CONST NStrings = 8;

TYPE
    IOPort = POINTER TO
                    RECORD
                        server: BOOLEAN;
                                    (* TRUE iff our end is the server end *)
                        negotiating, closing, inputinprogress: BOOLEAN;
                        sock: Socket;
                        logID: TransactionLogID;
                        InitLock: Semaphore;
                        inseqno, outseqno: CARD32;
                        session_id: ByteStr;
                        string: ARRAY [1..NStrings] OF ByteStr;
                        inCI, outCI: Encryption.CipherInfo;
                        incrypt, outcrypt: Encryption.CipherType;
                        inmac, outmac: MAC.MACType;
                        inMACkey: ARRAY [0..63] OF CARD8;
                        inMACkeylength: CARDINAL;
                        outMACkey: ARRAY [0..63] OF CARD8;
                        outMACkeylength: CARDINAL;
                        incompr, outcompr: Compression.CompressionType;
                    END (*RECORD*);

(********************************************************************************)
(*                                                                              *)
(* Each packet is in the following format:                                      *)
(*      uint32      packet_length                                               *)
(*      byte        padding_length                                              *)
(*      byte[n1]    payload; n1 = packet_length - padding_length - 1            *)
(*      byte[n2]    random padding; n2 = padding_length                         *)
(*      byte[m]     mac (Message Authentication Code - MAC); m = mac_length     *)
(*                                                                              *)
(* The packet length doesn't include the mac or the packet_length field itself. *)
(* Encryption applies to everything including the length but not the mac.       *)
(* (The mac of course has its own encryption.)                                  *)
(*                                                                              *)
(* Everything except the MAC should have a length that is a multiple of the     *)
(* cipher block size or 8, whichever is larger.  That is why we have padding.   *)
(* There must be at least 4 bytes of padding, and no more than 255 bytes.       *)
(*                                                                              *)
(********************************************************************************)


(********************************************************************************)
(*                          SETTING SESSION PARAMETERS                          *)
(********************************************************************************)

PROCEDURE OpenPacketSession (S: Socket;  server: BOOLEAN;
                                            LogID: TransactionLogID): IOPort;

    (* Initialises the parameters for an IOPort. *)

    VAR result: IOPort;
        j: CARDINAL;

    BEGIN
        NEW (result);
        result^.sock := S;
        result^.negotiating := TRUE;
        result^.closing := FALSE;
        result^.server := server;
        result^.logID := LogID;
        CreateSemaphore (result^.InitLock, 0);
        result^.inseqno := 0;
        result^.outseqno := 0;
        result^.inCI := Encryption.InitCipherInfo(TRUE);
        result^.outCI := Encryption.InitCipherInfo(FALSE);
        result^.incrypt := Encryption.none;
        result^.outcrypt := Encryption.none;
        result^.inmac := MAC.none;
        result^.outmac := MAC.none;
        result^.inMACkey[0] := 0;
        result^.inMACkeylength := 0;
        result^.outMACkey[0] := 0;
        result^.outMACkeylength := 0;
        result^.incompr := Compression.none;
        result^.outcompr := Compression.none;
        WITH result^.session_id DO
            allocated := 0;
            size := 0;
            data := NIL;
        END (*WITH*);
        FOR j := 1 TO NStrings DO
            WITH result^.string[j] DO
                allocated := 0;
                size := 0;
                data := NIL;
            END (*WITH*);
        END (*FOR*);
        (*Debug.DebugStart (server);*)
        RETURN result;
    END OpenPacketSession;

(********************************************************************************)

PROCEDURE DeactivatePort (VAR (*INOUT*) port: IOPort);

    (* The initial phase of shutting down a port. We set the port^.closing flag,*)
    (* and wait until no input operations are in progress, but we do not delete *)
    (* the port data structure.  This is left until ClosePacketSession is       *)
    (* called. We trust the caller not to initiate any output operations.       *)

    (* The reason for breaking the "close" operation into two phases is that    *)
    (* session termination can happen in multiple ways (request from this end,  *)
    (* request from the other end, broken connection, etc.), and we would like  *)
    (* the final deallocation to be localised to a single point.                *)

    BEGIN
        IF port <> NIL THEN
            port^.closing := TRUE;
            WHILE port^.inputinprogress DO
                so_cancel (port^.sock);
                Sleep (100);
            END (*WHILE*);
            soclose (port^.sock);
        END (*IF*);
    END DeactivatePort;

(********************************************************************************)

PROCEDURE ClosePacketSession (VAR (*INOUT*) port: IOPort);

    (* The final phase of shutting down a port. The termination of I/O          *)
    (* operations has been handled by procedure DeactivatePort, except in some  *)
    (* cases like a broken link.  We first wait, if necessary, until no input   *)
    (* operations are in progress, and then we close the port data structure.   *)
    (* We trust the caller not to initiate any output operations.               *)

    VAR j: CARDINAL;

    BEGIN
        IF port <> NIL THEN

            (* This first part should have been handled by DeactivatePort,      *)
            (* above, but we want to allow for unexpected scenarios.            *)

            port^.closing := TRUE;
            WHILE port^.inputinprogress DO
                so_cancel (port^.sock);
                Sleep (100);
            END (*WHILE*);
            soclose (port^.sock);

            (* Now, the final deallocation of the data structure. *)

            FOR j := 1 TO NStrings DO
                DiscardBS (port^.string[j]);
            END (*FOR*);
            DiscardBS (port^.session_id);
            DestroySemaphore (port^.InitLock);
            Encryption.CloseCipherInfo (port^.inCI);
            Encryption.CloseCipherInfo (port^.outCI);
            DISPOSE (port);
        END (*IF*);
    END ClosePacketSession;

(********************************************************************************)

PROCEDURE IsClosing (port: IOPort): BOOLEAN;

    (* Returns TRUE if this port has commenced shutting down. *)

    BEGIN
        RETURN port^.closing;
    END IsClosing;

(********************************************************************************)

PROCEDURE SetInitLock (port: IOPort;  starting: BOOLEAN);

    (* Sets a lock that says that key exchange is beginning (starting=TRUE) or  *)
    (* has completed (starting=FALSE).                                          *)

    BEGIN
        IF starting THEN
            IF NOT port^.negotiating THEN
                port^.negotiating := TRUE;
                Wait (port^.InitLock);
            END (*IF*);
        ELSE
            IF port^.negotiating THEN
                port^.negotiating := FALSE;
                Signal (port^.InitLock);
            END (*IF*);
        END (*IF*);
    END SetInitLock;

(********************************************************************************)

PROCEDURE InitWait (port: IOPort);

    (* Waits until key exchange is complete. *)

    BEGIN
        IF port^.negotiating THEN
            Wait (port^.InitLock);
            Signal (port^.InitLock);
        END (*IF*);
    END InitWait;

(********************************************************************************)

PROCEDURE PeerAddr (port: IOPort): CARDINAL;

    (* Returns the IP address of the peer, in network byte order. *)

    VAR peer: SockAddr;  size: CARDINAL;

    BEGIN
        size := SIZE(peer);
        IF getpeername (port^.sock, peer, size) THEN
            RETURN 0;
        ELSE
            RETURN peer.in_addr.addr;
        END (*IF*);
    END PeerAddr;

(********************************************************************************)

PROCEDURE SetSessionID (port: IOPort;  H: ByteStr): ByteStr;

    (* Makes H the session ID iff an ID has not already been set,   *)
    (* returns the result.                                          *)

    BEGIN
        IF port^.session_id.size = 0 THEN
            port^.session_id := CopyOfBS (H);
        END (*IF*);
        RETURN CopyOfBS (port^.session_id);
    END SetSessionID;

(********************************************************************************)

PROCEDURE LastSeqNo (port: IOPort): CARDINAL;

    (* Returns the sequence number of the last packet received. *)

    VAR result: CARDINAL;

    BEGIN
        result := port^.inseqno;
        IF result = 0 THEN result := MAX(CARDINAL);
        ELSE DEC (result);
        END (*IF*);
        RETURN result;
    END LastSeqNo;

(********************************************************************************)

PROCEDURE SetString (port: IOPort;  N: CARDINAL;  str: ByteStr;  copy: BOOLEAN);

    (* Sets the value of a string used to calculate the exchange hash.  *)
    (* Legal values of N are:                                           *)
    (*      1       client ID string (from initial greeting)            *)
    (*      2       server ID string (from initial greeting)            *)
    (*      3       client's KEXINIT message                            *)
    (*      4       server's KEXINIT message                            *)
    (*      5       server's public key                                 *)
    (*      6       exchange value sent by client                       *)
    (*      7       exchange value sent by server                       *)
    (*      8       the shared secret derived during key exchange       *)
    (* If copy = TRUE we make a copy, leaving the original for the      *)
    (* caller.                                                          *)

    BEGIN
        DiscardBS (port^.string[N]);
        WITH port^.string[N] DO
            size := str.size;
            allocated := str.allocated;
            IF copy THEN
                ALLOCATE (data, allocated);
                Copy (str.data, data, allocated);
            ELSE
                data := str.data;
                str.data := NIL;
                str.allocated := 0;
                str.size := 0;
            END (*IF*);
        END (*WITH*);
    END SetString;

(********************************************************************************)

PROCEDURE GetString (port: IOPort;  N: CARDINAL;  copy: BOOLEAN): ByteStr;

    (* Returns a string that is kept in the session record.             *)
    (* Legal values of N are:                                           *)
    (*      1       client ID string (from initial greeting)            *)
    (*      2       server ID string (from initial greeting)            *)
    (*      3       client's KEXINIT message                            *)
    (*      4       server's KEXINIT message                            *)
    (*      5       server's public key                                 *)
    (*      6       exchange value sent by client                       *)
    (*      7       exchange value sent by server                       *)
    (*      8       the shared secret derived during key exchange       *)
    (* If copy = TRUE we make a copy, leaving the original in place in  *)
    (* the session record.                                              *)

    VAR result: ByteStr;

    BEGIN
        result := port^.string[N];
        IF copy THEN
            ALLOCATE (result.data, result.allocated);
            Copy (port^.string[N].data, result.data, result.allocated);
        ELSE
            result.data := port^.string[N].data;
            port^.string[N].data := NIL;
            port^.string[N].allocated := 0;
        END (*IF*);
        RETURN result;
    END GetString;

(********************************************************************************)

PROCEDURE ClearStrings (port: IOPort);

    (* Clears strings 3..8.  The first two are retained for the whole session. *)

    VAR j: CARDINAL;

    BEGIN
        FOR j := 3 TO 8 DO
            DiscardBS (port^.string[j]);
        END (*IF*);
    END ClearStrings;

(********************************************************************************)
(*                      SWITCHING TO NEW ENCRYPTION OPTIONS                     *)
(********************************************************************************)

PROCEDURE NewKeys (port: IOPort;  incoming: BOOLEAN;
                        newcipher: Encryption.CipherType;
                                    VAR (*IN*) key, IV: ARRAY OF CARD8);

    (* Starts use of the new cipher, with associated key info.  Set incoming    *)
    (* to TRUE for incoming data, FALSE for outgoing.  It is the caller's       *)
    (* responsibility to dispose of the key and IV information afterwards.      *)

    BEGIN
        IF incoming THEN
            Encryption.ChangeEncryption (port^.inCI, newcipher, key, IV);
        ELSE
            Encryption.ChangeEncryption (port^.outCI, newcipher, key, IV);
        END (*IF*);

        (* The following assignments are, strictly speaking, redundant, because *)
        (* the ciphertypes are held in port^.inCI and port^.outCI, but it is    *)
        (* convenient to allow the redundancy.                                  *)

        IF incoming THEN
            port^.incrypt := newcipher;
        ELSE
            port^.outcrypt := newcipher;
        END (*IF*);

    END NewKeys;

(********************************************************************************)

PROCEDURE NewMAC (port: IOPort;  incoming: BOOLEAN;  newalg: MAC.MACType;
                                               VAR (*IN*) key: ARRAY OF CARD8);

    (* Starts use of the new MAC algorithm.  Set incoming    *)
    (* to TRUE for incoming data, FALSE for outgoing.        *)

    VAR L: CARDINAL;

    BEGIN
        L := MAC.MAClength (newalg);
        IF incoming THEN
            port^.inmac := newalg;
            port^.inMACkeylength := L;
            Copy (ADR(key), ADR(port^.inMACkey), L);
        ELSE
            port^.outmac := newalg;
            port^.outMACkeylength := L;
            Copy (ADR(key), ADR(port^.outMACkey), L);
        END (*IF*);
    END NewMAC;

(********************************************************************************)

PROCEDURE NewCompression (port: IOPort;
                            clnttosrv, srvtoclnt: Compression.CompressionType);

    (* Switches to a new compression algorithm. The first parameter is for  *)
    (* client-to-server packets, and the second is for server-to-client.    *)

    BEGIN
        IF port^.server THEN
            port^.incompr := clnttosrv;
            port^.outcompr := srvtoclnt;
        ELSE
            port^.outcompr := clnttosrv;
            port^.incompr := srvtoclnt;
        END (*IF*);
    END NewCompression;

(********************************************************************************)
(*                             INITIAL GREETING                                 *)
(********************************************************************************)

PROCEDURE putline (s: Socket;  message: ARRAY OF CHAR);

    (* Sends plain-text message to client. *)

    VAR buffer: ARRAY [0..511] OF CHAR;
        size: CARDINAL;

    BEGIN
        Strings.Assign (message, buffer);
        size := AddEOL (buffer);
        EVAL (send (s, buffer, size, 0));
    END putline;

(********************************************************************************)

PROCEDURE InitialGreeting (port: IOPort);

    (* Sends the initial greeting, and get the other machine's initial greeting. *)

    (* A trap for the unwary: if the server sends its greeting first, what it   *)
    (* will get back is the concatenation of the client greeting and the        *)
    (* client's next packet. To avoid this, we must be careful about the        *)
    (* order of operations.                                                     *)

    CONST CR = CHR(13);  LF = CHR(10);  Nul = CHR(0);

    VAR pos: CARDINAL;  tofind: ARRAY [0..0] OF CHAR;
        found: BOOLEAN;
        our_id, their_id: ARRAY [0..255] OF CHAR;

    BEGIN
        IF NOT port^.server THEN
            (* Send our id to the other machine. *)

            Strings.Assign ("SSH-2.0-sshpm-client_", our_id);
            Strings.Append (SFV.version, our_id);
            putline (port^.sock, our_id);
        END (*IF*);

        (* Get the other machine's greeting. *)

        EVAL (Receive (port^.sock, their_id, 256));

        (* The returned length appears to be unreliable, so I have to use a *)
        (* different method of finding the end of the string.               *)

        tofind[0] := LF;
        Strings.FindNext (tofind, their_id, 0, found, pos);
        IF found THEN their_id[pos] := Nul; END (*IF*);
        tofind[0] := CR;
        Strings.FindNext (tofind, their_id, 0, found, pos);
        IF found THEN their_id[pos] := Nul; END (*IF*);

        IF port^.server THEN
            LogTransaction (port^.logID, their_id);

            (* Send our id to the other machine. *)

            Strings.Assign ("SSH-2.0-sftpd_", our_id);
            Strings.Append (SFV.version, our_id);
            putline (port^.sock, our_id);
        END (*IF*);

        (* Sort out which is server and which is client.          *)
        (* The two IDs will be needed by the KeyExchange module.  *)

        IF port^.server THEN
            SetString (port, 1, CharStrToByteStr(their_id), TRUE);
            SetString (port, 2, CharStrToByteStr(our_id), TRUE);
        ELSE
            SetString (port, 1, CharStrToByteStr(our_id), TRUE);
            SetString (port, 2, CharStrToByteStr(their_id), TRUE);
        END (*IF*);

    END InitialGreeting;

(********************************************************************************)
(*                                   INPUT                                      *)
(********************************************************************************)

PROCEDURE ReceiveAndDecrypt (port: IOPort;  VAR (*OUT*) data: ARRAY OF CARD8;
                                                    amount: CARDINAL): CARDINAL;

    (* Receives and decrypts a specified number of bytes.  Returns the actual   *)
    (* number of bytes received, or 0 in the case of a communications failure.  *)

    VAR p, q: ByteStringPtr;
        length, togo, N: CARDINAL;

    BEGIN
        ALLOCATE (p, amount);
        IF p = NIL THEN
            WriteLn;  WriteLn;
            WriteString ("Out of memory");  WriteLn;
            RETURN 0;
        END (*IF*);
        q := p;  togo := amount;  length := 0;
        WHILE togo > 0 DO
            N := Receive (port^.sock, q^, togo);
            IF (N = 0) OR (N = MAX(CARDINAL)) THEN
                length := 0;  togo := 0;
            ELSE
                INC (length, N);
                DEC (togo, N);
                q := AddOffset (q, N);
            END (*IF*);
        END (*WHILE*);

        IF length > 0 THEN
            IF CAST(ADDRESS,port^.inCI) = NIL THEN
                (* Treat as communications failure. *)
                length := 0;
            ELSE
                Decrypt (port^.inCI, length, p^, data);
            END (*IF*);
        END (*IF*);

        DEALLOCATE (p, amount);
        RETURN length;
    END ReceiveAndDecrypt;

(********************************************************************************)

PROCEDURE ReceivePacket (port: IOPort;  VAR (*OUT*) result: ByteStr): BOOLEAN;

    (* Receives and decrypts one packet.  Returns FALSE on failure. *)

    CONST TooBig = 131072;          (* packet size too large to be plausible *)

    TYPE OneWord =  RECORD
                        CASE :BOOLEAN OF
                              FALSE: w: CARDINAL;
                            | TRUE: b: ARRAY [0..3] OF CARD8;
                        END (*CASE*);
                    END (*RECORD*);

    VAR length, padlength, ml, k: CARDINAL;
        tempbuff: ARRAY [0..15] OF CARDINAL;
        temp: OneWord;
        p: ByteStringPtr;
        rawdata: ByteStr;
        pdigest, qdigest: POINTER TO ARRAY [0..63] OF CARD8;
        message: ARRAY [0..63] OF CHAR;
        equal, success: BOOLEAN;

    BEGIN
        (* Set default value in case of failure. *)

        result.allocated := 0;
        result.size := 0;
        result.data := NIL;

        IF port^.closing THEN
            port^.inputinprogress := FALSE;
            RETURN FALSE;
        ELSE
            port^.inputinprogress := TRUE;
        END (*IF*);
        k := Encryption.BlockSize (port^.incrypt);
        IF k < 8 THEN k := 8 END (*IF*);

        (* Check for lost connection or aborted session. *)

        IF WaitForSocket (port^.sock, MAX(CARDINAL)) <= 0 THEN
            port^.inputinprogress := FALSE;
            RETURN FALSE;
        END (*IF*);

        (* Receive and decrypt the first k bytes, to find out the size. *)

        p := ADR(tempbuff);
        IF ReceiveAndDecrypt (port, p^, k) <> k THEN
            IF port^.server THEN
                LogTransactionL (port^.logID, "Client disconnected.");
            ELSE
                LogTransactionL (port^.logID, "Server disconnected.");
            END (*IF*);
            port^.inputinprogress := FALSE;
            RETURN FALSE;
        END (*IF*);

        length := Swap4 (tempbuff[0]);
        temp.w := tempbuff[1];
        padlength := temp.b[0];

        IF (length = 0) OR (length >= TooBig) THEN
            (* A faulty packet, so give up. *)

            port^.inputinprogress := FALSE;
            RETURN FALSE;
        END (*IF*);

        (* Remember that the length so far given does not include the space     *)
        (* taken by the length field itself.                                    *)

        INC (length, 4);

        (* The short section already read must be restored, because it will be  *)
        (* used in the MAC calculation.  Then receive and decrypt the remainder.*)

        MakeBS (rawdata, length);
        IF rawdata.data = NIL THEN
            LogTransactionL (port^.logID, "Out of memory in ReceivePacket");
            message := "Packet size = ";
            AppendCard (length, message);
            LogTransaction (port^.logID, message);
        END (*IF*);
        Copy (ADR(tempbuff), rawdata.data, k);
        IF length > k THEN
            p := ADR(rawdata.data^[k]);
            IF ReceiveAndDecrypt (port, p^, length-k) = 0 THEN
                LogTransactionL (port^.logID, "Failure while receiving packet.");
                DiscardBS (rawdata);
                port^.inputinprogress := FALSE;
                RETURN FALSE;
            END (*IF*);
        END (*IF*);

        (* Receive and check the MAC. *)

        ml := MAC.MAClength (port^.inmac);
        IF ml > 0 THEN
            ALLOCATE (pdigest, ml);
            MAC.ComputeMac (port^.inmac, port^.inMACkey, port^.inMACkeylength,
                                port^.inseqno, rawdata, pdigest^);
            ALLOCATE (qdigest, ml);
            k := Receive (port^.sock, qdigest^, ml);
            IF k <> ml THEN
                LogTransactionL (port^.logID, "Failed to get whole of MAC");
                port^.inputinprogress := FALSE;
                RETURN FALSE;
            END (*IF*);
            (*
            Debug.DebugMes ("MAC computed:");
            p := CAST (ByteStringPtr, pdigest);
            Debug.DumpBytes (ml, p);
            Debug.DebugMes ("MAC received:");
            p := CAST (ByteStringPtr, qdigest);
            Debug.DumpBytes (ml, p);
            *)

            equal := TRUE;

            FOR k := 0 TO ml DIV 4 - 1 DO
                IF pdigest^[k] <> qdigest^[k] THEN
                    equal := FALSE;
                END (*IF*);
            END (*FOR*);
            IF NOT equal THEN
                LogTransactionL (port^.logID, "MAC value not as expected");
            END (*IF*);
            DEALLOCATE (qdigest, ml);
            DEALLOCATE (pdigest, ml);
        END (*IF*);

        (* The first five bytes are no longer needed. *)

        BSDelete (rawdata, 0, 5);

        (* Remove padding before decompression, by decrementing size. *)

        DEC (rawdata.size, padlength);
        Decompress (port^.incompr, rawdata, result);

        <* COVERFLOW- *>
        INC (port^.inseqno);    (* Allow sequence number to wrap to 0 *)
        <* COVERFLOW+ *>

        success := TRUE;
        IF port^.closing THEN
            DiscardBS (result);
            success := FALSE;
        END (*IF*);

        port^.inputinprogress := FALSE;
        RETURN success;

    END ReceivePacket;

(********************************************************************************)
(*                                   OUTPUT                                     *)
(********************************************************************************)

PROCEDURE SendPacket (port: IOPort;  VAR (*INOUT*) data: ByteStr): BOOLEAN;

    (* Encrypts and sends one packet.  The original data are discarded. *)
    (* A FALSE result means that connection is lost.                    *)

    VAR intermediate, plaindata, outdata: ByteStr;
        length, k, minsize, blksize, ml: CARDINAL;
        padlength: CARD8;
        (*p: ByteStringPtr;*)
        pdigest: POINTER TO ARRAY [0..63] OF CARD8;

    BEGIN
        IF (port = NIL) OR (port^.sock = NotASocket)
                     OR (port^.outCI = CAST(Encryption.CipherInfo, NIL)) THEN
            RETURN FALSE;
        END (*IF*);

        (* To note:                                                         *)
        (*      data is the payload before compression and encryption       *)
        (*      intermediate is the result after compression                *)
        (*      plaindata is the assembled data including padding, etc.     *)
        (*      outdata is the encrypted form of plaindata.                 *)
        (* All of these must be discarded before this procedure returns.    *)

        Compress (port^.outcompr, data, intermediate);

        (* The packet length, including the length field itself but not     *)
        (* including the MAC, is                                            *)
        (*       payload length + 5 + padding length                        *)
        (* This must be a multiple of the cipher block size or 8, whichever *)
        (* is larger.  Minimum padding length is 4.  Minimum packet size is *)
        (* cipher block size or 16, whichever is larger.                    *)

        length := intermediate.size + 5;
        blksize := Encryption.BlockSize (port^.outcrypt);
        minsize := blksize;
        IF minsize < 16 THEN minsize := 16 END (*IF*);
        IF blksize < 8 THEN blksize := 8 END (*IF*);
        padlength := blksize - length MOD blksize;
        IF padlength < 4 THEN INC(padlength, blksize) END (*IF*);
        IF length + padlength < minsize THEN
            INC (padlength, blksize);
        END (*IF*);
        INC (length, padlength);

        (* At this stage, the length covers everything except the MAC.      *)

        MakeBS (plaindata, length);
        k := Swap4 (length - 4);
        Copy (ADR(k), plaindata.data, 4);
        plaindata.data^[4] := padlength;
        Copy (intermediate.data, ADR(plaindata.data^[5]),
                                        intermediate.size);
        RandomBytes (padlength, plaindata.data^[intermediate.size+5]);

        (* We have now assembled everything except the MAC, and the         *)
        (* intermediate version is no longer needed.  We'll still keep the  *)
        (* plaindata version until the MAC is computed.                     *)

        DiscardBS (intermediate);
        MakeBS (outdata, length);
        Encrypt (port^.outCI, length, plaindata.data^, outdata.data^);

        (* Send this much to the output socket. *)

        k := send (port^.sock, outdata.data^, outdata.size, 0);
        IF k <> outdata.size THEN
            LogTransactionL (port^.logID, "Not all output sent");
        END (*IF*);

        (* We still have to compute and send the MAC. *)

        ml := MAC.MAClength (port^.outmac);
        IF ml > 0 THEN
            ALLOCATE (pdigest, ml);
            MAC.ComputeMac (port^.outmac, port^.outMACkey, port^.outMACkeylength, port^.outseqno,
                            plaindata, pdigest^);
            IF send (port^.sock, pdigest^, ml, 0) <> ml THEN
                LogTransactionL (port^.logID, "Failed to send whole of MAC");
            END (*IF*);
            DEALLOCATE (pdigest, ml);
        END (*IF*);

        DiscardBS (plaindata);
        DiscardBS (outdata);
        <* COVERFLOW- *>
        INC (port^.outseqno);       (* allow sequence numbers to wrap *)
        <* COVERFLOW+ *>
        RETURN TRUE;

    END SendPacket;

(********************************************************************************)

END PacketIO.

