(**************************************************************************)
(*                                                                        *)
(*  Transferring data to higher layers of SSH                             *)
(*  Copyright (C) 2022   Peter Moylan                                     *)
(*                                                                        *)
(*  This program is free software: you can redistribute it and/or modify  *)
(*  it under the terms of the GNU General Public License as published by  *)
(*  the Free Software Foundation, either version 3 of the License, or     *)
(*  (at your option) any later version.                                   *)
(*                                                                        *)
(*  This program is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU General Public License for more details.                          *)
(*                                                                        *)
(*  You should have received a copy of the GNU General Public License     *)
(*  along with this program.  If not, see <http://www.gnu.org/licenses/>. *)
(*                                                                        *)
(*  To contact author:   http://www.pmoylan.org   peter@pmoylan.org       *)
(*                                                                        *)
(**************************************************************************)

IMPLEMENTATION MODULE PacketQueues;

        (********************************************************)
        (*                                                      *)
        (*               Interface between layers               *)
        (*                                                      *)
        (*  Programmer:         P. Moylan                       *)
        (*  Started:            29 May 2021                     *)
        (*  Last edited:        16 April 2022                   *)
        (*  Status:             OK so far                       *)
        (*                                                      *)
        (********************************************************)


FROM SYSTEM IMPORT CAST, ADDRESS;

IMPORT Strings;

FROM VarStrings IMPORT
    (* type *)  ByteStr,
    (* proc *)  DiscardBS;

FROM TransLog IMPORT
    (* type *)  TransactionLogID,
    (* proc *)  LogTransaction, LogTransactionL;

FROM Storage IMPORT
    (* proc *)  ALLOCATE, DEALLOCATE;

FROM Semaphores IMPORT
    (* type *)  Semaphore,
    (* proc *)  CreateSemaphore, DestroySemaphore, Wait, Signal;

FROM TaskControl IMPORT
    (* type *)  Lock,
    (* proc *)  CreateLock, DestroyLock, Obtain, Release;

(********************************************************************************)

TYPE
    (* Queues for transferring data packets between protocol layers. *)

    Qelt = POINTER TO
                RECORD
                    next: Qelt;
                    packet: ByteStr;
                END (*RECORD*);

    PacketQ = POINTER TO
                RECORD
                    access: Lock;
                    count: Semaphore;
                    head, tail: Qelt;
                END (*RECORD*);

(********************************************************************************)
(*                                                                              *)
(*                   TRANSFERRING PACKETS TO HIGHER LEVELS                      *)
(*                                                                              *)
(* The first byte of a packet says what sort of message it is.  The transport   *)
(* layer handles message numbers 1..49.  For higher message numbers we put the  *)
(* packet on a queue, from which it can be fetched by a higher-level layer of   *)
(* the SSH protocol.                                                            *)
(*                                                                              *)
(********************************************************************************)

PROCEDURE CreateQueue(): PacketQ;

    (* Creates a new queue. *)

    VAR Q: PacketQ;

    BEGIN
        NEW (Q);
        CreateLock (Q^.access);
        CreateSemaphore (Q^.count, 0);
        Q^.head := NIL;
        Q^.tail := NIL;
        RETURN Q;
    END CreateQueue;

(********************************************************************************)

PROCEDURE CloseQueue (VAR (*INOUT*) Q: PacketQ);

    (* Discards the queue, including any waiting packets. *)

    VAR elt, next: Qelt;

    BEGIN
        IF Q <> NIL THEN
            Obtain (Q^.access);
            elt := Q^.head;
            WHILE elt <> NIL DO
                next  := elt^.next;
                DiscardBS (elt^.packet);
                DISPOSE (elt);
                elt := next;
            END (*WHILE*);
            DestroySemaphore (Q^.count);
            Q^.count := NIL;
            Release (Q^.access);
            DestroyLock (Q^.access);
            DISPOSE (Q);
        END (*IF*);
    END CloseQueue;

(********************************************************************************)

PROCEDURE PutPacket (Q: PacketQ;  VAR (*INOUT*) packet: ByteStr);

    (* Moves the packet to the specified queue. If the queue doesn't exist we   *)
    (* throw away the packet content. In all cases, the data are gone from the  *)
    (* caller's point of view.                                                  *)

    CONST NilSem = CAST(Semaphore, NIL);

    VAR elt: Qelt;

    BEGIN
        IF (Q = NIL) OR (Q^.count = NilSem) THEN
            DiscardBS (packet);
        ELSE
            Obtain (Q^.access);
            NEW (elt);
            elt^.next := NIL;
            elt^.packet := packet;
            IF Q^.tail = NIL THEN
                Q^.head := elt;
            ELSE
                Q^.tail^.next := elt;
            END (*IF*);
            Q^.tail := elt;
            Release (Q^.access);
            Signal (Q^.count);

            (* For neatness, the caller must see that the data are gone. *)

            packet.size := 0;  packet.allocated := 0;
            packet.data := NIL;

        END (*IF*);
    END PutPacket;

(********************************************************************************)

PROCEDURE GetPacket (Q: PacketQ): ByteStr;

    (* Fetches the next input packet.    *)

    VAR elt: Qelt;
        result: ByteStr;

    BEGIN
        (* Set a default result. *)

        result.size := 0;  result.allocated := 0;  result.data := NIL;

        IF Q <> NIL THEN
            Wait (Q^.count);
            Obtain (Q^.access);
            elt := Q^.head;
            IF elt <> NIL THEN
                Q^.head := elt^.next;
                IF Q^.head = NIL THEN Q^.tail := NIL; END(*IF*);
                result := elt^.packet;
                DISPOSE (elt);
            END (*IF*);
            Release (Q^.access);
        END (*IF*);
        RETURN result;
    END GetPacket;

(********************************************************************************)

PROCEDURE MatchAddr (A, B: ADDRESS): BOOLEAN;

    (* Returns TRUE if A = B. *)

    BEGIN
        RETURN A = B;
    END MatchAddr;

(********************************************************************************)

PROCEDURE CheckConsistency (Q, originalQ: PacketQ;  label: ARRAY OF CHAR;
                                                    logID: TransactionLogID);

    (* Checks that Q is the same as originalQ, apart from the fact that *)
    (* originalQ is an empty queue and Q can have queue entries.        *)

    VAR message: ARRAY [0..127] OF CHAR;

    BEGIN
        Strings.Assign (label, message);
        IF NOT MatchAddr (Q, originalQ) THEN
            Strings.Append (": inconsistent queue pointer", message);
            LogTransaction (logID, message);
        ELSIF NOT MatchAddr (CAST(ADDRESS,Q^.access), CAST(ADDRESS, originalQ^.access)) THEN
            Strings.Append (": inconsistent access pointer", message);
            LogTransaction (logID, message);
        ELSIF NOT MatchAddr (CAST(ADDRESS,Q^.count), CAST(ADDRESS,originalQ^.count)) THEN
            Strings.Append (": inconsistent pointer to count semaphore", message);
            LogTransaction (logID, message);
        END (*IF*);
    END CheckConsistency;

(********************************************************************************)

PROCEDURE PutPacketCmp (Q, originalQ: PacketQ;  packet: ByteStr; logID: TransactionLogID);

    (* This is the same as PutPacket, except that we check for queue corruption *)
    (* by comparing with the original (empty) queue.                            *)

    BEGIN
        CheckConsistency (Q, originalQ, "PutPacketCmp, before PutPacket", logID);
        PutPacket (Q, packet);
        CheckConsistency (Q, originalQ, "PutPacketCmp, after PutPacket", logID);
    END PutPacketCmp;

(********************************************************************************)

PROCEDURE GetPacketCmp (Q, originalQ: PacketQ;  logID: TransactionLogID): ByteStr;

    (* This is the same as GetPacket, except that we check for queue corruption *)
    (* by comparing with the original (empty) queue.                            *)

    BEGIN
        CheckConsistency (Q, originalQ, "GetPacketCmp", logID);
        RETURN GetPacket (Q);
    END GetPacketCmp;

(********************************************************************************)

END PacketQueues.

