IMPLEMENTATION MODULE PublicKeys;

        (********************************************************)
        (*                                                      *)
        (*             Public key exchange algorithms           *)
        (*                                                      *)
        (*     (At present, only RSA keys are implemented)      *)
        (*                                                      *)
        (*  Programmer:         P. Moylan                       *)
        (*  Started:            19 January 2021                 *)
        (*  Last edited:        13 January 2021                 *)
        (*  Status:             OK                              *)
        (*                                                      *)
        (*  Relevant standard:  RFC 8332                        *)
        (*                                                      *)
        (********************************************************)


IMPORT sshRSA, HashAlgs, Strings, FileOps, Base64, OS2;

FROM NameLists IMPORT
    (* proc *)  MatchNames, ListAll;

FROM VarStrings IMPORT
    (* type *)  ByteStr;

FROM Inet2Misc IMPORT
    (* proc *)  NameIsNumeric, IPToString;

FROM Internet IMPORT
    (* proc *)  inet_addr;

FROM NetDB IMPORT
    (* type *)  HostEntPtr, AddressPointerArrayPointer,
    (* proc *)  gethostbyname;

FROM STextIO IMPORT
    (* proc *)  WriteChar, WriteString, WriteLn;

FROM TransLog IMPORT
    (* type *)  TransactionLogID, LogTransactionL;

(************************************************************************)

TYPE
    NameArray = ARRAY PKAType OF ARRAY [0..15] OF CHAR;

CONST
    Names = NameArray {"rsa-sha2-256", "rsa-sha2-512", "ssh-rsa", "none"};
    Label = NameArray {"ssh-rsa", "ssh-rsa", "ssh-rsa", '?'};

    emptybytestring = ByteStr {0, 0, NIL};

(************************************************************************)
(*                   IDENTIFIERS FOR PUBLIC KEY TYPES                   *)
(************************************************************************)

PROCEDURE NameOf (ka: PKAType;  VAR (*OUT*) name: ARRAY OF CHAR);

    (* Returns the RFC 4253 name of mt. *)

    BEGIN
        Strings.Assign (Names[ka], name);
    END NameOf;

(************************************************************************)

PROCEDURE AllNames (VAR (*OUT*) list: ARRAY OF CHAR);

    (* Returns a comma-separated list of our algorithm names.  *)

    CONST max = ORD(MAX(PKAType));

    BEGIN
        ListAll (Names, max, list);
    END AllNames;

(************************************************************************)

PROCEDURE FindMatch (VAR (*IN*) list: ARRAY OF CHAR): PKAType;

    (* Input is a comma-separated list of algorithm names.  *)
    (* Returns the first one we support.                    *)

    CONST max = ORD(MAX(PKAType));

    VAR result: CARDINAL;

    BEGIN
        result := MatchNames (Names, max, list);
        IF result > max THEN RETURN none;
        ELSE RETURN VAL(PKAType, result);
        END (*IF*);
    END FindMatch;

(************************************************************************)
(*                             CHECKING A KEY                           *)
(************************************************************************)

PROCEDURE IPaddr (host: ARRAY OF CHAR): CARDINAL;

    (* Translates a textual hostname into a numeric address. *)

    VAR result: CARDINAL;
        HostInfo: HostEntPtr;
        p: AddressPointerArrayPointer;

    BEGIN
        IF NameIsNumeric(host) THEN
            result := inet_addr(host);
        ELSE
            HostInfo := gethostbyname (host);
            IF HostInfo = NIL THEN p := NIL
            ELSE p := HostInfo^.h_addr_list
            END (*IF*);
            IF p = NIL THEN
                result := 0;
            ELSE
                result := p^[0]^;
            END (*IF*);
        END (*IF*);
        RETURN result;
    END IPaddr;

(************************************************************************)

PROCEDURE CheckPublicKey (ka: PKAType;  peeraddress: CARDINAL;
                                VAR (*IN*) key: ByteStr;
                                  logID: TransactionLogID): BOOLEAN;

    (* This is a client operation, where the client wants to know       *)
    (* whether key is valid for a server.   The peer address is in      *)
    (* network byte order.                                              *)

    (* We consult our own file of known peers.                          *)

    CONST
        CtrlZ = CHR(26);
        filename = ".ssh\known_hosts";
        MaxChars = 8192;

    VAR cid: FileOps.ChanId;
        CharData: OS2.KBDKEYINFO;
        line, key_b64: ARRAY [0..MaxChars-1] OF CHAR;
        address, kind: ARRAY [0..31] OF CHAR;
        pos: CARDINAL;
        ch: CHAR;
        match, found, duplicate: BOOLEAN;

    BEGIN
        duplicate := FALSE;  found := FALSE;
        Base64.Encode (key.data^, key.size, key_b64);
        cid := FileOps.OpenOldFile (filename, FALSE, FALSE);
        match := cid <> FileOps.NoSuchChannel;
        IF match THEN
            REPEAT
                match := FALSE;
                FileOps.ReadLine (cid, line);
                IF line[0] <> CtrlZ THEN
                    Strings.FindNext (' ', line, 0, found, pos);
                    IF found THEN
                        Strings.Extract (line, 0, pos, address);
                        Strings.Delete (line, 0, pos+1);
                        Strings.FindNext (' ', line, 0, found, pos);
                        IF found THEN
                            Strings.Extract (line, 0, pos, kind);
                            Strings.Delete (line, 0, pos+1);
                        END (*IF*);
                    END (*IF*);
                    found := found AND (peeraddress = IPaddr(address))
                                    AND Strings.Equal (kind, Label[ka]);
                    IF found THEN
                        match := Strings.Equal (line, key_b64);
                        duplicate := duplicate OR NOT match;
                    END (*IF*);
                END (*IF*);
            UNTIL match OR (line[0] = CtrlZ);
        END (*IF*);
        FileOps.CloseFile (cid);

        (* End of search.  If no match, give the user the option of     *)
        (* accepting this public key anyway.                            *)

        IF NOT match THEN
            IF duplicate THEN
                WriteString ("WARNING: We have a different public key for this site. ");
                WriteString ("That could indicate");  WriteLn;
                WriteString ("someone trying to hijack this connection.");  WriteLn;
            ELSE
                WriteString ("We have no public key record for this site. ");
            END (*IF*);
            WriteString ("Accept new key anyway (Y/N)?");
            OS2.KbdCharIn (CharData, 0, 0);
            ch := CharData.chChar;
            WriteChar (ch);  WriteLn;
            IF CAP(ch) = 'Y' THEN
                match := TRUE;
                cid := FileOps.OpenAtEnd (filename);
                IPToString (peeraddress, FALSE, address);
                FileOps.FWriteString (cid, address);
                FileOps.FWriteChar (cid, ' ');
                FileOps.FWriteString (cid, Label[ka]);
                FileOps.FWriteChar (cid, ' ');
                FileOps.FWriteString (cid, key_b64);
                FileOps.FWriteChar (cid, CHR(13));
                FileOps.FWriteChar (cid, CHR(10));
                FileOps.CloseFile (cid);
                IF found THEN
                    WriteString ("Suggestion: edit ");  WriteString (filename);
                    WriteString (" to remove the old public key.");  WriteLn;
                END (*IF*);
            ELSE
                WriteString ("Breaking connection");  WriteLn;
            END (*IF*);
        END (*IF*);

        RETURN match;

    END CheckPublicKey;

(************************************************************************)
(*                         RETURNING OUR OWN KEY                        *)
(************************************************************************)

PROCEDURE GetFormattedPublicKey (ka: PKAType): ByteStr;

    (* Returns a copy of our key in the key format of RFC 4253, section 6.6. *)

    BEGIN
        CASE ka OF
          | sshrsa, rsa_sha2_256, rsa_sha2_512:
                        RETURN sshRSA.GetFormattedPublicKey();
          | none:       RETURN emptybytestring;
        END (*CASE*);
    END GetFormattedPublicKey;

(************************************************************************)
(*                         GENERATING A SIGNATURE                       *)
(************************************************************************)

PROCEDURE Sign (ka: PKAType;  message: ByteStr): ByteStr;

    (* Creates a signature using a public key.  *)

    BEGIN
        CASE ka OF
          | sshrsa:
                    RETURN sshRSA.Sign (HashAlgs.sha1, message);
          | rsa_sha2_256:
                    RETURN sshRSA.Sign (HashAlgs.sha2_256, message);
          | rsa_sha2_512:
                    RETURN sshRSA.Sign (HashAlgs.sha2_512, message);
          | none:   RETURN emptybytestring;
        END (*CASE*);
    END Sign;

(************************************************************************)
(*                          CHECKING A SIGNATURE                        *)
(************************************************************************)

PROCEDURE SignatureValid (ka: PKAType;  message, publickey: ByteStr;
                            VAR (*INOUT*) signature: ByteStr;
                              logID: TransactionLogID): BOOLEAN;

    (* Checks that signature is a valid signature for message.  *)
    (* During processing the signature is discarded.            *)

    BEGIN
        CASE ka OF
          | sshrsa:
                RETURN sshRSA.CheckSignature (HashAlgs.sha1, message, publickey, signature, logID);
          | rsa_sha2_256:
                RETURN sshRSA.CheckSignature (HashAlgs.sha2_256, message, publickey, signature, logID);
          | rsa_sha2_512:
                RETURN sshRSA.CheckSignature (HashAlgs.sha2_512, message, publickey, signature, logID);
          | none:
                RETURN TRUE;
        END (*CASE*);
    END SignatureValid;

(************************************************************************)

END PublicKeys.

