(**************************************************************************)
(*                                                                        *)
(*  My own version of SSH                                                 *)
(*  Copyright (C) 2022   Peter Moylan                                     *)
(*                                                                        *)
(*  This program is free software: you can redistribute it and/or modify  *)
(*  it under the terms of the GNU General Public License as published by  *)
(*  the Free Software Foundation, either version 3 of the License, or     *)
(*  (at your option) any later version.                                   *)
(*                                                                        *)
(*  This program is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU General Public License for more details.                          *)
(*                                                                        *)
(*  You should have received a copy of the GNU General Public License     *)
(*  along with this program.  If not, see <http://www.gnu.org/licenses/>. *)
(*                                                                        *)
(*  To contact author:   http://www.pmoylan.org   peter@pmoylan.org       *)
(*                                                                        *)
(**************************************************************************)

IMPLEMENTATION MODULE TransportLayer;

        (********************************************************)
        (*                                                      *)
        (*                   Transport layer                    *)
        (*                                                      *)
        (*  Programmer:         P. Moylan                       *)
        (*  Started:            5 January 2021                  *)
        (*  Last edited:        3 October 2022                  *)
        (*  Status:             Working                         *)
        (*                                                      *)
        (********************************************************)


FROM SYSTEM IMPORT (*type*) CARD8, (*proc*) ADR, CAST;

IMPORT Strings, BigNum, KeyExchange, PublicKeys, HashAlgs,
            Encryption, MAC, Compression, PacketQueues, AuthLayer;

FROM PacketIO IMPORT
    (* type *)  IOPort,
    (* proc *)  InitialGreeting, OpenPacketSession, ClosePacketSession,
                SetInitLock, SetString, ClearStrings, NewKeys, NewMAC, NewCompression,
                ReceivePacket, SendPacket, LastSeqNo, DeactivatePort, IsClosing;

FROM Consts IMPORT
    (* const*)  SSH_MSG_DISCONNECT, SSH_MSG_IGNORE, SSH_MSG_UNIMPLEMENTED,
                SSH_MSG_DEBUG, SSH_MSG_SERVICE_REQUEST, SSH_MSG_SERVICE_ACCEPT,
                SSH_MSG_KEXINIT, SSH_MSG_KEXDH_INIT, SSH_MSG_KEXDH_REPLY,
                SSH_MSG_NEWKEYS, SSH_MSG_TERMINATE,
                SSH_DISCONNECT_KEY_EXCHANGE_FAILED;

FROM VarStrings IMPORT
    (* type *)  ByteStr,
    (* proc *)  ByteStrToStr, MakeBS, CopyOfBS, DiscardBS, AddCharString,
                GetCharString;

FROM BigNum IMPORT
    (* type *)  BN;

FROM MiscFuncs IMPORT
    (* proc *)  AppendCard, EVAL;

FROM Inet2Misc IMPORT
    (* proc *)  Swap4;

FROM TransLog IMPORT
    (* type *)  TransactionLogID,
    (* proc *)  LogTransaction, LogTransactionL;

FROM Semaphores IMPORT
    (* type *)  Semaphore,
    (* proc *)  Signal;

FROM RandomData IMPORT
    (* proc *)  RandomBytes;

FROM LowLevel IMPORT
    (* proc *)  Copy;

FROM Storage IMPORT
    (* proc *)  ALLOCATE, DEALLOCATE;

(********************************************************************************)

TYPE
    (* The options that are set by algorithm negotiation. *)

    Options = POINTER TO RECORD
                            kex:                KeyExchange.KexType;
                            kgctx:              HashAlgs.KeygenCtx;
                            pka:                PublicKeys.PKAType;
                            o2: ARRAY (server_to_client, client_to_server) OF
                                    RECORD
                                        cipher: Encryption.CipherType;
                                        mac:    MAC.MACType;
                                        compr:  Compression.CompressionType;
                                    END (*RECORD*);
                         END (*RECORD*);

(********************************************************************************)
(*                     NEGOTIATING WHAT ALGORITHMS TO USE                       *)
(********************************************************************************)

PROCEDURE SendAlgorithmList (port: IOPort;  ID: TransactionLogID);

    (* CLIENT CODE *)

    (* Sends the client SSH_MSG_KEXINIT packet. *)

    CONST LogSentAlgs = FALSE;

    VAR list: ARRAY [0..9] OF ARRAY [0..255] OF CHAR;
        message: ARRAY [0..2047] OF CHAR;
        packet: ByteStr;
        j, k, size: CARDINAL;

    BEGIN
        (* Calculate the size of the packet. *)

        size := 17;         (* for the command and cookie *)

        FOR k := 0 TO 9 DO
            CASE k OF
              | 0:  Strings.Assign ("Key exchange: ", message);
                    KeyExchange.AllNames (list[k]);
              | 1:  Strings.Assign ("Host key: ", message);
                    PublicKeys.AllNames (list[k]);
              | 2:  Strings.Assign ("Encryption (client to server): ", message);
                    Encryption.AllNames (list[k]);
              | 3:  Strings.Assign ("Encryption (server to client): ", message);
                    Encryption.AllNames (list[k]);
              | 4:  Strings.Assign ("MAC (client to server): ", message);
                    MAC.AllNames (list[k]);
              | 5:  Strings.Assign ("MAC (server to client): ", message);
                    MAC.AllNames (list[k]);
              | 6:  Strings.Assign ("Compression (client to server): ", message);
                    Compression.AllNames (list[k]);
              | 7:  Strings.Assign ("Compression (server to client): ", message);
                    Compression.AllNames (list[k]);
              | 8:  Strings.Assign ("Language client to server: ", message);
                    list[k] := "";
              | 9:  Strings.Assign ("Language server to client: ", message);
                    list[k] := "";
            END (*CASE*);
            IF LogSentAlgs AND (list[k][0] <> CHR(0)) THEN
                Strings.Append (list[k], message);
                LogTransaction (ID, message);
            END (*IF*);
            INC (size, 4 + LENGTH (list[k]));
        END (*FOR*);
        INC (size, 5);

        (* Allocate the space.  *)

        MakeBS (packet, size);

        (* Fill in the payload. *)

        packet.data^[0] := SSH_MSG_KEXINIT;
        RandomBytes (16, packet.data^[1]);
        j := 17;
        FOR k := 0 TO 9 DO
            AddCharString (packet, j, list[k]);
        END (*FOR*);

        packet.data^[j] := 0;  INC(j);  (* first_kex_packet_follows flag *)
        packet.data^[j] := 0;  INC(j);  (* four reserved bytes *)
        packet.data^[j] := 0;  INC(j);
        packet.data^[j] := 0;  INC(j);
        packet.data^[j] := 0;
        SetString (port, 3, packet, TRUE);

        EVAL(SendPacket (port, packet));

        (* We now expect the server to respond with choices from these.  *)

    END SendAlgorithmList;

(********************************************************************************)

PROCEDURE GetAlgorithmList (options: Options;  packet: ByteStr;
                            ID: TransactionLogID;  VAR (*OUT*) KEXpacketFollows: BOOLEAN);

    (* Decodes an SSH_MSG_KEXINIT packet.  This is used by both server and client. *)

    CONST LogAlgorithmLists = FALSE;

    VAR pos: CARDINAL;
        text: ARRAY [0..2047] OF CHAR;

    (****************************************************************************)

    PROCEDURE GetOneList (label: ARRAY OF CHAR);

        (* Gets a list of algorithm names starting at packet.data[pos], updates *)
        (* pos.  The result is stored in text.                                  *)

        BEGIN
            GetCharString (packet, pos, text);
            IF LogAlgorithmLists THEN
                LogTransaction (ID, label);
                LogTransaction (ID, text);
            END (*IF*);
        END GetOneList;

    (****************************************************************************)

    VAR cookie: ARRAY [0..15] OF CARD8;

    BEGIN
        IF LogAlgorithmLists THEN
            LogTransactionL (ID, "algorithm list from peer");
        END (*IF*);
        pos := 1;
        Copy (ADR(packet.data^[pos]), ADR(cookie), 16);
        INC (pos, 16);

        GetOneList ("key exchange algorithms");
        options^.kex := KeyExchange.FindMatch (text);
        IF options^.kex = KeyExchange.none THEN
            LogTransactionL (ID, "Cannot agree on a key exchange algorithm");
        END (*IF*);

        GetOneList ("server host key algorithms");
        options^.pka := PublicKeys.FindMatch (text);
        IF options^.pka = PublicKeys.none THEN
            LogTransactionL (ID, "Cannot agree on a server host key");
        END (*IF*);

        GetOneList ("encryption client to server");
        options^.o2[client_to_server].cipher := Encryption.FindMatch (text);
        IF options^.o2[client_to_server].cipher = Encryption.none THEN
            LogTransactionL (ID, "Cannot agree on an encryption algorithm");
        END (*IF*);

        GetOneList ("encryption server to client");
        options^.o2[server_to_client].cipher := Encryption.FindMatch (text);
        IF options^.o2[server_to_client].cipher = Encryption.none THEN
            LogTransactionL (ID, "Cannot agree on an encryption algorithm");
        END (*IF*);

        GetOneList ("MAC algorithms client to server");
        options^.o2[client_to_server].mac := MAC.FindMatch (text);
        IF options^.o2[client_to_server].mac = MAC.none THEN
            LogTransactionL (ID, "Cannot agree on a MAC algorithm");
        END (*IF*);

        GetOneList ("MAC algorithms server to client");
        options^.o2[server_to_client].mac := MAC.FindMatch (text);
        IF options^.o2[server_to_client].mac = MAC.none THEN
            LogTransactionL (ID, "Cannot agree on a MAC algorithm");
        END (*IF*);

        GetOneList ("compression algorithms client to server");
        options^.o2[client_to_server].compr := Compression.FindMatch (text);
        (* For compression, we don't mind a "none" result. *)

        GetOneList ("compression algorithms server to client");
        options^.o2[server_to_client].compr := Compression.FindMatch (text);
        (* For compression, we don't mind a "none" result. *)

        GetOneList ("languages client to server");
        GetOneList ("languages server to client");

        (* KEX packet will follow? *)

        KEXpacketFollows := packet.data^[pos] <> 0;

        IF KEXpacketFollows THEN
            LogTransactionL (ID, "first KEX packet will follow");
        END (*IF*);

    END GetAlgorithmList;

(********************************************************************************)

PROCEDURE SendAlgorithmReply (port: IOPort;  options: Options;  ID: TransactionLogID);

    (* Sends the server's reply to SSH_MSG_KEXINIT. *)

    CONST Nul = CHR(0);
        verbose = FALSE;

    VAR name: ARRAY [0..9] OF ARRAY [0..31] OF CHAR;
        message: ARRAY [0..2047] OF CHAR;
        packet: ByteStr;
        j, k, size: CARDINAL;

    BEGIN
        (* Calculate the size of the reply packet. *)

        size := 17;         (* for the command and cookie *)

        message := "";
        FOR k := 0 TO 9 DO
            CASE k OF
              | 0:  IF verbose THEN message := "Key exchange: " END(*IF*);
                    KeyExchange.NameOf (options^.kex, name[k]);
              | 1:  IF verbose THEN message := "Host key: " END(*IF*);
                    PublicKeys.NameOf (options^.pka, name[k]);
              | 2:  IF verbose THEN message := "Encryption (client to server): " END(*IF*);
                    Encryption.NameOf (options^.o2[client_to_server].cipher, name[k]);
              | 3:  IF verbose THEN message := "Encryption (server to client): " END(*IF*);
                    Encryption.NameOf (options^.o2[server_to_client].cipher, name[k]);
              | 4:  IF verbose THEN message := "MAC (client to server): " END(*IF*);
                    MAC.NameOf (options^.o2[client_to_server].mac, name[k]);
              | 5:  IF verbose THEN message := "MAC (server to client): " END(*IF*);
                    MAC.NameOf (options^.o2[server_to_client].mac, name[k]);
              | 6:  IF verbose THEN message := "Compression (client to server): " END(*IF*);
                    Compression.NameOf (options^.o2[client_to_server].compr, name[k]);
              | 7:  IF verbose THEN message := "Compression (server to client): " END(*IF*);
                    Compression.NameOf (options^.o2[server_to_client].compr, name[k]);
              | 8:  IF verbose THEN message := "Language client to server: " END(*IF*);
                    name[k] := "";
              | 9:  IF verbose THEN message := "Language server to client: " END(*IF*);
                    name[k] := "";
            END (*CASE*);
            IF (name[k][0] <> Nul) AND NOT Strings.Equal(name[k], "none") THEN
                IF verbose THEN
                    Strings.Append (name[k], message);
                    LogTransaction (ID, message);
                ELSIF (k = 1) OR NOT ODD(k) THEN
                    IF k > 0 THEN
                        Strings.Append (",", message);
                    END (*IF*);
                    Strings.Append (name[k], message);
                END (*IF*);
            END (*IF*);
            INC (size, 4 + LENGTH (name[k]));
        END (*FOR*);
        IF NOT verbose THEN
            LogTransaction (ID, message);
        END (*IF*);
        INC (size, 5);

        (* Allocate the space.  *)

        MakeBS (packet, size);

        (* Fill in the payload. *)

        packet.data^[0] := SSH_MSG_KEXINIT;
        RandomBytes (16, packet.data^[1]);
        j := 17;
        FOR k := 0 TO 9 DO
            AddCharString (packet, j, name[k]);
        END (*FOR*);

        packet.data^[j] := 0;  INC(j);  (* first_kex_packet_follows flag *)
        packet.data^[j] := 0;  INC(j);  (* four reserved bytes *)
        packet.data^[j] := 0;  INC(j);
        packet.data^[j] := 0;  INC(j);
        packet.data^[j] := 0;
        SetString (port, 4, packet, TRUE);

        EVAL(SendPacket (port, packet));

        (* Now, the standard says that these options should not take effect     *)
        (* until ???, but it seems to me that the key exchange and server host  *)
        (* key algorithms should take effect immediately, or there can be no    *)
        (* useful key exchange.                                                 *)

    END SendAlgorithmReply;

(********************************************************************************)
(*                           THE ACTUAL KEY EXCHANGE                            *)
(********************************************************************************)

PROCEDURE StartKeySwitch (port: IOPort);

    (* Sends an SSH_MSG_NEWKEYS message to the peer. *)

    VAR packet: ByteStr;

    BEGIN
        MakeBS (packet, 1);
        packet.data^[0] := SSH_MSG_NEWKEYS;
        EVAL(SendPacket (port, packet));
    END StartKeySwitch;

(********************************************************************************)

PROCEDURE ActivateNewKeys (port: IOPort;  server: BOOLEAN;  options: Options);

    (* Responds to an SSH_MSG_NEWKEYS message by switching to the specified     *)
    (* options, and also generating any keys or initialisation data needed.     *)

    VAR L: CARDINAL;
        cipher: Encryption.CipherType;
        mt: MAC.MACType;
        compr1, compr2: Compression.CompressionType;
        keydata, IVdata: ARRAY [0..127] OF CARD8;

    BEGIN
        ClearStrings (port);

        (* There are six options that we have to activate, identified   *)
        (* by codes 'A' to 'F'.                                         *)

        (* A and C: initial IV and key, client to server. *)

        cipher := options^.o2[client_to_server].cipher;
        L := Encryption.IVSize (cipher);
        HashAlgs.GenerateKeyData (options^.kgctx, 'A', IVdata, L);
        L := Encryption.KeySize (cipher);
        HashAlgs.GenerateKeyData (options^.kgctx, 'C', keydata, L);
        NewKeys (port, server, cipher, keydata, IVdata);

        (* Now do the same for the server-to-client direction. *)

        cipher := options^.o2[server_to_client].cipher;
        L := Encryption.IVSize (cipher);
        HashAlgs.GenerateKeyData (options^.kgctx, 'B', IVdata, L);
        L := Encryption.KeySize (cipher);
        HashAlgs.GenerateKeyData (options^.kgctx, 'D', keydata, L);
        NewKeys (port, NOT server, cipher, keydata, IVdata);

        (* MAC keys.    *)

        (* E: integrity key client to server. *)      (* MAC *)

        mt := options^.o2[client_to_server].mac;
        L := MAC.MAClength (mt);
        HashAlgs.GenerateKeyData (options^.kgctx, 'E', keydata, L);
        NewMAC (port, server, mt, keydata);

        (* F: integrity key server to client. *)      (* MAC *)

        mt := options^.o2[server_to_client].mac;
        L := MAC.MAClength (mt);
        HashAlgs.GenerateKeyData (options^.kgctx, 'F', keydata, L);
        NewMAC (port, NOT server, mt, keydata);

        (* The standard is vague on when compression is changed, but this   *)
        (* seems to be the logical place to do it.                          *)

        compr1 := options^.o2[client_to_server].compr;
        compr2 := options^.o2[server_to_client].compr;
        NewCompression (port, compr1, compr2);

        HashAlgs.EndKeyGeneration (options^.kgctx);

    END ActivateNewKeys;

(********************************************************************************)
(*                                END OF SESSION                                *)
(********************************************************************************)

PROCEDURE Disconnect (port: IOPort;  reason: CARDINAL;  message: ARRAY OF CHAR);

    (* Send an SSH_MSG_DISCONNECT message to the peer.  *)

    CONST
        langtag = "en";

    VAR pkt: ByteStr;
        N, k, pktlen: CARDINAL;

    BEGIN
        pktlen := 1 + 4 + 4 + LENGTH(message) + 4 + LENGTH(langtag);
        MakeBS (pkt, pktlen);
        pkt.data^[0] := SSH_MSG_DISCONNECT;
        N := Swap4(reason);
        Copy (ADR(N), ADR(pkt.data^[1]), 4);
        k := 5;
        AddCharString (pkt, k, message);
        AddCharString (pkt, k, langtag);
        EVAL(SendPacket (port, pkt));
    END Disconnect;

(********************************************************************************)

PROCEDURE ClosePort (port: IOPort);

    (* Shuts down this port.  As an acknowledgement, and also in the case of    *)
    (* an involuntary close when communications are lost, an SSH_MSG_TERMINATE  *)
    (* message will be sent to the next layer up.                               *)

    BEGIN
        DeactivatePort (port);

        (* This only begins the shutdown sequence. It will force an exit from   *)
        (* the main loop of RunTransportLayer, after which ClosePacketSession   *)
        (* will be called to finish the job.                                    *)

    END ClosePort;

(********************************************************************************)
(*                               RUN ONE SESSION                                *)
(********************************************************************************)

PROCEDURE RunTransportLayer (port: IOPort;  replyQ: PacketQueues.PacketQ;
                                KeepAlive: Semaphore; ID: TransactionLogID);

    (* Handles a complete SSH session.  All packets are encrypted and           *)
    (* optionally compressed once the encryption and compression algorithms     *)
    (* have been negotiated, but clear data are used up to the end of the       *)
    (* first negotiation.  If later renegotiations happen, the existing         *)
    (* encryption remains in force until a change is accepted by both parties.  *)

    (* This module, and in fact essentially all of the transport layer,     *)
    (* uses common code for both server and client use.  This is to take    *)
    (* advantage of the near-symmetry of the protocol.  Of course this      *)
    (* means some of the code will be unused, depending on whether this     *)
    (* module is linked into server code or client code, but that is a      *)
    (* minor drawback.                                                      *)

    (* In the server case, we must signal on KeepAlive regularly to say     *)
    (* that the server session is still in progress.  In the client case,   *)
    (* that parameter is ignored.                                           *)

    (* The replyQ parameter specifies a queue of incoming packets that will *)
    (* be redirected from the transport layer to higher protocol layers.    *)
    (* When running as a client, the caller must create this queue.  In the *)
    (* server case, this parameter must be NIL, because the queue will be   *)
    (* created internally.                                                  *)

    CONST
        NilQ = CAST(PacketQueues.PacketQ, NIL);
        Nilkgctx = CAST(HashAlgs.KeygenCtx, NIL);

    VAR message, name: ARRAY [0..255] OF CHAR;
        pkt, pkt2: ByteStr;
        options: Options;
        N: CARDINAL;
        opcode: CARD8;
        server, KEXpacketFollows, discardinput: BOOLEAN;
        kxprivate: BN;
        upQ: PacketQueues.PacketQ;

    BEGIN
        InitialGreeting (port);
        NEW (options);

        (* Initialise all options to "none".  *)

        kxprivate := NIL;
        WITH options^ DO
            kex := KeyExchange.none;
            kgctx := NIL;
            pka := PublicKeys.none;
            WITH o2[server_to_client] DO
                cipher := Encryption.none;
                mac    := MAC.none;
                compr :=  Compression.none;
            END (*WITH*);
            WITH o2[client_to_server] DO
                cipher := Encryption.none;
                mac    := MAC.none;
                compr :=  Compression.none;
            END (*WITH*);
        END (*WITH*);
        upQ := replyQ;
        server := upQ = NilQ;

        (* The client should begin by offering an algorithm list. *)

        IF NOT server THEN
            SendAlgorithmList (port, ID);
        END (*IF*);

        LOOP
            IF IsClosing(port) OR NOT ReceivePacket (port, pkt) THEN
                EXIT (*LOOP*);
            END (*IF*);
            IF server THEN
                Signal (KeepAlive);
            END (*IF*);
            discardinput := TRUE;
            opcode := pkt.data^[0];  N := 0;

                (*
                Strings.Assign ("Received packet type ", message);
                AppendCard (opcode, message);
                LogTransaction (ID, message);
                *)

            CASE opcode OF
                SSH_MSG_DISCONNECT:
                    (*
                    message := "SSH_MSG_DISCONNECT received, reason code ";
                    Copy (ADR(pkt.data^[1]), ADR(N), 4);
                    N := Swap4(N);
                    AppendCard (N, message);
                    LogTransaction (ID, message);
                    *)
                    Copy (ADR(pkt.data^[5]), ADR(N), 4);
                    N := Swap4(N);
                    IF N > 0 THEN
                        Copy (ADR(pkt.data^[9]), ADR(message), N);
                        message[N] := CHR(0);
                        LogTransaction (ID, message);
                    END (*IF*);
                    EXIT (*LOOP*);
                |
                SSH_MSG_IGNORE:
                    DiscardBS (pkt);
                |
                SSH_MSG_UNIMPLEMENTED:
                    message := "SSH_MSG_UNIMPLEMENTED received, packet number ";
                    Copy (ADR(pkt.data^[1]), ADR(N), 4);
                    N := Swap4(N);
                    AppendCard (N, message);
                    LogTransaction (ID, message);
                |
                SSH_MSG_DEBUG:
                    message := "SSH_MSG_DEBUG received";
                    IF pkt.data^[1] <> 0 THEN
                        Copy (ADR(pkt.data^[2]), ADR(N), 4);
                        N := Swap4(N);
                        Copy (ADR(pkt.data^[6]), ADR(message), N);
                        LogTransaction (ID, message);
                    END (*IF*);
                |
                SSH_MSG_SERVICE_REQUEST:
                    IF server THEN
                        (* The subsequent packets are probably going to be for a    *)
                        (* higher protocol level, so we need to prepare to start    *)
                        (* sending messages upstairs.                               *)

                        N := 1;
                        GetCharString (pkt, N, name);
                        message := "Service request: ";
                        Strings.Append (name, message);
                        LogTransaction (ID, message);
                        IF upQ = NilQ THEN
                            upQ := PacketQueues.CreateQueue();
                            IF AuthLayer.NewSession (port, upQ, ID) THEN
                                (* OK, no further action needed. *)
                            ELSE
                                (* Will have to think about this. *)
                            END (*IF*);
                        END (*IF*);

                        pkt2 := CopyOfBS (pkt);
                        pkt2.data^[0] := SSH_MSG_SERVICE_ACCEPT;
                        EVAL(SendPacket (port, pkt2));
                    END (*IF*);
                |
                SSH_MSG_SERVICE_ACCEPT:

                    (* Let higher-level client software deal with this. *)
                    (* Note: only a client would be receiving this.     *)

                    discardinput := FALSE;
                    IF upQ = NilQ THEN
                        LogTransactionL (ID, "Have created new queue");
                        upQ := PacketQueues.CreateQueue();
                    END (*IF*);
                    PacketQueues.PutPacket (upQ, pkt);
                    LogTransactionL (ID, "Have queued SERVICE_ACCEPT message");

                |
                SSH_MSG_KEXINIT:
                    (* The peer has sent us lists of algorithms, and we have to *)
                    (* reach agreement on which ones to use.                    *)

                    SetInitLock (port, TRUE);
                    IF server THEN
                        SetString (port, 3, pkt, TRUE);
                        GetAlgorithmList (options, pkt, ID, KEXpacketFollows);
                        SendAlgorithmReply (port, options, ID);
                    ELSE
                        SetString (port, 4, pkt, TRUE);
                        GetAlgorithmList (options, pkt, ID, KEXpacketFollows);
                        kxprivate := KeyExchange.ClientStartKeyExchange (port,
                                                  options^.kex, options^.pka, ID);
                    END (*IF*);
                |
                SSH_MSG_NEWKEYS:
                    (* Both parties to the communication issue SSH_MSG_NEWKEYS at   *)
                    (* approximately the same time, and this code is common to      *)
                    (* both client and server.                                      *)

                    ActivateNewKeys (port, server, options);
                    LogTransactionL (ID, "Have activated the new keys");
                    SetInitLock (port, FALSE);

                    (* Important point: since a call to ReceivePacket happens only  *)
                    (* in this main loop, we will not see an incoming packet with   *)
                    (* the new encryption options until we have passed this point,  *)
                    (* i.e. until ActivateNewKeys has completed execution.          *)

                |
                SSH_MSG_KEXDH_INIT:
                    (* The client initiates the Diffie-Helman key exchange. *)

                    IF server THEN
                        LogTransactionL (ID, "Client has started Diffie-Helman exchange");
                        options^.kgctx := KeyExchange.SrvDoKeyExchange (port, options^.kex,
                                                        options^.pka, pkt);

                        (* Procedure StartKeySwitch simply sends a NEWKEYS message to   *)
                        (* the peer (so to the client in this case). At this point we   *)
                        (* have saved the K and H values needed to calculate the new    *)
                        (* keys, but we will not do the calculation until the client    *)
                        (* sends its own NEWKEYS message.                               *)

                        StartKeySwitch (port);
                    END (*IF*);
                |
                SSH_MSG_KEXDH_REPLY:
                    IF NOT server THEN
                        options^.kgctx := KeyExchange.ClientKxphase2
                                (port, ID, pkt, options^.kex, options^.pka, kxprivate);
                        IF options^.kgctx <> Nilkgctx THEN
                            StartKeySwitch (port);
                        ELSE
                            (* Abandon the session. *)
                            SetInitLock (port, FALSE);
                            Disconnect (port, SSH_DISCONNECT_KEY_EXCHANGE_FAILED,
                                                "Key exchanged failed");
                            LogTransactionL (ID, "Key exchange failed");
                            EXIT (*LOOP*);
                        END (*IF*);
                    END (*IF*);
                |
                50..255:
                    IF upQ = NilQ THEN
                        Strings.Assign ("Unexpected packet type ", message);
                        AppendCard (opcode, message);
                        LogTransaction (ID, message);
                        MakeBS (pkt2, 5);
                        pkt2.data^[0] := SSH_MSG_UNIMPLEMENTED;
                        N := Swap4(LastSeqNo(port));
                        Copy (ADR(N), ADR(pkt2.data^[1]), 4);
                        EVAL(SendPacket (port, pkt2));
                    ELSE
                        (*
                        Strings.Assign ("Processing packet type ", message);
                        AppendCard (opcode, message);
                        LogTransaction (ID, message);
                        *)
                        discardinput := FALSE;
                        PacketQueues.PutPacket (upQ, pkt);
                    END (*IF*);

            ELSE
                Strings.Assign ("Cannot currently handle packet type ", message);
                AppendCard (opcode, message);
                LogTransaction (ID, message);
            END (*CASE*);
            IF discardinput THEN
                DiscardBS (pkt);
            END (*IF*);
        END (*LOOP*);

        (* In terminating this layer we should also terminate the threads that  *)
        (* have been created in upper layers.  The simplest way to do this is   *)
        (* to send a message upwards.                                           *)

        IF upQ <> NilQ THEN
            MakeBS (pkt, 1);
            pkt.data^[0] := SSH_MSG_TERMINATE;
            PacketQueues.PutPacket (upQ, pkt);
        END (*IF*);

        IF server THEN
            LogTransactionL (ID, "End of SSH session");
        END (*IF*);
        BigNum.Discard (kxprivate);
        IF options <> NIL THEN
            IF options^.kgctx <> Nilkgctx THEN
                HashAlgs.EndKeyGeneration (options^.kgctx);
            END (*IF*);
            DISPOSE (options);
        END (*IF*);
        ClosePacketSession (port);

    END RunTransportLayer;

(********************************************************************************)

END TransportLayer.

