    (**************************************************************************)
(*                                                                        *)
(*  Decoding an ASN.1 coded file                                          *)
(*  Copyright (C) 2018   Peter Moylan                                     *)
(*                                                                        *)
(*  This program is free software: you can redistribute it and/or modify  *)
(*  it under the terms of the GNU General Public License as published by  *)
(*  the Free Software Foundation, either version 3 of the License, or     *)
(*  (at your option) any later version.                                   *)
(*                                                                        *)
(*  This program is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU General Public License for more details.                          *)
(*                                                                        *)
(*  You should have received a copy of the GNU General Public License     *)
(*  along with this program.  If not, see <http://www.gnu.org/licenses/>. *)
(*                                                                        *)
(*  To contact author:   http://www.pmoylan.org   peter@pmoylan.org       *)
(*                                                                        *)
(**************************************************************************)

MODULE DumpKey;

        (********************************************************)
        (*                                                      *)
        (*          Extracting key data from a file             *)
        (*                                                      *)
        (*  Programmer:         P. Moylan                       *)
        (*  Started:            3 April 2021                    *)
        (*  Last edited:        3 April 2021                    *)
        (*  Status:             Working                         *)
        (*                                                      *)
        (********************************************************)


IMPORT Strings, TextIO, IOChan, BigNum, Base64;

FROM SYSTEM IMPORT CARD8;

FROM ASN1 IMPORT
    (* proc *)  ASN1toTxt;

FROM VarStrings IMPORT
    (* type *)  VarStringPtr,
    (* proc *)  WriteVarString;

FROM BigNum IMPORT BN;

FROM ProgramArgs IMPORT
    (* proc *)  ArgChan, IsArgPresent;

FROM FileOps IMPORT
    (* type *)  ChanId, FilePos,
    (* proc *)  OpenOldFile, OpenNewFile, CloseFile, Exists, DeleteFile,
                ReadLine, ReadRaw, WriteRaw, CurrentPosition, SetPosition;

FROM STextIO IMPORT
    (* proc *)  WriteChar, WriteString, WriteLn;

FROM LowLevel IMPORT
    (* proc *)  IAND, IANDB;

(************************************************************************)

TYPE FilenameString = ARRAY [0..511] OF CHAR;

CONST
    Nul = CHR(0);
    CtrlZ = CHR(26);
    DefaultFile = ".ssh\id_rsa";

(************************************************************************)
(*                    MISCELLANEOUS USEFUL PROCEDURES                   *)
(************************************************************************)

PROCEDURE WriteCard (N: CARDINAL);

    (* Write N in decimal.  *)

    BEGIN
        IF N > 9 THEN
            WriteCard (N DIV 10);
            N := N MOD 10;
        END (*IF*);
        WriteChar (CHR(ORD('0') + N));
    END WriteCard;

(************************************************************************)
(*                                INPUT                                 *)
(************************************************************************)

PROCEDURE GetParameter (VAR (*OUT*) result: ARRAY OF CHAR);

    VAR args: IOChan.ChanId;

    BEGIN
        args := ArgChan();
        IF IsArgPresent() THEN
            TextIO.ReadString (args, result);
        ELSE
            (*result[0] := Nul;*)
            Strings.Assign (DefaultFile, result);
        END (*IF*);
    END GetParameter;

(************************************************************************)
(*                        TRANSLATE PEM TO BINARY                       *)
(************************************************************************)

PROCEDURE ReadPEM (VAR (*IN*) filename: FilenameString;
                                VAR (*OUT*) buffer: ARRAY OF CHAR);

    (* Loads a Base64 string from a file, deleting any line that    *)
    (* begins with the three characters '---'.                      *)

    VAR line: ARRAY [0..32767] OF CHAR;
        cid: ChanId;
        k: CARDINAL;
        found: BOOLEAN;
        fpos: FilePos;

    BEGIN
        buffer[0] := Nul;
        cid := OpenOldFile (filename, FALSE, FALSE);

        (* Skip past first blank line.  If no blank line then go back   *)
        (* to the most plausible starting point.                        *)

        fpos := CurrentPosition (cid);
        ReadLine (cid, line);
        Strings.FindNext ("---", line, 0, found, k);
        IF found THEN
            fpos := CurrentPosition (cid);
        END (*IF*);
        LOOP
            ReadLine (cid, line);
            IF line[0] = Nul THEN
                EXIT (*LOOP*);
            ELSIF line[0] = CtrlZ THEN
                SetPosition (cid, fpos);
                EXIT (*LOOP*);
            END (*IF*);
        END (*LOOP*);

        (* Now the real data. *)

        LOOP
            ReadLine (cid, line);
            IF line[0] = CtrlZ THEN
                EXIT (*LOOP*);
            ELSIF (line[0] <> '-') AND (line[1] <> '-') AND (line[2] <> '-') THEN
                Strings.Append(line, buffer);
            END (*IF*);
        END (*LOOP*);
        CloseFile (cid);

        WriteCard (Strings.Length(buffer));
        WriteString (" base64 characters");  WriteLn;

    END ReadPEM;

(************************************************************************)

PROCEDURE PEMtoBinary (VAR (*IN*) infilename: FilenameString;
                        VAR (*OUT*) bindata: ARRAY OF CARD8): CARDINAL;

    (* Reads a PEM file, translates the data to binary in bindata.      *)
    (* Returns the number of output bytes.                              *)

    CONST
        InBufSize = 16384;

    VAR inbuf: ARRAY [0..InBufSize-1] OF CHAR;
        N: CARDINAL;

    BEGIN
        ReadPEM (infilename, inbuf);
        N := Base64.Decode (inbuf, bindata);
        RETURN N;
    END PEMtoBinary;

(************************************************************************)
(*                              MAIN PROGRAM                            *)
(************************************************************************)

PROCEDURE DoDecode;

    CONST BufSize = 16384;

    VAR infilename: FilenameString;
        binbuff: ARRAY [0..BufSize-1] OF CARD8;
        txtptr: VarStringPtr;
        M, N: CARDINAL;

    BEGIN
        (* Get the input file name. *)

        GetParameter (infilename);
        N := Strings.Length (infilename);
        WHILE (N > 0) AND (infilename[N-1] = ' ') DO
            DEC(N);
        END (*WHILE*);
        infilename[N] := Nul;
        IF infilename[0] = Nul THEN
            WriteString ("No input file was specified.");  WriteLn;
            RETURN;
        END (*IF*);

        (* Do the decoding.  *)

        WriteString ("Reading from file ");
        WriteString (infilename);  WriteLn;
        N := PEMtoBinary (infilename, binbuff);
        IF N = 0 THEN
            WriteString ("Could not load input file ");
            WriteString (infilename);
            WriteLn;
        ELSE
            WriteCard (N);  WriteString (" bytes total");  WriteLn;
            M := ASN1toTxt (binbuff, N, txtptr);
            WriteVarString (txtptr^, M);
        END (*IF*);

    END DoDecode;

(************************************************************************)

BEGIN
    DoDecode;
END DumpKey.

