    (**************************************************************************)
(*                                                                        *)
(*  Decoding an ASN.1 coded file                                          *)
(*  Copyright (C) 2021   Peter Moylan                                     *)
(*                                                                        *)
(*  This program is free software: you can redistribute it and/or modify  *)
(*  it under the terms of the GNU General Public License as published by  *)
(*  the Free Software Foundation, either version 3 of the License, or     *)
(*  (at your option) any later version.                                   *)
(*                                                                        *)
(*  This program is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU General Public License for more details.                          *)
(*                                                                        *)
(*  You should have received a copy of the GNU General Public License     *)
(*  along with this program.  If not, see <http://www.gnu.org/licenses/>. *)
(*                                                                        *)
(*  To contact author:   http://www.pmoylan.org   peter@pmoylan.org       *)
(*                                                                        *)
(**************************************************************************)

MODULE DumpPublicKey;

        (********************************************************)
        (*                                                      *)
        (*          Extracting key data from a file             *)
        (*                                                      *)
        (*  Programmer:         P. Moylan                       *)
        (*  Started:            6 May 2021                      *)
        (*  Last edited:        27 July 2021                    *)
        (*  Status:             Working                         *)
        (*                                                      *)
        (********************************************************)


FROM SYSTEM IMPORT CARD8, CARD32, ADR;

IMPORT IOChan, TextIO, Strings, Base64;

FROM STextIO IMPORT
    (* proc *)  WriteChar, WriteString, WriteLn;

FROM ProgramArgs IMPORT
    (* proc *)  ArgChan, IsArgPresent;

FROM FileOps IMPORT
    (* const*)  NoSuchChannel,
    (* type *)  ChanId,
    (* proc *)  OpenOldFile, ReadRaw, CloseFile;

FROM LowLevel IMPORT
    (* proc *)  SwapIt, EVAL, Copy;

(************************************************************************)

TYPE FilenameString = ARRAY [0..511] OF CHAR;

CONST
    Nul = CHR(0);
    CtrlZ = CHR(26);
    DefaultFile = ".ssh\id_rsa.pub";

(************************************************************************)
(*                    MISCELLANEOUS USEFUL PROCEDURES                   *)
(************************************************************************)

PROCEDURE WriteCard (N: CARDINAL);

    (* Writes N in decimal. *)

    BEGIN
        IF N > 9 THEN
            WriteCard (N DIV 10);
            N := N MOD 10;
        END (*IF*);
        WriteChar (CHR(ORD('0') + N));
    END WriteCard;

(************************************************************************)

PROCEDURE WriteHex1 (N: CARDINAL);

    (* Writes a one-digit hexadecimal number. *)

    BEGIN
        IF N > 9 THEN
            WriteChar (CHR(ORD('a') + N - 10));
        ELSE
            WriteChar (CHR(ORD('0') + N));
        END (*IF*);
    END WriteHex1;

(********************************************************************************)

PROCEDURE WriteLongHexStr (VAR (*IN*) buffer: ARRAY OF CARD8;  pos, size: CARDINAL);

    (* Writes number in decimal that starts at buffer[pos] and has size bytes. *)

    VAR val: CARD8;  j: CARDINAL;

    BEGIN
        FOR j := 1 TO size DO
            val := buffer[pos];
            WriteHex1 (val DIV 16);
            WriteHex1 (val MOD 16);
            INC (pos);
        END (*FOR*);
    END WriteLongHexStr;

(********************************************************************************)

PROCEDURE Swap4 (val: CARD32): CARD32;

    (* Returns the argument value in byte-reversed order.  This is needed       *)
    (* because network byte order is most significant byte first, whereas our   *)
    (* local host order is least significant byte first.                        *)

    VAR temp: CARD32;

    BEGIN
        temp := val;
        SwapIt (temp);
        RETURN temp;
    END Swap4;

(************************************************************************)
(*                                INPUT                                 *)
(************************************************************************)

PROCEDURE GetParameter (VAR (*OUT*) result: ARRAY OF CHAR);

    VAR args: IOChan.ChanId;

    BEGIN
        args := ArgChan();
        IF IsArgPresent() THEN
            TextIO.ReadString (args, result);
        ELSE
            (*result[0] := Nul;*)
            Strings.Assign (DefaultFile, result);
        END (*IF*);
    END GetParameter;

(************************************************************************)
(*                              MAIN PROGRAM                            *)
(************************************************************************)

PROCEDURE DoDecode;

    CONST
        MaxChars = 8192;

    VAR cid: ChanId;
        actual, pos, N, size, j: CARDINAL;
        found, error: BOOLEAN;
        name: ARRAY [0..63] OF CHAR;
        string: ARRAY [0..MaxChars-1] OF CHAR;
        bytes: ARRAY [0..2047] OF CARD8;
        infilename: FilenameString;

    BEGIN
        (* Get the input file name. *)

        GetParameter (infilename);
        N := Strings.Length (infilename);
        WHILE (N > 0) AND (infilename[N-1] = ' ') DO
            DEC(N);
        END (*WHILE*);
        infilename[N] := Nul;
        IF infilename[0] = Nul THEN
            WriteString ("No input file was specified.");  WriteLn;
            RETURN;
        END (*IF*);

        (* Do the decoding.  *)

        WriteString ("Reading from file ");
        WriteString (infilename);  WriteLn;

        cid := OpenOldFile (infilename, FALSE, FALSE);
        IF cid = NoSuchChannel THEN
            WriteString ("ERROR: Cannot open ");
            WriteString (infilename);  WriteLn;
            RETURN;
        END (*IF*);

        (* The content should be one big line. *)

        ReadRaw (cid, string, MaxChars, actual);
        CloseFile (cid);

        (* Leading string should be "ssh-rsa", but for now I'm not checking.  *)

        Strings.FindNext (' ', string, 0, found, pos);
        error := NOT found;
        IF NOT error THEN
            Strings.Delete (string, 0, pos+1);
        END (*IF*);

        (* Third substring is of no interest to us for now.  *)

        Strings.FindNext (' ', string, 0, found, pos);
        IF found THEN
            string[pos] := CHR(0);
        END (*IF*);

        N := Base64.Decode (string, bytes);
        WriteCard (N);  WriteString (" total bytes");  WriteLn;

        (* Array 'bytes' has internal structure.  *)

        size := 0;

        (* Name. *)

        Copy (ADR(bytes[0]), ADR(size), 4);
        size := Swap4 (size);
        Copy (ADR(bytes[4]), ADR(name), size);
        j := size + 4;
        WriteString (name);  WriteLn;

        (* Exponent (e). *)

        WriteString ("e=");
        Copy (ADR(bytes[j]), ADR(size), 4);  INC (j, 4);
        size := Swap4 (size);
        WriteLongHexStr (bytes, j, size);  WriteLn;
        INC (j, size);

        (* Modulus (n). *)

        WriteString ("n=");
        Copy (ADR(bytes[j]), ADR(size), 4);  INC (j, 4);
        size := Swap4 (size);
        WriteLongHexStr (bytes, j, size);  WriteLn;

        (* Tell user the key size. *)

        WriteString ("Key size = ");  WriteCard (size);
        WriteString (" bytes = ");  WriteCard (8*size);
        WriteString (" bits");  WriteLn;

    END DoDecode;

(************************************************************************)

BEGIN
    DoDecode;
END DumpPublicKey.

