(**************************************************************************)
(*                                                                        *)
(*  My own version of SFTP (over SSH)                                     *)
(*  Copyright (C) 2022   Peter Moylan                                     *)
(*                                                                        *)
(*  This program is free software: you can redistribute it and/or modify  *)
(*  it under the terms of the GNU General Public License as published by  *)
(*  the Free Software Foundation, either version 3 of the License, or     *)
(*  (at your option) any later version.                                   *)
(*                                                                        *)
(*  This program is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU General Public License for more details.                          *)
(*                                                                        *)
(*  You should have received a copy of the GNU General Public License     *)
(*  along with this program.  If not, see <http://www.gnu.org/licenses/>. *)
(*                                                                        *)
(*  To contact author:   http://www.pmoylan.org   peter@pmoylan.org       *)
(*                                                                        *)
(**************************************************************************)

MODULE sftpd;

        (********************************************************)
        (*                                                      *)
        (*                 My version of SFTP                   *)
        (*                                                      *)
        (*  Programmer:         P. Moylan                       *)
        (*  Started:            2 January 2021                  *)
        (*  Last edited:        27 September 2022               *)
        (*  Status:             Working                         *)
        (*                                                      *)
        (********************************************************)


IMPORT OS2, Strings, SFV, INIData, STextIO;

FROM SYSTEM IMPORT CAST, LOC;

FROM sshSession IMPORT
    (* proc *)  NewSession, SetSameIPLimit, SetMaxUsers, SetTimeout, SetTransactionLogging;

FROM PacketIO IMPORT
    (* type *)  IOPort;

FROM FtpdINI IMPORT
    (* proc *)  SetINIFileName, SetHashMax, OpenINIFile, CloseINIFile,
                GetINIFileName, INIFileExists;

FROM IPFilter IMPORT
    (* proc *)  UpdateMasterHostFilter, CheckMasterIPFilter;

FROM MiscFuncs IMPORT
    (* proc *)  AppendCard;

FROM Inet2Misc IMPORT
    (* proc *)  Swap2, WaitForSocket, IPToString;

FROM Internet IMPORT
    (* const*)  Zero8, INADDR_ANY,
    (* proc *)  inet_addr;

FROM Sockets IMPORT
    (* const*)  NotASocket,
    (* type *)  Socket, SockAddr, AddressFamily, SocketType,
    (* proc *)  sock_init, socket, sock_errno, psock_errno, setsockopt,
                bind, listen, accept, soclose;

FROM Exceptq IMPORT
    (* proc *)  InstallExceptq, UninstallExceptq;

FROM Names IMPORT
    (* type *)  HostName;

FROM ProgName IMPORT
    (* proc *)  GetProgramName, WeAreServer;

FROM LogCtx IMPORT
    (* var  *)  sshctx;

FROM TransLog IMPORT
    (* type *)  TransactionLogID,
    (* proc *)  (*StartTransactionLogging,*) CreateLogID, DiscardLogID, LogTransaction, LogTransactionL,
                UpdateTopScreenLine;

FROM STextIO IMPORT
    (* proc *)  WriteChar, WriteString, WriteLn;

FROM SplitScreen IMPORT
    (* proc *)  NotDetached;

FROM Timer IMPORT
    (* proc *)  TimedWait, Sleep;

FROM Semaphores IMPORT
    (* type *)  Semaphore,
    (* proc *)  CreateSemaphore, Wait, Signal;

FROM TaskControl IMPORT
    (* proc *)  CreateTask;

FROM CtrlC IMPORT
    (* proc *)  SetBreakHandler, IsCtrlCException;

FROM OS2Sem IMPORT
    (* type *)  SemKind,
    (* proc *)  WaitOnSemaphore;

FROM Names IMPORT
    (* type *)  FilenameString;

FROM LowLevel IMPORT
    (* proc *)  EVAL, IOR, IAND;

(************************************************************************)

CONST
    Nul = CHR(0);
    DefaultPort = 22;
    DefaultMaxUsers = 10;
    DefaultTimeout = 900;               (* seconds   *)
    DefaultLogLevel = 32;
    DefaultGuestLimit = DefaultMaxUsers-1;

VAR
    (* Main socket, server port. *)

    MainSocket: Socket;
    ServerPort, BindAddr: CARDINAL;

    (* Our INI file name. *)

    ININame: ARRAY [0..31] OF CHAR;
    UseTNI: BOOLEAN;

    (* Flag to say that the user has typed Ctrl/C. *)

    ShutdownInProgress: BOOLEAN;

    (* Semaphore that we use to start the shutdown process. *)

    CommenceShutdown: Semaphore;

    (* Termination semaphore used in shutdown processing. *)

    TaskDone: Semaphore;

    (* Event flag by which another process can request us to shut down. *)

    ShutdownSignal: OS2.HEV;

    (* Event semaphore to trigger updater task. *)

    UpdaterFlag: OS2.HEV;

    (* Flag to say that we're not running detached. *)

    ScreenEnabled: BOOLEAN;

(************************************************************************)
(*                          SHUTDOWN OPERATIONS                         *)
(************************************************************************)

PROCEDURE ShutterDowner;

    (* A task whose job is to shut down the program. *)

    (* Unsolved mystery: I could not make Ctrl/C shut down the program      *)
    (* reliably until I included the Sleep() call.  Experiments suggests    *)
    (* that the actual sleep duration is not critical.                      *)
    (* I should take a closer look at my TimedWait implementation.          *)

    (*VAR LogID: TransactionLogID;*)

    BEGIN
        (*LogID := CreateLogID (sshctx, "       ");*)
        Wait (CommenceShutdown);
        ShutdownInProgress := TRUE;
        Sleep (100);
        Signal (TaskDone);
    END ShutterDowner;

(************************************************************************)

PROCEDURE ["C"] ControlCHandler(): BOOLEAN;

    (* Intercepts a Ctrl/C from the keyboard. *)

    BEGIN
        Signal (CommenceShutdown);
        RETURN TRUE;
    END ControlCHandler;

(************************************************************************)

PROCEDURE ExternalShutdownRequestDetector;

    (* Runs as a separate task.  Responds to the public event semaphore that    *)
    (* requests a shutdown of this program.                                     *)

    CONST semName = "\SEM32\SFTPD\SHUTDOWN";

    BEGIN
        ShutdownSignal := 0;
        IF OS2.DosOpenEventSem (semName, ShutdownSignal) = OS2.ERROR_SEM_NOT_FOUND THEN
            OS2.DosCreateEventSem (semName, ShutdownSignal, OS2.DC_SEM_SHARED, FALSE);
        END (*IF*);

        WHILE NOT ShutdownInProgress DO
            WaitOnSemaphore (event, ShutdownSignal);
            Signal (CommenceShutdown);
        END (*WHILE*);

        OS2.DosCloseEventSem(ShutdownSignal);
        Signal (TaskDone);

    END ExternalShutdownRequestDetector;

(************************************************************************)
(*                          LOADING INI DATA                            *)
(************************************************************************)

PROCEDURE LoadUpdateableINIData;

    (* Loads configuration parameters from the INI file.  These are not *)
    (* all of the parameters; only those that may be modified while the *)
    (* server is running.  The parameters that can be set only at       *)
    (* program startup are set by LoadINIData (see below).  The present *)
    (* procedure is called by LoadINIData, but it may also be called    *)
    (* at other times.                                                  *)

    VAR hini: INIData.HINI;
        SameIPLimit, MaxUsers, TimeoutLimit, TransLevel: CARDINAL;
        TransLogName: FilenameString;
        SyslogHost: HostName;
        SYSapp: ARRAY [0..4] OF CHAR;

    (********************************************************************)

    PROCEDURE GetItem (name: ARRAY OF CHAR;
                            VAR (*OUT*) variable: ARRAY OF LOC): BOOLEAN;

        BEGIN
            RETURN INIData.INIGet (hini, SYSapp, name, variable);
        END GetItem;

    (********************************************************************)

    BEGIN
        SYSapp := "$SYS";
        MaxUsers := DefaultMaxUsers;
        TimeoutLimit := DefaultTimeout;
        TransLevel := 0;
        TransLogName := "SFTPTRANS.LOG";
        SyslogHost := "";

        hini := OpenINIFile();
        EVAL(GetItem ("SameIPLimitS", SameIPLimit));
        SetSameIPLimit (SameIPLimit);
        EVAL(GetItem ("MaxUsers", MaxUsers));
        SetMaxUsers (MaxUsers);

        EVAL(GetItem ("TimeOut", TimeoutLimit));
        SetTimeout (TimeoutLimit);

        EVAL(GetItem ("TransLevel", TransLevel));
        IF NOT INIData.INIGetString (hini, SYSapp, "TransLogNameS", TransLogName)
                   OR (TransLogName[0] = CHR(0)) THEN
            TransLogName := "SFTPTRANS.LOG";
        END (*IF*);
        EVAL (INIData.INIGetString (hini, SYSapp, "SyslogHost", SyslogHost));

        UpdateMasterHostFilter (hini);

        CloseINIFile (hini);

        (* TransLevel is actually an array of eight bits:               *)
        (*    bit 0          transaction log to disk   (obsolete)       *)
        (*    bit 1          transaction log to screen (obsolete)       *)
        (*    bit 2          enable user transfer logging               *)
        (*    bit 3          enable common.log transfer logging         *)
        (*    bit 4          transaction log to disk                    *)
        (*    bit 5          transaction log to screen                  *)
        (*    bit 6          transaction log to pipe                    *)
        (*    bit 7          transaction log to syslog                  *)
        (* That is, TransLevel DIV 16 controls the transaction logging, *)
        (* while TransLevel MOD 16 DIV 4 controls the transfer logging. *)
        (* Note that these bits can be overridden by CmdLevel.          *)
        (* The redundant bits 0 and 1 are a transition arrangement;     *)
        (* originally bits 0 and 1 controlled transaction logging, so   *)
        (* we have to take the OR of two ways to specify the disk and   *)
        (* screen options.                                              *)

        TransLevel := IOR (TransLevel DIV 16, TransLevel MOD 4);
        IF NOT ScreenEnabled THEN
            TransLevel := IAND (TransLevel, 13);
        END (*IF*);
        SetTransactionLogging (TransLevel, TransLogName, SyslogHost);

    END LoadUpdateableINIData;

(************************************************************************)

PROCEDURE LoadINIData;

    (* Sets the INI file name, and loads setup parameters. *)

    VAR hini: INIData.HINI;
        SYSapp: ARRAY [0..4] OF CHAR;

    (********************************************************************)

    PROCEDURE GetItem (name: ARRAY OF CHAR;
                            VAR (*OUT*) variable: ARRAY OF LOC): BOOLEAN;

        BEGIN
            RETURN INIData.INIGet (hini, SYSapp, name, variable);
        END GetItem;

    (********************************************************************)

    BEGIN
        IF NOT INIData.ChooseDefaultINI("Ftpd", UseTNI) THEN
            UseTNI := TRUE;
        END (*IF*);
        IF UseTNI THEN
            ININame := "FTPD.TNI";
        ELSE
            ININame := "FTPD.INI";
        END (*IF*);
        SetINIFileName (ININame);
        SYSapp := "$SYS";

        hini := OpenINIFile();
        IF INIData.INIValid(hini) THEN
            EVAL(GetItem ("SFTPPort", ServerPort));
            EVAL(GetItem ("BindAddrS", BindAddr));
            CloseINIFile (hini);
        ELSE
            WriteString ("Can't open ");
            IF UseTNI THEN
                WriteChar ('T');
            ELSE
                WriteChar ('I');
            END (*IF*);
            WriteString ("NI file");
            WriteLn;
            HALT;
        END (*IF*);
        LoadUpdateableINIData;
    END LoadINIData;

(********************************************************************************)
(*                    TASK TO CATCH UPDATES TO THE INI DATA                     *)
(********************************************************************************)

PROCEDURE INIChangeDetector;

    (* Runs as a separate task.  Rereads some of the configuration data each    *)
    (* time a public event semaphore tells us that there's been a change.       *)

    CONST semName = "\SEM32\FTPSERVER\UPDATED";

    BEGIN
        UpdaterFlag := 0;
        IF OS2.DosOpenEventSem (semName, UpdaterFlag) = OS2.ERROR_SEM_NOT_FOUND THEN
            OS2.DosCreateEventSem (semName, UpdaterFlag, OS2.DC_SEM_SHARED, FALSE);
        END (*IF*);

        WHILE NOT ShutdownInProgress DO
            WaitOnSemaphore (event, UpdaterFlag);
            IF NOT ShutdownInProgress THEN
                LoadUpdateableINIData;
            END (*IF*);
        END (*WHILE*);

        OS2.DosCloseEventSem(UpdaterFlag);
        Signal (TaskDone);

    END INIChangeDetector;

(************************************************************************)
(*                           THE MAIN SERVER CODE                       *)
(************************************************************************)

PROCEDURE WriteError (LogID: TransactionLogID);

    VAR LogLine: ARRAY [0..255] OF CHAR;

    BEGIN
        Strings.Assign ("Socket error ", LogLine);
        AppendCard (sock_errno(), LogLine);
        LogTransaction (LogID, LogLine);
    END WriteError;

(************************************************************************)

PROCEDURE RunTheServer;

    (* Keeps accepting connection requests until shutdown is requested. *)

    CONST NilSession = CAST (IOPort, NIL);

    VAR
        LogID: TransactionLogID;
        temp: CARDINAL;
        ns: Socket;  myaddr, client: SockAddr;
        ExceptqActive: BOOLEAN;
        StartupSuccessful: BOOLEAN;
        optval: CARDINAL;
        exRegRec: OS2.EXCEPTIONREGISTRATIONRECORD;
        message: ARRAY [0..127] OF CHAR;
        addr: ARRAY [0..16] OF CHAR;
        ProgVersion: ARRAY [0..31] OF CHAR;

    BEGIN
        (* Initial setup and opening log messages. *)

        LogID := CreateLogID (sshctx, "       ");
        IF sock_init() <> 0 THEN
            LogTransactionL (LogID, "No network.");
            DiscardLogID (LogID);
            RETURN;
        END (*IF*);

        ExceptqActive := InstallExceptq (exRegRec);
        IF NOT ExceptqActive THEN
            LogTransactionL (LogID, "Failed to load the Exceptq handler.");
        END (*IF*);

        message := "Getting configuration data from ";
        Strings.Append (ININame, message);
        LogTransaction (LogID, message);

        message := "[I] ";
        message[1] := CAP(ININame[LENGTH(ININame)-3]);
        GetProgramName (ProgVersion);
        Strings.Append (ProgVersion, message);
        LogTransaction (LogID, message);

        IF ScreenEnabled THEN
            Strings.Append ("            ", message);
            UpdateTopScreenLine (0, message);
            Strings.Assign ("(C) 2021-", message);
            Strings.Append (SFV.year(), message);
            Strings.Append (" Peter Moylan", message);
            UpdateTopScreenLine (25, message);
            UpdateTopScreenLine (62, "Users: 0 ");

            EVAL (SetBreakHandler (ControlCHandler));
        END (*IF*);

        message := "exceptq support is ";
        IF ExceptqActive THEN
            Strings.Append ("present", message);
        ELSE
            Strings.Append ("absent", message);
        END (*IF*);
        LogTransaction (LogID, message);

        (* Now start up the server. *)

        StartupSuccessful := FALSE;

        MainSocket := socket (AF_INET, SOCK_STREAM, AF_UNSPEC);

        (* Allow reuse of the port we're binding to. *)

        optval := 1;
        setsockopt (MainSocket, 0FFFFH, 4, optval, SIZE(optval));

        Strings.Assign ("listening on ", message);
        IF BindAddr = 0 THEN
            Strings.Append ("all interfaces", message);
        ELSE
            IPToString (BindAddr, TRUE, addr);
            Strings.Append (addr, message);
        END (*IF*);
        Strings.Append (", port ", message);
        AppendCard (ServerPort, message);
        LogTransaction (LogID, message);

        (* Now have the socket, bind to our machine. *)

        WITH myaddr DO
            family := AF_INET;
            WITH in_addr DO
                port := Swap2 (ServerPort);
                addr := BindAddr;
                zero := Zero8;
            END (*WITH*);
        END (*WITH*);

        IF bind (MainSocket, myaddr, SIZE(myaddr)) THEN

            WriteError (LogID);
            LogTransactionL (LogID, "Cannot bind to server port.");

        ELSE

            (* Go into listening mode. *)

            IF listen (MainSocket, 5) THEN
                WriteError (LogID);
            ELSE
                StartupSuccessful := TRUE;
            END (*IF*);

        END (*IF bind*);

        IF StartupSuccessful THEN

            Strings.Assign (ProgVersion, message);
            Strings.Append (" started.", message);
            LogTransaction (LogID, message);

            (* HERE'S THE MAIN SERVICE LOOP. *)

            WHILE WaitForSocket (MainSocket, MAX(CARDINAL)) > 0 DO
                temp := SIZE(client);
                ns := accept (MainSocket, client, temp);
                IF ns <> NotASocket THEN

                    (* Start the session that talks to this client. *)

                    EVAL (NewSession (ns, client));

                END (*IF*);
            END (*WHILE*);

            (* Close the main socket. *)

            IF soclose(MainSocket) THEN
                psock_errno ("");
            END (*IF*);

        END (*IF*);

        (* End of operation, shut down the server. *)

        Signal (CommenceShutdown);

        LogTransactionL (LogID, "sftpd closing down");
        DiscardLogID (LogID);

        UninstallExceptq (exRegRec);
        Signal (TaskDone);

    END RunTheServer;

(********************************************************************************)
(*                                 MAIN PROGRAM                                 *)
(********************************************************************************)

VAR TransLevel: CARDINAL;
    TransLogName: FilenameString;

BEGIN
    WeAreServer (TRUE);
    ScreenEnabled := NotDetached();
    ShutdownInProgress := FALSE;
    UseTNI := TRUE;
    ServerPort := DefaultPort;
    TransLogName := "SFTPD.LOG";
    TransLevel := 2;
    CreateSemaphore (TaskDone, 0);
    CreateSemaphore (CommenceShutdown, 0);
    EVAL (SetBreakHandler (ControlCHandler));
    EVAL (CreateTask (ShutterDowner, 2, "shutterdowner"));
    EVAL (CreateTask (ExternalShutdownRequestDetector, 2, "shutdown"));
    EVAL (CreateTask (INIChangeDetector, 2, "update"));
    LoadINIData;

    RunTheServer;

FINALLY
    OS2.DosPostEventSem (ShutdownSignal);       (* external shutdown detector *)
    Wait (TaskDone);
    Wait (TaskDone);
    Wait (TaskDone);
END sftpd.

