(**************************************************************************)
(*                                                                        *)
(*  SSH sessions that run on the SSH Connection Protocol                  *)
(*  Copyright (C) 2022   Peter Moylan                                     *)
(*                                                                        *)
(*  This program is free software: you can redistribute it and/or modify  *)
(*  it under the terms of the GNU General Public License as published by  *)
(*  the Free Software Foundation, either version 3 of the License, or     *)
(*  (at your option) any later version.                                   *)
(*                                                                        *)
(*  This program is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU General Public License for more details.                          *)
(*                                                                        *)
(*  You should have received a copy of the GNU General Public License     *)
(*  along with this program.  If not, see <http://www.gnu.org/licenses/>. *)
(*                                                                        *)
(*  To contact author:   http://www.pmoylan.org   peter@pmoylan.org       *)
(*                                                                        *)
(**************************************************************************)

IMPLEMENTATION MODULE sshApps;

        (********************************************************)
        (*                                                      *)
        (*              ssh user-level applications             *)
        (*                                                      *)
        (*  Programmer:         P. Moylan                       *)
        (*  Started:            7 November 2021                 *)
        (*  Last edited:        24 March 2022                   *)
        (*  Status:             Possibly redundant module       *)
        (*                                                      *)
        (********************************************************)


IMPORT Strings;

FROM PacketIO IMPORT
    (* type *)  IOPort;

FROM STextIO IMPORT
    (* proc *)  WriteString, WriteLn;

FROM VarStrings IMPORT
    (* type *)  ByteStr,
    (* proc *)  DiscardBS, GetCharString, GetCard;

FROM TransLog IMPORT
    (* type *)  TransactionLogID,
    (* proc *)  LogTransaction, LogTransactionL;

FROM MiscFuncs IMPORT
    (* proc *)  AppendCard;

FROM Storage IMPORT
    (* proc *)  ALLOCATE, DEALLOCATE;

(************************************************************************)

TYPE
    SessionData = POINTER TO
                    RECORD
                        dummy: CARDINAL;
                    END (*RECORD*);

(************************************************************************)
(*                 CHECKING FOR SUPPORTED SESSION TYPE                  *)
(************************************************************************)

(*
PROCEDURE EstablishSession (): SessionData;

    (* Establishes a new session.  Returns NIL if unsupported type. *)

    VAR sessinfo: SessionData;

    BEGIN
        NEW (sessinfo);
        RETURN sessinfo;
    END EstablishSession;
*)
(************************************************************************)


(************************************************************************)
(*                          CHANNEL REQUEST                             *)
(************************************************************************)

TYPE ptydatatype =
            RECORD
                term: ARRAY [0..31] OF CHAR;
                wchar, hchar: CARDINAL;
                wpix, hpix: CARDINAL;
                modes: ARRAY [0..127] OF CHAR;
            END (*RECORD*);

PROCEDURE GetPtySpec (pkt: ByteStr;  VAR (*INOUT*) pos: CARDINAL;
                                        logID: TransactionLogID): ptydatatype;

    (* Picks up the parameters of a "pty-req" request. *)

    VAR res: ptydatatype;
        message: ARRAY [0..127] OF CHAR;

    BEGIN
        GetCharString (pkt, pos, res.term);
        res.wchar := GetCard (pkt, pos);
        res.hchar := GetCard (pkt, pos);
        res.wpix := GetCard (pkt, pos);
        res.hpix := GetCard (pkt, pos);
        GetCharString (pkt, pos, res.modes);

        Strings.Assign (res.term, message);
        Strings.Append (" terminal size = ", message);  AppendCard (res.wchar, message);
        Strings.Append ("x", message);  AppendCard (res.hchar, message);
        Strings.Append (", or in pixels ", message);  AppendCard (res.wpix, message);
        Strings.Append ("x", message);  AppendCard (res.hpix, message);
        LogTransaction (logID, message);
        message := "  modes = ";  Strings.Append (res.modes, message);
        LogTransaction (logID, message);

        RETURN res;
    END GetPtySpec;

(************************************************************************)

PROCEDURE ParseChannelRequest (VAR (*INOUT*) SD: SessionData;
                                pkt: ByteStr;  port: IOPort;
                                        logID: TransactionLogID): BOOLEAN;

    (* Handles an SSH_MSG_CHANNEL_REQUEST.  On entry the caller has     *)
    (* already read the recipient channel field, so pos is pointing to  *)
    (* the request type in pkt.                                         *)

    VAR pos: CARDINAL;
        success: BOOLEAN;
        reqtype, message: ARRAY [0..63] OF CHAR;
        ptydata: ptydatatype;

    BEGIN
        IF SD = NIL THEN
            NEW (SD);
        END (*IF*);
        pos := 5;
        GetCharString (pkt, pos, reqtype);
        INC(pos);    (* skip the want-reply field *)

        Strings.Assign ("Request for channel, type ", message);
        Strings.Append (reqtype, message);
        LogTransaction (logID, message);
        success := TRUE;       (* Until I know more *)

        (* There may be other parameters depending on request type. *)

        IF Strings.Equal (reqtype, "pty-req") THEN
            ptydata := GetPtySpec (pkt, pos, logID);
        ELSIF Strings.Equal (reqtype, "x11-req") THEN
            success := FALSE;           (* not currently supported *)
        ELSIF Strings.Equal (reqtype, "env") THEN
            success := FALSE;           (* not currently supported *)
        ELSIF Strings.Equal (reqtype, "shell") THEN
            success := FALSE;           (* not currently supported *)
        ELSIF Strings.Equal (reqtype, "exec") THEN
            success := FALSE;           (* not currently supported *)
        ELSIF Strings.Equal (reqtype, "subsystem") THEN
            success := FALSE;           (* not currently supported *)
        END (*IF*);

        DiscardBS (pkt);

        RETURN success;

    END ParseChannelRequest;

(************************************************************************)

PROCEDURE DiscardSessionData (VAR (*INOUT*) SD: SessionData);

    (* Discards SD. *)

    BEGIN
        DISPOSE (SD);
    END DiscardSessionData;

(************************************************************************)

END sshApps.

