IMPLEMENTATION MODULE sshRSA;

        (********************************************************)
        (*                                                      *)
        (*                 ssh-rsa public keys                  *)
        (*                                                      *)
        (*  Programmer:         P. Moylan                       *)
        (*  Started:            9 April 2021                    *)
        (*  Last edited:        26 September 2021               *)
        (*  Status:             Complete, untested              *)
        (*                                                      *)
        (*          Our signature is being rejected, but        *)
        (*          that could also be a fault in H calculation *)
        (*                                                      *)
        (********************************************************)


IMPORT RSAKeys, RSA, Strings, Base64, VarStrings, HashAlgs, BigNum, ASN1;

FROM SYSTEM IMPORT CARD8, ADR, FILL;

FROM BigNum IMPORT
    (* type *)  BN;

FROM VarStrings IMPORT
    (* type *)  ByteStr, VarStringPtr, ByteStringPtr,
    (* proc *)  MakeBS, CopyOfBS, DiscardBS;

FROM FileOps IMPORT
    (* const*)  NoSuchChannel,
    (* type *)  ChanId,
    (* proc *)  OpenOldFile, CloseFile, CreateDir, ReadRaw, ReadLine;

FROM Inet2Misc IMPORT
    (* proc *)  Swap4;

FROM TransLog IMPORT
    (* type *)  TransactionLogID,
    (* proc *)  LogTransactionL;

FROM LowLevel IMPORT
    (* proc *)  Copy, EVAL;

FROM STextIO IMPORT
    (* proc *)  WriteChar, WriteString, WriteLn;

FROM Storage IMPORT
    (* proc *)  ALLOCATE, DEALLOCATE;

IMPORT Debug, MiscFuncs;         (* while debugging *)

(************************************************************************)

CONST
    keydir = ".ssh";

VAR
    (* RSAKey holds all fields of the key.  *)

    RSAKey: RSAKeys.RSAKeyType;

    (* PublicKeyEncoded is a binary string containing 'ssh-rsa'         *)
    (* followed by the two parts of the public key.                     *)

    PublicKeyEncoded: ByteStr;

(************************************************************************)
(*               EXTRACTING THE COMPONENTS OF A PUBLIC KEY              *)
(************************************************************************)

PROCEDURE DissectPublicKey (VAR (*IN*) key: ByteStr;  VAR (*OUT*) e, n: BN);

    (* Gets the two numbers contained in the key.       *)

    VAR name: ARRAY [0..128] OF CHAR;
        size, j: CARDINAL;

    BEGIN
        size := 0;

        (* Name - although we don't actually use it so far. *)

        Copy (ADR(key.data^[0]), ADR(size), 4);
        size := Swap4 (size);
        Copy (ADR(key.data^[4]), ADR(name), size);
        name[size] := CHR(0);
        j := size + 4;

        (* Exponent (e). *)

        Copy (ADR(key.data^[j]), ADR(size), 4);  INC (j, 4);
        size := Swap4 (size);
        e := BigNum.BinToBN (key.data^, j, size);
        INC (j, size);

        (* Modulus (n). *)

        Copy (ADR(key.data^[j]), ADR(size), 4);  INC (j, 4);
        size := Swap4 (size);
        n := BigNum.BinToBN (key.data^, j, size);

    END DissectPublicKey;

(************************************************************************)
(*                        RETURNING OUR OWN KEY                         *)
(************************************************************************)

PROCEDURE GetFormattedPublicKey (): ByteStr;

    (* Returns a copy of our key in the key format of RFC 4253, section 6.6. *)

    BEGIN
        RETURN CopyOfBS (PublicKeyEncoded);
    END GetFormattedPublicKey;

(************************************************************************)
(*                        GENERATING A SIGNATURE                        *)
(************************************************************************)

PROCEDURE PKCS1_v1_5 (hashf: HashAlgs.HashAlgorithm;
                                    message: ByteStr;  k: CARDINAL): BN;

    (* Implements the EMSA-PKCS1-v1_5-ENCODE operation as specified in  *)
    (* section 9.2 (page 45) of RFC 8017.   This is the initial step    *)
    (* needed in creating a signature.  The result, as a byte string,   *)
    (* must be k bytes long.                                            *)

    (* There is obvious scope here to do the job more efficiently, but  *)
    (* until this is known to be working it's best to stay with the     *)
    (* step-by-step following of the standard.                          *)

    TYPE
        asn1str1 = ARRAY [0..14] OF CARD8;
        asn1str2 = ARRAY [0..18] OF CARD8;

    CONST
        asn1start1 = asn1str1 {30H, 21H, 30H, 09H, 06H, 05H, 2BH, 0EH,
                               03H, 02H, 1AH, 05H, 00H, 04H, 14H};

        asn1start2 = asn1str2 {30H, 31H, 30H, 0DH, 06H, 09H, 60H, 86H,
                               48H, 01H, 65H, 03H, 04H, 02H, 01H, 05H,
                               00H, 04H, 20H};

        asn1start3 = asn1str2 {30H, 51H, 30H, 0DH, 06H, 09H, 60H, 86H,
                               48H, 01H, 65H, 03H, 04H, 02H, 03H, 05H,
                               00H, 04H, 40H};

    VAR j, L, PSLen, tLen: CARDINAL;
        ctx: HashAlgs.HashCtx;
        digest: ARRAY [0..63] OF CARD8;
        T: ARRAY [0..82] OF CARD8;
        EM: ByteStr;
        M: BN;

    BEGIN
        (* Step 1: compute hash of message.  *)

        ctx := HashAlgs.HashInit (hashf);
        HashAlgs.HashUpdate (ctx, message.data^, message.size);
        HashAlgs.HashFinal (ctx, digest);

        (* hLen = 20 to 64 bytes after SHA encoding.    *)
        (* emLen = parameter k = length of RSAKey.n     *)

        (* Step 2: ASN1-encode the digest.  This only requires  *)
        (* appending the hash to the asn1start string.          *)

        CASE hashf OF
            HashAlgs.sha1:
                FOR j := 0 TO 14 DO
                    T[j] := asn1start1[j];
                END (*FOR*);
                j := 15;
                PSLen := k - 38;  tLen := 35;

          | HashAlgs.sha2_256:
                FOR j := 0 TO 18 DO
                    T[j] := asn1start2[j];
                END (*FOR*);
                j := 19;
                PSLen := k - 54;  tLen := 51;

          | HashAlgs.sha2_512:
                FOR j := 0 TO 18 DO
                    T[j] := asn1start3[j];
                END (*FOR*);
                j := 19;
                PSLen := k - 86;  tLen := 83;

        ELSE
                WriteString ("ERROR: unhandled hash function");
                WriteLn;
                HALT;
        END (*CASE*);

        L := HashAlgs.HashLength(hashf);
        Copy (ADR(digest), ADR(T[j]), L);

        (* We require k >= tLen + 11.  The length of PS (see standard)  *)
        (* is k - tLen - 3 bytes, and this must be at least 8.  For the *)
        (* hashing algorithms I am currently using, the numbers are:    *)
        (*                    tLen     PSLen   min k    min bits        *)
        (*   sha1           15+20=35   k-38     46        368           *)
        (*   sha2_256       19+32=51   k-54     62        496           *)
        (*   sha2_512       19+64=83   k-86     94        752           *)

        (* PS = emLen - tLen - 3 bytes of 0FFH. *)
        (* This is k - 38 bytes.                *)

        (* Precede T with some padding. *)

        MakeBS (EM, k);
        EM.data^[0] := 0;
        EM.data^[1] := 1;
        FILL (ADR(EM.data^[2]), 0FFH, PSLen);
        EM.data^[k-tLen-1] := 0;

        (* Move T to the tail of EM.  *)

        Copy (ADR(T), ADR(EM.data^[k-tLen]), tLen);

        (* Convert EM to a Bignum. *)

        M := BigNum.BinToBN (EM.data^, 0, EM.size);
        DiscardBS (EM);
        RETURN M;

    END PKCS1_v1_5;

(************************************************************************)

PROCEDURE Sign (hashf: HashAlgs.HashAlgorithm;  message: ByteStr): ByteStr;

    (* Creates an RSA signature using the specified hash function and   *)
    (* our private key.  RFC 4253 requires RSASSA-PKCS1-v1_5 signature  *)
    (* method as specified in RFC 8017 section 8.2 (page 35).           *)

    CONST bufsize = 512;

    VAR buffer: ARRAY [0..bufsize-1] OF CHAR;

    VAR M, V: BN;
        k, L, j: CARDINAL;
        p: VarStrings.ByteStringPtr;
        tempstr: ARRAY [0..12] OF CHAR;
        result: ByteStr;

    BEGIN
        (* Let k = length in bytes of the RSA modulus n.  This will     *)
        (* also be the length of the result.                            *)

        k := BigNum.NbytesNLZ (RSAKey.n);

        (*
        (* WHILE DEBUGGING *)
        buffer := "Public key length in bytes = ";
        MiscFuncs.AppendCard (k, buffer);
        Debug.DebugMes (buffer);
        *)

        (* Initial encoding of the message. *)

        M := PKCS1_v1_5 (hashf, message, k);

        (*
        (* WHILE DEBUGGING *)
        (* Dump the unencrypted message. *)

        Debug.DebugMes ("Signature without the encryption");
        BigNum.ToHex (M, buffer, bufsize);
        Debug.DebugMes (buffer);
        (* END DEBUG CODE *)
        *)

        (* Process with our private key.  *)

        V := RSA.Decrypt (M, RSAKey);
        BigNum.Discard (M);

        (*
        (* WHILE DEBUGGING *)
        (* Dump the message to be sent. *)

        Debug.DebugMes ("Signature after RSA processing");
        BigNum.ToHex (V, buffer, bufsize);
        Debug.DebugMes (buffer);
        (* END DEBUG CODE *)
        *)

        (* V is what is known as the "signature blob".  The complete    *)
        (* signature is, in RFC 4253 notation,                          *)
        (*      string   algorithm name                                 *)
        (*      string   rsa_signature_blob                             *)

        CASE hashf OF
            HashAlgs.sha1:
                tempstr := "ssh-rsa";
          | HashAlgs.sha2_256:
                tempstr := "rsa-sha2-256";
          | HashAlgs.sha2_512:
                tempstr := "rsa-sha2-512";
        END (*CASE*);
        L := Strings.Length (tempstr);

        (* Convert V to ByteStr form.  This must occupy exactly k   *)
        (* bytes, by inserting leading zeroes if necessary.         *)

        MakeBS (result, k + L + 8);

        j := Swap4(L);
        Copy (ADR(j), result.data, 4);
        Copy (ADR(tempstr), ADR(result.data^[4]), L);
        j := Swap4(k);
        Copy (ADR(j), ADR(result.data^[4+L]), 4);
        p := ADR(result.data^[8+L]);
        BigNum.BNtoBytesExact (V, k, p^);
        BigNum.Discard (V);

        (*
        (* WHILE DEBUGGING *)
        (* Dump the message to be sent. *)

        Debug.DebugMes ("Final version of signature");
        Debug.DumpByteString (result);
        (* END DEBUG CODE *)
        *)

        RETURN result;

    END Sign;

(************************************************************************)

PROCEDURE CheckSignature (hashf: HashAlgs.HashAlgorithm;
                            message, publickey: ByteStr;
                               VAR (*INOUT*) signature: ByteStr;
                                logID: TransactionLogID): BOOLEAN;

    (* Checks that signature is a valid signature of the message,       *)
    (* using the given public key.  The verification method is          *)
    (* specified in RFC 8017 section 8.2.2 (page 37).                   *)
    (* During processing the signature is discarded.                    *)

    CONST bufsize = 2048;

    VAR k, L: CARDINAL;
        S, V, M, key, prime: BN;
        match: BOOLEAN;
        buffer: ARRAY [0..bufsize-1] OF CHAR;

    BEGIN
        (* The signature blob starts with an algorithm name.  This is   *)
        (* redundant and may be removed.                                *)

        L := 0;  k := 0;
        Copy (signature.data, ADR(L), 4);
        L := Swap4 (L);
        Copy (ADR(signature.data^[4+L]), ADR(k), 4);
        k := Swap4 (k);
        VarStrings.BSDelete (signature, 0, L+8);

        (* We need to extract two numbers from the public key. *)

        DissectPublicKey (publickey, key, prime);

        (* k should also be the length in bytes of the RSA modulus.  *)

        IF k <> BigNum.NbytesNLZ(prime) THEN
            LogTransactionL (logID, "Wrong signature size");
            RETURN FALSE;
        END (*IF*);

        (* Encrypt the signature using the public key. *)

        S := BigNum.BinToBN (signature.data^, 0, signature.size);
        VarStrings.DiscardBS (signature);
        V := BigNum.ModularPower (S, key, prime);
        BigNum.Discard (key);
        BigNum.Discard (prime);

        (*
        (* WHILE DEBUGGING *)

        (* Dump the message as supplied. *)

        WriteString ("Signature encrypted");  WriteLn;
        BigNum.ToHex (S, buffer, bufsize);
        WriteString (buffer);  WriteLn;

        (* Dump the unencrypted message. *)

        WriteString ("Signature without the encryption");  WriteLn;
        BigNum.ToHex (V, buffer, bufsize);
        WriteString (buffer);  WriteLn;
        (* END DEBUG CODE *)
        *)
        BigNum.Discard (S);

        (* Encode the original message. *)

        M := PKCS1_v1_5 (hashf, message, k);

        (* Now M should be equal to V. *)

        match := BigNum.Cmp (M,V) = 0;

        IF NOT match THEN
            LogTransactionL (logID, "Mismatch of signature content, dumping to debug file");
            Debug.DebugMes ("Mismatch of signature content.  Signature as supplied:");
            BigNum.ToHex (V, buffer, bufsize);
            Debug.DebugMes (buffer);
            Debug.DebugMes ("Our encoding for comparison:");
            BigNum.ToHex (M, buffer, bufsize);
            Debug.DebugMes (buffer);
        END (*IF*);

        BigNum.Discard (V);
        BigNum.Discard (M);
        RETURN match;

    END CheckSignature;

(************************************************************************)
(*                           LOAD RSA PUBLIC KEY                        *)
(************************************************************************)

PROCEDURE LoadRSAPublicKey;

    (* Gets RSA public key from a file.   *)

    CONST
        filename = ".ssh\id_rsa.pub";
        MaxChars = 8192;

    VAR cid: ChanId;
        actual, pos, N: CARDINAL;
        found, error: BOOLEAN;
        string: ARRAY [0..MaxChars-1] OF CHAR;
        bytes: ARRAY [0..2047] OF CARD8;

    BEGIN
        cid := OpenOldFile (filename, FALSE, FALSE);
        IF cid = NoSuchChannel THEN
            (* Fatal error *)
            WriteString ("FATAL ERROR: Cannot open ");
            WriteString (filename);  WriteLn;
            WriteString ("Program halted");  WriteLn;
            HALT;
        END (*IF*);

        (* The content should be one big line. *)

        ReadRaw (cid, string, MaxChars, actual);

        CloseFile (cid);

        (* Leading string should be "ssh-rsa", but for now I'm not checking.  *)

        Strings.FindNext (' ', string, 0, found, pos);
        error := NOT found;
        IF NOT error THEN
            Strings.Delete (string, 0, pos+1);
        END (*IF*);

        (* Third substring is of no interest to us for now.  *)

        Strings.FindNext (' ', string, 0, found, pos);
        IF found THEN
            string[pos] := CHR(0);
        END (*IF*);

        N := Base64.Decode (string, bytes);

        (* What we have now is the formatted public key.  If we need    *)
        (* its components, see DissectPublicKey above.                  *)

        MakeBS (PublicKeyEncoded, N);
        Copy (ADR(bytes), PublicKeyEncoded.data, N);

        IF error THEN
            (* Fatal error *)
            WriteString ("FATAL ERROR: Unexpected data format in ");
            WriteString (filename);  WriteLn;
            WriteString ("Program halted");  WriteLn;
            HALT;
        END (*IF*);

    END LoadRSAPublicKey;

(************************************************************************)
(*                          LOAD RSA PRIVATE KEY                        *)
(************************************************************************)

PROCEDURE LoadRSAPrivateKey;

    (* Gets RSA private key from a file.  *)

    CONST
        filename = ".ssh\id_rsa";
        MaxChars = 32768;
        MaxBytes = 32768;

    VAR cid: ChanId;
        count: CARDINAL;
        line: ARRAY [0..255] OF CHAR;
        pstring: VarStringPtr;
        pbytes: ByteStringPtr;
        numbers: ARRAY [0..9] OF BN;

    BEGIN
        cid := OpenOldFile (filename, FALSE, FALSE);
        IF cid = NoSuchChannel THEN
            (* Fatal error *)
            WriteString ("FATAL ERROR: Cannot open ");
            WriteString (filename);  WriteLn;
            WriteString ("Program halted");  WriteLn;
            HALT;
        END (*IF*);

        (* Skip past first blank line. *)

        REPEAT
            ReadLine (cid, line);
        UNTIL line[0] = CHR(0);

        (* Read the Base64 string. *)

        ALLOCATE (pstring, MaxChars);
        pstring^[0] := CHR(0);
        ReadLine (cid, line);
        WHILE line[0] <> '-' DO
            Strings.Append (line, pstring^);
            ReadLine (cid, line);
        END (*WHILE*);
        CloseFile (cid);

        ALLOCATE (pbytes, MaxBytes);
        EVAL (Base64.Decode (pstring^, pbytes^));
        DEALLOCATE (pstring, MaxChars);

        count := ASN1.GetNumericArray (pbytes^, numbers);
        DEALLOCATE (pbytes, MaxBytes);
        WriteString ("count = ");
        WriteChar (CHR(ORD('0')+count));

        (* Assign the values to the fields of our key. *)

        BigNum.Discard (numbers[0]);
        RSAKey.n       := numbers[1];
        RSAKey.public  := numbers[2];
        RSAKey.private := numbers[3];
        RSAKey.p       := numbers[4];
        RSAKey.q       := numbers[5];
        RSAKey.dp      := numbers[6];
        RSAKey.dq      := numbers[7];
        RSAKey.qinv    := numbers[8];

    END LoadRSAPrivateKey;

(************************************************************************)
(*                          MODULE INITIALISATION                       *)
(************************************************************************)

BEGIN
    CreateDir (keydir);         (* does nothing if directory exists *)
    LoadRSAPublicKey;
    LoadRSAPrivateKey;
FINALLY
    DiscardBS (PublicKeyEncoded);
    RSAKeys.DiscardKey (RSAKey);
END sshRSA.

