(**************************************************************************)
(*                                                                        *)
(*  My own version of SSH                                                 *)
(*  Copyright (C) 2022   Peter Moylan                                     *)
(*                                                                        *)
(*  This program is free software: you can redistribute it and/or modify  *)
(*  it under the terms of the GNU General Public License as published by  *)
(*  the Free Software Foundation, either version 3 of the License, or     *)
(*  (at your option) any later version.                                   *)
(*                                                                        *)
(*  This program is distributed in the hope that it will be useful,       *)
(*  but WITHOUT ANY WARRANTY; without even the implied warranty of        *)
(*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *)
(*  GNU General Public License for more details.                          *)
(*                                                                        *)
(*  You should have received a copy of the GNU General Public License     *)
(*  along with this program.  If not, see <http://www.gnu.org/licenses/>. *)
(*                                                                        *)
(*  To contact author:   http://www.pmoylan.org   peter@pmoylan.org       *)
(*                                                                        *)
(**************************************************************************)

IMPLEMENTATION MODULE sshSession;

        (********************************************************)
        (*                                                      *)
        (*                   Session handler                    *)
        (*                                                      *)
        (*  This is the server version. Although the client and *)
        (*  server code are almost identical, we separate them  *)
        (*  out because we need more legality checks in the     *)
        (*  server case.                                        *)
        (*                                                      *)
        (*  Programmer:         P. Moylan                       *)
        (*  Started:            4 January 2021                  *)
        (*  Last edited:        1 October 2022                  *)
        (*  Status:             OK                              *)
        (*                                                      *)
        (********************************************************)


IMPORT OS2, Strings;

FROM SYSTEM IMPORT
    (* type *)  ADDRESS, CARD32,
    (* proc *)  CAST;

FROM TransportLayer IMPORT
    (* proc *)  RunTransportLayer;

FROM PacketIO IMPORT
    (* type *)  IOPort,
    (* proc *)  OpenPacketSession, ClosePacketSession;

FROM IPFilter IMPORT
    (* proc *)  CheckMasterIPFilter;

FROM Sockets IMPORT
    (* const*)  NotASocket,
    (* type *)  Socket, SockAddr,
    (* proc *)  getpeername, soclose, so_cancel;

FROM Inet2Misc IMPORT
    (* proc *)  IPToString;

FROM MiscFuncs IMPORT
    (* proc *)  ConvertCard;

FROM Conversions IMPORT
    (* proc *)  CardinalToString;

FROM LogCtx IMPORT
    (* var  *)  sshctx;

FROM TransLog IMPORT
    (* type *)  TransactionLogID,
    (* proc *)  CreateLogID, DiscardLogID, SetSyslogHost, StartTransactionLogging,
                LogTransaction, LogTransactionL, UpdateTopScreenLine;

FROM Queues IMPORT
    (* type *)  Queue,
    (* proc *)  CreateQueue, DestroyQueue, AddToQueue, TakeFromQueue;

FROM Timer IMPORT
    (* proc *)  Sleep, TimedWaitSpecial;

FROM Semaphores IMPORT
    (* type *)  Semaphore,
    (* proc *)  CreateSemaphore, DestroySemaphore, Signal;

FROM TaskControl IMPORT
    (* type *)  Lock,
    (* proc *)  CreateLock, DestroyLock, Obtain, Release, CreateTask, CreateTask1,
                TaskExit;

FROM Storage IMPORT
    (* proc *)  ALLOCATE, DEALLOCATE;

FROM MiscFuncs IMPORT
    (* proc *)  EVAL;

(********************************************************************************)

CONST
    Nul = CHR(0);
    ClientPriority = 3;

TYPE
    (* String form of a session identifier. *)

    SevenChar = ARRAY [0..7] OF CHAR;

    (* Data used in creating a new instance of the session handler task. *)

    NewSessionPointer = POINTER TO NewSessionRecord;
    NewSessionRecord = RECORD
                           port: IOPort;
                           socket: Socket;
                           clientaddr: CARDINAL;
                           LogID: TransactionLogID;
                           logcode: SevenChar;
                       END (*RECORD*);

    (* Data needed by the timeout checker task. *)

    KeepAlivePointer = POINTER TO
                           RECORD
                               SocketOpen, dying: BOOLEAN;
                               WatchdogRunning: BOOLEAN;
                               TimedOut: BOOLEAN;
                               sem: Semaphore;
                               socket: Socket;
                               SessionID: SevenChar;
                           END (*RECORD*);
    (* A list of currently active client addresses. *)

    SessionListPtr = POINTER TO
                         RECORD
                             next: SessionListPtr;
                             addr: CARDINAL;
                             count: CARDINAL;
                         END (*RECORD*);

    SessionListInfo =   RECORD
                            access: Lock;
                            head: SessionListPtr;
                        END (*RECORD*);

VAR
    (* A queue of KeepAlive records, which are passed to the    *)
    (* TimeoutChecker tasks as they are created.                *)

    KeepAliveQueue: Queue;

    (* Timeout delay, in milliseconds, and its mutual exclusion lock. *)

    MaxTime: CARDINAL;
    MaxTimeLock: Lock;

    (* Lock to slow down excessive thread creation during a denial      *)
    (* of service attack.                                               *)

    ThreadCreationLock: Lock;

    (* Maximum allowed number of simultaneous users.  For critical *)
    (* section we use lock UserCountLock, see below.               *)

    MaxUsers: CARDINAL;

    (* Maximum allowed number of clients with same IP address, and  a   *)
    (* list used to check that condition.                               *)

    SameIPLimit: CARDINAL;
    SessionList: SessionListInfo;

    (* Count of active users, and a lock to protect it. *)

    UserCount: CARDINAL;
    UserCountLock: Lock;

    (* Flag to control whether we write the number of users to the screen. *)

    LogToScreen: BOOLEAN;

(************************************************************************)
(*                UPDATING THE COUNT OF ACTIVE SESSIONS                 *)
(************************************************************************)

PROCEDURE UpdateCount (increment: INTEGER): CARDINAL;

    (* Updates the count of the number of users, and returns the new    *)
    (* count.  Special case: if this would take us beyond the MaxUsers  *)
    (* limit, then the count is not updated and the returned value      *)
    (* is zero.                                                         *)

    VAR value, pos: CARDINAL;  Buffer: ARRAY [0..15] OF CHAR;

    BEGIN
        Obtain (UserCountLock);
        IF increment > 0 THEN INC (UserCount, increment);
        ELSIF increment < 0 THEN DEC (UserCount, -increment)
        END (*IF*);
        value := UserCount;
        IF value > MaxUsers THEN
            DEC (UserCount, increment);  value := 0;
        ELSIF LogToScreen THEN
            pos := 0;
            ConvertCard (value, Buffer, pos);
            Buffer[pos] := ' ';  INC(pos);
            Buffer[pos] := CHR(0);
            UpdateTopScreenLine (69, Buffer);
        END (*IF*);
        Release (UserCountLock);
        RETURN value;
    END UpdateCount;

(************************************************************************)
(*                         THE WATCHDOG TIMER                           *)
(************************************************************************)

PROCEDURE TimeoutChecker;

    (* A new instance of this task is created for each client session.  *)
    (* It kills the corresponding SessionHandler task if more than      *)
    (* MaxTime milliseconds have passed since the last Signal() on the  *)
    (* session's KeepAlive semaphore.                                   *)

    (* This is a workaround.  I would have preferred to set the         *)
    (* timeout in the socket options, but I haven't yet figured out     *)
    (* how to do it.  An older version of IBM's sockets documentation   *)
    (* gave details on the send and receive timeouts, but this seems    *)
    (* to have disappeared from later versions of the documentation.    *)

    VAR p: KeepAlivePointer;  LogID: TransactionLogID;
        TimeLimit, KillCount: CARDINAL;

    BEGIN
        p := TakeFromQueue (KeepAliveQueue);
        p^.WatchdogRunning := TRUE;
        p^.SessionID[0] := 'W';
        LogID := CreateLogID (sshctx, p^.SessionID);
        (*LogTransactionL (LogID, "Watchdog timer started");*)
        REPEAT
            Obtain (MaxTimeLock);
            TimeLimit := MaxTime;
            Release (MaxTimeLock);
            TimedWaitSpecial (p^.sem, TimeLimit, p^.TimedOut);
        UNTIL p^.TimedOut OR p^.dying;
        IF p^.TimedOut THEN
            LogTransactionL (LogID, "Cancelling client socket");
            IF p^.socket <> NotASocket THEN
                so_cancel (p^.socket);
            END (*IF*);
            (*p^.socket := NotASocket;*)
        ELSE
            (*LogTransactionL (LogID, "Session terminated without timing out");*)
            Sleep(50);     (* Let session close socket itself *)
        END (*IF*);

        (* Wait for the socket to be closed. *)

        IF p^.SocketOpen THEN
            LogTransactionL (LogID, "Waiting for client socket to be closed");
        END (*IF*);
        KillCount := 0;
        WHILE p^.SocketOpen DO
            Sleep (5000);

            (* For the really stubborn cases, keep trying to cancel *)
            (* the sockets.                                         *)

            IF p^.SocketOpen THEN
                LogTransactionL (LogID, "Still trying to cancel client socket");
                so_cancel (p^.socket);
            END (*IF*);
            INC (KillCount);

            IF KillCount >= 10 THEN
                LogTransactionL (LogID, "Unable to cancel client socket, giving up");
                EVAL(soclose (p^.socket));
                p^.SocketOpen := FALSE;
                p^.socket := NotASocket;
            END (*IF*);

        END (*WHILE*);

        DiscardLogID (LogID);

        (* Note that the KeepAlive record still exists.  We leave it    *)
        (* up to the client thread to discard it.                       *)

        p^.WatchdogRunning := FALSE;

    END TimeoutChecker;

(********************************************************************************)
(*                                LEGALITY CHECKS                               *)
(********************************************************************************)

PROCEDURE AddToSessionList (IPaddr: CARDINAL): BOOLEAN;

    (* Adds addr to the list of active client addresses, also checks for too    *)
    (* many clients with the same address.  Returns TRUE if all OK.             *)

    VAR SP: SessionListPtr;
        current: SessionListPtr;
        searching, acceptable: BOOLEAN;

    BEGIN
        acceptable := TRUE;
        Obtain (SessionList.access);
        SP := SessionList.head;  searching := TRUE;
        WHILE searching AND (SP <> NIL) DO
            IF SP^.addr = IPaddr THEN
                searching := FALSE;
                acceptable := SP^.count < SameIPLimit;
                IF acceptable THEN
                    INC (SP^.count);
                END (*IF*);
            END (*IF*);
            SP := SP^.next;
        END (*WHILE*);

        (* If no record found, create a new one. *)

        IF searching THEN
            NEW (current);
            current^.next := SessionList.head;
            current^.addr := IPaddr;
            current^.count := 1;
            SessionList.head := current;
        END (*IF*);
        Release (SessionList.access);
        RETURN acceptable;

    END AddToSessionList;

(********************************************************************************)

PROCEDURE RemoveFromSessionList (IPaddr: CARDINAL);

    (* Removes addr from the list of active client addresses.  *)

    VAR prev, SP, next: SessionListPtr;
        searching: BOOLEAN;

    BEGIN
        Obtain (SessionList.access);
        prev := NIL;  SP := SessionList.head;  searching := TRUE;
        WHILE searching AND (SP <> NIL) DO
            IF SP^.addr = IPaddr THEN
                searching := FALSE;
                IF SP^.count > 0 THEN
                    DEC (SP^.count);
                END (*IF*);
                IF SP^.count = 0 THEN
                    next := SP^.next;
                    IF prev = NIL THEN
                        SessionList.head := next;
                    ELSE
                        prev^.next := next;
                    END (*IF*);
                    DISPOSE (SP);
                    SP := next;
                END (*IF*);
            END (*IF*);
            prev := SP;
            IF SP <> NIL THEN
                SP := SP^.next;
            END (*IF*);
        END (*WHILE*);

        Release (SessionList.access);

    END RemoveFromSessionList;

(********************************************************************************)

PROCEDURE RejectClient (addr: CARDINAL;  LogID: TransactionLogID): BOOLEAN;

    (* Returns TRUE iff we find this client unacceptable. *)

    CONST BannedHostDelay = 240*1000 (*milliseconds*);

    VAR LogMessage: ARRAY [0..127] OF CHAR;
        IPBuffer: ARRAY [0..16] OF CHAR;

    BEGIN
        IF NOT CheckMasterIPFilter(addr) THEN
            Sleep (BannedHostDelay);
            Strings.Assign ("Client address rejected ", LogMessage);
            IPToString (addr, TRUE, IPBuffer);
            Strings.Append (IPBuffer, LogMessage);
            LogTransaction (LogID, LogMessage);
            RETURN TRUE;
        END (*IF*);

        IF NOT AddToSessionList (addr) THEN
            Sleep (BannedHostDelay);
            LogTransactionL (LogID, "Too many users with same address");
            RETURN TRUE;
        END (*IF*);

        RETURN FALSE;

    END RejectClient;

(********************************************************************************)
(*                                RUN ONE SESSION                               *)
(********************************************************************************)

PROCEDURE RunSession (arg: ADDRESS);

    (* Runs as a separate thread, keeps looping until told to exit. *)

    VAR NSP: NewSessionPointer;
        sess: IOPort;
        S: Socket;
        clientaddr: CARDINAL;
        UserNumber: CARDINAL;
        LogFilePrefix: SevenChar;
        LogID: TransactionLogID;
        KeepAliveSemaphore: Semaphore;
        KA: KeepAlivePointer;

    BEGIN
        NSP := arg;
        OS2.DosError (OS2.FERR_DISABLEHARDERR);              (* disable hard error popups *)

        (* Copy the NewSessionPointer^ information. *)

        Obtain (ThreadCreationLock);
        sess := NSP^.port;
        LogID := NSP^.LogID;
        LogFilePrefix := NSP^.logcode;
        S := NSP^.socket;
        clientaddr := NSP^.clientaddr;
        Release (ThreadCreationLock);
        DEALLOCATE (NSP, SIZE(NewSessionRecord));

        UserNumber := UpdateCount (+1);

        (* Check for too many users. *)

        IF UserNumber = 0 THEN
            LogTransactionL (LogID, "Too many users");
            DiscardLogID (LogID);
            RemoveFromSessionList (clientaddr);
            RETURN;
        END (*IF*);

        (* Create an instance of the TimeoutChecker task. *)

        CreateSemaphore (KeepAliveSemaphore, 0);
        NEW (KA);
        WITH KA^ DO
            SocketOpen := TRUE;  socket := S;
            WatchdogRunning := TRUE;
            dying := FALSE;
            sem := KeepAliveSemaphore;
            SessionID := LogFilePrefix;
            TimedOut := FALSE;
        END (*WITH*);

        IF CreateTask (TimeoutChecker, ClientPriority+2, "sftpd timeout") THEN
            AddToQueue (KeepAliveQueue, KA);
        ELSE
            LogTransactionL (LogID, "Failed to create watchdog thread");
            soclose (S);
            DestroySemaphore (KeepAliveSemaphore);
            DISPOSE (KA);
            EVAL (UpdateCount(-1));
            DiscardLogID (LogID);
            RemoveFromSessionList (clientaddr);
            TaskExit;
        END (*IF*);

        (* Now we start talking to the client.  *)

        RunTransportLayer (sess, NIL, KA^.sem, LogID);

        IF KA^.TimedOut THEN
            LogTransactionL (LogID, "Timed out");
        END (*IF*);

        KA^.dying := TRUE;  Signal (KA^.sem);
        EVAL(soclose (S));
        KA^.SocketOpen := FALSE;
        WHILE KA^.WatchdogRunning DO
            Signal (KA^.sem);
            Sleep (500);
        END (*WHILE*);
        DestroySemaphore (KA^.sem);
        DISPOSE (KA);
        EVAL (UpdateCount (-1));
        DiscardLogID (LogID);
        RemoveFromSessionList (clientaddr);

    END RunSession;

(********************************************************************************)
(*                            STARTING A NEW SESSION                            *)
(********************************************************************************)

PROCEDURE NewSession (S: Socket;  client: SockAddr): IOPort;

    (* Launches a session. *)

    (* Starts and runs one instance of the SSH transport layer.  The session    *)
    (* runs in a separate thread; this procedure returns after starting the     *)
    (* session, it does not wait until the session is over.                     *)

    VAR NSP: NewSessionPointer;
        ourport: IOPort;
        peer: SockAddr;
        clientaddr: CARDINAL;
        IPBuffer: ARRAY [0..16] OF CHAR;
        message: ARRAY [0..31] OF CHAR;
        LogFilePrefix: SevenChar;
        ID: TransactionLogID;
        size: CARDINAL;
        success: BOOLEAN;

    BEGIN
        Obtain (ThreadCreationLock);

        clientaddr := client.in_addr.addr;

        (* Create the log file ID, and log session start. *)

        CardinalToString (S, LogFilePrefix, 7);
        ID := CreateLogID (sshctx, LogFilePrefix);
        size := SIZE(SockAddr);
        IF NOT getpeername (S, peer, size) THEN
            message := "New client ";
            IPToString (peer.in_addr.addr, TRUE, IPBuffer);
            Strings.Append (IPBuffer, message);
            LogTransaction (ID, message);
        END (*IF*);

        IF RejectClient(clientaddr, ID) THEN
            soclose (S);
            DiscardLogID (ID);
            Release (ThreadCreationLock);
            RETURN NIL;
        END (*IF*);

        (* Open the encrypted port. *)

        ourport := OpenPacketSession (S, TRUE, ID);

        NEW (NSP);
        WITH NSP^ DO
            LogID := ID;
            logcode := LogFilePrefix;
            port := ourport;
            socket := S;
        END (*WITH*);
        NSP^.clientaddr := clientaddr;
        success := CreateTask1 (RunSession, 3, "ssh session", NSP);
        Release (ThreadCreationLock);
        IF NOT success THEN
            LogTransactionL (ID, "Failed to create new thread");
            RemoveFromSessionList (peer.in_addr.addr);
            ClosePacketSession (ourport);
            soclose (S);
            DiscardLogID (ID);
            DEALLOCATE (NSP, SIZE(NewSessionRecord));
            RETURN NIL;
        END (*IF*);
        RETURN ourport;
    END NewSession;

(************************************************************************)
(*                          SETTING LIMITS                              *)
(************************************************************************)

PROCEDURE SetSameIPLimit (limit: CARDINAL);

    (* Sets the limit on maximun number of simultaneous clients. *)

    BEGIN
        Obtain (SessionList.access);
        SameIPLimit := limit;
        Release (SessionList.access);
    END SetSameIPLimit;

(************************************************************************)

PROCEDURE SetMaxUsers (max: CARDINAL);

    (* Sets the limit on maximun number of simultaneous clients. *)

    BEGIN
        Obtain (UserCountLock);
        MaxUsers := max;
        Release (UserCountLock);
    END SetMaxUsers;

(************************************************************************)

PROCEDURE SetTimeout (seconds: CARDINAL);

    (* Specifies how long a session can be idle before it is forcibly   *)
    (* closed.                                                          *)

    CONST
        limit_in_ms = MAX(CARD32) DIV 2;
        limit = limit_in_ms DIV 1000;

    BEGIN
        Obtain (MaxTimeLock);
        IF seconds > limit THEN
            MaxTime := limit_in_ms;
        ELSE
            MaxTime := 1000*seconds;
        END (*IF*);
        Release (MaxTimeLock);
    END SetTimeout;

(************************************************************************)

PROCEDURE SetTransactionLogging (level: CARDINAL;
                                 VAR (*IN*) logfile: ARRAY OF CHAR;
                                 VAR (*IN*) SyslogHost: ARRAY OF CHAR);

    (* Option to control transaction logging: 0 for none, 1 for disk,   *)
    (* 2 for screen, 4 for pipe, 8 for translog, and combinations.      *)

    BEGIN
        IF level > 15 THEN
            level := 15;
        END (*IF*);
        LogToScreen := ODD (level DIV 2);
        IF SyslogHost[0] <> Nul THEN
            SetSyslogHost (SyslogHost);
        END (*IF*);
        StartTransactionLogging (sshctx, logfile, level);
    END SetTransactionLogging;

(********************************************************************************)
(*                            MODULE INITIALISATION                             *)
(********************************************************************************)

BEGIN
    CreateLock (MaxTimeLock);
    CreateLock (UserCountLock);
    Obtain (UserCountLock);
    UserCount := 0;  SetMaxUsers (10);
    Release (UserCountLock);
    CreateLock (SessionList.access);
    SetSameIPLimit (MAX(CARDINAL));
    SessionList.head := NIL;
    CreateLock (ThreadCreationLock);
    CreateQueue (KeepAliveQueue);
    LogToScreen := TRUE;
FINALLY
    DestroyLock (ThreadCreationLock);
    DestroyLock (UserCountLock);
    DestroyLock (MaxTimeLock);
END sshSession.

