MODULE keygen;

        (********************************************************)
        (*                                                      *)
        (*                Generating an RSA key                 *)
        (*                                                      *)
        (*  Programmer:         P. Moylan                       *)
        (*  Last edited:        28 July 2021                    *)
        (*  Status:             OK                              *)
        (*                                                      *)
        (*      In case you're wondering what RSA stands for,   *)
        (*      it's Rivest-Shamir-Adleman, after the           *)
        (*      inventors of the algorithm.                     *)
        (*                                                      *)
        (********************************************************)


IMPORT IOChan, TextIO, BigNum, Base64, ASN1, Strings;

FROM SYSTEM IMPORT CARD8, ADR;

FROM RSAKeys IMPORT
    (* type *)  RSAKeyType,
    (* proc *)  DiscardKey;

FROM RSAGenerate IMPORT
    (* proc *)  RSA_Generate;

FROM BigNum IMPORT
    (* type *)  BN;

FROM VarStrings IMPORT
    (* type *)  ByteStr, ByteStringPtr,
    (* proc *)  DiscardBS;

FROM TimeConv IMPORT
    (* proc *)  time;

FROM FileOps IMPORT
    (* type *)  ChanId,
    (* proc *)  Exists, DeleteFile, MoveFile, CreateDir, OpenNewFile, CloseFile,
                FWriteString, FWriteLn;

FROM ProgramArgs IMPORT
    (* proc *)  ArgChan, IsArgPresent;

FROM Inet2Misc IMPORT
    (* proc *)  Swap4, AddressToHostName;

FROM Sockets IMPORT
    (* proc *)  gethostid;

FROM STextIO IMPORT
    (* proc *)  WriteChar, WriteString, WriteLn;

FROM Names IMPORT
    (* type *)  HostName;

FROM LowLevel IMPORT
    (* proc *)  EVAL, Copy, AddOffset;

(************************************************************************)

CONST
    TESTING = FALSE;
    keydir = ".ssh";
    defaultkeylength = 1536;         (* bits *)
    testOptions = "-b 512";        (* while testing *)

(************************************************************************)
(*                       GETTING THE PARAMETERS                         *)
(************************************************************************)

PROCEDURE GetParams (VAR (*OUT*) keylength: CARDINAL);

    (* Only option in this version is   *)
    (*      -b bits                     *)

    TYPE CharSet = SET OF CHAR;

    CONST Digits = CharSet {'0'..'9'};

    VAR Options: ARRAY [0..63] OF CHAR;
        j: CARDINAL;

    (********************************************************************)

    PROCEDURE SkipBlanks;

        BEGIN
            LOOP
                IF Options[j] <> ' ' THEN EXIT(*LOOP*) END(*IF*);
                IF j = 63 THEN
                    Options[j] := CHR(0);  EXIT (*LOOP*);
                ELSE
                    INC (j);
                END (*IF*);
            END (*LOOP*);
        END SkipBlanks;

    (********************************************************************)

    VAR args: IOChan.ChanId;

    BEGIN
        keylength := 0;
        args := ArgChan();
        IF IsArgPresent() OR TESTING THEN
            IF TESTING THEN
                Options := testOptions;
            ELSE
                TextIO.ReadString (args, Options);
            END (*IF*);
            j := 0;  SkipBlanks;
            LOOP
                CASE Options[j] OF
                    CHR(0): EXIT (*LOOP*);
                  | '0'..'9':
                            (* Be tolerant of a missing -b *)

                            IF keylength = 0 THEN
                                WHILE Options[j] IN Digits DO
                                    keylength := 10*keylength
                                                + ORD(Options[j]) - ORD('0');
                                    INC (j);
                                END (*WHILE*);
                            END (*IF*);
                  | '-':    INC (j);
                  | 'b':    INC (j);  SkipBlanks;
                            WHILE Options[j] IN Digits DO
                                keylength := 10*keylength
                                            + ORD(Options[j]) - ORD('0');
                                INC (j);
                            END (*WHILE*);
                ELSE
                    WriteString ("Unknown option ");
                    WriteChar (Options[j]);  WriteLn;
                    INC(j);
                END (*CASE*);
                SkipBlanks;
            END (*LOOP*);
        END (*IF*);
        IF keylength = 0 THEN
            keylength := defaultkeylength;
        END (*IF*);
    END GetParams;

(************************************************************************)
(*                      WRITING A PUBLIC KEY FILE                       *)
(************************************************************************)

PROCEDURE WriteCard (N: CARDINAL);

    (* Writes N to standard output in decimal.  *)

    BEGIN
        IF N > 9 THEN
            WriteCard (N DIV 10);
            N := N MOD 10;
        END (*IF*);
        WriteChar (CHR(ORD('0') + N));
    END WriteCard;

(************************************************************************)

PROCEDURE WritePublicKeyFile (key: RSAKeyType;  keylength: CARDINAL);

    (* keylength is specified in bytes. *)

    CONST filename = ".ssh/id_rsa.pub";  bakname = ".ssh/id_rsa.pub.bak";

    VAR cid: ChanId;
        N, size, pos: CARDINAL;
        p: ByteStringPtr;
        name: ARRAY [0..7] OF CHAR;
        bytes: ARRAY [0..16383] OF CARD8;
        encoded: ARRAY [0..16383] OF CHAR;
        ourhostname: HostName;

    BEGIN
        IF Exists(filename) THEN
            IF Exists(bakname) THEN
                DeleteFile (bakname);
            END (*IF*);
            EVAL (MoveFile (filename, bakname));
        END (*IF*);
        cid := OpenNewFile (filename, FALSE);
        FWriteString (cid, "ssh-rsa ");

        (* The next part has three components, which have to be *)
        (* concatenated before the Base64 encoding.             *)

        p := ADR(bytes);
        pos := 0;

        (* Name = "ssh-rsa". *)

        size := Swap4 (7);  name := "ssh-rsa";
        Copy (ADR(size), p, 4);
        p := AddOffset (p, 4);
        Copy (ADR(name), ADR(bytes[4]), 7);
        p := AddOffset (p, 7);
        INC (pos, 4 + 7);

        (* Public exponent. *)

        size := BigNum.Nbytes (key.public);
        N := Swap4(size);
        Copy (ADR(N), p, 4);
        p := AddOffset (p, 4);
        N := BigNum.BNtoBytes (key.public, p^);
        p := AddOffset (p, N);
        INC (pos, 4 + N);

        (* Modulus. *)

        size := BigNum.Nbytes (key.n);
        N := Swap4(keylength);
        Copy (ADR(N), p, 4);
        p := AddOffset (p, 4);
        BigNum.BNtoBytesExact (key.n, keylength, p^);
        INC (pos, 4 + keylength);

        (* Encode as base64.  *)

        WriteString ("Public key (base64) = ");
        Base64.Encode (bytes, pos, encoded);
        WriteString (encoded);
        WriteLn;

        (* Finish writing to file.  *)

        FWriteString (cid, encoded);
        EVAL (AddressToHostName(Swap4(gethostid()), ourhostname));
        FWriteString (cid, " root@");  FWriteString (cid, ourhostname);
        FWriteLn (cid);
        CloseFile (cid);

    END WritePublicKeyFile;

(************************************************************************)
(*                     WRITING A PRIVATE KEY FILE                       *)
(************************************************************************)

PROCEDURE WritePrivateKeyFile (key: RSAKeyType);

    CONST filename = ".ssh/id_rsa";  bakname = ".ssh/id_rsa.bak";
        CharsPerLine = 64;  Nul = CHR(0);

    VAR cid: ChanId;
        amount: CARDINAL;
        Zero: BN;
        asn: ByteStr;
        keydata: ARRAY [0..8] OF BN;
        encoded: ARRAY [0..16383] OF CHAR;
        line: ARRAY [0..CharsPerLine-1] OF CHAR;

    BEGIN
        (* Copy the information we want to put in the file.  *)

        Zero := BigNum.MakeBignum(0);

        keydata[0] := Zero;
        keydata[1] := key.n;
        keydata[2] := key.public;
        keydata[3] := key.private;
        keydata[4] := key.p;
        keydata[5] := key.q;
        keydata[6] := key.dp;
        keydata[7] := key.dq;
        keydata[8] := key.qinv;

        (* Convert to ASN.1 notation.  *)

        asn := ASN1.ASNEncodeArray (9, keydata);
        BigNum.Discard (Zero);

        (* Further encoded it as Base64. *)

        Base64.Encode (asn.data^, asn.size, encoded);
        DiscardBS (asn);

        (* Write the final result to the file.  *)

        IF Exists(filename) THEN
            IF Exists(bakname) THEN
                DeleteFile (bakname);
            END (*IF*);
            EVAL (MoveFile (filename, bakname));
        END (*IF*);
        cid := OpenNewFile (filename, FALSE);
        FWriteString (cid, "-----BEGIN RSA PRIVATE KEY-----");  FWriteLn (cid);
        FWriteLn (cid);
        REPEAT
            amount := Strings.Length (encoded);
            IF amount > CharsPerLine THEN
                amount := CharsPerLine;
            END (*IF*);
            Strings.Extract (encoded, 0, amount, line);
            Strings.Delete (encoded, 0, amount);
            FWriteString (cid, line);
            FWriteLn (cid);
        UNTIL encoded[0] = Nul;
        FWriteString (cid, "-----END RSA PRIVATE KEY-----");  FWriteLn (cid);
        CloseFile (cid);
    END WritePrivateKeyFile;

(************************************************************************)
(*                            KEY GENERATION                            *)
(************************************************************************)

PROCEDURE WriteKey (key: RSAKeyType);

    (* Writes public and private key to standard output. *)

    BEGIN
        WriteString ("PUBLIC KEY");  WriteLn;
        WriteString ("n = ");     BigNum.WriteBignum(key.n);  WriteLn;
        WriteString ("e = ");     BigNum.WriteBignum (key.public);  WriteLn;

        WriteString ("PRIVATE KEY");  WriteLn;
        WriteString ("d = ");     BigNum.WriteBignum (key.private);  WriteLn;
        WriteString ("p = ");     BigNum.WriteBignum (key.p);  WriteLn;
        WriteString ("q = ");     BigNum.WriteBignum (key.q);  WriteLn;
        WriteString ("dp = ");    BigNum.WriteBignum (key.dp);  WriteLn;
        WriteString ("dq = ");    BigNum.WriteBignum (key.dq);  WriteLn;
        WriteString ("qinv = ");  BigNum.WriteBignum (key.qinv);  WriteLn;

    END WriteKey;

(************************************************************************)

PROCEDURE CreateTheKey (keylength: CARDINAL);

    (* Parameter is desired key length in bits.  *)

    VAR key: RSAKeyType;
        t0, t1, N: CARDINAL;

    BEGIN
        WriteString ("Generating an RSA key, please be patient");  WriteLn;
        t0 := time();
        key := RSA_Generate(keylength);
        t1 := time();

        WriteKey (key);
        N := BigNum.Nbits(key.n);
        keylength := (keylength + 4) DIV 8;     (* bytes *)
        WriteString ("Key length is ");  WriteCard (N);
        WriteString (" significant bits, but we will round it up to ");
        WriteCard (8*keylength);  WriteString (" bits");  WriteLn;
        IF N < 368 THEN
            WriteString ("WARNING: at least 368 bits are needed to create a RSASSA-PKCS1-V1_5 signature");
            WriteLn;
        END (*IF*);

        IF t1 < t0 THEN t1 := t0 END (*IF*);
        WriteString ("Generation took ");  WriteCard (t1-t0);
        WriteString (" seconds");  WriteLn;

        WritePublicKeyFile (key, keylength);
        WritePrivateKeyFile (key);
        DiscardKey (key);

    END CreateTheKey;

(************************************************************************)

VAR keylength: CARDINAL;

BEGIN
    CreateDir (keydir);
    GetParams (keylength);
    CreateTheKey (keylength);
END keygen.

